/*
 ============================================================================
 Name        : global.h
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */
/****************************************************************
*
*                 global.h (global include file for Drodi project)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

#ifndef GLOBAL_H_
#define GLOBAL_H_
/* Include for platform independent modules */

/*--- Standard Includes ----------------------------------------*/
#include <stddef.h>                     /* Standard definitions         */
#include <string.h>                     /* strcpy/memset                */
#include <ctype.h>                      /* isdigit/toupper...           */
#include <stdio.h>                      /* sprintf/fopen...             */
#include <stdlib.h>                     /* malloc/free...               */
#include <time.h>                       /* clock(),...                  */
#include <sys/time.h>                   /* gettimeofday()               */
#include <limits.h>                     /* CHAR_BIT                     */
#include <stdint.h>                     /* intxx_t data types           */
#include <locale.h>                     /* setlocale()                  */
#include <math.h>                       /* sqrt()                       */
#include <signal.h>                     /* signal(),...                 */
#include <errno.h>                      /* strerror()                   */
#include <unistd.h>                     /* getpid(), sysconf()...       */
#include <sys/types.h>                  /* getpid()                     */
#include <sys/resource.h>               /* getrlimit(), setrlimit()     */
#include <sys/mman.h>                   /* mmap()                       */
#include <semaphore.h>                  /* sem_xxx()                    */
#include <sys/wait.h>                   /* wait()                       */

/* Define's for compilation time code generation control. */
//#define DEBUGMEMORY                         /* Generate debug code for memory problems detected by Valgrind.*/
                                            /* See MYFREE() comments for information. */
//#define DEBUGCODE                           /* Generate additional debug code. */
//#define VERBOSE                             /* Display inferences if DEBUGCODE is defined and only one core enabled. */
//#define DEBUGTREE                             /* Generate debug code for discrimination tree integrity.*/
//#define DEBUGTREE2                            /* Generate debug code for discrimination tree queries. The checks will */
                                              /* only be done if DEBUGTREE is not defined. See ChecDscTQueries() */
                                              /* function global comments for additional information. */
#ifdef DEBUGTREE2
#define NBOFTESTS 3                           /* Number of literals to be tested by ChecDscTQueries() function. */
#endif
//#define SATDEBUGCODE1                       /* Generate additional debug code specific to check coherence */
                                            /* of hashchcmp elements ordered sub-queue. */
//#define SATDEBUGCODE2                       /* Generate additional debug code specific to check coherence */
                                            /* of satclause elements ordered sub-queue. */
//#define VERIFYSATPROOF                      /* Verify that the SAT part of the proof is coherent. */
#define DEBUGKBDATA 0                       /* Set to 0 for normal operations. Set to 1 to force print KB data even in */
                                            /* TIMEOUT or NOMEMORY conditions. */
//#define MURMUR3                             /* Use Murmur3 hash function (default is to use KRM hash function). */
//#define BENCHMARKING                        /* Include code for benchmarking summary output. */
                                            /* VERY IMPORTANT: If BENCHMARKING is defined then the BENCHDIMENSION parameter */
                                            /* below must be adjusted appropriately. */
#ifdef BENCHMARKING
#define BENCHTYPE 3                         /* 1 -> report solving times for problems solved. */
                                            /* 2 -> report THM, SAT, CSA, CAX or UNS for problems solved. */
                                            /* 3 -> report 1 and 2 above for problems solved. */
#define VERIFYSATFINALSTATUS 0              /* If this equals one then SAT final status is verified after a problem solving attempt. */
                                            /* Only if BENCHMARKING is defined this can be set to one, otherwise it must be set */
                                            /* to zero. See VerifySatStatus() function for details. */
#define BENCHDIMENSION 381                  /* Adjust this to the number of problems per run. */
#else
#define VERIFYSATFINALSTATUS 0              /* Don't modify this. See comments above. */
#endif
//#define MEMORYTUNNING                       /* If this is defined some code is generated for memory limit fine tuning and the */
                                            /* SymMalloc() function is defined and called from Initialize_pgm() and Clean4Exit() */
                                            /* functions to simulate huge memory allocation. */
#ifdef MEMORYTUNNING
#define ADDTLMEMORY 4000000000              /* Additional memory to be allocated if MEMORYTUNNING is defined. */
#endif
#define ALLOWDOTSINFNAME                     /* If this is defined then dots are allowed in "name" field and possibly in other */
                                             /* positions inside the of a TPTP annotated formula. */

/* Other algorithm parameter definitions. These are candidates to be adjusted via command options. The definitions */
/* of some of them may significantly change the code generation. */
#define MAXDIST2CONJ 3                  /* Maximum distance to conjecture. See global comments in ClauseWieght() function for details. */
#define VARWEIGHT 1.0                   /* Relative weight of variable symbols for clause weight evaluation. */
#define PREDWEIGHT 10.0                 /* Relative weight of predicate symbols for clause weight evaluation. This weight is taken */
                                        /* as a reference and should be always equal to 10.0. */
#define EQUWEIGHT 2.0                   /* Relative weight of equalities for clause weight evaluation. If this is commented out then */
                                        /* equalities are not accounted for in clause weight calculations. */
#define FUNCWEIGHT 5.0                  /* Relative weight of function symbols for clause weight evaluation. */
#define PNOTINCONJFACTOR1 1.5           /* Factor that multiplies weight of predicates not in the conjecture for first factor choice. */
#define PNOTINCONJFACTOR2 1.5           /* Same as above for second factor choice. */
#define ENOTINCONJFACTOR1 1.0           /* Factor that multiplies weight of equalities if there are no equalities in the conjecture */
                                        /* for first factor choice. */
#define ENOTINCONJFACTOR2 1.0           /* Same as above for second factor choice. */
#define FNOTINCONJFACTOR1 1.4           /* Factor that multiplies weight of functions not in the conjecture for first factor choice. */
#define FNOTINCONJFACTOR2 1.4           /* Same as above for second factor choice. */
#define POSLITERALFACTOR1 2.0           /* Factor that multiplies weight of items in positive literals of non input clauses for first */
                                        /* factor choice. */
#define POSLITERALFACTOR2 1.0           /* Same as above for second factor choice. */
#define POSLITERALFACTOR1I 1.0          /* Factor that multiplies weight of items in positive literals of input clauses for first */
                                        /* factor choice. It seems that it produces better results to assign lower values to this factor, */
                                        /* even to set it to 1.0. */
#define POSLITERALFACTOR2I 1.0          /* Same as above for second factor choice. */
//#define ORIENTEDEQFACTOR 0.5            /* Factor to apply to weights of positive oriented equalities and their sub-terms. */
                                        /* If this is not defined then no factor will be applied. */
#define SINGLEANY                       /* If this is defined then the SelectItems() function will select the first */
                                        /* negative literal if selection option is set to SINGLENEG or the first positive */
                                        /* literal if selection is set to SINGLEPOS. If this is NOT defined then the */
                                        /* SelectItems() function will select the MAXIMUM (if it exist) or first MAXIMAL */
                                        /* negative (if SINGLENEG) or positive (if SINGLEPOS) literal with respect to all */
                                        /* the negative literals. The results obtained is that it is better to define SINGLEANY. */
#define DEMODPOSITION 1                 /* If this equals 1 then demodulations are done first in FwdSimplify() function. */
                                        /* If this equals 2 then demodulations are done just after FirstSimplify() call. */
                                        /* If this equals 3 then demodulations are done just after SecondSimplify() call. */
                                        /* Value 1 is a bit better than the other values. */
#define KEEPSATMODEL 1                  /* If this equals 1 then each execution of SatSolver() function keeps the previous model */
                                        /* until a node is found without a forced option match. This is a slightly better option. */
                                        /* If this doesn't equal 1 then each execution of SatSolver() function keeps the previous */
                                        /* model until a node is found without a forced option match and there are still some  */
                                        /* forced options available. */
#define PRECEDENCE 1                    /* If this is zero then the symbol precedence (precedence field in haschain structure) */
                                        /* is set increasingly as new symbols are defined. Otherwise the symbols that appear */
                                        /* more times are assigned a lower precedence. */
//#define TREENODESFIFO                   /* If this is defined then nodes that are not variables will be added FIFO to the peer */
                                        /* nodes chain, otherwise they will be added LIFO. See TREEVARLAST dfinition for the */
                                        /* position of nodes that are variables in the peer nodes chain. */
//#define TREELEAFSFIFO                   /* If this is defined then leafs will be added FIFO to the peer leafs chain, otherwise */
                                        /* they will be added LIFO. */
#define TREEVARLAST                     /* If this is defined then discrimination tree nodes that are variables are always */
                                        /* in the last position of the peer nodes chain. Otherwise they are placed in the */
                                        /* first position. */
//#define PROGCLAUSESELECT                /* If this is defined then the clause selection in each round robin cycle is */
                                        /* done progressively through all the queues involved. If this is not defined */
                                        /* then the clause selection in each round robin cycle is done depleting each */
                                        /* queue one after the other. */
//#define PREFERPROCESSED                 /* If this is defined then clauses that are added to PASSIVE more than once are placed at */
                                        /* the top of the age queue. */
//#define UNITCLRESOLUTION                /* If this is defined then a resolution between a unit clause without assertions being */
                                        /* forward simplified will be attempted with active unit clauses without assertions. */
                                        /* It is much better not to use this, but the code has been left just in case it may be */
                                        /* useful in the future. */

/* Definitions for parse functions return types. */
#define NEGATION 1                          /* Negation. */
#define UNIVERSALQ 2                        /* Valid universal quantifier. */
#define EXISTENTIALQ 3                      /* Valid existential quantifier. */
#define BRACKET 4                           /* Open parenthesis "(". */
#define VALIDATOMIC 5                       /* Valid atomic sentence. */
#define UNDECLAREDTERM 6                    /* Predicate or function not previously declared. */
#define DECLAREDFUNC 7                      /* Term has been previously declared as a function. */
#define DECLAREDPRED 8                      /* Term has been previously declared as a predicate. */
#define STDVARNAME 9                        /* Standardized variable reserved name. */
#define DECLAREDVAR 10                      /* Declared variable. */
#define VALIDVARLIST 11                     /* Valid variable list. */
#define VALIDSENTENCE 12                    /* Valid sentence. */
#define VALIDNAME 13                        /* Valid name. */

/* Define's for syntax error codes. */
#define NOMEMORY 100000100                  /* No memory. */
#define EMPTYSENTENCE 100000101             /* Empty sentence. */
#define EMPTYBLOCK 100000102                /* Empty parenthesis block. */
#define INVTPTPVARNAME 100000103            /* Invalid TPTP variable name. */
#define INVTPTPNAME 100000104               /* Invalid TPTP predicate or function name. */
#define INVALIDATOMIC 100000105             /* Invalid atomic sentence. */
#define INVALIDVARLIST 100000106            /* Invalid variable list. */
#define INVALIDNAME 100000107               /* Invalid or too long variable, constant, predicate or function name. */
#define INVALIDTERM 100000109               /* Invalid term. */
#define MISSINGARGS 100000110               /* Missing arguments. */
#define ALREADYDECLARED 100000111           /* Variable name already declared in equal scope quantifier. */
#define MISSINGVARINLIST 100000112          /* Missing variable in a variable list. */
#define INVALIDDECLAREDARGS 100000113       /* Predicate or function doesn't match previous declared number of arguments. */
#define INVALIDDECLAREDTYPE 100000114       /* Predicate or function doesn't match previous declared type. */
#define TOOMANYARGS 100000115               /* Too many arguments. */
#define RESERVEDNAME 100000116              /* Use of standard variable name for a function or predicate. */
#define SYNTAXERROR 100000117               /* Generic syntax error. */

/* Definitions for flags field of kbase structure. */
#define RETENTIONUSED 1                     /* LRS or weight limit retention processes have been used and some clauses have */
                                            /* been discarded. The KB now is not complete. */

/* Definitions for flags field of cmprefix structure and type argument of simplify functions. */
/* These flags must be compatible with SELECTED flag and also with HORN flag except FROMNOTAXIOM */
/* that may have the same value as HORN flag. The HORN flag is set in clauses if appropriate but */
/* it is not used by the moment. The flag FROMCONJECTURE.  must be compatible with flags in type */
/* field of hashchain structure. */
#define TEXTFORMULA 1                       /* Formula is of text type. */
                                            /* These formulas have their own linked queue. */
#define UNPROC 2                            /* Formula is an unprocessed binary clause. */
                                            /* These formulas have their own linked queue. */
#define PASSIVE 4                           /* Formula is a passive binary clause. */
                                            /* These formulas have their own linked queue. */
#define ACTIVE 8                            /* Formula is an active binary clause. */
                                            /* These formulas have their own linked queue. */
#define LOCKED 16                           /* Formula is a locked binary clause. */
                                            /* These formulas have their own linked queue. */
#define NOFACTORS 32                        /* Clause have no factors. */
#define INDISPOSALQUEUE 128                 /* Clause is in disposal queue for later deletion or locking. */
#define FROMPASSIVE 256                     /* Clause is coming from PASSIVE, so only subsumption must be checked */
                                            /* in FirstSimplify() function. This is only used as a flag argument in */
                                            /* some simplifying functions. This is not used in the flags field of the */
                                            /* cmprefix structure. */
#define NUMBERED 512                        /* Clause has been assigned a number. */
#define INPROOFQUEUE 1024                   /* Clause or hashchcmp component is already in proof queue. This flag must be */
                                            /* compatible with hashchcmp structure flags. */
#define ASSERTSCHAINED 2048                 /* Assertions for this clause has been already chained. Only used for */
                                            /* binary clauses. */
#define FROMCONJECTURE 4096                 /* Formula was obtained from a negated conjecture in the pre-processing. */
                                            /* Used in hashchain type field flags for symbols appearing in these formulas. */
#define SELECTEDCLAUSE 8192                 /* Clause has been selected from PASSIVE to ACTIVE. Clauses with this flag */
                                            /* that are added to PASSIVE are placed at the top of the age queue. */
                                            /* With the AVATAR architecture a clause can go from PASSIVE to ACTIVE */
                                            /* to UNPROC to PASSIVE and this cycle can be repeated several times. */
#define KEPTCLAUSE 16384                    /* Clause was kept in the relevance filtering pre-processing. */
#define FROMNOTAXIOM 32768                  /* Symbol appears in axiom like formulas hypothesis, definition, assumption, */
                                            /* lemma, theorem and corollary, but not in formulas with axiom role. */
											/* Used only in hashchain type field flags for symbols appearing in these clauses. */
#define FROMMAINFILE 65536                  /* Clause is coming from the main file and can be considered as a conjecture */
                                            /* to the effects of detecting contradictory axioms. */
#define PICKED 131072                       /* Clause has been selected as ACTIVE. Due to AVATAR architecture an active clause */
                                            /* may later be an unprocessed clause so any type of clause may have this flag. */
#define WEIGHTED 262144                     /* Clause weights have been already assigned. */
#define GOODCHOICE 524288                   /* Clause has been classified as a good choice by the learning algorithms. */
#define SIMPLIFIED 1048576                  /* Clause has been simplified. See FirstSimplify() function global comments. */

/* Definitions for inference field of cmprefix structure. */
#define AXIOM 0                             /* Formula is an axiom. */
#define HYPOTHESIS 1                        /* Formula is an hypothesis (axiom like). */
#define DEFINITION 2                        /* Formula is a definition (axiom like). */
#define ASSUMPTION 3                        /* Formula is an assumption (axiom like). */
#define NEGCONJ 4                           /* Formula is a negated conjecture. This is also used as a flag and must be */
                                            /* compatible with FROMMAINFILE above. */
#define CONJECTURE 5                        /* Formula was an input conjecture to the problem. */
#define LEMMA 6                             /* Formula is a lemma (axiom like, redundant). */
#define THEOREM 7                           /* Formula is a theorem (axiom like, redundant). */
#define COROLLARY 8                         /* Formula is a corollary (axiom like, redundant). */
#define PLAIN 9                             /* Formula has a plain role, accepted as an axiom just in case . */
#define CLAUSIFY 10                         /* Formula comes from clausifying an input formula. */
#define RESOLUTION 11                       /* Formula was inferred by resolution. */
#define PARAMODULATION 12                   /* Formula was inferred by paramodulation. */
#define FACTORING 13                        /* Formula was inferred by factoring. */
#define FWDEMODULATION 14                   /* Formula was inferred by forward demodulation. */
#define BKDEMODULATION 15                   /* Formula was inferred by backward demodulation. */
#define REMOVEDUPS 16                       /* Formula is the result of removing duplicate literals from another formula. */
#define TRIVEQRES 17                        /* Formula was inferred by trivial equality resolution. */
#define EQRESOLUTION 18                     /* Formula was inferred by equality resolution. */
#define DESTEQRES 19                        /* Formula was inferred by destructive equality resolution. */
#define EQFACTORING 20                      /* Formula was inferred by equality factoring. */
#define FWSUBSRESLTNS 21                    /* Formula was inferred by forward subsumption resolution. */
#define BKSUBSRESLTNS 22                    /* Formula was inferred by backward subsumption resolution. */
#define SPLIT 23                            /* Formula was inferred by clause splitting. */
#define COMPONENT 24                        /* Formula is a clause component. */
#define CONTRADICTION 25                    /* Formula was inferred by first order solver contradiction. */
#define PRENNFCONV 26                       /* Formula is the result of a pre-nnf conversion. */
#define FORMULARENAMING 27                  /* Formula is the result of renaming a subformula. */
#define PREDICATEDEF 28                     /* Formula is a predicate definition for subformula renaming. */
#define NNFCONV 29                          /* Formula is the result of a full-nnf conversion. */
#define MINISCOPING 30                      /* Formula is the result of miniscoping. */
#define SKOLEMIZATION 31                    /* Formula is the result of skolemization. */
#define TFSIMPLIFICATION 32                 /* Formula is the result of a $true and $false simplification. */

/* Definitions for byte markers in compiled clauses and discrimination trees, type field */
/* of hashchain structures and type parameter of CreateGlobalSymbol() function. These flags must be compatible */
/* with FROMCONJECTURE and FROMMAINFILE also used in type field of hashchain structures for learning process. */
#define UNITEND 1                       /* End of clause marker. Only used for byte markers. */
                                        /* This flag must be compatible with flags for Generalize() and */
                                        /* Unify() functions and with flags in hashchcmp and satnode structures */
                                        /* and assertions signbyte (see below). */
#define EQUALITY 2                      /* Item is an equality literal. Also used in UpdateSymbolParms() function flags. */
#define PREDICATE 4                     /* Item is a predicate literal. */
#define NEGATED 8                       /* Item is a negated literal. Only used for byte markers and in */
                                        /* UpdateSymbolParms() function flags. */
                                        /* This flag must be compatible with ORIENTEDEQ flag and flags for Generalize() */
                                        /* and Unify() functions (see below). */
#define VARIABLE 16                     /* Item is a variable. Only used for byte markers. */
#define FUNCTION 32                     /* Item is a function (may be a constant, i.e. arity zero). */
#define SELECTED 64                     /* Item is a selected literal or equality root term. */
                                        /* This flag must be compatible with flags for QueryGeneric() function. */
#define MAXIMUM 128                     /* Item is a maximum equality root term. This must be different than ITEM1SEC, STANDARD */
                                        /* and NONRECURSIVE flags. */
#define PMSELECTED 128                  /* If this is set to the same value as MAXIMUM then a positive equality will be selected */
                                        /* for paramodulating from it only if it is in a clause without negative literals. */
                                        /* If this is set to the same value as SELECTED then the above condition will not */
                                        /* be required. It works a bit better setting it to the same value as MAXIMUM. */
#define DEFINEDPRED 256                 /* Used only in "type" parameter of CreateGlobalSymbol() function */
                                        /* and type field of hashchain structure. */
#define INCONJECTURE 512                /* Symbol is used in the conjecture. Used only in "type" parameter */
                                        /* of CreateGlobalSymbol() function. */
#define RELEVANT 1024                   /* Symbol is in a relevant clause. Used only in type field of hashchain */
                                        /* structure for clause relevance filtering. */
#define SKOLEM 2048                     /* Symbol is a skolem. Used only in type field of hashchain structure for */
                                        /* learning  process. */
#define DISTINCTOBJECT 8192             /* Symbol is a "distinct object". Used only in type field of hashchain structure */
                                        /* and type parameter of CreateGlobalSymbol() function. */

/* Definitions for "type" parameter in NextItem() function. */
#define OVERSUBTERMS 1                  /* Jump to next item that is not a subterm of the given item. */
#define IMMED 2                         /* Jumpt to item immediately following the given item. */

/* Definitions for Unify() and Generalize() flags. */
/* These flags must be compatible with the UNITEND and NEGATED flags (see above). */
#define LIST 2                          /* Parts to be unified are a list of terms. */
#define ITEM1SEC 16                     /* The first Unify() function argument or the variable in a substitution */
                                        /* belongs to a secondary clause. They belong to the main clause by default. */
                                        /* Also used for flag value of CheckEqRootOrder() function so it must be */
                                        /* different than MAXIMUM, STANDARD and NONRECURSIVE values defined elsewhere. */
#define ITEM2SEC ITEM1SEC<<1            /* The second Unify() function argument or the substituting term in a */
                                        /* substitution  belongs to a secondary clause. They belong to the main */
                                        /* clause by default. This must be ITEM1SEC shifted by 1 bit. */
#define INITSUBST 64                    /* Initialize substitution to null. */
#define DEBUG 128                       /* Call BuildTxtSubstData() function to show results in text form. */

/* Definitions for status fields of kbase and mpctl structures. They must be compatible with NOMEMORY above. */
#define STOPPED 0                       /* Process is stopped. */
#define RUNNING 1                       /* Process is running a (possibly multi-processing) first order solver algorithm */
                                        /* with or without a SAT solver. */
#define RUNNINGSAT 2                    /* Process is running a single process SAT solver algorithm. */
#define SATISFIABLE 3                   /* Formulas in KB are satisfiable, if there are conjectures */
                                        /* they are not refutable. */
#define UNSATISFIABLE 4                 /* Formulas in KB are unsatisfiable, if there are conjectures */
                                        /* they are refutable. */
#define UNKNOWN 5                       /* It is unknown if formulas in KB are satisfiable or unsatisfiable. */
#define TIMEOUT 6                       /* Saturation algorithm timed out. This cannot be 0 or 1. */

/* Definitions for ordtype values and CompareTerms() flag. */
#define STANDARD 1                      /* Use standard KBO ordering. */
#define NONRECURSIVE 2                  /* Use non recursive KBO ordering. This flag must be compatible  with */
                                        /* MAIMUM and ITEM1SEC flags defined elsewhere. */
#define LEXICOGRAPHIC 4                 /* Use >> symbol precedence relation for atom ordering. */

/* Definitions for demodulation and destructive equality resolution options (DEMODULAITON option). */
#define DEMODON 0                       /* Demodulation is enabled without restrictions. */
#define DEMODOFF 1                      /* Demodulation is disabled. */
#define DEMODCPL 2                      /* Demodulation and destructive equality resolution are restricted */
                                        /* to cases that don't compromise completeness. */

/* Definitions for weight functions. */
#define UNIFORM 0                       /* Use a uniform weight function assigned a weight=1 to all functions. */
#define ARITY 1                         /* Use weight function = 1 + 10·arity(function). */

/* Definitions for ordering functions return values and for the result field of the */
/* equpnode structure. See the functions description for the meaning of each one, */
/* as it may differ from one function to another. */
#define EQUAL 1                         /* item1 = item2. */
#define GREATER 2                       /* item1 > item2. */
#define LESSER 4                        /* item2 > item1. */
#define NOGREATER 8                     /* item2 ·> item1 but neither item2 = item1 nor item2 > item1. */
#define NOLESSER 16                     /* item1 ·> item2 but neither item1 = item2 nor item1 > item2. */
#define FAILURE 32                      /* None of the above. */
#define EQUALW 64                       /* weight(item1) = weight(item2). */
#define FAILUREW 128                    /* weight(item1) is not comparable with weight(item2). */

/* Definitions for algorithm field of kbase structure. These cannot be 0 nor 1 for compatibility */
/* with other values of split field in options structure. */
#define OTTER 2                         /* Otter algorithm. */
#define DISCOUNT 3                      /* Discount algorithm. */

/* Definitions for EditDiscrTrees() function flag parameter. */
#define ADDTREES 1                      /* Add tree branches for specified clause. */
#define DELETETREES 2                   /* Delete tree branches for specified clause. */

/* Definitions for type argument of QueriDscTree() function. */
#define UNIFICATION 0                   /* Search is for an item that unifies with the query term. */
#define INSTANCE 1                      /* Search is for an item that is an instance of the query term. */
#define GENERALIZATION 2                /* Search is for an item that is a generalization of the query term. */

/* Definitions for flags of AddBinClause2KB() function. */
#define SELECT 1                        /* Appropriate clause items will be selected. */
#define SETFNUMBER 2                    /* Set formula number according to the nformulas field of KB. */

/* Definitions for flags of JumpOverBlockOrTerm() function. */
#define TOEND 0                         /* Jump to the last item in the block or term. */
#define PASTEND 1                       /* Jump to the item following the last item in the block or term. */

/* Definitions for mxwghtopt field of options structure. */
#define MXWEIGHTOFF 1                   /* Maximum weight option deactivated for algorithms without LRS enabled. */
#define MXWEIGHTLRS 2                   /* Maximum weight calculated using Limited Resource Strategy enabled */
                                        /* for all algorithms. */
#define MXWEIGHTLRSOTT 4                /* Maximum weight calculated using Limited Resource Strategy enabled for Otter */
                                        /* algorithm, disabled for Discount algorithm. */
#define MXWEIGHTLRSDSC 8                /* Maximum weight calculated using Limited Resource Strategy enabled for Discount */
                                        /* algorithm, disabled for Otter algorithm. */
#define MXWEIGHTLRSOFF 16               /* Maximum weight calculated using Limited Resource Strategy id disabled */
                                        /* for all algorithms. */
#define MXWEIGHTNUM 32                  /* Maximum weight selected by user for algorithms without LRS enabled. */

/* Definitions for SYNTAX and PROOF options. */
#define TPTP 1                          /* TPTP mode syntax. */
#define DRODI 2                         /* Drodi mode syntax. */
#define PROOFOFF 3                      /* Don't print proof. */

/* Definitions for program modes (pgmmode global variable). */
#define NONINTERACTIVE 0                /* non interactive mode, entered if some arguments are given to the program. */
#define INTERACTIVE 1                   /* Interactive mode, entered if no arguments are given to the program. */

/* Definitions for select field of options structure. The definitions are set as if they were flags */
/* to speed up some tests, but only one of the values (flags) is allowed. */
#define MAXIMAL 1                       /* All maximal literals are selected. */
#define SINGLENEG 2                     /* If there are negative literals then the first one is selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define MULTINEG 4                      /* If there are negative literals then the negative literals */
                                        /* that are maximal in the set of negative literals are selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define SINGLEPOS 8                     /* If there are positive literals then the first one is selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define MULTIPOS 16                     /* If there are positive literals then the positive literals */
                                        /* that are maximal in the set of positive literals are selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */

/* Definitions for flags field in hashchcmp and satnode structures and assertions signbyte */
/* (see AddAssertions() function comments). These definitions must be compatible with UNITEND */
/* that is also used in assertions signbyte and with INPROOFQUEUE cmprefix structure flag */
/* that is also used in hashchcmp flags. */
#define POSITIVE 2                      /* The component is true in SAT interpretation model or in assertion_data. */
#define NEGATIVE 4                      /* The component is false in SAT interpretation model or in assertion_data. */
#define UNDEFINED 8                     /* The component is undefined in SAT interpretation model. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define GROUND 16                       /* The component is ground. This is only used in flags field of hashchcmp structure. */
                                        /* This flag must be compatible with flags for pbmflags global variable and serves */
                                        /* to identify ground problems. */
#define PRPOSITIVE 32                   /* The component was true in previous model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define PRNEGATIVE 64                   /* The component was false in previous model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define PRUNDEFINED 128                 /* The component was undefined in previous model of SAT interpretation. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define LASTOPTION 256                  /* The assignment of node is the last possibility available, no other possibility */
                                        /* (true or false) is left or valid to be explored.*/
#define ORDERED 512                     /* The hashchcmp or satclause structure element has been linked to the ordered */
                                        /* ordered sub-queue. */
#define FREENODE 2048                   /* The node is allocated but it is not in use and its data is no longer necessary. */
                                        /* The node can be reused. This way unnecessary memory allocation and freeing is */
                                        /* avoided. This is only used in flags field of satnode structure. */
#define MODELPOSITIVE 4096              /* The component is true in current model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define MODELNEGATIVE 8192              /* The component is false in current model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define MODELUNDEFINED 16384            /* The component is undefined in current model of SAT interpretation. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define POSCOMPADDED 32768              /* The component clause C <- [C] has been added to first order solver. This is */
                                        /* only used in flags field of hashchcmp structure. */
#define NEGCOMPADDED 65536              /* The component clause ~C <- ~[C] has been added to first order solver. This is */
                                        /* only used in flags field of hashchcmp structure. */

/* Definitions for flags in pbmflags global variable. The GROUND flag defined above is also used for pbmflags. */
#define EXISTCONJ 1                     /* A conjecture exist for the current problem. */
#define EXISTNEGCONJ 2                  /* A negated conjecture exist for the current problem. */
#define WITHHORNCLAUSES 4               /* Problem is not horn but there are horn clauses. */
#define UNITEQUALITY 8                  /* Problem has some positive unit equalities. */
#define NONUNITEQUALITY 32              /* Problem has some equalities none of which are positive unit equalities. */
#define UNITCLAUSE 64                   /* All clauses in problem are unit clauses. */
#define NOTCAX 128                      /* The proof includes some formulas from main file and therefore has not */
                                        /* contradictory axioms. */
#define SATREFUTATION 256               /* There was a Sat refutation. */
#define HORN 32768                      /* Problem is horn. This flag must be compatible with flags field */
                                        /* of cmprefix structure. */
#define CNFFORMAT 65536                 /* Problem was read in CNF TPTP format. */
#define EXISTAXIOMFILE 131072           /* Problem included an axiom file. */

/* Definitions for type field in proofnode structure. */
#define SYMBOLDEFNTN 1                  /* Symbol definition. */
#define SATCLAUSE 2                     /* Sat clause. */
#define ACLAUSE 3                       /* A-Clause. */

/* Definitions for flags field in lkedfitem structure. */
#define OPENBRACKET 1                   /* Opening bracket type item. */
#define CLOSEBRACKET 2                  /* Closing bracket type item. */
#define EQUPRED 4                       /* Equality predicate type item. */
#define NAMEDPRED 8                     /* Normal predicate type item. */
#define NAMEDFUNC 16                    /* Function type item. */
#define NAMEDVAR 32                     /* Variable type item. */
#define UQUANTIFIER 64                  /* Universal quantifier type item. */
#define EQUANTIFIER 128                 /* Existential quantifier type item. */
#define ANDOPER 256                     /* AND connective type item. */
#define OROPER 512                      /* OR connective type item. */
#define IMPOPER 1024                    /* Implication connective type item. */
#define REVIMPOPER 2048                 /* Reverse implication connective type item. */
#define EQUIVOPER 4096                  /* Equivalence connective type item. */
#define XOROPER 8192                    /* XOR connective type item. */
#define NOROPER 16384                   /* NOR connective type item. */
#define NANDOPER 32768                  /* NAND connective type item. */
#define NEGITEM 65536                   /* Negated item. */
#define POSPOL 131072                   /* Positive polarity item. */
#define NEGPOL 262144                   /* Negative polarity item. */
#define NEUTRALPOL 524288               /* Neutral polarity item. */
#define RENAMED 1048576                 /* Variable renamed in a quantifier or variable item. */
#define FREEVAR 2097152                 /* The quantifier variable is a free variable in the renamed block. */
#define TRUEPRED 4194304                /* The predicate is a $true predefined predicate. */
#define FALSEPRED 8388608               /* The predicate is a $false predefined predicate. */

/* Definitions for childtype field of dstrnode structure. */
#define NODECHILD 0                     /* Child is a tree node (dstrnode structure). */
#define LEAFCHILD 1                     /* Child is a tree leaf (dstrleaf structure). */

/* Definitions for clause relevance filtering program option (prproflags global variable). */
#define CLFILTERINGRESIGN 1             /* Clause relevance filtering is not performed performed in case of timeout condition. */
#define CLFILTERINGOFF 2                /* No clause relevance filtering is performed. */
#define CLFILTERINGON 4                 /* Clause relevance filtering is always performed. */
#define CLFILTERINGNONEQU 8             /* Clause relevance filtering is performed in problems without equalities only. */
                                        /* Only one of the CLFILTERINGOFF, CLFILTERINGON or CLFILTERINGNONEQU flags can be set. */
                                        /* However CLFILTERINGRESIGN flag is compatible with the other three flags. */
#define CLFILTERINGITERLIMIT 16         /* The threshold in ppthrshld global variable is the maximum number of iterations */
                                        /* allowed in the relevance filtering algorithm. Otherwise the threshold in ppthrshld */
                                        /* global variable is the maximum time in seconds allowed for the algorithm. */
#define CLFILTERINGLIMITREACHED 32      /* A timeout condition occurred or maximum number of iterations reached during */
                                        /* relevance filtering pre-processing before */
                                        /* the filtering process was completed. */
#define CLFILTERINGPRUNNED 64           /* The relevance filtering deleted one or more clauses. */

/* Definitions of masks for each type of adjustable parameter. */
#define ALGMASK 1                       /* ALGORITHM parameter mask. */
#define SELECTMASK 0b1000000000110      /* SELECT parameter mask. */
#define WFACTSETMASK 0b1000             /* Weight factor index parameter mask. */
#define DEMODMASK 0b100000000010000     /* Demodulation parameter mask. */
#define FACTORINGMASK 0b100000          /* FACTORING parameter mask. */
#define TERMORDRMASK 0b1000000          /* TERMORDER parameter mask. */
#define LITORDRMASK 0b110000000         /* LITORDER parameter mask. */
#define FUNCWMASK 0b1000000000          /* FUNCTIONW parameter mask. */
#define MAXWMASK 0b110000000000         /* MAXWEIGHT parameter mask. */

/* Definitions of values for each type of adjustable parameter. */
#define OTTER_P 0                       /* Otter algorithm. */
#define DISCOUNT_P 1                    /* Discount algorithm. */
#define SINGLENEG_P 0                   /* SINGLENEG SELECT option. */
#define MULTINEG_P 0b10                 /* MULTINEG SELECT option. */
#define MAXIMAL_P 0b100                 /* MAXIMAL SELECT option. */
#define SINGLEPOS_P 0b110               /* SINGLEPOS SELECT option. */
#define MULTIPOS_P 0b1000000000000      /* MULTIPOS SELECT option. */
#define WFACT1_P 0                      /* Weight factor index 1. */
#define WFACT2_P 0b1000                 /* Weight factor index 2. */
#define DEMODON_P 0                     /* Demodulation is ON. */
#define DEMODOFF_P 0b10000              /* Demodulation is OFF. */
#define FACTON_P 0                      /* ON FACTORING option. */
#define FACTOFF_P 0b100000              /* OFF FACTORING option. */
#define TRMORDSTD_P 0                   /* STANDARD TERMORDER option. */
#define TRMORDNR_P 0b1000000            /* NONRECURSIVE TERMORDER option. */
#define LITORDSTD_P 0                   /* STANDARD LITORDER option. */
#define LITORDNR_P 0b10000000           /* NONRECURSIVE LITORDER option. */
#define LITORDLEX_P 0b100000000         /* LEXICOGRAPHIC LITORDER option. */
#define UNIFORM_P 0                     /* UNIFORM FUNCTIONW option. */
#define ARITY_P 0b1000000000            /* ARITY FUNCTIONW option. */
#define LRSOTT_P 0                      /* LRSOTTER MAXWEIGHT option. */
#define LRSOFF_P 0b10000000000          /* LRSOFF MAXWEIGHT option. */
#define LRSALL_P 0b100000000000         /* LRS MAXWEIGHT option. */
#define LRSDSC_P 0b110000000000         /* LRSDISCOUNT MAXWEIGHT option. */
#define LEARN_P 0b10000000000000        /* Use learning option. This doen't need an adjustable parameter mask. */
#define DEMODCPL_P 0b100000000000000    /* Redundancy check is on for demodulation and destructive equality resolution. */

/* Definitions to be used if VERIFYSATSOLVER is not zero. They are always defined anyway. */
#define SATOK 0                         /* SAT solver status is OK. */
#define SATSAT 1                        /* SAT solver has reported a refutation but all SAT clauses are satisfied. */
#define SATUNSAT 2                      /* SAT solver has reported a model but not all SAT clauses are satisfied. */
#define SATUNPR1 4                      /* There are unprocessed clauses that must be locked. */
#define SATUNPR2 8                      /* There are clauses in unprocessed queue not flagged as unprocessed. */
#define SATUNPR12 16                    /* Conditions SATUNPR1 and SATUNPR2. */
#define SATPASSIVE1 32                  /* There are passive clauses that must be locked. */
#define SATPASSIVE2 64                  /* There are clauses in passive queue not flagged as passive. */
#define SATPASSIVE12 128                /* Conditions SATPASSIVE1 and SATPASSIVE2. */
#define SATACTIVE1 256                  /* There are active clauses that must be locked. */
#define SATACTIVE2 512                  /* There are clauses in active queue not flagged as active. */
#define SATACTIVE12 1024                /* Conditions SATACTIVE1 and SATACTIVE2. */
#define SATLOCKED1 2048                 /* There are active clauses that must be locked. */
#define SATLOCKED2 4096                 /* There are clauses in active queue not flagged as active. */
#define SATLOCKED12 8192                /* Conditions SATLOCKED1 and SATLOCKED2. */
#define SATUNK 16384                    /* SAT solver status is unknown because the problem was not reported as solved. */

/* Definitions for the automatic learning process. */
#define VECTORLENGTH 2471               /* Number of bytes in a clause vector used for learning. */
#define CONJECTURECOMPONENT 10.0        /* Value of feature if a conjecture exist. */
#define AXIOMFILECOMPONENT 10.0         /* Value of feature if an axiom file exist. */
#define PENALTYPARAMETER 0.1            /* Penalty parameter for learning process. */
#define LEARNPRECISION 0.005            /* Learning result vector relative error threshold. */
#define TOLERANCE 26.0                  /* Duality gap tolerance before starting to check learning result vector relative error */
                                        /* in dual DCDescentMethod() function. */
#define BADCLAUSEWEIGHT 100             /* Weight penalty if the clause is not classified as good by the learning algorithm. */
#define WEIGHT0FACTOR 1.0               /* Factor that multiplies clause clweight[1] to get clause clweight[0]. */
#define SHRINKING                       /* If defined then a shrinking DCDescentMethod() function will be used. */
                                        /* Otherwise a DCDescentMethod() function without shrinking will be used. */

/* Miscellanea definitions. */
#define VERSION "Drodi V3.3.3"          /* Program version. */
#define SPREAD 0                        /* Use spread core affinity bind option. */
#define CLOSE 1                         /* Use close core affinity bind option. */
#define LRSTHRESHOLD 200                /* Number of passive clauses moved to active per LRS cycle. */
#define LRSCOEF 1.1                     /* Coefficient to apply to number of LRS reachable clauses. */
#define TIMEOUTCCLDSP 10                /* Number of cycles for timeout check in clause disposal. */
#define INPUT_CHUNK_SIZE 1024           /* Chunk size in bytes for command input. */
#define VAR_CHUNK_SIZE 100              /* Chunk size in bytes for text formula names. */
#define STR_CHUNK_SIZE 200              /* Chunk size in bytes for text buffers. */
                                        /* This must be bigger or equal to 10. */
#define TREEPTR_CHUNK_SIZE 4            /* Pointer set chunk size in tree pointer elements. */
#define SUBST_CHUNK_SIZE 1000           /* Chunk size in bytes for substitution buffer. */
#define CLAUSE_CHUNK_SIZE 200           /* Compiled auxiliary clauses chunk size in bytes. */
#define PNDARGS_CHUNK_SIZE 10           /* Pending arguments elements chunk size in number of elements. */
#define VARRENUM_CHUNK_SIZE 50          /* Variable renumbering buffer size in number of elements. */
#define ARGSMASK 0x1fffffff             /* Mask for number of arguments. */
#define MAXRECURSIVE 60000              /* Maximum recursive calls in QueryGeneric() and QueryUI() functions. */
#define ABSMAXWEIGHT 0xffffffff         /* Absolute maximum clause weight. This is to prevent exhausting the stack space by */
                                        /* too many calls to QueryGeneric() for too long clauses. It is rare but it may happen. */
#define MAXWEIGHT 1000                  /* Maximum number of weight queue pointers hold in each weight queue. */
#define WEIGHTQUEUES 5                  /* Number of weight queues used for clause selection. */
#define SYMBOLDIM1 3                    /* Number of different symbol weights used to compute the different clause weights. */
                                        /* This is the dimension of the weight set field of hashchain structure. */
#define SYMBOLDIM2 2                    /* Number of different total symbol weights in a clause. This is the dimension of the */
                                        /* symbolw set field of binprefix structure. */
#define MAXPOSITIONS 400                /* Maximum number of hash values for item positions. */
#define MAXSTRATEGIES 4320              /* Number of predefined strategies, this depends on the number of parameters that */
                                        /* are automatically adjusted depending on the problem type. See InitStrategies() */
                                        /* function for details. */
#define BKSTRATEGIESPERPROC 8           /* Number of backup strategies per process to cover early returns due to incompleteness. */
#define TUNEDSTRATEGIES 200             /* Number of tuned strategies for each problem type. */
#ifdef UNITCLRESOLUTION
#define DSCTREETYPES 8                  /* Number of discr[] set field elements in hashchain structure. */
#else
#define DSCTREETYPES 6                  /* Number of discr[] set field elements in hashchain structure. */
#endif
#define HASHVALUES 0x10000              /* Number of different hash values for symbol names and clause variants. */
#define MAXMINISCOPETIME 1.0            /* Miniscope time limit in seconds. */
#define EXITWAITSECONDS 1               /* Number of seconds to wait for exit of a non solution winner children process. */
                                        /* Child exit is indicated by clearing procsem[] process specific semaphore. */
#define KILLWAITITERATIONS 20           /* Number of iterations to wait for state changes in child processes via waitpid(). */
#define KILLWAITNANOSECS 10000000       /* Number of nanoseconds to wait between iterations for state changes in child */
                                        /* processes via waitpid(). This must be less than 1000000000. */
#define KILLWAITNANOSECS2 100000000     /* Number of nanoseconds to wait before waiting for state changes in child processes */
                                        /* via waitpid() if an interruption occurs. This must be less than 1000000000. */
#define TMOWAITMICROSECS 500000         /* Additional time in microseconds to wait for timeouts not caught by normal means. */
                                        /* This must be less than 1000000. */
#ifdef WINTYPE
#define WPATHSEPARATOR L'\\'            /* Windows type wide char file path separator. */
#else
#define WPATHSEPARATOR L'/'             /* Linux & AIX type wide char file path separator. */
#endif

/* Useful macros */
#ifdef DEBUGMEMORY
#define MYFREE(X) myfree(X,__FILE__,__LINE__)
#define DBGMEMELEMS 2000
#else
#define MYFREE(X) myfree(X)
#define DBGMEMELEMS 0
#endif
#define ABS(X) ((X)>=0?X:-(X))
#define MAX(X,Y) ((X)>=(Y)?X:Y)
#define MIN(X,Y) ((X)<=(Y)?X:Y)
#define NUMELEMS(X) sizeof(X)/sizeof(X[0])
#define MATCHFLGS11(X,Y) (((X)&ITEM1SEC)==((Y)&ITEM1SEC))
#define MATCHFLGS12(X,Y) (((X)&ITEM1SEC)==(((Y)&ITEM2SEC)>>1))
#define SHORTSYMBOLSIZE sizeof(symbol)-(2*sizeof(void *))
#ifdef MURMUR3
#define ROT32(x, y) ((x << y) | (x >> (32 - y)))
#endif

/* STRUCTURES */
/* Structure for chaining names buffers for told sentences. */
/* See AddNametoChain() function for full format description of names buffer. */
typedef struct snameschain {
	struct snameschain *nxtnameschain;      /* Pointer to next nameschain structure, NULL if last. */
	struct slkedfitem *quantifier;          /* Pointer to quantifier linked list item corresponding to a variable item. */
	char *names;                            /* Pointer to names list buffer. */
	int32_t buffsize;                       /* Size of buffer. */
} nameschain;

/* Structure for formula common prefix. */
typedef struct scmprefix {
	struct scmprefix *next;             /* Pointer to next formula for formulas of the same kind (TEXT, ACTIVE, PASSIVE, */
	                                    /* UNPROC) in the KB. NULL if last formula of its kind. */
	struct scmprefix *prev;             /* Pointer to previous formula for formulas of the same kind (TEXT, ACTIVE, PASSIVE, */
	                                    /* UNPROC) in the KB. NULL if first formula of its kind. */
	                                    /* This is intended to free formula memory. */
	struct scmprefix *parent1;          /* First parent formula pointer. */
	struct scmprefix *parent2;          /* Second parent formula pointer, NULL if there is no second parent. */
	                                    /* For predicate definition formulas this pointer is used to chain */
	                                    /* the set of predicate definition formulas used in a single formula */
	                                    /* renaming inference. */
	struct scmprefix *prevdisp;         /* Pointer to previous formula in disposal queue. NULL if last formula */
	                                    /* or formula not in disposal queue. This queue interconnects clauses */
	                                    /* from other queues.*/
	union {
		char *text;                     /* Pointer to formula in plain text form. Used for text type formulas. */
		struct sbinprefix *bin;         /* Pointer to binary clause prefix structure. Used for compiled clauses. */
	} part2;
	uint64_t number;                    /* Number assigned to formula. */
	int32_t inference;                  /* Inference from where the formula comes from (see definitions above). */
	uint32_t flags;                     /* See flags above. This must be the last field of this structure. */
} cmprefix;

/* Structure for binary clause prefix. */
typedef struct sbinprefix {
	cmprefix *clause;                   /* Pointer to cmprefix structure of the formula itself. */
	cmprefix *nextweight[WEIGHTQUEUES]; /* Set of pointers to next formula in each weight queue. NULL if last formula. */
	cmprefix *prevweight[WEIGHTQUEUES]; /* Set of pointers to previous formula in each weight queue. NULL if last formula. */
	uint8_t *asserts;                   /* Pointer to buffer with clause assertions. */
	uint8_t *lockasserts;               /* Pointer to buffer with clause locking assertions. */
	uint64_t signature;                 /* Signature of last inference process. */
	uint64_t locksignature;             /* Signature for lock and unlock processes. */
	uint32_t literals;                  /* Number of literals in clause. */
	float symbolw[SYMBOLDIM2];          /* Set of symbol weights used to compute clause weight for the different weight queues. */
	                                    /* Index 0 is the pure symbol weight. */
	                                    /* Index 1 is the symbol weight with a factor applied if the symbol doesn't appear */
	                                    /* in the conjecture. */
	uint32_t dist2conj;                 /* Distance to conjecture. See global comments in ClauseWieght() function for details. */
	uint32_t clweight[WEIGHTQUEUES];    /* Clause weight for each weight queue. */
	int32_t depth;                      /* Max nesting depth of predicate/terms. */
	int32_t size;                       /* Size in bytes of formula including ending UNITEND byte. */
	int32_t asize;                      /* Size in bytes of assertions including ending UNITEND byte. */
	int32_t lasize;                     /* Size in bytes of locking assertions assertions including ending UNITEND byte. */
	int32_t maxvarnb;                   /* Maximum variable number, -1 if no variables. This is useful for */
	                                    /* memory allocations in ordering functions. */
	int32_t equflg;                     /* 0 if equality ordering data has not been set in equalities. */
	                                    /* 1 if equality ordering data has been set in SELECTED equalities */
	                                    /* by calls from SelectItems() function. */
	                                    /* 2 if equality ordering data has been set in the only equality */
	                                    /* of a clause used for backward demodulation. */
	                                    /* See comments in FreeEquData() function for indications on how */
	                                    /* equality ordering data is allocated and freed with the help of this flag. */
	uint8_t formula[1];                 /* Dummy for start of formula items. This must be the last field of this structure. */
} binprefix;

/* Structure for SAT clauses. */
typedef struct ssatclause {
	struct ssatclause *glblnext;        /* Pointer to next SAT clause of the same global queue, NULL if last. */
	struct ssatclause *glblprev;        /* Pointer to previous SAT clause of the same global queue, NULL if last. */
	struct ssatclause *next;            /* Pointer to next SAT clause of the SAT solver ordered clauses queue, NULL if last. */
	struct ssatclause *prev;            /* Pointer to previous SAT clause of the SAT solver ordered clauses queue, NULL if last. */
	struct scmprefix *parent;           /* Pointer to parent A-clause. */
	uint64_t number;                    /* Number assigned to the SAT clause. */
	int32_t inference;                  /* Inference from which the SAT clause was obtained. It can be SPLIT or CONTRADICTION. */
	                                    /* See definitions above. */
	int32_t literals;                   /* Number of literals in clause. */
	int32_t satlits;                    /* Number of satisfied literals in clause. */
	int32_t undeflits;                  /* Number of undefined literals in clause. */
	int32_t inproof;                    /* Set to 1 if clause is in the unsatisfiable core but is not the contradiction clause */
	                                    /* that caused the SAT refutation, set to 2 if clause is in the unsatisfiable core and */
	                                    /* it is the contradiction clause that caused the SAT refutation, zero otherwise. */
	uint8_t formula[1];                 /* Dummy for start of formula items. This must be the last field of this structure. */
} satclause;

/* Structure for statistic information. */
/* If new fields are added to this structure the following changes must be made */
/* to the involved functions: */
/* - Search for an existing field and add the appropriate equivalent code */
/*   for the new field. */
/* - Update totals calculation in PrintStatData() function if appropriate. */
/* - Add code for printing proof lines in PrintProof() function if appropriate. */
typedef struct sstats {
	uint64_t tautologies;               /* Number of tautologies detected. */
	uint64_t fwdsubsum;                 /* Number of forward subsumptions performed. */
	uint64_t bcksubsum;                 /* Number of backward subsumptions performed. */
	uint64_t factors;                   /* Number of factoring inferences. */
	uint64_t removedups;                /* Number of duplicate literal removing simplifications. */
	uint64_t fwdemodulations;           /* Number of forward demodulation inferences. */
	uint64_t bkdemodulations;           /* Number of backward demodulation inferences. */
	uint64_t fwsubsresltns;             /* Number of forward subsumtion resolution inferences. */
	uint64_t bksubsresltns;             /* Number of forward subsumtion resolution inferences. */
	uint64_t triveqres;                 /* Number of trivial equality resolution inferences. */
	uint64_t desteqres;                 /* Number of destructive equality resolution inferences. */
	uint64_t resolutions;               /* Number of resolution inferences. */
	uint64_t paramodulations;           /* Number of paramodulation inferences. */
	uint64_t equresolutions;            /* Number of equality resolution inferences. */
	uint64_t equfactorings;             /* Number of equality factoring inferences. */
	uint64_t splits;                    /* Number of split inferences. */
	uint64_t prennfxform;               /* Number of pre-nnf inferences. */
	uint64_t nnfxform;                  /* Number of nnf inferences. */
	uint64_t miniscoping;               /* Number of miniscoping inferences. */
	uint64_t skolemize;                 /* Number of skolemization inferences. */
	uint64_t predicatedef;              /* Number of predicate definitions for sub-formula renaming. */
	uint64_t formularenaming;           /* Number of formula renaming inferences. */
	uint64_t cnfconversion;             /* Number of CNF conversion inferences. */
	uint64_t tfsimplif;                 /* Number of $true and $false simplifications. */
	uint64_t sattautologies;            /* Number of tautologies detected in SAT clauses. */
	uint64_t satremovedups;             /* Number of duplicate literal removing simplifications in SAT clauses. */
	uint64_t satsubsum;                 /* Number of subsumptions performed for SAT clauses. */
	uint64_t nbactive;                  /* Number of active clauses. */
	uint64_t nbpassive;                 /* Number of passive clauses. */
	uint64_t nbunproc;                  /* Number of unprocessed clauses. */
	float memory;                       /* Memory used. */
	double elapsed_time;                /* Elapsed time in seconds. */
} stats;

/* Structure for program options. */
typedef struct soptions {
	int32_t fweight;                    /* Type of function weight (UNIFORM or ARITY). */
	int32_t termord;                    /* Terms ordering type (STANDARD or NONRECURSIVE). */
	int32_t litord;                     /* Literals ordering type (STANDARD, NONRECURSIVE or LEXICOGRAPHIC). */
	int32_t select;                     /* Type of literal selection (MAXIMAL, SINGLENEG, MULTINEG, SINGLEPOS or MULTIPOS). */
	int32_t demodulation;               /* Demodulation options (DEMODON, DEMODOFF, COMPLETE). */
	int32_t factoring;                  /* Factoring is enabled if this is not zero. */
	int32_t algorithm;                  /* Selected algorithm (OTTER or DISCOUNT). */
	int32_t mxwghtopt;                  /* MXWEIGHTxxx (see definitions above). */
	uint32_t maxweight;                 /* Maximum weight allowed. */
	int32_t factoridx;                  /* Index to weight factors to be used in clause weight calculations. */
	                                    /* See PNOTINCONJFACTOR, ENOTINCONJFACTOR, FNOTINCONJFACTOR and POSLITERALFACTOR definitions. */
	int32_t split;                      /* Splitting and interface between first order prover and */
	                                    /* SAT solver is enabled if this is not zero. */
	float timeout;                      /* Timeout setting. */
} options;

/* Structure for hash separate symbol chaining. */
typedef struct shashchain {
	struct shashchain *nextelem;        /* Pointer to next hashchain element or NULL if last. This pointer is */
	                                    /* used for freeing memory allocated to hashchain elements. */
	struct shashchain *nextsymbol;      /* Pointer to next hashchain element of the same hash value */
	                                    /* or NULL if last. */
	void *discr[DSCTREETYPES];          /* Set of pointers to discrimination trees. See EditDiscrTrees() */
	                                    /* global function comments for an index description. */
	#ifdef DEBUGTREE
	uint32_t numleafs[DSCTREETYPES];    /* Number of leafs of each discrimination tree. */
	#endif
	int32_t arity;                      /* Symbol arity. */
	int32_t type;                       /* Type flags, see flag definitions above. */
	float weight[SYMBOLDIM1];           /* Symbol weights for weight queues for evaluating the clause weight used to check */
	                                    /* clause retention. See SYMBOLWEIGHT definition above. */
	                                    /* Index 0 refers to pure symbol weight, depending on its kind (variable,predicate or */
	                                    /* function, see VARWEIGHT, PREDWEIGHT and FUNCWEIGHT definitions). */
	                                    /* Indices 1 and 2 refers to the weight as defined above multiplied by appropriate factors */
	                                    /* if the symbol doesn't appear in the conjecture for two different weight parameterization. */
	int32_t precedence;                 /* Symbol precedence value for ordering purposes. */
	#if PRECEDENCE != 0
	int32_t usecount;                   /* Number of times the symbol was used in the input formulas. */
	#endif
	uint32_t hash;                      /* Symbol 32 bits hash value corresponding to symbol name. */
} hashchain;

/* Structure for hash component symbol chaining. */
typedef struct shashchcmp {
	struct shashchcmp *nextelem;        /* Pointer to next hashchcmp element or NULL if last. This pointer is */
	                                    /* used for freeing memory allocated to hashchcmp elements. */
	struct shashchcmp *nextsymbol;      /* Pointer to next hashchcmp element of the same hash value */
	                                    /* or NULL if last. */
	struct shashchcmp *nextundef;       /* Pointer to next undefined hashchcmp element of the same component. */
	struct shashchcmp *prevundef;       /* Pointer to previous undefined hashchcmp element of the same component. */
	struct sasymbol *frstposassrt;      /* Pointer to first positive asymbol in assertions of first order prover. */
	struct sasymbol *frstnegassrt;      /* Pointer to first negative asymbol in assertions of first order prover. */
	struct sasymbol *firstlock;         /* Pointer to first asymbol in locks of first order prover. */
	struct sasymbol *frstsat;           /* Pointer to first asymbol in SAT solver. */
	struct ssatnode *satnode;           /* Pointer to satnode that assigned a value to this component. */
	uint8_t *formula;                   /* Pointer to clause binary formula of component clause associated */
	                                    /* with the P-predicate. */
	int32_t size;                       /* Size in bytes of formula including ending UNITEND byte. */
	int32_t literals;                   /* Number of literals in formula. */
	int32_t number;                     /* Number assigned to symbol name, the xxx in spl_xxx name. */
	uint64_t defnumber;                 /* Number assigned to the symbol definition in the proof. The component */
	                                    /* clause is assigned the number defnumber+1. */
	uint32_t hash;                      /* Symbol 32 bits hash value corresponding to clause component. */
	uint32_t flags;                     /* Flags for this component, see definitions above. */
	uint32_t npositive;                 /* Number of clauses in SAT solver with a positive literal for this component. */
	uint32_t nnegative;                 /* Number of clauses in SAT solver with a negative literal for this component. */
	uint32_t absdif;                    /* Absolute value of npositive field minus nnegative field. */
} hashchcmp;

/* Structure for coefficients of weight ordering polynomials. */
typedef struct scoeff {
	int32_t alpha0;                     /* First coefficient of weight polynomial. */
	int32_t *alpha;                     /* Pointer to second and subsequent coefficients of weight polynomial. */
	struct scoeff *eqterm1;             /* Pointer to coeff structure of first root term in an equality. */
	                                    /* NULL if current structure is not for an equality literal. */
	struct scoeff *eqterm2;             /* Pointer to coeff structure of second root term in an equality. */
	                                    /* NULL if current structure is not for an equality literal. */
} coeff;

/* Structure for coefficients of weight ordering polynomials, compact format. */
/* These are used for equality data saved in long term storage. */
typedef struct scoeffcpt {
	int32_t coefnb;                     /* Coefficient number. Number 0 corresponds to symbols coefficient. */
	                                    /* Number n corresponds to variable n-1 for n>1. A negative value */
	                                    /* indicates the end of the list. */
	int32_t coef;                       /* Coefficient value. Not used in the last list element. */
} coeffcpt;

/* Structure for coefficients of weight ordering polynomials, compact format and linked list. */
/* These are used for polynomials saved in short term storage. The linked list is ordered */
/* for increasing values of the coefnb field. */
typedef struct slkedcoef {
	int32_t coefnb;                     /* Coefficient number. Number 0 corresponds to symbols coefficient. */
	                                    /* Number n corresponds to variable n-1 for n>1. */
	int32_t coef;                       /* Coefficient value. */
	struct slkedcoef *next;             /* Pointer to next polynomial term, NULL if last term. */
} lkedcoef;

/* Structure for linear polynomial node of tree structure for evaluating */
/* ordering constraints of equality root terms. */
typedef struct sequpnode {
	coeffcpt *coefs;                    /* Pointer to linear coefficients of term in compact form. */
	int32_t result;                     /* Indicates the result (EQUAL, GREATER, LESSER or FAILURE) of an execution */
	                                    /* of CompareEquTerms(), that is, with no substitution applied. It also */
	                                    /* indicates the result of the call to CompareItemWeight2() done */
	                                    /* from CompareEquTerms() with  EQUALW, NOGREATER, NOLESSER */
	                                    /* or FAILUREW flags. */
	struct sequpnode *subterms[1];      /* Set of pointers to child (subterms) nodes. This must be the last item of this structure. */
} equpnode;

/* Structure for symbols (literals and functions) items data. */
typedef struct ssymbol {
	int32_t offset;                     /* Offset from beginning of formula field buffer to start of current item. */
	int32_t nextoff;                    /* Offset from beginning of formula field buffer to next item */
	                                    /* that is not a subterm of current item. */
	hashchain *symbol;                  /* Pointer to hashchain structure for the item symbol. */
	equpnode *leftnode;                 /* Pointer to left node equpnode structure. */
	equpnode *rightnode;                /* Pointer to right node equpnode structure. */
} symbol;

/* Structure for assertion symbols (P-predicates, component names). */
typedef struct sasymbol {
	hashchcmp *symbol;                  /* Pointer to hashchain structure for the item symbol. */
	struct sasymbol *next;              /* Pointer to next asymbol structure for the same component of the same class. */
	                                    /* (assertion, lock or SAT clause literal). */
	struct sasymbol *prev;              /* Pointer to previous asymbol structure for the same component of the same class. */
	union {
		binprefix *binp;                /* Pointer to A-clause binary formula to which the asymbol belongs if asymbol */
		                                /* belongs to assertions or lock assertions of an A-clause. */
		satclause *satcl;               /* Pointer to SAT clause binary formula to which the asymbol belongs if asymbol */
		                                /* belongs to a SAT clause. */
	} part2;
} asymbol;

/* Discrimination tree node structure. */
typedef struct sdstrnode {
	struct sdstrnode *parent;           /* Pointer to parent node, NULL if first level node. The discr[] field of the */
	                                    /* symbols hashchain structure points to a first level node or a leaf. */
	struct sdstrnode *nxtpeer;          /* Pointer to next peer node. */
	struct sdstrnode *prevpeer;         /* Pointer to previous peer node. */
	union {
		struct sdstrnode *childnode;    /* Pointer to first child node. */
		struct sdstrleaf *childleaf;    /* Pointer to first child leaf. */
	} child;
	hashchain *symbol;                  /* Pointer to node symbol for nodes that are not for variables. Not used */
	                                    /* for nodes that are variables. */
	uint8_t *item;                      /* This is used for temporary storage. DsTermIndexing() function uses it */
	                                    /* to store the pointer to the next clause term that is not a sub-term */
	                                    /* of the clause item matching the node if it has sub-terms or NULL otherwise. */
	                                    /* This is used later to update pndbrkts field. QueryDscTree() function uses it */
	                                    /* to store the pointer to clause item that matches the node or NULL if this is */
	                                    /* a sub-term of a node item that matches a variable. This is used later for */
	                                    /* backtracking. See DsTermIndexing() */
	                                    /* global comments. */
	uint16_t pndbrkts;                  /* Number of brackets to be closed just after the node symbol, see DsTermIndexing() */
	                                    /* global comments. */
	uint8_t nodetype;                   /* VARIABLE if node is for a variable, zero otherwise. */
	int8_t childtype;                   /* Type of child, it can be NODECHILD or LEAFCHILD (see definitions above). */
} dstrnode;

/* Discrimination tree node structure. */
typedef struct sdstrleaf {
	struct sdstrnode *parent;           /* Pointer to parent node. */
	struct sdstrleaf *peer;             /* Pointer to next peer (same parent) leaf. */
	#ifdef TREELEAFSFIFO
	struct sdstrleaf *lstpeer;          /* Pointer to last peer leaf in the chain. Only used by first leaf in the chain. */
	#endif
	uint8_t *item;                      /* Pointer to the tree indexed clause item. */
} dstrleaf;

/* Structure for indexing of positive equalities selected for paramodulating from them */
/* in active clauses that have maximal root terms that are variables. */
typedef struct sdendnode {
	int32_t size;                       /* Number of total symbol pointers allocated. */
	int32_t used;                       /* Number of used symbol pointers. */
	uint8_t **symbol;                   /* Pointer to set of pointers to item in the clause that matches the tree leave. */
} dsendnode;

/* Structure for SAT solver node. */
typedef struct ssatnode {
	struct ssatnode *nextnode;          /* Pointer to next SAT solver node, NULL if last.*/
	struct ssatnode *prevnode;          /* Pointer to previous SAT solver node, NULL if first.*/
	hashchcmp *symbol;                  /* SAT symbol that was assigned a value true or false in this node. */
	int32_t flags;                      /* Flags for this node: POSITIVE or NEGATIVE and possibly LASTOPTION. */
	satclause *forcedcl;                /* Pointer to unit clause that forced the value of the node symbol field, */
	                                    /* NULL if the node was not forced by a unit clause. */
	uint32_t number;                    /* Sequence number for this node. */
} satnode;

/* Structure for proof node. */
typedef struct sproofnode {
	struct sproofnode *nextnode;        /* Pointer to next proof node, NULL if last.*/
	struct sproofnode *prevnode;        /* Pointer to previous proof node, NULL if first.*/
	struct sproofnode *backptr;         /* This is used by AddAclause2Proof() for backtracking. */
	int32_t type;                       /* Type of node, see definitions above. */
	union {
		hashchcmp *symbol;              /* Pointer to component symbol. */
		satclause *satclause;           /* Pointer to SAT clause. */
		cmprefix *Aclause;              /* Pointer to A-clause. */
	} ptr;
} proofnode;

/* Structure for vertical features scan node. */
typedef struct svertnode {
	struct svertnode *nextnode;         /* Pointer to next node, NULL if last.*/
	struct svertnode *prevnode;         /* Pointer to previous node, NULL if first.*/
	uint8_t *item;                      /* Pointer to clause item. */
	uint8_t *nextitem;                  /* Next item jumping over subterms of current item. */
	int32_t depth;                      /* Node depth, starting with zero. */
} vertnode;

/* Structure for learning record. */
typedef struct slearnrecord {
	uint64_t header;                    /* Record header with flags indicating the relevant part of the strategy */
	                                    /* to which the learn record applies. The relevant strategy flags are DISCOUNT_P, */
	                                    /* MULTINEG_P, MAXIMAL_P, SINGLEPOS_P, MULTIPOS_P TRMORDNR_P, LITORDNR_P, LITORDLEX_P, */
	                                    /* ARITY_P and GOODCHOICE. The most significant byte of the header contains the */
	                                    /* problem type, a value between 0 and 7. */
	double vector[VECTORLENGTH];        /* Record features vector. See GetClauseVector() function for a description */
	                                    /* of each feature in vector. */
} learnrecord;

/* Structure for learning feature vector. */
typedef struct slearnvector {
	double vector[VECTORLENGTH];        /* Features vector. See GetClauseVector() function for a description */
	                                    /* of each feature in vector. */
} learnvector;

/* Structure for KB definition. */
typedef struct skbase {
	cmprefix *firsttxt;                 /* Pointer to first text formula, NULL if no text formulas added yet. */
	                                    /* Older formulas are placed first and this pointer points to the oldest */
	                                    /* text formula. */
	cmprefix *lasttxt;                  /* Pointer to last text formula, NULL if no text formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* text formula. */
	cmprefix *frstpassive;              /* Pointer to first passive formula, NULL if no passive formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* passive formula. */
	cmprefix *lstpassive;               /* Pointer to last passive formula, NULL if no passive formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* passive formula. */
	cmprefix *frstactive;               /* Pointer to first active formula, NULL if no active formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* active formula. */
	cmprefix *lstactive;                /* Pointer to last active formula, NULL if no active formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* active formula. */
	cmprefix *frstunproc;               /* Pointer to first unprocessed formula, NULL if no unprocessed formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* unprocessed formula. */
	cmprefix *lstunproc;                /* Pointer to last unprocessed formula, NULL if no unprocessed formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* unprocessed formula. */
	cmprefix *frstlocked;               /* Pointer to first locked formula, NULL if no locked formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* locked formula. */
	cmprefix *lstlocked;                /* Pointer to last locked formula, NULL if no locked formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* locked formula. */
	cmprefix *frstglweight[WEIGHTQUEUES];  /* Pointers to first formula in weight queue for each weight queue, */
	                                    /* each pointer is NULL if no formulas of that weight added yet. */
	                                    /* Only passive clauses are linked to weight queue. */
	cmprefix *frstweight[WEIGHTQUEUES][MAXWEIGHT+1];  /* Pointers to first formula in weight queue for each weight queue and for */
	                                    /* each weight, each pointer is NULL if no formulas of that weight added yet. */
	                                    /* Only passive clauses are linked to weight queue. */
	cmprefix *lstweight[WEIGHTQUEUES][MAXWEIGHT+1];  /* Pointers to last formula in weight queue for each weight queue and for */
	                                    /* each weight, each pointer is NULL if no formulas of that weight added yet. */
	                                    /* Only passive clauses are linked to weight queue. */
	cmprefix *lastdisp;                 /* Pointer to last formula in disposal queue, NULL if no formulas */
	                                    /* added yet. This queue interconnects clauses from other queues. */
	proofnode *frstprfnode;             /* Pointer to first proof node. */
	satclause *firstsatq;               /* Pointer to first SAT clause in the SAT queue. */
	satclause *lastsatq;                /* Pointer to last SAT clause in the SAT queue. */
	satclause *frstsatglbl;             /* Pointer to first SAT clause in the SAT solver global queue. */
	satclause *lastsatglbl;             /* Pointer to last SAT clause in the SAT solver global queue. */
	satclause *frstunsats;              /* Pointer to first SAT clause in the SAT solver ordered clause queue. */
	satclause *lstunsats;               /* Pointer to last SAT clause in the SAT solver ordered clause queue. */
	satclause *frstunsats2;             /* Pointer to start of the part of SAT solver ordered clause queue with clauses */
	                                    /* that are unsatisfied and have two or more undefined literals. */
	hashchcmp *frsthshundf;             /* Pointer to first undefined hashchcmp element for this KB in the hashchcmp ordered queue. */
	hashchcmp *frsthunbp;               /* Pointer to first undefined hashchcmp element with clauses with negative and */
	                                    /* positive literals (both polarities) for this KB in the hashchcmp ordered queue. */
	hashchcmp *lasthshundf;             /* Pointer to last undefined hashchcmp element for this KB in the hashchcmp ordered queue. */
	satnode *satrootnode;               /* SAT solver root node pointer. */
	double *lrnvector;                  /* Pointer to learning vector or NULL if no learning vector is available. */
	dsendnode vartrmidx;                /* Indexes of positive equalities selected for paramodulating from them in active clauses */
	                                    /* that have maximal root terms that are variables. */
	uint64_t signature;                 /* Last signature used for inference process. */
	uint64_t locksignature;             /* Last signature used for lock and unlock processes. */
	int32_t flags;                      /* KB flags, see definitions above. */
	uint64_t nformulas;                 /* Number of formulas in KB so far. */
	options opts;                       /* Options for KB. */
	int32_t selectcnt;                  /* Number of extractions done so far in the current round-robin clause extraction cycle. */
	int32_t lrscnt;                     /* Count of clauses removed from the passive queue in the current LRS cycle. */
	uint32_t splsymbols;                /* Number of P-predicates for P-predicate name assignment. */
	int32_t status;                     /* KB status (see flags above). */
	struct timeval time;                /* Start time of current saturation algorithm. */
	struct timeval endtime;             /* End time of current saturation algorithm. */
	stats prstats;                      /* KB statistics related to this KB. This depends on the context (process, main or working KB). */
} kbase;

/* Structure for substitutions. See Unify() function for substitution format. */
typedef struct ssubst {
	uint8_t *buffer;                    /* Pointer to buffer with substitution. May be NULL. */
	                                    /* See Unify() function for format description. */
	uint32_t buffsize;                  /* Total buffer size in bytes. */
	uint32_t substsize;                 /* Size in bytes of substitutions including the UNITEND byte. */
} subst;

/* Structure to store subsuming related data for each literal in a subsuming or subsumed formula. */
typedef struct ssubsumdata {
	int32_t lits1;                      /* Number of literals in candidate subsumed formula. */
	int32_t lits2;                      /* Number of literals in candidate subsuming clause. */
	uint8_t ***pinst;                   /* Address of set of pointers, one pointer for each literal in the candidate */
	                                    /* subsuming clause. Each pointer is the address of the set of pointers to */
	                                    /* instance literals in candidate subsumed formula for each candidate subsuming */
	                                    /* literal. */
	int32_t **instnb;                   /* Address of set of pointers, one pointer for each literal in the candidate */
                                        /* subsuming clause. Each pointer is the address of the set of instance literal */
	                                    /* numbers in candidate subsumed formula for each candidate subsuming literal. */
	int32_t *nbofinst;                  /* Address of set containing the number of instance literals for each literal */
	                                    /* in the subsuming clause. */
	int32_t *usedinst;                  /* Address of set containing one element for each literals in candidate subsumed */
	                                    /* clause that is set to 1 if the literal has been already assigned to a literal */
	                                    /* in the candidate subsuming clause or 0 otherwise. */
} subsumdata;

/* Structure for linked list formula format. */
typedef struct slkedfitem {
	struct slkedfitem *next;            /* Pointer to next linked list item. */
	struct slkedfitem *prev;            /* Pointer to previous linked list item. */
	union {
		struct slkedfitem *mtchbrckt;   /* Pointer to matching bracket item. */
		nameschain *vardata;            /* Pointer to nameschain structure of a quantifier or variable item. */
		hashchain *symbol;              /* Pointer to hashchain structure of a predicate or function item. */
	} ptr;
	union {
		hashchain *sksymbol;            /* Pointer to hashchain skolem symbol in existential quantifier or variable items. */
		int16_t varnumber;              /* Number assigned to standard name of variable in a quantifier or variable item. */
		struct {
			int32_t posclauses;         /* Lower bound of number of clauses generated for a block for the */
			                            /* open bracket or literal starting the block. */
			int32_t negclauses;         /* Lower bound of number of clauses generated for the negation of a block*/
			                            /* for the open bracket or literal starting the block. */
		} clauses;
	} data;
	struct slkedfitem *dupq;            /* Auxiliary pointer to duplicate quantifier, used when duplicating blocks. */
	int32_t flags;                      /* Item flags. */
} lkedfitem;

/* Structure for multi-processing control. */
typedef struct smpctl {
	clock_t start;                      /* CPU ticks at start of saturation process. */
	clock_t end;                        /* CPU ticks at end of saturation process. */
	int32_t status;                     /* Process status. See definitions above. */
	int32_t prfprntinproc;              /* Set to non zero if proof printing is in process, zero otherwise. */
	int32_t solvekb;                    /* Process number for problems solved in the saturation stage (see procnb variable), */
	                                    /* active_cores for problems solved in the pre-process, -1 if problem was not solved. */
	sem_t procctl;                      /* Semaphore for process status updates. */
	sem_t statistics;                   /* Semaphore for statistics update. */
	sem_t strategy;                     /* Semaphore for process strategy selection. */
	sem_t sync;                         /* Semaphore for process synchronization. */
} mpctl;

/*----------------------------------------------------------------
 *
 *  Global variables shared with other modules that are initialized here.
 *  WARNING: Don't use "static" as then a different variable will be created in
 *  each module with the same name.
 *
 *--------------------------------------------------------------*/
#ifndef VARTYPE
#define VARTYPE extern
extern double tot_cpu_time; /* Total CPU time used in seconds. */
extern int32_t tot_pbms; /* Total number of problems examined. */
extern int32_t solved_pbms; /* Total number of problems solved. */
extern int32_t queuerts[WEIGHTQUEUES+1]; /* Ratio numbers for weight queues and age queue with learning disabled. */
extern int32_t queuerts2[WEIGHTQUEUES+1]; /* Ratio numbers for weight queues and age queue with learning queue. */
extern int32_t *slctordr; /* Pointer to clause selection order queue set with learning disabled. The I-th clause selection */
                          /* is picking the clause from the slctordr[I] weight queue if slctordr[I]<WEIGHTQUEUES, */
                          /* otherwise the clause is picked from the age queue. */
extern int32_t *slctordr2; /* Pointer to clause selection order queue set with learning enabled. */
extern int32_t wqdeplord[WEIGHTQUEUES]; /* Weight queues depletion order with learning disabled. If wqdelord[I]=J then */
                                        /* the weight queue with J index is depleted I-th. */
extern int32_t wqdeplord2[WEIGHTQUEUES]; /* Weight queues depletion order with learning enabled. */
extern hashchain *kb_hash; /* First chained hashchain element. */
extern hashchcmp *kb_hashcmp; /* First chained hashchcmp element. */
extern int32_t syntaxmode; /* Syntax type: TPTP or DRODI. */
extern int32_t proofmode; /* Proof printing type: TPTP or DRODI. */
extern int32_t prtkbdata; /* PRINTKBDATA type: PRTWINNER, PRTALL or PRTOFF. */
extern int32_t verbose; /* Set to 1 if VERBOSE option is ON, zero otherwise. */
extern char *importfile; /* Pointer to IMPORT file name or NULL. */
extern char *learnto; /* Pointer to LEARNTO file name or NULL. */
extern char *learnfrom; /* Pointer to LEARNFROM file name or NULL. */
extern learnrecord *lrnrecords; /* Learning records with headers and vectors from learning file. */
extern int32_t learnnew; /* If zero then learn-to output will be appended to learn-to file. Otherwise a new file */
                         /* will be created. */
extern int32_t loadredundant; /* If not zero then add formulas with lemma, theorem and corollary roles, */
                       /* otherwise these formulas are not loaded. */
extern int32_t timeslices; /* Number of time slices. */
extern double ppthrshld; /* Threshold for the preprocessing clause relevance filtering. */
                  /* See GlobalPreprocess() global comments for details. */
extern int32_t prproflags; /* Preprocessing clause filtering option flags. This can be CLFILTERINGOFF, CLFILTERINGON, */
                    /* CLFILTERINGNONEQU and/or CLFILTERINGRESIGN (see definitions above). */
extern int16_t pbmtypestrats[2][8][TUNEDSTRATEGIES]; /* Strategies to use for each type of problem. */
                                     /* See InitStrategies() for details. */
extern uint32_t usrparam; /* User parameter options. */
extern uint32_t usrparam_mask; /* User parameter options mask. */
extern volatile sig_atomic_t signalflag; /* Not zero if a program error caught signal process is in progress. */
extern uint64_t allcdmemory; /* Currently allocated memory. */
extern pid_t *pids; /* Pointer to set of child process pids. */
extern uint64_t memorylimit; /* User specified memory limit. */
extern void *mmapptr; /* Pointer to mmap shared memory. */
extern char *logpath; /* Pointer to path for exceptions logging. */
#if defined(DEBUGCODE) && defined(VERBOSE)
extern int32_t verbfrom; /* Initial clause number for verbose tracking. */
extern int32_t verbto; /* Final clause number for verbose tracking. */
#endif
#else
double tot_cpu_time=0.0;
int32_t tot_pbms=0;
int32_t solved_pbms=0;
int32_t queuerts[WEIGHTQUEUES+1]={0,1,2,2,0,1};
int32_t queuerts2[WEIGHTQUEUES+1]={6,1,2,2,0,1};
int32_t *slctordr=NULL;
int32_t *slctordr2=NULL;
int32_t wqdeplord[WEIGHTQUEUES]={0,1,2,3,4};
int32_t wqdeplord2[WEIGHTQUEUES]={0,1,2,3,4};
hashchain *kb_hash=NULL;
hashchcmp *kb_hashcmp=NULL;
int32_t loadredundant=1;
int32_t syntaxmode=TPTP;
int32_t proofmode=TPTP;
int32_t prtkbdata=0;
int32_t verbose=0;
char *importfile=NULL;
char *learnto=NULL;
int32_t learnnew=0;
char *learnfrom=NULL;
learnrecord *lrnrecords=NULL;
int32_t timeslices=3;
double ppthrshld=0.1;
int32_t prproflags=CLFILTERINGNONEQU;
int16_t pbmtypestrats[2][8][TUNEDSTRATEGIES]=
	{{{95,611,1318,691,1170,1287,146,673,454,509,1379,135,2101,1891,523,1884,477,2080,1774,805,1331,1258,621,930,1297,1166,1383,1902,1239,539,
			195,1741,1471,1253,898,417,1725,1278,815,601,2099,681,210,577,206,1263,2027,519,529,255,492,422,533,1128,407,1720,2134,145,85,1321,
			538,1229,1900,1351,866,365,1907,213,1249,237,1681,1243,1008,663,560,627,1804,335,481,528,631,2025,567,446,553,1164,569,649,315,1332,
			912,643,2055,1127,958,548,1045,1201,588,1401,200,687,216,1393,1662,1299,1369,5,2041,1951,653,2050,887,1854,1268,35,1211,598,511,2020,
			697,386,1273,1542,1289,1259,1095,602,902,1363,1197,58,1290,1411,1779,25,491,2069,571,1160,1417,1116,543,934,579,659,1025,1746,792,
			1373,1187,1188,2085,1341,2074,745,1381,1704,30,1407,1176,1281,1896,214,965,1429,957,1686,535,1914,1785,925,1398,325,671,1734,450,1227,
			1231,926,245,62,1547,719,467,875,482,1375,2039,510,678,1055,661,2094,920,507,1221,936,975,1831},
	{325,1254,982,1415,335,584,512,565,1055,1244,802,1285,265,594,322,1225,1045,1707,572,1405,985,1314,992,705,275,524,1606,575,995,1304,1282,1435,
			85,534,752,595,35,2067,2086,515,805,2062,502,505,25,2092,1601,1315,755,2097,1571,1245,745,1702,812,585,815,1732,272,1295,95,1737,1961,
			1305,295,1726,1235,1005,32,715,1015,2051,1425,285,1691,1375,1075,1576,1365,345,1841,645,355,1456,655,1065,332,685,775,1222,625,105,1232,
			695,115,1481,635,765,1816,1355,55,2056,1345,45,562,525,835,1966,535,825,742,1255,1043,1185,2081,754,1265,803,945,82,1452,675,93,225,1292,
			94,1335,333,465,1042,1937,1205,323,475,1486,1812,485,813,955,1846,1847,545,83,1195,1936,564,665,1053,175,92,574,1215,753,895,1696,24,
			1385,263,885,1721,1932,555,993,405,1052,34,495,983,165,1931,344,605,743,1125,1451,84,615,273,415,1811,2057,1275,33,1135,262,514,1395,23,
			235,22,1482,1325,823,395,994,965,1073,215,1697,65},
	{31,1212,655,987,854,225,813,6,411,39,1475,1369,339,198,581,1059,1684,57,749,310,887,815,2074,593,505,16,221,91,1922,163,1055,12,357,99,254,103,
			515,1535,1065,753,2109,175,1285,1022,1019,97,70,647,83,674,835,1053,1148,1313,501,1473,957,579,184,955,561,2133,1191,759,62,779,1045,1142,
			413,273,1623,1199,27,1507,1257,95,1569,1127,741,1742,1075,325,1508,1007,35,188,827,507,1024,401,511,1984,539,1047,966,119,263,978,409,803,
			1276,237,277,2136,1319,271,1386,1071,519,1470,291,1225,432,775,1041,246,643,517,738,525,23,366,465,999,0,107,991,732,475,1281,728,469,981,
			556,1133,807,1444,653,81,1715,49,93,964,587,1287,1020,767,569,908,645,575,422,1077,331,540,461,809,318,353,573,850,1125,1221,1447,467,805,
			1477,589,337,1863,949,279,780,1303,571,1925,53,743,1890,407,565,434,659,811,900,1435,33,1140,521,1057,1026,117,995,1567,1129,333,542,951,
			989,1749},
	{569,588,810,975,1751,1347,1237,772,504,1033,1356,935,807,1610,754,311,1104,865,819,586,982,971,686,871,1283,2090,506,1039,1540,201,31,288,812,
			547,690,1161,1231,1460,998,1031,938,449,1047,1240,2083,541,2115,1417,981,830,574,729,926,1359,809,1076,808,797,2110,443,25,44,1286,317,
			1513,205,747,1708,2056,1271,1418,877,97,1823,512,305,868,453,989,1304,1967,1205,936,1413,565,1318,1813,301,142,1403,327,1948,1048,77,
			386,451,987,1250,814,737,1905,689,1059,1977,1236,789,1518,1111,815,1709,1932,9,202,459,89,530,750,1027,2021,921,1229,356,1282,15,1176,
			1175,1225,344,1968,497,1994,1345,563,1300,1488,1273,1636,141,1299,1705,502,19,1514,145,23,1703,986,731,1118,925,755,1258,1960,1037,1549,
			621,1291,770,20,1263,1876,639,1051,292,1937,551,1416,203,1239,1704,758,489,156,687,1221,1942,1292,315,148,937,277,1496,1842,255,2149,
			1343,995,778,1935,75,218,391,745,532,262,969,1661,1119,759,584},
	{1011,1389,815,617,623,963,1634,1114,933,424,1421,1399,1045,323,497,1999,1353,384,373,1657,461,811,549,215,1207,1223,1083,808,1992,333,1597,861,
			1301,1279,1175,803,1337,1974,633,744,723,434,441,1311,669,325,1163,1327,1939,1343,1452,973,1957,921,711,679,445,923,487,1008,503,274,733,
			1154,771,1431,559,805,93,1217,1994,1233,1932,13,2012,331,701,1269,1055,1173,607,1914,513,1637,243,304,751,591,1407,263,1105,992,1964,2133,
			1313,3,453,831,1952,581,687,863,265,1162,1459,1713,1423,133,813,1652,991,381,619,935,853,527,1949,1576,1295,2138,1433,213,1034,1161,609,
			85,843,597,838,22,565,1386,583,203,2017,1241,1287,873,995,1876,218,1396,703,1093,1043,1024,391,1251,1297,1103,35,32,1934,676,593,363,83,
			1592,741,1371,1417,743,865,982,768,2073,713,253,443,1144,271,521,697,983,985,1816,928,1276,1303,123,1053,314,81,1361,577,23,25,1112,1554,
			2078,1041,641,567,273,875,1456,1944,1266,1191,531},
	{1011,1277,1902,591,557,1174,111,1397,814,101,1387,1054,1071,677,1542,1251,547,934,1311,1207,1662,1301,667,1962,1061,1267,324,761,1269,84,821,
			1329,1547,351,617,1487,291,607,1602,341,497,1907,831,1389,804,51,499,334,581,1317,204,531,559,1842,41,537,924,1001,549,1607,1241,1327,
			1044,281,1399,214,521,619,454,771,609,1847,1191,1337,94,811,1121,1219,1967,801,941,489,1482,1291,1431,487,1667,741,1181,1347,2027,261,
			1421,1209,1164,81,161,1279,2022,561,891,669,444,1281,881,1427,1932,91,951,1217,994,991,701,689,744,1221,461,1407,754,571,221,699,264,1231,
			1131,1287,1457,1041,231,599,1452,511,1371,687,1577,981,641,1289,1572,751,411,1339,984,331,401,1247,34,321,651,679,274,1051,171,579,1937,
			271,711,1429,1812,501,1361,707,1817,21,471,1439,24,31,451,1409,874,1171,507,154,201,509,1637,1411,1237,864,1101,1377,1104,1351,569,384,
			681,717,1114,871,567,1632,441,1419,1997,931,637,394,1111,517},
	{1279,771,2152,307,1086,1951,1326,2151,1309,891,1797,347,1126,1591,636,1222,569,741,714,1067,846,1062,1218,1282,1359,11,1274,807,136,1042,1206,
			1741,489,751,2072,1077,1016,1656,548,632,1259,851,1394,827,896,1052,1338,2126,699,861,1712,317,46,1142,1328,2076,659,171,1727,837,26,472,
			496,1332,1269,841,505,927,176,1152,1236,1721,1329,21,525,1157,866,1192,1288,1202,1399,41,635,447,756,1616,1406,1342,1229,731,1355,1177,
			856,1032,1426,1701,589,871,1235,197,766,422,578,1686,1299,161,495,907,386,1611,678,1761,1319,151,1225,947,1116,462,1246,702,689,121,615,
			1187,246,1666,516,672,629,721,1205,1057,886,1651,1378,592,639,1,625,67,1503,1676,686,2156,549,881,1335,1027,56,1601,546,1262,1369,141,535,
			1047,1443,1966,518,2086,1349,991,515,207,146,1172,646,1242,559,51,1375,237,1588,2011,498,1786,1289,981,1245,1037,976,2031,1248,2081,709,
			31,1255,797,166,1596,1336,1696,719,131,485,1197,876,1022,608,2066},
	{881,1947,1752,1285,353,2061,2071,1736,1241,1527,1214,1425,203,1731,2076,341,1532,1787,665,433,2116,1756,861,854,654,565,963,2081,2146,611,1024,
			1234,1405,463,1711,532,681,864,2087,1415,453,2041,851,374,2112,655,733,1791,401,954,1284,695,313,1696,311,1892,1204,1305,343,921,54,1697,
			1395,953,731,1482,704,1275,323,811,1094,1364,1063,685,191,1977,614,923,705,101,1044,684,303,1295,481,1867,1777,93,1385,801,1952,514,823,
			585,1431,1837,1254,893,1205,981,1557,2057,933,605,41,254,1414,793,505,1121,2037,1374,783,535,901,264,1742,943,1215,691,2017,1424,1023,555,
			1181,1967,1732,63,635,1261,214,1702,1193,1435,571,894,1404,83,1265,1071,814,1687,883,495,951,754,1384,753,1345,491,114,1772,833,525,561,
			1054,604,223,1255,701,234,544,1153,1225,661,194,2077,723,575,91,64,2042,183,1315,201,1832,554,73,1355,1081,94,1782,773,485,1051,1652,1792,
			393,545,911,74,1294,853,1365,1021,1857,1762,293,645,931}},
	{{2193,4034,2251,95,3599,3470,3268,509,371,3507,3203,3473,2961,2766,3175,2741,3973,3910,2471,2451,2733,2247,91,3548,3005,3277,2851,3501,1379,4165,
			3025,3979,2345,2697,3023,3361,3348,3214,2666,2182,3454,2243,4039,2241,755,3258,3569,3410,361,3537,3165,3463,2751,81,3204,2376,2461,2441,
			2723,2725,2841,2776,2257,3033,3327,3015,2388,3571,4160,2936,2968,2295,2747,3035,3371,3358,2676,3444,2253,3278,2181,25,3527,3400,611,1239,
			3673,3403,2348,2691,811,2995,2693,3915,2401,2735,2511,3431,2267,4008,3083,3511,3337,2836,3075,2560,2781,4009,3488,2285,2757,3858,3145,3533,
			3288,2264,4212,2183,3328,2191,85,3587,3413,2901,3460,2358,1289,3,2786,3153,3421,2681,2985,3940,2501,2277,2976,3561,1249,2791,3267,1051,
			3065,3248,2630,2100,2826,3974,2411,3498,2355,2687,3155,3523,3853,3093,2274,4217,2311,3028,805,2261,3579,3443,4154,4220,2963,365,3597,2641,
			571,3223,2421,3522,2288,3978,3235,131,529,4180,3581,2918,3800,2354,3558,3357,2667,2187,3307,3013,3486,3298,2531,2314,2227,2301,3038,815,
			2271,3519},
	{3175,2911,3200,2485,3441,594,3565,3227,2491,3165,2901,3140,2505,3451,1254,3505,3167,2431,3215,2961,2445,3210,1244,3595,3391,3177,2481,3145,2971,
			4120,2495,3381,584,3525,3217,2421,3205,2931,3150,2425,3575,3474,3473,3147,3461,3155,2921,4125,2455,3535,3404,3403,3157,3471,3235,2435,
			2991,4090,3585,3411,3414,3237,3225,2981,4095,2515,3515,3401,4257,3207,325,2667,3230,534,2373,2625,2785,3453,2963,2745,335,2677,3220,1707,
			2363,2615,3463,2875,2973,2755,1055,2727,3170,1737,2313,2605,2845,3413,2913,2735,265,2737,4135,1304,2393,2635,3383,2805,2903,2725,1045,
			2747,3160,524,2383,2555,2815,3393,2993,2695,985,2697,4100,1314,2323,2545,2865,3443,2983,2685,275,2757,4130,2067,2575,2337,3153,2621,2511,
			2665,995,2687,4105,2092,2565,2367,3143,2611,2501,2675,2195,3213,2441,2303,2097,2601,3765,2307,3095,2185,3203,2451,2333,2062,4250,2561,
			2377,3025,2975,2429,3735,2631,2327,1702,3085,3163,2734,2965,2439,3760,2541,2387,1732,2241,3035,2905,2499,3730,2551,2397,3173,3464,2915,
			2489,2180,2571,2317,3460,4222,85},
	{2305,31,411,2435,2315,987,163,2425,2375,813,225,2365,39,1369,2335,339,1212,2385,1059,2109,2325,505,6,2395,749,2133,3529,815,2020,2751,2294,3593,
			2284,3349,1535,1348,3359,198,2112,3587,91,70,3289,1860,682,3261,1922,2817,2873,1055,1511,3049,515,1904,2339,753,2116,2389,99,452,2569,83,
			4278,3581,579,2114,3119,1053,1172,3537,1285,854,2577,97,1104,3401,561,446,2877,27,440,2399,759,2023,2503,273,1475,3271,1656,3589,2639,501,
			692,2803,95,2028,3531,325,1358,2627,1045,1178,2329,35,448,3117,1047,2143,3293,511,1513,3299,254,2145,2579,741,2117,3411,519,632,3599,277,
			1106,3343,271,444,3109,507,218,3283,263,206,2869,803,1541,2629,1225,1507,2879,1281,1148,2867,23,1663,3265,1742,3059,3107,991,1754,2819,
			517,1418,3057,1041,698,3353,999,1544,3591,93,1669,3321,552,3521,3539,981,1902,2813,807,1623,2807,81,16,3461,1287,1414,2863,1239,1447,2637,
			809,1685,3275,548,2937,3325,378,2337,2809,569,1907,2567,331,310,3047,265,2074,3113,743},
	{3159,4250,3287,3398,3799,2425,569,2195,2989,2339,2913,2697,2521,2456,2211,2301,146,3925,2673,2453,2604,3219,4225,3297,3458,3794,2485,2245,759,
			2851,2929,2963,2687,2531,2446,3353,2271,2311,440,3950,2733,2433,2614,3209,4255,3347,2480,3829,2495,2185,31,2841,2939,2757,2903,2591,3323,
			2371,2201,144,2506,3920,2483,2723,2554,3149,4220,3357,3448,3824,2435,2255,1059,2791,2999,2973,2747,2581,2516,3333,2261,2361,872,3955,2663,
			2493,2544,3239,4210,3337,3388,3839,2505,2781,747,2215,2993,2737,2601,3857,218,2799,2979,2191,2753,3945,2423,3460,3792,2321,3169,4215,3277,
			3456,3809,2455,2399,27,2275,2933,2667,2611,3852,874,2919,2241,2743,2801,3940,2513,3400,2634,2381,3229,4245,3267,3396,3834,2445,2389,571,
			2205,2983,2677,2551,3887,452,2969,2181,2693,2871,3915,2503,3892,2574,2391,3179,4240,3327,3386,3804,2515,2329,87,2265,2923,2727,2571,3882,
			200,2909,2251,2683,2811,3910,2443,3410,3797,2331,3451,3743,2847,2684,2437,2600,2665,2661,85,2861,3141,2631,512,2850,2431,2674,2624,2816,
			2691,3441,3748},
	{2301,1389,2995,3053,3767,2349,2716,2491,3430,4071,3765,3354,3025,3151,3180,2275,3869,2722,2212,3094,2311,1399,2985,3043,2484,2359,2646,2431,3112,
			3420,2480,3264,3095,3141,3120,2265,2698,2672,2202,2291,2361,669,2615,3321,3647,2656,2369,2925,4230,3102,3023,3760,3344,2411,3035,3190,
			2437,2205,3864,3221,2682,2351,2371,549,2605,3331,3642,2706,2379,2935,4235,4076,3033,2430,3334,2401,3085,3130,2427,2688,2195,2732,3171,
			2341,3580,2687,2494,2443,3211,2281,2389,3175,3579,3103,4152,3735,4121,3210,2461,2345,3062,2321,619,2453,3762,3201,3068,3165,3843,3360,
			2339,3113,4187,2420,4126,3150,2471,2841,2355,3012,3089,2331,679,2971,2254,1163,2309,3581,3155,3569,2673,4260,3083,3324,3730,3220,2811,
			2481,3002,3099,3231,2295,3590,2697,2905,2399,3737,2421,3092,623,3509,3093,2391,4167,2490,3202,2851,3140,2185,2662,3069,3161,2285,3530,
			2727,2915,2398,2424,2501,2621,1353,3888,3212,2381,4182,2801,3170,2215,2752,3079,2300,3520,2737,2423,2244,2319,2451,513,2261,3519,3294,
			3145,2791,3200,3072,933,2310,3039,3470,3386},
	{2241,2671,2723,2706,2191,2731,2733,4298,2251,2661,3556,2693,2181,2721,3909,2683,2271,2741,2716,2201,2751,3938,2211,2681,3864,2261,2691,3953,2371,
			2161,3171,2301,2171,3411,2361,2221,2751,2311,2231,2271,2381,2707,2451,2391,2647,2511,2321,2717,2931,2331,2657,3401,2743,3461,2737,4062,
			2753,2211,2747,3334,2851,3221,2757,2265,2781,2441,2727,2205,2791,3231,2697,2245,2841,2741,2667,2974,2871,2921,2677,2275,2801,2261,2687,
			3214,2861,3471,2255,2673,2811,2201,2215,2663,2863,2691,2195,2653,2873,2501,2185,2713,2813,2981,2395,2643,2165,2803,3161,2385,2703,2175,
			2783,2991,2335,2235,2793,2681,2325,2225,2853,3141,2375,2843,3201,2315,2771,3581,4122,2971,2831,3281,2484,2961,2761,3291,2244,2421,2821,
			2321,3762,2481,2341,3101,3094,3451,2351,3521,3702,3381,2281,3041,3822,2721,2291,3341,3647,3151,2833,2801,2365,3391,2823,3351,2305,3441,
			2763,3531,3707,2251,2773,2621,2964,2431,2865,2561,3767,2671,2701,2875,2861,3204,2731,2711,2855,2811,2651,4007,3211,2845,2631,2641,2494,
			2241,3591,2815,2254,2901,3051,2805,4002,2911},
	{3568,3557,2760,3359,2627,2611,3583,4091,3409,2309,2677,2421,2472,3548,3567,2840,3249,2587,2591,3553,4106,3369,2299,2647,2491,3771,3598,3597,2850,
			3339,2637,3142,2621,3533,2688,2339,2667,2511,3776,3578,3587,2770,3259,2607,4081,2601,3483,3389,2319,2657,2431,3751,3558,3577,2810,3319,
			2547,4086,3644,2581,2701,3399,2697,2177,3761,4279,771,3547,3309,2202,2678,2527,3639,2551,1721,2247,2501,3756,4264,741,3487,3329,3503,2567,
			2192,2698,2521,2399,2481,2512,4284,11,3537,3349,3493,2557,2182,2658,2531,2289,2441,2482,4274,891,3507,3289,3513,2537,2162,2648,2561,2389,
			2451,2462,4269,861,3497,3299,3523,2212,2227,2631,2719,2727,2577,3529,2461,3766,4289,21,3527,3269,1938,2680,2237,2571,3132,3467,3589,2597,
			2151,2411,2492,3538,41,3517,2606,3279,2203,2257,2541,2663,3419,1731,2617,2471,2216,3518,721,2797,2629,2650,1086,3579,2197,2173,1776,3281,
			2401,2298,3508,731,2817,2589,2690,1126,3479,2207,2193,672,3241,2751,2288,3498,751,2787,2599,2670,846,3569,2187,4101,3311,1222,2711,2338,
			3528,871,2777,2619},
	{2720,2985,2557,2670,2750,2995,2627,2690,2955,2617,2730,2975,2607,2280,2640,2945,2637,2320,2660,2965,2547,2350,3880,2895,2597,2370,2680,2915,2527,
			2300,2700,2935,2537,2340,2710,2925,2587,2620,3845,2595,2567,2260,3870,2905,2577,2580,3875,2615,3457,3415,2560,2650,2605,3447,3435,2180,
			3890,2885,3467,3445,2570,3855,2585,3377,3585,2190,3865,2635,3367,3119,3895,2625,3397,3405,3885,2385,3417,3595,3840,2375,3407,2755,3850,
			2365,3437,2695,2740,2575,3387,3069,3860,2325,3169,3535,3570,2525,2633,2725,3480,2535,2593,2735,3540,2545,2613,3555,3500,2565,2553,2745,
			3590,2555,2583,3515,3550,2395,2543,2645,3560,2295,2523,2715,3490,2345,2563,2705,3510,2285,2603,2665,3520,2315,1423,3495,3580,2305,583,
			3575,3530,2355,1293,3505,4315,2335,573,3525,4290,3105,593,3485,4270,3065,1403,2685,4295,3095,1393,2675,4300,3085,1263,2655,4275,3115,693,
			3009,4305,2244,1273,3019,4265,2234,713,3039,4285,2414,543,3455,4260,2404,673,3565,4280,2454,703,3079,4310,2434,663,3545,2492,2357,3869,
			2444,3375,2372,2347,3859,2424,3425}}};
uint32_t usrparam=0;
uint32_t usrparam_mask=0;
volatile sig_atomic_t signalflag=0;
uint64_t allcdmemory=0;
pid_t *pids=NULL;
uint64_t memorylimit=0xFFFFFFFFFFFFFFFF;
void *mmapptr=NULL;
char *logpath=NULL;
#if defined(DEBUGCODE) && defined(VERBOSE)
int32_t verbfrom=0;
int32_t verbto=0x7fffffff;
#endif
#endif

/*----------------------------------------------------------------
 *
 *  Global variables shared with other modules that are not initialized here.
 *  WARNING: Don't use "static" as then a different variable will be created in
 *  each module with the same name.
 *
 *--------------------------------------------------------------*/
VARTYPE kbase kbset; /* Working KB. */
VARTYPE kbase mainkb; /* Main KB with axioms (theory) being analyzed. */
VARTYPE int num_cores;  /* Number of physical cores available. */
VARTYPE int active_cores;  /* Number of enabled cores */
VARTYPE int32_t procnb; /* Internal process number of current process. It is in the range 0 to active_cores-2 */
                        /* for child processes and equals active_cores-1 for parent process. */
VARTYPE int proc_affinity; /* Core affinity option. */
VARTYPE uint64_t numskolem; /* Number of skolem functions defined in main KB so far. */
VARTYPE int32_t skolemid; /* Number to identify unique skolem prefix. */
VARTYPE char skfprefix[15]; /* Prefix for unique skolem function names. */
VARTYPE uint64_t numpdefs; /* Number of predicate definitions in main KB so far. */
VARTYPE int32_t pdefid; /* Number to identify unique predicate definition prefix. */
VARTYPE char prdprefix[15]; /* Prefix for unique predicate definition names. */
VARTYPE int32_t splid; /* Number to identify unique P-predicate prefix. */
VARTYPE char splprefix[15]; /* Prefix for unique P-predicates. */
VARTYPE uint32_t numsymbols; /* Number of symbols for symbol weight assignment. */
#if PRECEDENCE != 0
VARTYPE uint32_t maxusecount; /* Maximum value of usecount field in hashchain symbol elements. */
#endif
VARTYPE hashchain *hashkeys[HASHVALUES]; /* Hash keys indexed by hash values of symbol names. */
VARTYPE hashchcmp *hashsplit[HASHVALUES]; /*Hash keys indexed by hash values of clause variants. */
VARTYPE uint32_t hashpos[MAXPOSITIONS]; /* Hash values for item positions. */
VARTYPE hashchain equkey; /* Key corresponding to equality literal. */
VARTYPE int32_t binpsize; /* Effective size of binprefix structure. */
VARTYPE int32_t satclsize; /* Effective size of satclause structure. */
VARTYPE mpctl *procctl; /* Pointer to multi-processing control data to be placed in mmap area. */
VARTYPE double cpu_time; /* Process CPU time used in seconds. */
VARTYPE double *secprc_cpu_time; /* Pointer to total child processes CPU time used in seconds to be placed in mmap area. */
VARTYPE sem_t *procsem; /* Pointer to set of process synchronization semaphores to be placed in mmap area. */
VARTYPE stats glblstats; /* GLobal statistics. */
VARTYPE options glblopts; /* Global options. */
VARTYPE stats *pbmstats; /* Pointer to problem global statistics to be placed in mmap area. */
VARTYPE int32_t pbmflags; /* Flags for current problem. See definitions above. */
VARTYPE int32_t pbmtype; /* Problem type number. */
VARTYPE int32_t pgmmode; /* Problem mode: INTERACTIVE or NONINTERACTIVE. */
VARTYPE int32_t slctordrsize; /* Number of elements in slctordr set. */
VARTYPE int32_t slctordrsize2; /* Number of elements in slctordr2 set. */
VARTYPE uint16_t *strategies; /* Pointer to strategies selected for problem being solved placed in mmap area. */
VARTYPE int32_t *nextstrat; /* Pointer to index to next free strategy in the strategies[] set. */
VARTYPE int32_t nbstrats; /* Total number of strategies in the strategies[] set. */
VARTYPE int32_t numlrnrecs; /* Number of record elements in set pointed by learnrecord. */
VARTYPE uint64_t *maxprcmemory; /* Pointer to process memory limit in mmap region to be used for saturation processes. */
VARTYPE uint64_t maxmemory; /* Memory limit to be used for saturation processes. This equals to *maxprcmemory. */
VARTYPE uint64_t mmapsize; /* Size of mmap shared memory. */
VARTYPE uint16_t stratflgs; /* Active strategy flags. */
#ifdef DEBUGMEMORY
VARTYPE void *memptrs[DBGMEMELEMS]; /* Pointers to memory allocations that must be tracked when DEBUGMEMORY is defined. */
                            /* The actual size of this set must be adjusted for each case. */
VARTYPE int32_t memtrack[DBGMEMELEMS]; /* Status of each entry in memptrs[]. The actual size of this set must be equal to the */
                               /* size of memptrs[]. The content is: */
                               /* 0 -> Entry not used. */
                               /* 1 -> Entry for memory allocated but not yet freed. */
                               /* 2 -> Entry for memory allocated and freed. */
VARTYPE int32_t ndemods;               /* Number of entries used in the above sets. */
VARTYPE int32_t auxcounter;            /* Auxiliary counter. */
#endif
#ifdef DEBUGCODE
VARTYPE int32_t dbgnb; /* For setting appropriate breakpoint in Otter and Discount functions. */
#endif
#ifdef BENCHMARKING
VARTYPE double *bresults[3]; /* Pointers to benchmarking results to be placed in mmap area. */
VARTYPE int32_t *bncounter; /* Pointer to benchmarking counter to be placed in mmap area. */
VARTYPE char *benchfname; /* Benchmarking output file name. */
#endif

/*----------------------------------------------------------------
 *
 *  Platform independent functions common to all platform independent modules.
 *
 *--------------------------------------------------------------*/
/* Drodi module functions. */
int main(int argc,char *argv[]); /* Entry point. */
int32_t Initialize_pgm(void); /* Initialize program */
void Clean4Exit(void); /* Performs the needed cleaning before exit */
int32_t CheckAnswer(char *,char *); /* Confirm answer with user */
int32_t Create_KB(kbase *); /* Create a new KB */
int32_t Initialize_KB(kbase *); /* Initialize a KB */
void Delete_KB(kbase *); /* Delete a KB */
void Free_Formula(kbase *); /* Free formula and tree memory and initialize queue pointers */
void FreeInfFormulas(void); /* Free inferred formulas in main KB */
void Initialize_hash(void); /* Initialize symbol hash tables for names in input formulas */
void Initialize_hashcmp(void); /* Initialize symbol hash tables for component names */
int32_t Tell(char *,int32_t,int32_t); /* Add formula to main KB */
int32_t InputSentence(char **,int32_t *); /* Input formula lines */
int32_t CheckSyntax(char **,int32_t *,char **,nameschain *,int32_t); /* Check sentence syntax */
char *SkipBlanks(char *); /* Skips space characters */
int32_t ParseSentenceStart(char *,char **,nameschain *,int32_t); /* Get formula start type */
int32_t ParseVariableList(char *string,char **,nameschain *); /* Variable list parser */
int32_t AddNametoChain(char *,nameschain *,int32_t); /* Add name to text formulas names chain */
int32_t ParseAtomic(char *,char **,nameschain *,int32_t); /* Atomic formula parser */
int32_t ParseName(char *,char **); /* Parses a variable, constant, function or predicate name */
int32_t ParseTerm(char *,char **,char **,nameschain *,int32_t *,int32_t); /* Parse term */
int32_t CheckDeclared(char *,nameschain *,int32_t,int32_t *); /* Check if name is locally declared */
int32_t Ask(char *); /* Include a negated conjecture in main KB */
void AddTxtFormula2KB(cmprefix *,int32_t,kbase *,int32_t); /* Add text formula to a KB */
void ResetStats(stats *); /* Reset statistics structure */
void AddInferenceStats(stats *,stats *); /* Add inference statistics */
int32_t TPTPQuantifiers(char **,int32_t *); /* Set scope of quantifiers */
float FormatMemory(float,char **); /* Format memory in KB, MB and GB */
char *JumpOverQuotes(char *); /* Jump over a quoted block */
void CatchSignal1(int); /* Catch program exception signals. */
char *strdupmn(char *); /* Memory tracking version of strdup() function */
void *mallocmn(size_t); /* Allocate and track total memory */
void *reallocmn(void *,size_t); /* Reallocate and track total memory */
uint64_t GetAvailableMemory(void); /* Get available memory */
#ifndef DEBUGMEMORY
void myfree(void *); /* Free and track total memory */
int32_t Init_mmap(void); /* Initialize mmap common memory area */
#else
void myfree(void *,char *,int); /* Performs memory debug tasks and calls free() function */
#endif
#ifdef MEMORYTUNNING
int32_t SymMalloc(uint64_t); /* Allocate additional memory for MAXMEMORY tuning */
#endif

/* clausify module functions. */
cmprefix *PreprocessFormula(cmprefix *,char **,int32_t *,int32_t); /* Text formula pre-processing */
int32_t PutBrackets(lkedfitem **ptopitem); /* Insert parenthesis according to precedence */
int32_t SafeAppend(char **,int32_t *,char *,int32_t); /* Append a string safely */
cmprefix *Convert2PreNNF(lkedfitem **,cmprefix *,char **,int32_t *); /* Perform pre-nnf conversion */
cmprefix *Convert2NNF(lkedfitem **,cmprefix *,char **,int32_t *); /* Perform nnf conversion */
void DropImplication(lkedfitem *); /* Translate an implication */
void DropRevImplication(lkedfitem *); /* Translate a reverse implication */
lkedfitem *DropEquivalence(lkedfitem *,lkedfitem **); /* Translate an equivalence */
lkedfitem *DropXOR(lkedfitem *,lkedfitem **); /* Translate a XOR connective */
void DropNAND(lkedfitem *); /* Translate a NAND */
void DropNOR(lkedfitem *); /* Translate a NOR */
int32_t IndentNegations(lkedfitem *); /* Move negations inwards */
void DropQuantifiers(lkedfitem **); /* Delete quantifiers in formula */
void RemoveLinkedListChain(lkedfitem **,lkedfitem *,lkedfitem *); /* Remove a chain of linked list items */
int32_t Convert2CNF(char *,char **,char **,int32_t *); /* Convert sentence to CNF */
int32_t CnfOrCnf(char **,int32_t *,char *); /* Merge two CNF's separated by an OR operator */
lkedfitem *Text2LinkedList(char *); /* Convert formula text to linked list format */
void FreeLkedLstFormula(lkedfitem *); /* Free formula storage in linked list format */
void RemoveRedundantBrackets(lkedfitem **); /* Remove redundant brackets */
lkedfitem *RemoveAdjBrackets(lkedfitem *,lkedfitem **); /* Remove set of adjacent redundant brackets */
void SetPolarity(lkedfitem *); /* Set polarity of linked list items */
void AddItem2LkdLstF(lkedfitem **,lkedfitem **,lkedfitem **); /* Add item to formula in linked list format */
int32_t AddOpenBracket(lkedfitem **,lkedfitem **,lkedfitem **); /* Add open bracket to formula in linked list format */
int32_t AddQuantifier(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain **); /* Add quantifier to formula in linked list format */
int32_t AddName(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add name to formula in linked list format */
int32_t AddPredOrFunc(hashchain *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add a predicate or function in linked list format */
int32_t AddEquality(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add an (in)equality in linked list format */
int32_t AddEquRootTerm(char *,char **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add an (in)equality root term in linked list format */
int32_t DuplicateBlock(lkedfitem *,lkedfitem *,lkedfitem **,lkedfitem **); /* Duplicate a block of linked list items */
void InsertBlock(lkedfitem *block,lkedfitem *,lkedfitem *); /* Insert a block of linked list items */
lkedfitem *JumpOverBlockOrTerm(lkedfitem *,int32_t); /* Jump over a block or term */
cmprefix *MiniscopeFormula(lkedfitem **,cmprefix *,char **,int32_t *); /* Miniscope a formula */
int32_t RemoveUnneededQfiers(lkedfitem **); /* Remove redundant quantifiers */
int32_t MiniscopeQuantifier(lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a quantifier */
int32_t MiniscopeSegment(lkedfitem *,lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a formula segment */
int32_t MiniscopeBlock(lkedfitem *,lkedfitem *,lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a formula block */
int32_t IsVarInBlock(lkedfitem *,lkedfitem *); /* Check if a variable exists in a block */
void RelinkVars(lkedfitem *,lkedfitem *,lkedfitem *); /* Rebuild links of variables in a block */
lkedfitem *DuplicateQuantifier(lkedfitem *,lkedfitem **,lkedfitem **,nameschain *); /* Duplicate and insert a quantifier */
cmprefix *Skolemize(lkedfitem **,cmprefix *,char **,int32_t *); /* Skolemize a formula */
cmprefix *TFSimplify(lkedfitem **,cmprefix *,char **,int32_t *); /* Simplify $true and $false predicates */
void RenameVariables(lkedfitem *); /* Rename variables in a formula */
cmprefix * AddLkdLstFormula2KB(lkedfitem *,cmprefix *,cmprefix *,int32_t,char **,
		int32_t *); /* Add linked list formula to main KB in text form */
cmprefix *FormulaRenaming(lkedfitem **,cmprefix *,char **,int32_t *); /* Formula renaming */
void GetClauseNumberBounds(lkedfitem *,int32_t *,int32_t *); /* Get lower bounds of number of clauses in a segment */
void CalcSegmentBounds(int32_t *,int32_t *,int32_t,int32_t,int32_t,int32_t,
		int32_t); /* Calculate lower bounds of number of clauses in a segment */
int32_t SegmentRenaming(lkedfitem **,lkedfitem *,int32_t,int32_t,cmprefix **,
		cmprefix *,nameschain *); /* Rename blocks within a segment */
void GetaABCoefs(int32_t,int32_t,lkedfitem *,lkedfitem *,int32_t,int32_t *,
		int32_t *); /* Get "a" and "b" coefficients of a block */
int32_t RenameBlock(lkedfitem **,lkedfitem *,cmprefix **,cmprefix *,nameschain *); /* Rename a block in a formula */

/* compile module functions. */
int32_t CompileCNF(char *formula,cmprefix *); /* Compile a CNF text formula */
int32_t CompileClause(char *,cmprefix **,cmprefix *); /* Compile a clause */
void CompileLiteral(char *,int32_t *,cmprefix *,int32_t *); /* Literal compilation */
void CompileTerm(char *,int32_t *,cmprefix *,int32_t *,int32_t); /* Term compilation */
void CompileSymbol(char *,int32_t *,cmprefix *,int32_t *,int32_t); /* Symbol compilation */
int32_t BytesRequired(char *); /* Compute the number of bytes needed for a compiled clause */
int32_t AddBinClause2KB(cmprefix *,kbase *,int32_t); /* Add binary clause to a KB. */
int32_t RenumberVars(binprefix *); /* Renumber variables in a binary formula */
int32_t SafeAppendF(char **,int32_t *,char *,int32_t *); /* Append a string fast and safely */

/* indexing module functions. */
int32_t AddGlobalSymbols(nameschain *); /* Add locally defined symbols to global symbol list */
hashchain *CreateGlobalSymbol(char *,int32_t,int32_t); /* Create global symbol entry */
#ifdef MURMUR3
uint32_t Murmur3(char *); /* Murmur3 hash function */
#else
uint32_t KRMultHash(char *); /* Kernighan and Ritchie hash function */
#endif
hashchain *CheckGlblDeclared(char *); /* Check if a name is globally declared */
int32_t EditDiscrTrees(cmprefix *,int32_t); /* Edit discrimination trees for clause */
int32_t DsTermIndexing(uint8_t *,void **); /* Add item to a discrimination tree */
void DeleteBranch(uint8_t *,void **); /* Delete item in a discrimination tree */
void FreeDiscrBranch(void *,int32_t); /* Free discrimination tree */
int32_t QueryDscTree(uint8_t *,uint8_t **,void *,int32_t,int32_t); /* Query a discrimination tree */
int32_t Unlink(cmprefix *,kbase *); /* Unlink binary clause from KB */
void FreeTreeMemory(void); /* Free KB tree memory */
uint32_t HashClause(uint8_t *); /* Calculate 32 bits hash value for a clause key */
#ifdef DEBUGTREE
void CheckDscTrees(void); /* Discrimination trees integrity check */
#else
#ifdef DEBUGTREE2
void CheckDscTQueries(void); /* Discrimination trees queries check */
#endif
#endif

/* infere module functions. */
uint8_t *NextItem(uint8_t *,int32_t); /* Jump to next item in clause */
int32_t IsEqual(uint8_t *,uint8_t *,int32_t); /* Check if two items are identical */
int32_t IsEqual2(uint8_t *,uint8_t *); /* Check if two items in different clauses are identical */
int32_t Unify(uint8_t *,uint8_t *,subst *,int32_t); /* Perform literal or term unification */
int32_t UnifyVar(uint8_t *,uint8_t *,subst *,int32_t); /* Unify a variable with a term */
int32_t SymEquUnify(uint8_t *,uint8_t *,subst *,int32_t); /* Perform symmetric equality unification */
int32_t Generalize(uint8_t *,uint8_t *,subst *,int32_t); /* Perform literal, term or list generalization */
int32_t GeneralizeVar(uint8_t *,uint8_t *,subst *); /* Generalize a variable with a term */
int32_t SymEquGeneralize(uint8_t *,uint8_t *,subst *,int32_t); /* Perform symmetric equality generalization */
int32_t IsCyclicSubst(subst *); /* Cyclic substitution check */
int32_t CheckClosedLoop(uint8_t *,uint8_t *,int16_t *,int32_t *,int16_t *,
		int32_t *); /* Check if a variable is in a closed loop */
uint32_t CheckVarInTerm(uint8_t *,uint8_t *,subst *,
		int32_t); /* Check if a variable occurs in a term with substitution */
uint32_t CheckVarInTerm2(uint8_t *,uint8_t *); /* Check if a variable occurs in a term */
int32_t CopyClause(cmprefix *,int32_t *,cmprefix *); /* Copy a clause */
void DeleteLiteral(binprefix *,uint8_t *); /* Deletes a literal in a clause */
uint8_t *GetLiteralPtr(cmprefix *,int32_t); /* Gets a pointer to literal given its order number */
int32_t GetLiteralNbr(binprefix *,uint8_t *); /* Gets a literal number given a literal pointer */
int32_t ApplySubst2Clause(cmprefix *,int32_t *,subst *,int32_t); /* Apply substitution to a clause */
void GetSubstTermData(subst *,uint8_t **,int32_t); /* Initialize set of pointers to terms in a substitution */
int32_t ApplySubst2Item(uint8_t **,binprefix **,int32_t *,uint8_t *,
		int32_t,uint8_t **); /* Apply substitution to an item */
void UpdateParams(binprefix *); /* Update binary clause parameters */
void UpdateParams2(uint8_t *); /* Update binary formula parameters */
uint8_t *UpdateSymbolParms(binprefix *,uint8_t *,int32_t,
		int32_t); /* Update binprefix and symbol parameters in an item */
uint8_t *UpdateSymbolParms2(uint8_t *,uint8_t *); /* Update symbol parameters in an item */
void UpdateClauseSymbolWeights(binprefix *); /* Update clause symbolw field */
void Infere(cmprefix *); /* Perform saturation inferences */
void Resolution(cmprefix *); /* Perform resolution inference */
int32_t BuildResolvent(cmprefix *,cmprefix *,uint8_t *,uint8_t *,
		cmprefix **,subst *,int32_t); /* Build clause inferred by resolution */
void PM2Clause(cmprefix *); /* Perform Paramodulation to a clause */
void PMFromClause(cmprefix *); /* Perform Paramodulation from a clause */
int32_t BuildPMClause(cmprefix *,cmprefix *,uint8_t *,uint8_t *,uint8_t *,uint8_t *,
		cmprefix **,subst *); /* Build clause inferred by paramodulation */
void EqualityResolution(cmprefix *); /* Perform equality resolution inference */
void EqualityFactoring(cmprefix *); /* Perform equality factoring inference */
void Factor(cmprefix *); /* Infer factors of a clause */

/* simplify module functions. */
int32_t IsTautology(cmprefix *); /* Check if clause is a tautology. */
int32_t RemoveDuplicates(cmprefix *,cmprefix **,int32_t); /* Remove literal duplicates from clause */
int32_t TrivEqResolution(cmprefix *,cmprefix **); /* Trivial equality resolution inference */
int32_t DestrEqResolution(cmprefix *,cmprefix **); /* Destructive equality resolution inference */
int32_t FwSubsuming(cmprefix *,int32_t); /* Check if clause is subsumed by a KB */
int32_t BckSubsuming(cmprefix *,int32_t); /* Perform backward subsumption inference */
int32_t IsSubsumedByClause(uint8_t *,int32_t,binprefix *,subst *,
		uint8_t **); /* Check if clause is subsumed by another clause */
int32_t LiteralSubsuming(uint8_t *,uint8_t *,int32_t,subsumdata *,subst *,
		struct timeval *,uint8_t **); /* Check literal subsuming */
int32_t FwdSimplify(cmprefix **,int32_t); /* Forward and backward simplifications */
int32_t FirstSimplify(cmprefix **,int32_t); /* First clause simplification */
int32_t SecondSimplify(cmprefix **,int32_t); /* Forward clause simplification */
int32_t BckSimplify(cmprefix *,int32_t); /* Backward simplification */
int32_t ResolFactor(binprefix *,uint8_t *,binprefix **,uint8_t **,
		subst *,int32_t); /* Factoring for resolution inferences */
int32_t FwdDemodulation(cmprefix **,int32_t); /* Perform forward demodulation */
int32_t BckDemodulation(cmprefix *,int32_t); /* Perform backward demodulation */
int32_t Demodulate(cmprefix *,uint8_t *,binprefix *,uint8_t *,
		cmprefix **,subst *); /* Demodulate a clause */
int32_t RedundancyCheck(uint8_t *,uint8_t *,binprefix *,binprefix *); /* Perform demodulation redundancy check */
int32_t FwSubsResltn(cmprefix *,cmprefix **,int32_t); /* Perform forward subsumption resolution */
int32_t BkSubsResltn(cmprefix *,int32_t); /* Perform backward subsumption resolution */
int32_t SubsResltn(cmprefix *,binprefix *,uint8_t *,cmprefix **); /* Perform subsumption resolution inference */
#ifdef UNITCLRESOLUTION
int32_t UCResolution(cmprefix *); /* Perform unit clause resolution inference */
#endif

/* ordering module functions. */
int32_t CompareLiterals(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Literals comparison */
int32_t CompareEqualities(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Equality literals comparison */
int32_t ComparePredicates(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Predicates comparison */
int32_t CompareItems(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Terms and predicates comparison */
int32_t CompareEquTerms(binprefix *,uint8_t *,uint8_t *,equpnode **,
		equpnode **,int32_t); /* Equality root terms comparison */
int32_t CompSubstEquTrms(binprefix *,uint8_t *,uint8_t *,equpnode **,equpnode **,subst *,
		int32_t,int32_t); /* Equality root terms comparison with substitution */
int32_t CompareSubstItems(uint8_t *,uint8_t *,int32_t,subst *,int32_t,
		int32_t); /* Terms comparison with substitution */
int32_t SecondKBOCheck(uint8_t *,uint8_t *,int32_t); /* Perform second KBO check */
int32_t SecondKBOCheck2(binprefix *,uint8_t *,uint8_t *,equpnode *,
		equpnode *,int32_t); /* Perform second KBO check for polynomial trees */
int32_t SecondKBOCheck3(binprefix *,uint8_t *,uint8_t *,equpnode *,
		equpnode *,subst *,int32_t,int32_t); /* Perform second KBO check with substitution */
int32_t CompareLexic(binprefix *,uint8_t *,uint8_t *,equpnode *,equpnode *,subst *,int32_t,
		int32_t); /* Perform lexicographical comparison with substitution */
int32_t CompareItemWeight(uint8_t *,uint8_t *,coeff *,coeff *); /* Terms and predicates weight comparison */
int32_t CompareItemWeight2(coeffcpt *,coeffcpt *); /* Equality subterms weight comparison */
int32_t CompItmWeightSubst(equpnode *,equpnode *,subst *,int32_t,
		int32_t);/* Equality subterms weight comparison with substitution. */
int32_t ApplySubst2Pol(lkedcoef **,subst *,int32_t,int32_t); /* Apply substitution to polynomial. */
int32_t ApplySbst2TwoItm(uint8_t *,uint8_t *,uint8_t **,uint8_t **,cmprefix **,
		subst *,int32_t,int32_t); /* Apply substitution to two terms */
int32_t CalcCoeffs(uint8_t *,coeff *,int32_t); /* Calculate weight polynomial coefficients */
int32_t SelectItems(cmprefix *,int32_t); /* Select items in a clause */
int32_t CheckEqRootOrder(binprefix *,uint8_t *,uint8_t *,uint8_t *,subst *,int32_t,
		int32_t); /* Check ordering constraints in equality root terms */
uint8_t *UpdateOffsets(binprefix *,uint8_t *); /* Update item offsets */
int32_t SetEquData(binprefix *,uint8_t *,uint8_t *,uint8_t *); /* Initialize equality data for ordering constraints. */
int32_t SetEquChildNode(binprefix *,uint8_t *,
		equpnode **); /* Initialize and set equality tree node for ordering constraints. */
int32_t SetEqTrmPol(binprefix *,uint8_t *,coeffcpt **); /* Set equality subterm compact polynomial. */
int32_t SetLftRhtPols(binprefix *,coeffcpt **,coeffcpt **); /* Convert compact polynomials to left and right format. */
void FreeEquData(binprefix *); /* Free equality ordering data storage allocated for a compiled clause */
void FreeEquNode(uint8_t *,equpnode *); /* Free equality ordering node and its data. */
void FreeEquNodeData(uint8_t *,equpnode *); /* Free data in an equality ordering node. */

/* multith module functions. */
void Saturate(void); /* Start saturation algorithms */
void SetMTParams(int32_t); /* Set strategy specific KB parameters */
void InferenceProcess(void); /* Main inference function for the process */
void Otter(void); /* Otter saturation algorithm */
void Discount(void); /* Discount saturation algorithm */
cmprefix *SelectClause(void); /* Select current clause from passive clauses */
int32_t Retained(cmprefix *); /* Check if clause is retained */
void ClauseWeight(cmprefix *); /* Compute clause weight */
int32_t LRSClean(int32_t); /* Delete unreachable passive clauses */
int32_t SetClSelectOrder(void); /* Set select order for clause selection queues */
int32_t Preprocess(int32_t); /* Perform problem pre-processing */
uint64_t ClauseRelevance(void); /* Perform relevance filtering */
void GoType(void); /* Print problem type flags */
void Clausify(void); /* Convert problem formulas to CNF and print them in TPTP format */
int32_t InitStrategies(void); /* Select strategies for current problem */
void GetPbmTypeNumber(void); /* Calculate problem type number */
uint16_t StratNb2StratFlags(int32_t); /* Convert a strategy number to strategy flags */

/* filemgmt module functions. */
char *Decompile(binprefix *); /* Translate a binary clause formula to text format */
char *Decompile2(uint8_t *); /* Translate a binary hashchcmp formula to text format */
char *DecompileSat(satclause *); /* Translate a SAT clause formula to text format */
char *DecompileAsserts(uint8_t *,int32_t); /* Translate an A-clause assertions to text format */
int32_t AddTermText(uint8_t *,char **,int32_t *,int32_t *); /* Translate a binary term to text format */
#ifdef DEBUGCODE
void BuildTxtSubstData(uint8_t *,uint8_t *,subst *,int32_t); /* Build substitution data in text form */
#endif
void PrintStats(int32_t); /* Display statistics */
void PrintStatData(stats *,int32_t); /* Display statistics data */
void PrintProcessSettings(void); /* Print process option settings */
int32_t BuildProof(void); /* Print theorem proof */
int32_t AddSatCl2Proof(void); /* Add SAT clauses to proof queue */
int32_t AddAclause2Proof(proofnode *); /* Add A-clause to proof queue */
void LinkProofNode(proofnode *,proofnode *); /* Link node to proof queue */
int32_t PrintProof(void); /* Print theorem proof */
int32_t PrintModel(void); /* Print model from a GOSAT run */
int32_t Import(char *,char *,int32_t,int32_t); /* Import a TPTP problem file */
int32_t ReadDirective(char **,int32_t *,FILE *); /* Read a directive from a TPTP file */
int32_t CheckBrackets(char **); /* Check brackets and parenthesis balance */
int32_t Cnf2Fof(char **); /* Convert a CNF formula to FOF */
int32_t FileExist(char *); /* Check if a file exist */
void PrintKBData(void); /* Print current data of a knowledge base */
int32_t PrintAClauses(void); /* Print A-clauses of a knowledge base */
int32_t PrintSingleAClause(cmprefix *); /* Print a single A-clause */
int32_t PrintCompSymbols(void); /* Print the component symbols of a knowledge base */
int32_t PrintSATClauses(void); /* Print SAT clauses of a knowledge base */
int32_t PrintSingleSATClause(satclause *); /* Print a single SAT clause */
int32_t LinkedList2Text(lkedfitem *,char **,int32_t *,int32_t); /* Convert formula in linked list format to text */
int32_t Equ2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert (in)equality in linked list formula to text */
int32_t PredOrFunc2Text(lkedfitem **,char **,int32_t *,
		int32_t *); /* Convert predicate or function in linked list formula to text */
int32_t Term2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert term in linked list formula to text */
int32_t Quantifier2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert a quantifier in linked list formula to text */
int32_t PrintClausifiedFormulas(int32_t); /* Print main KB formulas in TPTP CNF format */
char *ParseFilename(char *,char **); /* Parse and build a filename */
void Help(char*); /* Display general help or a help topic */

/* satsolver module functions. */
int32_t SplitClause(cmprefix *,int32_t); /* Get maximal split of an A-clause */
int32_t AddAssertions(binprefix *,binprefix *,binprefix *); /* Add assertions from two clauses to a third clause */
int32_t GetLocksAsserts(binprefix *,binprefix *,binprefix *); /* Get simplification assertions and locking assertions */
void ChainAsserts(uint8_t *); /* Chain assertions to appropriate queue */
void ChainLocks(uint8_t *); /* Chain locking assertions to appropriate queue */
void UnchainAsserts(uint8_t *); /* Unchain assertions */
void UnchainLocks(uint8_t *); /* Unchain locking assertions */
int32_t GetSplitSymbol(uint8_t *,int32_t,int32_t,hashchcmp **,int32_t); /* Get a P-predicate symbol for a clause */
int32_t IsVariant(uint8_t *,int32_t,hashchcmp *); /* Check if clause is a variant of another clause */
int32_t LiteralVariant(uint8_t *,hashchcmp *,int8_t *,int32_t,subst *,struct timeval *); /* Literal variant check */
int32_t ItemVariant(uint8_t *,uint8_t *,subst *); /* Item variant check */
int32_t SatQ2SatSolver(void); /* Send SAT queue to SAT solver */
int32_t SatSolver(void); /* SAT solver algorithm */
void ChainOrdComponent(hashchcmp *); /* Chain a hashchcmp element to the ordered sub-queue */
void UnchainOrdComponent(hashchcmp *); /* Remove a hashchcmp element from the ordered sub-queue */
void ChainSatClause(satclause *); /* Chain a SAT solver clause to ordered sub-queue */
void UnchainSatClause(satclause *); /* Remove a satclause element from the ordered sub-queue */
int32_t DoSymbolAssgnmnt(satnode *,satclause **); /* Perform symbol assignment tasks */
void UndoSymbolAssgnmnt(satnode *); /* Undo symbol assignment */
void GoSat(void); /* Call SAT solver without first order prover */
int32_t ApplyModelChanges(void); /* Apply changes of current interpretation */
int32_t RemoveLocks(hashchcmp *); /* Remove locks unlocked by model change */
int32_t DoLocks(hashchcmp *); /* Do new locks produced by model change */
int32_t AddCompClause(hashchcmp *); /* Add non existing component clauses */
int32_t BuildContrClause(cmprefix *); /* Add contradiction clause to SAT queue */
int32_t SatSimplify(satclause **); /* Check tautologies and remove duplicates in SAT clause */
int32_t SatSubsumption(satclause *); /* Perform by and to Sat clause subsumptions */
int32_t IsSatClSubsumed(satclause *,satclause *,uint8_t *); /* Check if a Sat clause subsumes another Sat clause */
int32_t Backtrack(satclause *,satnode **); /* Backtrack from a contradiction */
#ifdef VERIFYSATPROOF
void VerifySatProof(void); /* Verify SAT part of the proof */
#endif
#if VERIFYSATFINALSTATUS == 1
void VerifySatStatus(void); /* Verify SAT final status after a problem solving run */
#endif

/* satsolver module functions. */
int32_t ProcessLearnData(char *,char *,int32_t); /* Build learn data from output learn file */
double *DCDescentMethod(learnvector *,uint8_t *,int32_t); /* Dual Coordinate Descent Method */
double *GetClauseVector(cmprefix *); /* Build a clause vector */
int32_t VerticalFeatures(cmprefix *,double *); /* Compute vertical features of a clause vector */
void HorizontalFeatures(cmprefix *,double *); /* Compute horizontal features of a clause vector */
void StaticFeatures(cmprefix *clause,double *); /* Compute static features of a clause vector */
void LearnFromProblem(); /* Extract learning information from solved problem */
int32_t ReadLearnRecords(void); /* Read learn records generated by PROCLEARN command */

/* parsing module functions. */
void CommandProc(void); /* Command processor */
void ParseArgs(int,char **); /* Parse program arguments */
int32_t ChkConvertOption(char *); /* Check and convert option to interactive command */
int32_t ParseCommand(char *,char **); /* Command parser */
#endif /* GLOBAL_H_ */
