/*
 ============================================================================
 Name        : learn.c
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */
/****************************************************************
*
*                        Drodi (Knowledge base management module)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

/*-------------------------------------------------------------
 *
 *  DESCRIPTION:
 *
 *    This source contains the learning related functions for Drodi
 *    package.
 *
 *    Drodi is a multi-platform Open Source Artificial Intelligence application.
 *    See Drodi.c for additional information.
 *
 *
 *  IMPLEMENTATION:
 *
 *    Multi-platform
 *
 *  NOTES:
 *
 *
 *--------------------------------------------------------------*/

/* Include for all C sources */
#include "global.h"

/** Global variables for this module: only those that need initialization in their definition. */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Build learn data from output learn file)  OCJ
 *
 *    This function process a learned data file produced by the
 *    LEARNTO command option and generates a learn file usable
 *    by the LEARNFROM command option.
 *
 *    See LearnFromProblem() for a description of the input file.
 *    The output file formats is a set of learn records, with one
 *    record for each type of problem to which the learn record
 *    applies. Each learn record is a learnrecord structure.
 *
 *
 *  ARGUMENTS:
 *
 *    infile: Input file produced by the LEARNTO command option.
 *    outfile: Output file usable by the LEARNFROM command option.
 *    boost: Boost factor. Positive clauses will be boosted so that
 *           boost instances of each clause will be included.
 *    mxrecords: Maximum number of records of each type that will be
 *               accepted from the input file. If this parameter is
 *               zero then the number of records will not be limited.
 *               This parameter is used to avoid using swap space for
 *               memory allocations and prevent severe performance
 *               degradation.
 *
 *  RETURNS:
 *
 *    0 if OK or 1 of error.
 *
 *
 *
 *--------------------------------------------------------------*/
int32_t ProcessLearnData(char *infile,char *outfile,int32_t boost){
	auto FILE *filein,*fileout;                          /* Input an output file handles */
	auto learnvector *vectors;                           /* Pointer to feature vectors buffer */
	auto uint8_t *flags;                                 /* Good/bad flags for each feature vector */
	auto int32_t mxrecords;                              /* Maximum number of records of each type that will be loaded */
	auto int32_t numrecords;                             /* Number of records allocated for records buffer */
	auto int32_t readrecords;                            /* Number of records used in buffer */
	auto int32_t recordsnotread;                         /* Records not read because maximum records limit */
	auto uint64_t recordtypes[960];                      /* Record types fully processed */
	auto int32_t types;                                  /* Number of record types fully processed */
	auto int64_t offset;                                 /* Current infile read offset */
	auto int64_t nxtoffset;                              /* Offset of first record of next type of records to be processed. */
	auto double *result;                                 /* Resulting weight vector. */
	auto uint64_t blocklength;                           /* Number of flag/vector entries in block */
	auto uint64_t gooditems;                             /* Number of flag/vector entries in block corresponding to good clauses */
	auto uint64_t header;                                /* Block header */
	auto uint64_t filesize;                              /* Size of input file */
	auto learnvector *ptr1;                              /* Auxiliary pointer */
	auto uint8_t *ptr2;                                  /* Auxiliary pointer */
	auto char *ptr3;                                     /* Auxiliary pointer */
	auto size_t ii;                                      /* Auxiliary */
	auto uint64_t jj;                                    /* Auxiliary */
	auto int32_t kk,mm;                                  /* Auxiliary */

	/* Initialize maximum number of records that will be loaded. */
	/* this is limited by the maximum memory available. */
	mxrecords=(GetAvailableMemory()-allcdmemory-VECTORLENGTH*sizeof(double))
			/(sizeof(learnvector) + (2*sizeof(int8_t)+sizeof(double)+sizeof(int32_t)));

	/* Set prefix for printed lines. */
	if (pgmmode==NONINTERACTIVE){
		ptr3="% ";
	} else {
		ptr3="";
	}

	/* Open input file and get file size. */
	printf("\n%s**********************************************\n",ptr3);
	printf("%sGenerating file %s from file %s\n",ptr3,outfile,infile);
	if (NULL==(filein=fopen(infile,"rb"))){
		printf("%s%s opening input file %s.\n",ptr3,strerror(errno),infile);
		printf("%s**********************************************\n",ptr3);
		return(1);
	}
	fseek(filein,0,SEEK_END);
	filesize=ftell(filein);
	fseek(filein,0,SEEK_SET);

	/* Open output file and check its file size. */
	if (NULL==(fileout=fopen(outfile,"wb"))){
		fclose(filein);
		printf("%s%s opening output file %s.\n",ptr3,strerror(errno),outfile);
		printf("%s**********************************************\n",ptr3);
		return(1);
	}

	/* Initialize control data. */
	numrecords=0;
	vectors=NULL;
	flags=NULL;
	memset(&recordtypes[0],255,sizeof(recordtypes));
	types=0;
	nxtoffset=0;

	/* Record type outer loop. */
	while (1){

		/* Block reading inner loop. */
		readrecords=0;
		recordsnotread=0;
		offset=nxtoffset;
		fseek(filein,offset,SEEK_SET);
		while (1){

			/* Read block length. */
			ii=fread(&blocklength,sizeof(blocklength),1,filein);
			if (ii==0){

				/* Read error. */
				if (ferror(filein)){
					fclose(filein);
					fclose(fileout);
					if (vectors==NULL){
						MYFREE(vectors);
						MYFREE(flags);
					}
					printf("%s%s reading input file %s.\n",ptr3,strerror(errno),infile);
					printf("%s**********************************************\n",ptr3);
					return(1);
				}

				/* End of file, leave block reading loop. */
				break;
			}

			/* Read number of good vectors. */
			ii=fread(&gooditems,sizeof(gooditems),1,filein);
			if (ii==0){
				if (ferror(filein)){
					printf("%s%s reading input file %s.\n",ptr3,strerror(errno),infile);
				} else {
					printf("%sInvalid input file %s.\n",ptr3,infile);
				}
				fclose(filein);
				fclose(fileout);
				if (numrecords>0){
					MYFREE(vectors);
					MYFREE(flags);
				}
				printf("%s**********************************************\n",ptr3);
				return(1);
			}

			/* Validate block length. */
			jj=(blocklength*(1+(VECTORLENGTH*sizeof(double))))+(2*sizeof(uint64_t))+(blocklength?sizeof(uint64_t):0);
			if ((gooditems>blocklength)||(((offset+jj)!=filesize)&&((offset+jj+(2*sizeof(uint64_t)))>filesize))){
				fclose(filein);
				fclose(fileout);
				if (numrecords>0){
					MYFREE(vectors);
					MYFREE(flags);
				}
				printf("%sInvalid input file %s.\n",ptr3,infile);
				printf("%s**********************************************\n",ptr3);
				return(1);
			}

			/* Iterate if empty block. */
			if (blocklength==0){
				offset+=(2*sizeof(uint64_t));
				continue;
			}

			/* Read block header. */
			ii=fread(&header,sizeof(header),1,filein);
			if (ii==0){
				printf("%s%s reading input file %s.\n",ptr3,strerror(errno),infile);
				printf("%s**********************************************\n",ptr3);
				fclose(filein);
				fclose(fileout);
				if (numrecords>0){
					MYFREE(vectors);
					MYFREE(flags);
				}
				return(1);
			}

			/* Validate block header. */
			if (((0xffffffffffffff&header)!=((ALGMASK|SELECTMASK|TERMORDRMASK|
					LITORDRMASK|FUNCWMASK|GOODCHOICE)&header))
					||((0xff00000000000000&header)>0x700000000000000)){
				fclose(filein);
				fclose(fileout);
				if (numrecords>0){
					MYFREE(vectors);
					MYFREE(flags);
				}
				printf("%sInvalid record header detected at offset %ld.\n",ptr3,offset);
				printf("%s**********************************************\n",ptr3);
				return(1);
			}

			/* Block type is the first type or the type currently in process */
			/* and maximum number of records has not been allocated. */
			if ((((readrecords==0)&&(types==0))||(header==recordtypes[types]))
					&&((mxrecords==0)||(readrecords<(mxrecords-boost)))){

				/* Allocate memory and initialize recordtype[0] if necessary. */
				if (numrecords==0){
					numrecords=blocklength+((boost-1)*gooditems);
					if (((mxrecords!=0))&&(numrecords>mxrecords)){
						numrecords=mxrecords;
					}
					if (NULL==(vectors=mallocmn(numrecords*sizeof(learnvector)))){
						fclose(filein);
						fclose(fileout);
						printf("%sNot enough memory.\n",ptr3);
						printf("%s**********************************************\n",ptr3);
						return(1);
					}
					if (NULL==(flags=mallocmn(numrecords*sizeof(int8_t)))){
						fclose(filein);
						fclose(fileout);
						MYFREE(vectors);
						printf("%sNot enough memory.\n",ptr3);
						printf("%s**********************************************\n",ptr3);
						return(1);
					}
					recordtypes[0]=header;

				/* Reallocate memory if necessary. */
				} else if (numrecords<(readrecords+blocklength+((boost-1)*gooditems))){
					numrecords=readrecords+blocklength+((boost-1)*gooditems);
					if (((mxrecords!=0))&&(numrecords>mxrecords)){
						numrecords=mxrecords;
					}
					if (NULL==(ptr1=reallocmn(vectors,numrecords*sizeof(learnvector)))){
						fclose(filein);
						fclose(fileout);
						MYFREE(vectors);
						MYFREE(flags);
						printf("%sNot enough memory.\n",ptr3);
						printf("%s**********************************************\n",ptr3);
						return(1);
					}
					vectors=ptr1;
					if (NULL==(ptr2=reallocmn(flags,numrecords*sizeof(int8_t)))){
						fclose(filein);
						fclose(fileout);
						MYFREE(vectors);
						MYFREE(flags);
						printf("%sNot enough memory.\n",ptr3);
						printf("%s**********************************************\n",ptr3);
						return(1);
					}
					flags=ptr2;
				}

				/* Read block vectors and flags. Loop for each flag/vector pair. */
				for (mm=0;(mm<blocklength)&&((mxrecords==0)||(readrecords<=(mxrecords-boost)));mm++){

					/* Read block vectors and flags. */
					ii=fread(&flags[readrecords],sizeof(uint8_t),1,filein);
					if (ii==0){
						fclose(filein);
						fclose(fileout);
						MYFREE(vectors);
						MYFREE(flags);
						printf("%s%s reading input file %s.\n",ptr3,strerror(errno),infile);
						printf("%s**********************************************\n",ptr3);
						return(1);
					}
					ii=fread(&vectors[readrecords],sizeof(learnvector),1,filein);
					if (ii==0){
						fclose(filein);
						fclose(fileout);
						MYFREE(vectors);
						MYFREE(flags);
						printf("%s%s reading input file %s.\n",ptr3,strerror(errno),infile);
						printf("%s**********************************************\n",ptr3);
						return(1);
					}

					/* Good clauses are boosted boost times. */
					if (flags[readrecords]){
						for (kk=1;kk<boost;kk++){
							memcpy(&vectors[readrecords+kk],&vectors[readrecords],sizeof(learnvector));
							memcpy(&flags[readrecords+kk],&flags[readrecords],sizeof(uint8_t));
						}
						readrecords+=boost;
					} else {
						readrecords++;
					}
				}

				/* Not all the records in the block were read. Update recordsnotread */
				/* and move file pointer to the end of the block. */
				if (mm<blocklength){
					recordsnotread+=(blocklength-mm);
					fseek(filein,(blocklength-mm)*(1+(VECTORLENGTH*sizeof(double))),SEEK_CUR);
				}

			/* Record type is not the first type nor the type currently in process */
			/* or maximum number of records has been already allocated. */
			} else {

				/* There is not a next record type yet. */
				if (recordtypes[types+1]==0xffffffffffffffff){

					/* Check that the record type has not been already processed. */
					for (kk=mm=0;kk<=types;kk++){
						if (recordtypes[kk]==header){
							mm=1;
							break;
						}
					}

					/* If record type has not been processed yet then set this record */
					/* type as next record type and remember its position. */
					if (mm==0){

						/* Set this record type as next record type and remember its position. */
						recordtypes[types+1]=header;
						nxtoffset=offset;

						/* If maximum number of records has been achieved then leave */
						/* the inner loop. */
						if (recordsnotread>0){
							break;
						}
					}
				}

				/* Position the file pointer to the end of the block. */
				fseek(filein,blocklength*(1+(VECTORLENGTH*sizeof(double))),SEEK_CUR);
			}

			/* Update current file offset. */
			offset+=jj;
		}

		/* If some records were read then process current records, add resulting */
		/* record to output file, update number of record types fully processed, */
		/* free result, report progress and update record types. */
		if (readrecords>0){
			if (NULL==(result=DCDescentMethod(vectors,flags,readrecords))){
				fclose(filein);
				fclose(fileout);
				MYFREE(vectors);
				MYFREE(flags);
				printf("%sNot enough memory.\n",ptr3);
				printf("%s**********************************************\n",ptr3);
				return(1);
			}
			jj=(~GOODCHOICE)&recordtypes[types];
			if (1>fwrite(&jj,sizeof(jj),1,fileout)){
				fclose(filein);
				fclose(fileout);
				MYFREE(vectors);
				MYFREE(flags);
				printf("%sError writing to output file %s.\n",ptr3,outfile);
				printf("%s**********************************************\n",ptr3);
				return(1);
			}
			if (VECTORLENGTH>fwrite(result,sizeof(result[0]),VECTORLENGTH,fileout)){
				fclose(filein);
				fclose(fileout);
				MYFREE(vectors);
				MYFREE(flags);
				printf("%sError writing to output file %s.\n",ptr3,outfile);
				printf("%s**********************************************\n",ptr3);
				return(1);
			}
			MYFREE(result);
			printf("%sProcessed %d records of type %d.\n",ptr3,readrecords,types+1);
			if (recordsnotread>0){
				printf("%s===> Maximum records limit exceeded, %d input records not processed.\n",ptr3,recordsnotread);
			}
			types++;
		}

		/* Leave record type loop if no new record type available. */
		if (recordtypes[types]==0xffffffffffffffff){
			break;
		}
	}

	/* Free memory, close files, reset memory limit and return. */
	MYFREE(vectors);
	MYFREE(flags);
	fclose(filein);
	fclose(fileout);
	printf("%sFile %s has been created.\n",ptr3,outfile);
	printf("%s**********************************************\n",ptr3);

	return(0);
} /* ProcessLearnData */

#ifdef SHRINKING
/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Dual Coordinate Descent Method)  OCJ
 *
 *    This function process records from a learned data file
 *    produced by the LEARNTO command option and generates a learn
 *    file usable by the LEARNFROM command option using the Dual
 *    Coordinate Descent Method for Large-Scale Linear SVM with
 *    random permutation and shrinking for L2-SVM loss function
 *    as described in document "A Dual Coordinate Descent Method
 *    for Large-Scale Linear SVM" by Cho-Jui Hsieh, Kai-Wei Chang,
 *    Chih-Jen Lin, S. Sathiya Keerthi and S. Sundararajan.
 *
 *    Current code should have some problems because the algorithm
 *    is in fact slower than the algorithm without shrinking.
 *
 *
 *  ARGUMENTS:
 *
 *    vectors: Pointer to learnvector set of structures.
 *    flags: Pointer to set of flags. Each flag equals 1 if the
 *           corresponding vector comes from a good clause,
 *           0 otherwise.
 *    numrecords: Number of vectors in vectors set.
 *
 *  RETURNS:
 *
 *    Pointer to vector with resulting weights or NULL if not enough
 *    memory. If not NULL the result must be freed by the caller.
 *
 *
 *
 *--------------------------------------------------------------*/
double *DCDescentMethod(learnvector *vectors,uint8_t *flags,int32_t numrecords){
	auto double *result;                                 /* Resulting weight vector. */
	auto double *alpha;                                  /* Working vector dual to result */
	auto int32_t *indices;                               /* Set of random indices */
	auto double error;                                   /* Actual relative error in dual vector coordinates */
	auto int8_t *shrkstatus;                             /* Active (1) or shrunk (0) index indicator */
	auto double gradient;                                /* Gradient of minimized function */
	auto double prgrad;                                  /* Projected gradient */
	auto double bQii;                                    /* Diagonal element of barred Q matrix */
	auto double maxpg;                                   /* Maximum projected gradient */
	auto double minpg;                                   /* Minimum projected gradient */
	auto double bmaxpg;                                  /* Barred maximum projected gradient */
	auto int32_t shrunk;                                 /* 1 if some indices were shrunk, 2 if shrinking is inactive, 0 otherwise */
	auto double oldval,newval;                           /* Old and new objective function values */
	auto int32_t ii,jj,kk;                               /* Auxiliary */
	auto double aa,bb;                                   /* Auxiliary */

	/* Allocate memory and initialize. */
	if (NULL==(result=mallocmn(VECTORLENGTH*sizeof(result[0])))){
		return(NULL);
	}
	if (NULL==(alpha=mallocmn(numrecords*sizeof(alpha[0])))){
		MYFREE(result);
		return(NULL);
	}
	if (NULL==(indices=mallocmn(numrecords*sizeof(*indices)))){
		MYFREE(result);
		MYFREE(alpha);
		return(NULL);
	}
	if (NULL==(shrkstatus=mallocmn(numrecords*sizeof(*shrkstatus)))){
		MYFREE(result);
		MYFREE(alpha);
		MYFREE(indices);
		return(NULL);
	}
	memset(result,0,VECTORLENGTH*sizeof(result[0]));
	memset(alpha,0,numrecords*sizeof(alpha[0]));
	memset(shrkstatus,1,numrecords*sizeof(shrkstatus[0]));
	newval=PENALTYPARAMETER*numrecords;
	shrunk=0;
	bmaxpg=1.0e308;
	error=2.0*LEARNPRECISION;
	for (ii=0;ii<numrecords;ii++){
		indices[ii]=ii;
	}

	/* Main loop. */
	do {

		/* Initialize random indices. */
		for (ii=0;ii<numrecords;ii++){
			jj=(((int64_t)numrecords)*((int64_t)rand()))/RAND_MAX;
			if (jj==numrecords){
				jj--;
			}
			kk=indices[ii];
			indices[ii]=indices[jj];
			indices[jj]=kk;
		}

		/* Initialize maximum and minimum projected gradient. */
		maxpg=-1.0e308;
		minpg=1.0e308;

		/* Loop through the random indices. */
		for (ii=0;ii<numrecords;ii++){
			jj=indices[ii];

			/* Index is not shrunk. */
			if (shrkstatus[jj]){

				/* Calculate gradient and diagonal element of barred Q matrix. */
				for (gradient=0.0,bQii=1/(2.0*PENALTYPARAMETER),kk=0;kk<VECTORLENGTH;kk++){
					gradient+=(result[kk]*vectors[jj].vector[kk]);
					bQii+=(vectors[jj].vector[kk]*vectors[jj].vector[kk]);
				}
				if (0==flags[jj]){
					gradient=-gradient;
				}
				gradient+=((alpha[jj]/(2*PENALTYPARAMETER))-1.0);

				/* Calculate projected gradient and update shrunk indices. */
				prgrad=0.0;
				if (alpha[jj]==0.0){
					if ((shrunk!=2)&&(gradient>bmaxpg)){
						shrkstatus[jj]=0;
						shrunk=1;
						continue;
					}
					if (gradient<0.0){
						prgrad=gradient;
					}
				} else {
					prgrad=gradient;
				}

				/* Update maximum and minimum projected gradient. */
				if (maxpg<prgrad){
					maxpg=prgrad;
				}
				if (minpg>prgrad){
					minpg=prgrad;
				}

				/* Projected gradient is not zero. */
				if (prgrad!=0){

					/* Update alpha. */
					aa=alpha[jj]; /* Remember old alpha. */
					alpha[jj]-=(gradient/bQii);
					if (alpha[jj]<0.0){
						alpha[jj]=0.0;
					}

					/* Update result. */
					bb=alpha[jj]-aa;
					if (flags[jj]){
						for (kk=0;kk<VECTORLENGTH;kk++){
							result[kk]+=(bb*vectors[jj].vector[kk]);
						}
					} else {
						for (kk=0;kk<VECTORLENGTH;kk++){
							result[kk]-=(bb*vectors[jj].vector[kk]);
						}
					}
				}
			}
		}

		/* Start to calculate learning precision. */
		if ((maxpg-minpg)<TOLERANCE){

			/* Stop shrinking. */
			if (shrunk==1){
				shrunk=2;
				memset(shrkstatus,1,numrecords*sizeof(shrkstatus[0]));
				bmaxpg=1.0e308;
			}

			/* Calculate learning precision. */
			oldval=newval;
			for (newval=0.0,kk=0;kk<VECTORLENGTH;kk++){
				newval+=(result[kk]*result[kk]);
			}
			newval/=2.0;
			for (aa=0.0,ii=0;ii<numrecords;ii++){
				for (bb=0.0,kk=0;kk<VECTORLENGTH;kk++){
					bb+=(result[kk]*vectors[ii].vector[kk]);
				}
				if (flags[ii]){
					bb=1.0-bb;
				} else {
					bb=1.0+bb;
				}
				if (bb>0.0){
					aa+=(bb*bb);
				}
			}
			newval+=(aa*PENALTYPARAMETER);
			aa=newval-oldval;
			error=ABS(aa)/newval;
		}

		/* Update barred maximum projected gradient. */
		if (maxpg<=0.0){
			bmaxpg=1.0e308;
		} else {
			bmaxpg=maxpg;
		}
	} while (error>LEARNPRECISION);

	/* Free memory and return. */
	MYFREE(alpha);
	MYFREE(indices);
	MYFREE(shrkstatus);
	return(result);
} /* DCDescentMethod */

#else
/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Dual Coordinate Descent Method)  OCJ
 *
 *    This function process a records from a learned data file
 *    produced by the LEARNTO command option and generates a learn
 *    file usable by the LEARNFROM command option using the Dual
 *    Coordinate Descent Method for Large-Scale Linear SVM with
 *    L2-SVM loss function as described in document "A Dual
 *    Coordinate Descent Method for Large-Scale Linear SVM" by
 *    Cho-Jui Hsieh, Kai-Wei Chang, Chih-Jen Lin, S. Sathiya Keerthi
 *    and S. Sundararajan.
 *
 *    The random permutation technique is implemented to improve
 *    speed. However the shrinking technique is not implemented
 *    by the moment.
 *
 *
 *  ARGUMENTS:
 *
 *    vectors: Pointer to learnvector set of structures.
 *    flags: Pointer to set of flags. Each flag equals 1 if the
 *           corresponding vector comes from a good clause,
 *           0 otherwise.
 *    numrecords: Number of vectors in vectors set.
 *
 *  RETURNS:
 *
 *    Pointer to vector with resulting weights or NULL if not enough
 *    memory. If not NULL the result must be freed by the caller.
 *
 *
 *
 *--------------------------------------------------------------*/
double *DCDescentMethod(learnvector *vectors,uint8_t *flags,int32_t numrecords){
	auto double *result;                                 /* Resulting weight vector. */
	auto double *alpha;                                  /* Working vector dual to result */
	auto int32_t *indices;                               /* Set of random indices */
	auto double error;                                   /* Actual relative error in dual vector coordinates */
	auto double gradient;                                /* Gradient of minimized function */
	auto double prgrad;                                  /* Projected gradient */
	auto double bQii;                                    /* Diagonal element of barred Q matrix */
	auto double maxpg;                                   /* Maximum projected gradient */
	auto double minpg;                                   /* Minimum projected gradient */
	auto double oldval,newval;                           /* Old and new objective function values */
	auto int32_t ii,jj,kk;                               /* Auxiliary */
	auto double aa,bb;                                   /* Auxiliary */

	/* Allocate memory and initialize. */
	if (NULL==(result=mallocmn(VECTORLENGTH*sizeof(result[0])))){
		return(NULL);
	}
	if (NULL==(alpha=mallocmn(numrecords*sizeof(alpha[0])))){
		MYFREE(result);
		return(NULL);
	}
	if (NULL==(indices=mallocmn(numrecords*sizeof(*indices)))){
		MYFREE(result);
		MYFREE(alpha);
		return(NULL);
	}
	memset(result,0,VECTORLENGTH*sizeof(result[0]));
	memset(alpha,0,numrecords*sizeof(alpha[0]));
	newval=PENALTYPARAMETER*numrecords;
	error=2.0*LEARNPRECISION;
	for (ii=0;ii<numrecords;ii++){
		indices[ii]=ii;
	}

	/* Main loop. */
	do {

		/* Initialize random indices. */
		for (ii=0;ii<numrecords;ii++){
			jj=(((int64_t)numrecords)*((int64_t)rand()))/RAND_MAX;
			if (jj==numrecords){
				jj--;
			}
			kk=indices[ii];
			indices[ii]=indices[jj];
			indices[jj]=kk;
		}

		/* Initialize maximum and minimum projected gradient. */
		maxpg=-1.0e308;
		minpg=1.0e308;

		/* Loop through the random indices. */
		for (ii=0;ii<numrecords;ii++){
			jj=indices[ii];

			/* Calculate gradient and diagonal element of barred Q matrix. */
			for (gradient=0.0,bQii=1/(2.0*PENALTYPARAMETER),kk=0;kk<VECTORLENGTH;kk++){
				gradient+=(result[kk]*vectors[jj].vector[kk]);
				bQii+=(vectors[jj].vector[kk]*vectors[jj].vector[kk]);
			}
			if (0==flags[jj]){
				gradient=-gradient;
			}
			gradient+=((alpha[jj]/(2*PENALTYPARAMETER))-1.0);

			/* Calculate projected gradient. */
			if ((alpha[jj]==0.0)&&(gradient>0.0)){
				prgrad=0.0;
			} else {
				prgrad=gradient;
			}

			/* Update maximum and minimum projected gradient. */
			if (maxpg<prgrad){
				maxpg=prgrad;
			}
			if (minpg>prgrad){
				minpg=prgrad;
			}

			/* Projected gradient is not zero. */
			if (prgrad!=0){

				/* Update alpha. */
				aa=alpha[jj]; /* Remember old alpha. */
				alpha[jj]-=(gradient/bQii);
				if (alpha[jj]<0.0){
					alpha[jj]=0.0;
				}

				/* Update result and error. */
				bb=(alpha[jj]-aa);
				if (flags[jj]){
					for (kk=0;kk<VECTORLENGTH;kk++){
						result[kk]+=(bb*vectors[jj].vector[kk]);
					}
				} else {
					for (kk=0;kk<VECTORLENGTH;kk++){
						result[kk]-=(bb*vectors[jj].vector[kk]);
					}
				}
			}
		}

		/* Start to calculate learning precision. */
		if ((maxpg-minpg)<TOLERANCE){
			oldval=newval;
			for (newval=0.0,kk=0;kk<VECTORLENGTH;kk++){
				newval+=(result[kk]*result[kk]);
			}
			newval/=2.0;
			for (aa=0.0,ii=0;ii<numrecords;ii++){
				for (bb=0.0,kk=0;kk<VECTORLENGTH;kk++){
					bb+=(result[kk]*vectors[ii].vector[kk]);
				}
				if (flags[ii]){
					bb=1.0-bb;
				} else {
					bb=1.0+bb;
				}
				if (bb>0.0){
					aa+=(bb*bb);
				}
			}
			newval+=(aa*PENALTYPARAMETER);
			aa=newval-oldval;
			error=ABS(aa)/newval;
		}
	} while (error>LEARNPRECISION);

	/* Free memory and return. */
	MYFREE(alpha);
	MYFREE(indices);
	return(result);
} /* DCDescentMethod */
#endif

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Build a clause vector)  OCJ
 *
 *    This function builds a vector with clause features used in
 *    the ENIGMA learning process. The following documents describe
 *    the methodology used:
 *    - ENIGMA: Efficient Learning-based Inference Guiding Machine
 *    - Clause Features for Theorem Prover Guidance by Jan Jakubův
 *    - Enhancing ENIGMA Given Clause Guidance
 *    All three documents are by Jan Jakubův and Josef Urban.
 *
 *    The main difference with the Enigma methodology is that no
 *    specific symbols are used for predicates and functions, only
 *    generic classes:
 *    - Those that are in the conjecture.
 *    - Those that are not in the conjecture but are in the main
 *      input file.
 *    - Those that are only in axiom files.
 *    The exception are equality predicates that have their own symbol.
 *    Predicate and equality types have two selected and unselected
 *    sub-types. Also the literal polarity (+/-), skolems and variables
 *    are symbols themselves. Therefore there is a total of 15 different
 *    symbols.
 *
 *    The following additional global problem features are included:
 *    - Indication that a conjecture exist.
 *    - Indication that axiom files are used.
 *
 *    It is the caller's responsibility to free vector memory.
 *
 *    IMPORTANT: This function must be called only in the winning process.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Clause for which the vector will be computed.
 *
 *  RETURNS:
 *
 *    Pointer to vector or NULL if not enough memory.
 *
 *
 *
 *--------------------------------------------------------------*/
double *GetClauseVector(cmprefix *clause){
	auto double *vector;                                /* Pointer to clause vector */

	/* Ensure that appropriate clause literals are SELECTED. */
	if (0==(ACTIVE&clause->flags)){
		if (SelectItems(clause,1)){
			return(NULL);
		}
	}

	/* Allocate and initialize memory for the vector. */
	if (NULL==(vector=mallocmn(VECTORLENGTH*sizeof(double)))){
		return(NULL);
	}
	memset(vector,0,VECTORLENGTH*sizeof(double));

	/* Vertical and horizontal features. */
	if (NOMEMORY==VerticalFeatures(clause,vector)){
		MYFREE(vector);
		return(NULL);
	}
	HorizontalFeatures(clause,vector);

	/* Static features. */
	StaticFeatures(clause,vector);

	/* Global features: conjecture existence and axiom file existence. */
	if ((EXISTCONJ|EXISTNEGCONJ)&pbmflags){
		vector[2469]=CONJECTURECOMPONENT;
	}
	if (EXISTAXIOMFILE&pbmflags){
		vector[2470]=AXIOMFILECOMPONENT;
	}

	return(vector);
} /* GetClauseVector */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Compute vertical features of a clause vector)  OCJ
 *
 *    This function computes the vertical features of clause and
 *    adds them to the clause vector. See GetClauseVector()
 *    function for the methodology and bibliography.
 *
 *    Each feature is a three terms walk. The first term cannot be
 *    a variable so there are 14 different types. The second term
 *    cannot be a variable nor a polarity so there are 12 different
 *    symbols. The third term cannot be a polarity nor a predicate
 *    so there are 5 different symbols. Therefore the total number
 *    of vertical features is 14*12*5=840.
 *
 *    Also the maximum depth of functions, skolems and variables
 *    by polarity of the literal where they appear is recorded as
 *    another "vertical" feature. There are 5*2=10 additional
 *    "vertical" (depth) features of this kind.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Clause for which the vector will be computed.
 *    vector: Pointer to vector.
 *
 *  RETURNS:
 *
 *    NOMEMORY if not enough memory or 0 if OK.
 *
 *
 *
 *--------------------------------------------------------------*/
int32_t VerticalFeatures(cmprefix *clause,double *vector){
	auto vertnode *node;                                 /* Pointer to current vertical feature node */
	auto vertnode *newnode;                              /* Pointer to new vertical feature node */
	auto vertnode *topnode;                              /* Pointer to top vertical feature node */
	auto int32_t vectidx;                                /* Vector index for the three terms walk */
	auto uint8_t *ptr1;                                  /* Auxiliary pointer */

	/* Loop through clause items. */
	topnode=node=NULL;
	vectidx=0; /* Just to avoid compile warnings. */
	for (ptr1=&clause->part2.bin->formula[0];ptr1[0]!=UNITEND;
				ptr1=NextItem(ptr1,IMMED)){

		/* Initialize node: top node or new branch. */
		if ((node==NULL)||(ptr1==node->nextitem)){

			/* Top node. Allocate new node. */
			if (node==NULL){
				if (NULL==(topnode=node=mallocmn(sizeof(vertnode)))){
					return(NOMEMORY);
				}
				node->depth=0;
				node->nextnode=node->prevnode=NULL;

			/* New branch. Backtrack to branching point. */
			} else {
				for (;(node->prevnode!=NULL)&&(ptr1==node->prevnode->nextitem);node=node->prevnode){
				}
			}

		/* Initialize node: continue in the same branch. */
		} else {

			/* No deeper nodes are available. Allocate new node. */
			if (node->nextnode==NULL){
				if (NULL==(newnode=mallocmn(sizeof(vertnode)))){
					for (node=topnode;node!=NULL;node=newnode){
						newnode=node->nextnode;
						MYFREE(node);
					}
					return(NOMEMORY);
				}
				node->nextnode=newnode;
				newnode->prevnode=node;
				newnode->nextnode=NULL;
				newnode->depth=1+node->depth;
				node=newnode;

			/* A deeper node is available. */
			} else {
				node=node->nextnode;
			}
		}

		/* Update item and nextitem node fields. */
		node->item=ptr1;
		node->nextitem=NextItem(ptr1,OVERSUBTERMS);

		/* Compute the vector component index for the feature that has */
		/* this node as last term. */
		if (node->depth>0){

			/* Last node. Initialize vector index for vertical feature */
			/* and update maximum depth feature. */
			switch ((VARIABLE|FUNCTION)&node->item[0]){
				case FUNCTION:
					switch ((FROMCONJECTURE|FROMNOTAXIOM|SKOLEM)
							&(((symbol *)&node->item[1])->symbol)->type){
						case FROMCONJECTURE:
						case FROMCONJECTURE|FROMNOTAXIOM:
							vectidx=0*14*12;
							if (NEGATED&topnode->item[0]){
								if (vector[2430]<node->depth){
									vector[2430]=node->depth;
								}
							} else {
								if (vector[2431]<node->depth){
									vector[2431]=node->depth;
								}
							}
							break;
						case FROMNOTAXIOM:
							vectidx=1*14*12;
							if (NEGATED&topnode->item[0]){
								if (vector[2432]<node->depth){
									vector[2432]=node->depth;
								}
							} else {
								if (vector[2433]<node->depth){
									vector[2433]=node->depth;
								}
							}
							break;
						case SKOLEM:
							vectidx=2*14*12;
							if (NEGATED&topnode->item[0]){
								if (vector[2434]<node->depth){
									vector[2434]=node->depth;
								}
							} else {
								if (vector[2435]<node->depth){
									vector[2435]=node->depth;
								}
							}
							break;
						case 0:
							vectidx=3*14*12;
							if (NEGATED&topnode->item[0]){
								if (vector[2436]<node->depth){
									vector[2436]=node->depth;
								}
							} else {
								if (vector[2437]<node->depth){
									vector[2437]=node->depth;
								}
							}
							break;
					}
					break;
				case VARIABLE:
					vectidx=4*14*12;
					if (NEGATED&topnode->item[0]){
						if (vector[2438]<node->depth){
							vector[2438]=node->depth;
						}
					} else {
						if (vector[2439]<node->depth){
							vector[2439]=node->depth;
						}
					}
					break;
			}

			/* Next to last node. */
			switch ((PREDICATE|EQUALITY|FUNCTION)&node->prevnode->item[0]){
				case FUNCTION:
					switch ((FROMCONJECTURE|FROMNOTAXIOM|SKOLEM)
							&(((symbol *)&node->prevnode->item[1])->symbol)->type){
						case FROMNOTAXIOM:
							vectidx+=(1*14);
							break;
						case SKOLEM:
							vectidx+=(2*14);
							break;
						case 0:
							vectidx+=(3*14);
							break;
					}
					break;
				case PREDICATE:
					switch ((FROMCONJECTURE|FROMNOTAXIOM)
							&(((symbol *)&node->prevnode->item[1])->symbol)->type){
						case FROMCONJECTURE:
						case FROMCONJECTURE|FROMNOTAXIOM:
							vectidx+=(4*14);
							break;
						case FROMNOTAXIOM:
							vectidx+=(5*14);
							break;
						default:
							vectidx+=(6*14);
							break;
					}
					if (SELECTED&node->prevnode->item[0]){
						vectidx+=(4*14);
					}
					break;
				case EQUALITY:
					vectidx+=(7*14);
					if ((SELECTED|PMSELECTED)&node->prevnode->item[0]){
						vectidx+=(4*14);
					}
					break;
			}

			/* Next to last node is the top node. Process sign of literal. */
			if (node->depth==1){
				if (NEGATED&topnode->item[0]){
					vectidx+=12;
				} else {
					vectidx+=13;
				}

			/* Next to last node is not the last node. Process the first */
			/* node in the three walk terms. */
			} else {
				switch ((PREDICATE|EQUALITY|FUNCTION)&node->prevnode->prevnode->item[0]){
					case FUNCTION:
						switch ((FROMCONJECTURE|FROMNOTAXIOM|SKOLEM)
								&(((symbol *)&node->prevnode->prevnode->item[1])->symbol)->type){
							case FROMNOTAXIOM:
								vectidx+=1;
								break;
							case SKOLEM:
								vectidx+=2;
								break;
							case 0:
								vectidx+=3;
								break;
						}
						break;
					case PREDICATE:
						switch ((FROMCONJECTURE|FROMNOTAXIOM)
								&(((symbol *)&node->prevnode->prevnode->item[1])->symbol)->type){
							case FROMCONJECTURE:
							case FROMCONJECTURE|FROMNOTAXIOM:
								vectidx+=4;
								break;
							case FROMNOTAXIOM:
								vectidx+=5;
								break;
							default:
								vectidx+=6;
								break;
						}
						if (SELECTED&node->prevnode->item[0]){
							vectidx+=4;
						}
						break;
					case EQUALITY:
						vectidx+=7;
						if ((SELECTED|PMSELECTED)&node->prevnode->item[0]){
							vectidx+=4;
						}
						break;
				}
			}

			/* Update vector component. */
			vector[vectidx]++;
		}
	}

	/* Free memory and return. */
	for (node=topnode;node!=NULL;node=newnode){
		newnode=node->nextnode;
		MYFREE(node);
	}
	return(0);
} /* VerticalFeatures */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Compute horizontal features of a clause vector)  OCJ
 *
 *    This function computes the horizontal features of clause and
 *    adds them to the clause vector. See GetClauseVector()
 *    function for the methodology and bibliography.
 *
 *    Only symbols with an arity bigger than 0 and less than or
 *    equal to 3 are considered. However ten vector indices are
 *    reserved for cases with arity bigger than 3, one for each symbol
 *    type not including polarity symbols, equalities and variables.
 *
 *    The total number of horizontal features is:
 *      10 (symbols with arity bigger than 3)
 *      2*(5+4+3+2+1) (equalities taking into account the symmetry)
 *      10*5 (symbols with arity=1)
 *      10*5*5 (symbols with arity=2)
 *      10*5*5*5 (symbols with arity=2)
 *      Total 1590 horizontal features.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Clause for which the vector will be computed.
 *    vector: Pointer to vector.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *
 *--------------------------------------------------------------*/
void HorizontalFeatures(cmprefix *clause,double *vector){
	auto uint8_t *ptr1,*ptr2;                            /* Auxiliary pointers */
	auto int32_t vectidx;                                /* Vector index for the horizontal feature */
	auto int32_t jj,kk,mm,pp,qq;                         /* Auxiliary */
	static int16_t offsets[3]={880,930,1180};            /* Offsets by arity */

	/* Loop through clause items. */
	for (ptr1=&clause->part2.bin->formula[0];ptr1[0]!=UNITEND;
				ptr1=NextItem(ptr1,IMMED)){

		/* This item is not a variable. */
		if (0==(VARIABLE&ptr1[0])){

			/* Arity is bigger than three. Update vector component. */
			kk=(((symbol *)&ptr1[1])->symbol)->arity;
			if (kk>3){
				switch ((PREDICATE|FUNCTION)&ptr1[0]){
					case PREDICATE:
						mm=840+(SELECTED&ptr1[0]?3:0);
						switch ((FROMCONJECTURE|FROMNOTAXIOM)
								&(((symbol *)&ptr1[1])->symbol)->type){
							case FROMCONJECTURE:
							case FROMCONJECTURE|FROMNOTAXIOM:
								vector[mm]++;
								break;
							case FROMNOTAXIOM:
								vector[mm+1]++;
								break;
							case 0:
								vector[mm+2]++;
								break;
						}
						break;
					case FUNCTION:
						switch ((FROMCONJECTURE|FROMNOTAXIOM|SKOLEM)
								&(((symbol *)&ptr1[1])->symbol)->type){
							case FROMCONJECTURE:
							case FROMCONJECTURE|FROMNOTAXIOM:
								vector[846]++;
								break;
							case FROMNOTAXIOM:
								vector[847]++;
								break;
							case SKOLEM:
								vector[848]++;
								break;
							case 0:
								vector[849]++;
								break;
						}
						break;
				}

			/* Arity is less than or equal to three and bigger than zero. */
			} else if ((((symbol *)&ptr1[1])->symbol)->arity>0){

				/* Set variable vectidx to index offset of root arguments */
				/* of functions and non equality predicates. Set variables */
				/* pp and qq to a numbers in the range 1 to 5 corresponding */
				/* to the first two root arguments. These will be used to */
				/* calculate the index corresponding to equalities. */
				/* Loop through first level arguments. */
				pp=qq=0; /* Just to prevent compiler warnings. */
				for (vectidx=mm=0,jj=10,ptr2=NextItem(ptr1,IMMED);mm<kk;mm++,ptr2=NextItem(ptr2,OVERSUBTERMS)){
					switch ((FUNCTION|VARIABLE)&ptr2[0]){
						case FUNCTION:
							switch ((FROMCONJECTURE|FROMNOTAXIOM|SKOLEM)
									&(((symbol *)&ptr2[1])->symbol)->type){
								case FROMCONJECTURE:
								case FROMCONJECTURE|FROMNOTAXIOM:
									switch (mm){
										case 0:
											pp=1;
											break;
										case 1:
											qq=1;
											break;
									}
									break;
								case FROMNOTAXIOM:
									vectidx+=(jj*1);
									switch (mm){
										case 0:
											pp=2;
											break;
										case 1:
											qq=2;
											break;
									}
									break;
								case SKOLEM:
									vectidx+=(jj*2);
									switch (mm){
										case 0:
											pp=3;
											break;
										case 1:
											qq=3;
											break;
									}
									break;
								case 0:
									vectidx+=(jj*3);
									switch (mm){
										case 0:
											pp=4;
											break;
										case 1:
											qq=4;
											break;
									}
									break;
							}
							break;
						case VARIABLE:
							vectidx+=(jj*4);
							switch (mm){
								case 0:
									pp=5;
									break;
								case 1:
									qq=5;
									break;
							}
							break;
					}
					jj*=5;
				}

				/* Calculate vector index. */
				switch ((EQUALITY|PREDICATE|FUNCTION)&ptr1[0]){
					case EQUALITY:
						if (pp>=qq){
							vectidx=849+qq+((pp*(pp-1))/2);
						} else {
							vectidx=849+pp+((qq*(qq-1))/2);
						}
						if ((SELECTED|PMSELECTED)&ptr1[0]){
							vectidx+=15;
						}
						break;
					case PREDICATE:
						switch ((FROMCONJECTURE|FROMNOTAXIOM)
								&(((symbol *)&ptr1[1])->symbol)->type){
							case FROMCONJECTURE:
							case FROMCONJECTURE|FROMNOTAXIOM:
								vectidx+=offsets[mm];
								break;
							case FROMNOTAXIOM:
								vectidx+=(1+offsets[mm]);
								break;
							case 0:
								vectidx+=(2+offsets[mm]);
								break;
						}
						if (SELECTED&ptr1[0]){
							vectidx+=3;
						}
						break;
					case FUNCTION:
						switch ((FROMCONJECTURE|FROMNOTAXIOM|SKOLEM)
								&(((symbol *)&ptr1[1])->symbol)->type){
							case FROMCONJECTURE:
							case FROMCONJECTURE|FROMNOTAXIOM:
								vectidx+=(6+offsets[mm]);
								break;
							case FROMNOTAXIOM:
								vectidx+=(7+offsets[mm]);
								break;
							case SKOLEM:
								vectidx+=(8+offsets[mm]);
								break;
							case 0:
								vectidx+=(9+offsets[mm]);
								break;
						}
						break;
				}

				/* Update vector component. */
				vector[vectidx]++;
			}
		}
	}
	return;
} /* HorizontalFeatures */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Compute static features of a clause vector)  OCJ
 *
 *    This function computes the static features of clause and
 *    adds them to the clause vector. See GetClauseVector()
 *    function for the methodology and bibliography.
 *
 *    The following features are calculated:
 *    - Symbol count by polarity (13*2=26 features).
 *    - Clause length and number of positive and negative literals
 *      (3 features).
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Clause for which the vector will be computed.
 *    vector: Pointer to vector.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *
 *--------------------------------------------------------------*/
void StaticFeatures(cmprefix *clause,double *vector){
	auto uint8_t *ptr1;                                 /* Auxiliary pointer */
	auto int32_t ii;                                    /* Auxiliary */

	/* Loop through clause items. */
	ii=0; /* To prevent compiler warnings. */
	for (ptr1=&clause->part2.bin->formula[0];ptr1[0]!=UNITEND;
				ptr1=NextItem(ptr1,IMMED)){
		switch ((EQUALITY|PREDICATE|FUNCTION|VARIABLE)&ptr1[0]){
			case EQUALITY:

				/* Clause length. */
				vector[2466]++;

				/* Symbol and literal count, negative polarity. */
				if (NEGATED&ptr1[0]){
					ii=1;
					if ((SELECTED|PMSELECTED)&ptr1[0]){
						vector[2440]++;
					} else {
						vector[2441]++;
					}
					vector[2467]++;

				/* Symbol and literal count, positive polarity. */
				} else {
					ii=0;
					if ((SELECTED|PMSELECTED)&ptr1[0]){
						vector[2442]++;
					} else {
						vector[2443]++;
					}
					vector[2468]++;
				}
				break;
			case PREDICATE:

				/* Clause length. */
				vector[2466]++;

				/* Symbol and literal count, negative polarity. */
				if (NEGATED&ptr1[0]){
					ii=1;
					switch ((FROMCONJECTURE|FROMNOTAXIOM)&(((symbol *)&ptr1[1])->symbol)->type){
						case FROMCONJECTURE:
						case FROMCONJECTURE|FROMNOTAXIOM:
							if (SELECTED&ptr1[0]){
								vector[2444]++;
							} else {
								vector[2445]++;
							}
							break;
						case FROMNOTAXIOM:
							if (SELECTED&ptr1[0]){
								vector[2446]++;
							} else {
								vector[2447]++;
							}
							break;
						case 0:
							if (SELECTED&ptr1[0]){
								vector[2448]++;
							} else {
								vector[2449]++;
							}
							break;
					}
					vector[2467]++;

				/* Symbol and literal count, positive polarity. */
				} else {
					ii=0;
					switch ((FROMCONJECTURE|FROMNOTAXIOM)&(((symbol *)&ptr1[1])->symbol)->type){
						case FROMCONJECTURE:
						case FROMCONJECTURE|FROMNOTAXIOM:
							if (SELECTED&ptr1[0]){
								vector[2450]++;
							} else {
								vector[2451]++;
							}
							break;
						case FROMNOTAXIOM:
							if (SELECTED&ptr1[0]){
								vector[2452]++;
							} else {
								vector[2453]++;
							}
							break;
						case 0:
							if (SELECTED&ptr1[0]){
								vector[2454]++;
							} else {
								vector[2455]++;
							}
							break;
					}
					vector[2468]++;
				}
				break;
			case FUNCTION:

				/* Clause length. */
				vector[2466]++;

				/* Symbol count, negative polarity. */
				if (ii){
					switch ((FROMCONJECTURE|FROMNOTAXIOM|SKOLEM)&(((symbol *)&ptr1[1])->symbol)->type){
						case FROMCONJECTURE:
						case FROMCONJECTURE|FROMNOTAXIOM:
							vector[2456]++;
							break;
						case FROMNOTAXIOM:
							vector[2457]++;
							break;
						case SKOLEM:
							vector[2458]++;
							break;
						case 0:
							vector[2459]++;
							break;
					}

				/* Symbol count, positive polarity. */
				} else {
					switch ((FROMCONJECTURE|FROMNOTAXIOM|SKOLEM)&(((symbol *)&ptr1[1])->symbol)->type){
						case FROMCONJECTURE:
						case FROMCONJECTURE|FROMNOTAXIOM:
							vector[2460]++;
							break;
						case FROMNOTAXIOM:
							vector[2461]++;
							break;
						case SKOLEM:
							vector[2462]++;
							break;
						case 0:
							vector[2463]++;
							break;
					}
				}
				break;
			case VARIABLE:

				/* Symbol count, negative polarity. */
				if (ii){
					vector[2464]++;

				/* Symbol count, positive polarity. */
				} else {
					vector[2465]++;
				}
				break;
		}
	}
	return;
} /* StaticFeatures */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Extract learning information from solved problem)  OCJ
 *
 *    This function extracts learning information from a solved problem.
 *    It is assumed that a problem has been just solved and that a proof
 *    has been built so that formula used in the proof is marked with
 *    INPROOFQUEUE flag.
 *
 *    The function scans text, passive, active, unprocessed and locked
 *    clauses. For every clause with the PICKED flag a record is built
 *    and added to the learnto file.
 *
 *    LEARNTO FILE FORMAT:
 *    -------------------
 *      block1 block2 ... blockn
 *
 *    where:
 *      blocki: A block corresponding to records of a specific type.
 *
 *    BLOCK FORMAT:
 *    ------------
 *      length goodchoices [header flag1 vector1 flag2 vector2 ... flagm vectorm]
 *
 *    where:
 *      length: 64 bit unsigned integer with the number of vectors in block.
 *      goodchoices: 64 bit unsigned integer with the number of vectors
 *                   in block corresponding to good clauses.
 *      header: 64 bit unsigned integer with the type of block. See header field
 *              of learnrecord structure for details about types of block. This
 *              field may be missing if length=0.
 *      flagi: 8 bit integer with one if feature vector belongs to
 *             a good clause, 0 otherwise. These fields may be missing if
 *             length=0.
 *      vectori: Feature vector in the block. It has VECTORLENGTH double
 *               floating point format numbers. It is also a learnvector
 *               structure. These block fields may be missing if length=0.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *
 *--------------------------------------------------------------*/
void LearnFromProblem(){
	auto FILE *file;                                    /* Learn to file handle */
	auto uint64_t header;                               /* Common part of the header */
	auto double *vector;                                /* Pointer to clause vector */
	auto cmprefix *tmpclause;                           /* Temporary binary clause built from a text clause. */
	auto cmprefix *ptr1;                                /* Auxiliary pointer */
	auto char *ptr2,*ptr3,*ptr4,*ptr5;                  /* Auxiliary pointers */
	auto char *ptr6,*ptr7,*ptr8;                        /* Auxiliary pointers */
	auto uint64_t ii,jj,gg,mm,nn,pp;                    /* Auxiliary */
	auto int8_t kk;                                     /* Auxiliary */

	/* Set prefix for printed lines. */
	if (pgmmode==NONINTERACTIVE){
		ptr4="% ";
	} else {
		ptr4="";
	}

	/* Get address of pointer to KB. No learning is extracted */
	/* if the problem was solved during pre-processing. */
	if (procctl->solvekb==active_cores){
		return;
	}

	/* Build the header according to the strategy parameters of the winning */
	/* process and the type of problem. */
	header=0;
	if (kbset.opts.algorithm==DISCOUNT){
		header|=DISCOUNT_P;
	}
	switch (kbset.opts.select){
		case MULTINEG:
			header|=MULTINEG_P;
			break;
		case MAXIMAL:
			header|=MAXIMAL_P;
			break;
		case SINGLEPOS:
			header|=SINGLEPOS_P;
			break;
		case MULTIPOS:
			header|=MULTIPOS_P;
			break;
	}
	if (kbset.opts.termord==NONRECURSIVE){
		header|=TRMORDNR_P;
	}
	switch (kbset.opts.litord){
		case NONRECURSIVE:
			header|=LITORDNR_P;
			break;
		case LEXICOGRAPHIC:
			header|=LITORDLEX_P;
			break;
	}
	if (kbset.opts.fweight==ARITY){
		header|=ARITY_P;
	}
	ii=pbmtype;
	ii<<=56;
	header|=ii;

	/* Open learnto file. */
	printf("%s**********************************************\n",ptr4);
	printf("%sExtracting learning information from problem...\n",ptr4);
	if (learnnew){
		if (NULL==(file=fopen(learnto,"w+b"))){
			printf("%s%s opening file %s.\n",ptr4,strerror(errno),learnto);
			printf("%sLearning extraction not completed.\n",ptr4);
			printf("%s**********************************************\n",ptr4);
			return;
		}
	} else {
		if (NULL==(file=fopen(learnto,"r+b"))){
			if (NULL==(file=fopen(learnto,"w+b"))){
				printf("%s%s opening file %s.\n",ptr4,strerror(errno),learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
		}
	}

	/* Get file length. */
	fseek(file,0,SEEK_END);
	jj=ftell(file);

	/* Scan file until last block. The variable mm is set to */
	/* the offset of the block to which data will be added. */
	/* The variable nn is set to the number of vectors initially */
	/* in the block to which the new vectors will be added. */
	/* The file position indicator will be set to the point */
	/* where the flag/vector pairs will be written. */
	mm=0;
	fseek(file,0,SEEK_SET);
	while (1){

		/* Read number of vectors in block. */
		if (0==fread(&nn,sizeof(nn),1,file)){

			/* Read error. */
			if (ferror(file)){
				fclose(file);
				printf("%s%s reading file %s.\n",ptr4,strerror(errno),learnto);
				printf("%sLearning extraction not performed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}

			/* Empty file. */
			fseek(file,0,SEEK_SET);
			nn=gg=0;
			if (1>fwrite(&nn,sizeof(nn),1,file)){
				fclose(file);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not performed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (1>fwrite(&gg,sizeof(gg),1,file)){
				fclose(file);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not performed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			break;

		/* Read number of good vectors in block. */
		} else if (0==fread(&gg,sizeof(gg),1,file)){
			if (ferror(file)){
				printf("%s%s reading file %s.\n",ptr4,strerror(errno),learnto);
			}
			fclose(file);
			printf("%s%s reading file %s.\n",ptr4,strerror(errno),learnto);
			printf("%sLearning extraction not performed.\n",ptr4);
			printf("%s**********************************************\n",ptr4);
			return;
		}

		/* Validate block. */
		pp=(nn*(1+(VECTORLENGTH*sizeof(double))))+(2*sizeof(uint64_t))+(nn?sizeof(uint64_t):0);
		if ((gg>nn)||(((mm+pp)!=jj)&&((mm+pp+(2*sizeof(uint64_t)))>jj))){
			fclose(file);
			printf("%sInvalid file %s.\n",ptr4,learnto);
			printf("%sLearning extraction not performed.\n",ptr4);
			printf("%s**********************************************\n",ptr4);
			return;
		}

		/* Last block. */
		if ((mm+pp)==jj){

			/* Non empty block. */
			if (nn!=0){

				/* Read block header and move to the end of the file. */
				if (0==fread(&ii,sizeof(ii),1,file)){
					fclose(file);
					printf("%sError reading file %s.\n",ptr4,learnto);
					printf("%sLearning extraction not performed.\n",ptr4);
					printf("%s**********************************************\n",ptr4);
					return;
				}
				fseek(file,0,SEEK_END);

				/* If headers doesn't match then start a new block and update */
				/* offset of the block to which the data will be added. */
				if (ii!=header){
					nn=gg=0;
					if (1>fwrite(&nn,sizeof(nn),1,file)){
						fclose(file);
						printf("%sError writing to file %s.\n",ptr4,learnto);
						printf("%sLearning extraction not performed.\n",ptr4);
						printf("%s**********************************************\n",ptr4);
						return;
					}
					if (1>fwrite(&gg,sizeof(gg),1,file)){
						fclose(file);
						printf("%sError writing to file %s.\n",ptr4,learnto);
						printf("%sLearning extraction not performed.\n",ptr4);
						printf("%s**********************************************\n",ptr4);
						return;
					}
					mm=jj;
				}
			}
			break;
		}

		/* Prepare next iteration. */
		fseek(file,pp-(2*sizeof(uint64_t)),SEEK_CUR);
		mm+=pp;
	}

	/* Loop through text clauses. */
	ii=0; /* Number of vectors added to block. */
	for (ptr1=kbset.firsttxt;ptr1!=NULL;ptr1=ptr1->next){
		if (PICKED&ptr1->flags){

			/* Write header if first record. Update number of written records. */
			if ((ii==0)&&(nn==0)){
				if (1>fwrite(&header,sizeof(header),1,file)){
					fclose(file);
					printf("%sError writing to file %s.\n",ptr4,learnto);
					printf("%sLearning extraction not performed.\n",ptr4);
					printf("%s**********************************************\n",ptr4);
					return;
				}
			}
			ii++;

			/* Remove the possible clause assertions. */
			ptr2=ptr1->part2.text;
			while (1){
				ptr3=strchr(ptr2,'\'');
				ptr8=strchr(ptr2,'"');
				if (ptr3==NULL){
					if (ptr8==NULL){
						break;
					}
					ptr2=ptr8+1;
				} else {
					if ((ptr8==NULL)||(ptr3>ptr8)){
						ptr2=ptr3+1;
					} else {
						ptr2=ptr8+1;
					}
				}
			}
			if ((proofmode==DRODI)||(syntaxmode==DRODI)){
				if (NULL!=(ptr3=strchr(ptr2,'<'))){
					jj=*(ptr3-1);
					*(ptr3-1)=0;
				}
			} else {
				while (NULL!=(ptr3=strchr(ptr2,'|'))){
					if (ptr3[1]=='~'){
						ptr7=ptr3+1;
					} else {
						ptr7=ptr3;
					}
					if ((strlen(ptr7)>5)&&(0==strncmp(splprefix,&ptr7[1],3))&&(NULL!=(ptr5=strchr(&ptr7[4],'_')))){
						kk=strtol(&ptr7[4],&ptr6,10);
						if ((ptr5==ptr6)&&(kk==splid)){
							ptr3[0]=0;
							break;
						}
					}
					ptr2=ptr3+1;
				}
			}

			/* Compile text clause. Variables must be renumbered as this is needed */
			/* because SelectItems() will be called by GetClauseVector(). */
			if (NOMEMORY==(CompileClause(ptr1->part2.text,&tmpclause,NULL))){
				fclose(file);
				printf("%sNot enough memory, learning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			tmpclause->flags=ptr1->flags;
			RenumberVars(tmpclause->part2.bin);

			/* Restore the possibly removed clause assertions. */
			if (ptr3!=NULL){
				if ((proofmode==DRODI)||(syntaxmode==DRODI)){
					*(ptr3-1)=jj;
				} else {
					ptr3[0]='|';
				}
			}

			/* Save learn record. */
			if (INPROOFQUEUE&ptr1->flags){
				kk=1;
				gg++;
			} else {
				kk=0;
			}
			if (1>fwrite(&kk,sizeof(kk),1,file)){
				fclose(file);
				MYFREE(tmpclause->part2.bin);
				MYFREE(tmpclause);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (NULL==(vector=GetClauseVector(tmpclause))){
				fclose(file);
				FreeEquData(tmpclause->part2.bin);
				MYFREE(tmpclause->part2.bin);
				MYFREE(tmpclause);
				printf("%sNot enough memory, learning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (VECTORLENGTH>fwrite(vector,sizeof(vector[0]),VECTORLENGTH,file)){
				fclose(file);
				FreeEquData(tmpclause->part2.bin);
				MYFREE(tmpclause->part2.bin);
				MYFREE(tmpclause);
				MYFREE(vector);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			FreeEquData(tmpclause->part2.bin);
			MYFREE(tmpclause->part2.bin);
			MYFREE(tmpclause);
			MYFREE(vector);
		}
	}

	/* Loop through unprocessed clauses and save learn records. */
	for (ptr1=kbset.frstunproc;ptr1!=NULL;ptr1=ptr1->next){
		if (PICKED&ptr1->flags){

			/* Write header if first record. Update number of written records. */
			if ((ii==0)&&(nn==0)){
				if (1>fwrite(&header,sizeof(header),1,file)){
					fclose(file);
					printf("%sError writing to file %s.\n",ptr4,learnto);
					printf("%sLearning extraction not performed.\n",ptr4);
					printf("%s**********************************************\n",ptr4);
					return;
				}
			}
			ii++;

			/* Save learn record. */
			if (INPROOFQUEUE&ptr1->flags){
				kk=1;
				gg++;
			} else {
				kk=0;
			}
			if (1>fwrite(&kk,sizeof(kk),1,file)){
				fclose(file);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (NULL==(vector=GetClauseVector(ptr1))){
				fclose(file);
				printf("%sNot enough memory, learning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (VECTORLENGTH>fwrite(vector,sizeof(vector[0]),VECTORLENGTH,file)){
				fclose(file);
				MYFREE(vector);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			MYFREE(vector);
		}
	}

	/* Loop through passive clauses and save learn records. */
	for (ptr1=kbset.frstpassive;ptr1!=NULL;ptr1=ptr1->next){
		if (PICKED&ptr1->flags){

			/* Write header if first record. Update number of written records. */
			if ((ii==0)&&(nn==0)){
				if (1>fwrite(&header,sizeof(header),1,file)){
					fclose(file);
					printf("%sError writing to file %s.\n",ptr4,learnto);
					printf("%sLearning extraction not performed.\n",ptr4);
					printf("%s**********************************************\n",ptr4);
					return;
				}
			}
			ii++;

			/* Save learn record. */
			if (INPROOFQUEUE&ptr1->flags){
				kk=1;
				gg++;
			} else {
				kk=0;
			}
			if (1>fwrite(&kk,sizeof(kk),1,file)){
				fclose(file);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (NULL==(vector=GetClauseVector(ptr1))){
				fclose(file);
				printf("%sNot enough memory, learning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (VECTORLENGTH>fwrite(vector,sizeof(vector[0]),VECTORLENGTH,file)){
				fclose(file);
				MYFREE(vector);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			MYFREE(vector);
		}
	}

	/* Loop through active clauses and save learn records. */
	for (ptr1=kbset.frstactive;ptr1!=NULL;ptr1=ptr1->next){
		if (PICKED&ptr1->flags){

			/* Write header if first record. Update number of written records. */
			if ((ii==0)&&(nn==0)){
				if (1>fwrite(&header,sizeof(header),1,file)){
					fclose(file);
					printf("%sError writing to file %s.\n",ptr4,learnto);
					printf("%sLearning extraction not performed.\n",ptr4);
					printf("%s**********************************************\n",ptr4);
					return;
				}
			}
			ii++;

			/* Save learn record. */
			if (INPROOFQUEUE&ptr1->flags){
				kk=1;
				gg++;
			} else {
				kk=0;
			}
			if (1>fwrite(&kk,sizeof(kk),1,file)){
				fclose(file);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (NULL==(vector=GetClauseVector(ptr1))){
				fclose(file);
				printf("%sNot enough memory, learning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (VECTORLENGTH>fwrite(vector,sizeof(vector[0]),VECTORLENGTH,file)){
				fclose(file);
				MYFREE(vector);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			MYFREE(vector);
		}
	}

	/* Loop through locked clauses and save learn records. */
	for (ptr1=kbset.frstlocked;ptr1!=NULL;ptr1=ptr1->next){
		if (PICKED&ptr1->flags){

			/* Write header if first record. Update number of written records. */
			if ((ii==0)&&(nn==0)){
				if (1>fwrite(&header,sizeof(header),1,file)){
					fclose(file);
					printf("%sError writing to file %s.\n",ptr4,learnto);
					printf("%sLearning extraction not performed.\n",ptr4);
					printf("%s**********************************************\n",ptr4);
					return;
				}
			}
			ii++;

			/* Save learn record. */
			if (INPROOFQUEUE&ptr1->flags){
				kk=1;
				gg++;
			} else {
				kk=0;
			}
			if (1>fwrite(&kk,sizeof(kk),1,file)){
				fclose(file);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (NULL==(vector=GetClauseVector(ptr1))){
				fclose(file);
				printf("%sNot enough memory, learning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			if (VECTORLENGTH>fwrite(vector,sizeof(vector[0]),VECTORLENGTH,file)){
				fclose(file);
				MYFREE(vector);
				printf("%sError writing to file %s.\n",ptr4,learnto);
				printf("%sLearning extraction not completed.\n",ptr4);
				printf("%s**********************************************\n",ptr4);
				return;
			}
			MYFREE(vector);
		}
	}

	/* Update block length and number of good vectors. */
	fseek(file,mm,SEEK_SET);
	nn+=ii;
	if (1>fwrite(&nn,sizeof(nn),1,file)){
		fclose(file);
		printf("%sError writing to file %s.\n",ptr4,learnto);
		printf("%sLearning extraction not completed.\n",ptr4);
		printf("%s**********************************************\n",ptr4);
		return;
	}
	if (1>fwrite(&gg,sizeof(gg),1,file)){
		fclose(file);
		printf("%sError writing to file %s.\n",ptr4,learnto);
		printf("%sLearning extraction not completed.\n",ptr4);
		printf("%s**********************************************\n",ptr4);
		return;
	}

	/* Close file and return. */
	fclose(file);
	printf("%s%ld records added to file %s\n",ptr4,ii,learnto);
	printf("%s**********************************************\n",ptr4);
	learnnew=0;
	return;
} /* LearnFromProblem */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Read learn records generated by PROCLEARN command)  OCJ
 *
 *    This function reads learn records generated by PROCLEARN command
 *    and initializes lrnrecords and numlrnrecs global variables. Only
 *    learn records with a header that matches the current problem type
 *    will be loaded.
 *
 *    It is assumed that learnfrom is not NULL but this is not verified.
 *
 *    If the file pointed by learnfrom is not valid then learnfrom
 *    global variable is freed and set to NULL, thus setting the
 *    LEARNFROM option to OFF.
 *
 *    It is the caller's responsibility to free memory allocated to
 *    lrnrecords.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    0 if OK or 1 of error or if there are no appropriate records
 *    for the current problem type.
 *
 *
 *
 *--------------------------------------------------------------*/
int32_t ReadLearnRecords(void){
	auto FILE *filein;                                   /* Input file handle */
	auto learnrecord *ptr1;                              /* Auxiliary pointer */
	auto char *ptr2;                                     /* Auxiliary pointer */
	auto uint64_t ii;                                    /* Auxiliary */

	/* Set prefix for printed lines. */
	if (pgmmode==NONINTERACTIVE){
		ptr2="% ";
	} else {
		ptr2="";
	}

	/* Open input file. */
	if (verbose){
		printf("%sReading learning file %s\n",ptr2,learnfrom);
	}
	if (lrnrecords!=NULL){
		MYFREE(lrnrecords);
		lrnrecords=NULL;
	}
	if (NULL==(filein=fopen(learnfrom,"rb"))){
		printf("%sLEARNFROM error: %s opening input file %s.\n",ptr2,strerror(errno),learnfrom);
		MYFREE(learnfrom);
		learnfrom=NULL;
		return(1);
	}

	/* Check infile length. */
	fseek(filein,0,SEEK_END);
	ii=ftell(filein);
	fseek(filein,0,SEEK_SET);
	if ((0==ii)||(0!=(ii%sizeof(learnrecord)))){
		fclose(filein);
		printf("%sLEARNFROM error: invalid %s input file length.\n",ptr2,learnfrom);
		MYFREE(learnfrom);
		learnfrom=NULL;
		return(1);
	}

	/* Allocate space for lrnrecords and initialize numlrnrecs. */
	/* The space allocated for lrnrecords is enough to fit all */
	/* possible records of one problem type. */
	if (NULL==(ptr1=mallocmn(120*sizeof(learnrecord)))){
		fclose(filein);
		printf("%sLEARNFROM error: not enough memory.\n",ptr2);
		MYFREE(learnfrom);
		learnfrom=NULL;
		return(1);
	}
	lrnrecords=ptr1;

	/* Read and validate learning records and close input file. */
	for (numlrnrecs=0;numlrnrecs<120;){

		/* Read learning record. */
		if (0==fread(&lrnrecords[numlrnrecs],sizeof(lrnrecords[0]),1,filein)){
			if (ferror(filein)){
				printf("%sLEARNFROM error: %s error reading input file %s.\n",ptr2,strerror(errno),learnfrom);
				MYFREE(learnfrom);
				MYFREE(lrnrecords);
				learnfrom=NULL;
				lrnrecords=NULL;
				return(1);
			}
			break;
		}

		/* Validate learning record. */
		if (((0xffffffffffffff&lrnrecords[numlrnrecs].header)!=((ALGMASK|SELECTMASK|TERMORDRMASK|
				LITORDRMASK|FUNCWMASK)&lrnrecords[numlrnrecs].header))
				||((0xff00000000000000&lrnrecords[numlrnrecs].header)>0x700000000000000)){
			printf("%sLEARNFROM error: invalid %s input file.\n",ptr2,learnfrom);
			MYFREE(learnfrom);
			MYFREE(lrnrecords);
			learnfrom=NULL;
			lrnrecords=NULL;
			return(1);
			break;
		}

		/* Check problem type match. */
		ii=lrnrecords[numlrnrecs].header>>56;
		if (ii==pbmtype){
			numlrnrecs++;
		}
	}

	/* Close file, reallocate lrnrecords to exact size and return. */
	fclose(filein);
	if (numlrnrecs==0){
		MYFREE(lrnrecords);
		lrnrecords=NULL;
		return(1);
	} else {
		if (NULL!=(ptr1=reallocmn(lrnrecords,numlrnrecs*sizeof(learnrecord)))){
			lrnrecords=ptr1;
		}
	}
	return(0);
} /* ReadLearnRecords */
