/*
 ============================================================================
 Name        : multith.c
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */

/****************************************************************
*
*               multith (Multi-processing functions module)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

/*-------------------------------------------------------------
 *
 *  DESCRIPTION:
 *
 *    This source contains the multi-processing functions, saturation algorithms
 *    and other related functions of the Drodi package.
 *
 *    Drodi is a multi-platform Open Source Artificial Intelligence application.
 *    See Drodi.c for additional information.
 *
 *
 *  IMPLEMENTATION:
 *
 *    Multi-platform
 *
 *  NOTES:
 *
 *
 *--------------------------------------------------------------*/

/* Include for all C sources */
#include "global.h"

/** Global variables for this module: only those that need initialization in their definition. */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Start saturation algorithms)  OCJ
 *
 *    This function perform the following tasks:
 *    - Converts the input formulas to conjunctive normal form, compiles
 *      the produced clauses and adds them to the main KB.
 *    - Makes copies of the resulting KB.
 *    - Start a saturation algorithm with each KB. Each algorithm is run
 *      in a separate process and there are as many processes as available
 *      cores.
 *    - Prints the resulting output.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void Saturate(void){
	auto uint64_t avmemory;                          /* Physical memory available */
	auto clock_t start;                              /* CPU ticks at start of child process */
	auto struct timespec starttime;                  /* To specify wait time for semaphores */
	auto struct timespec endtime;                    /* To specify wait time for semaphores */
	auto int32_t timeleft;                           /* Time left after a proof was found */
	auto struct itimerval alarmtime;                 /* To define a global timeout alarm */
	auto char *ptr1,*ptr2;                           /* Auxiliary pointers */
	#if PRECEDENCE != 0
	auto hashchain *ptr3;                            /* Auxiliary pointer */
	#endif
	auto float dd;                                   /* Auxiliary */
	auto int32_t ii,jj,kk;                           /* Auxiliary */

	#ifndef DEBUGTREE
	#ifdef DEBUGTREE2
	/* Discrimination trees query check. */
	if (NULL==mainkb.frstunproc){
		CheckDscTQueries();
		return;
	}
	#endif
	#endif

	/* Start time measurement. */
	start=clock();
	gettimeofday(&mainkb.time,NULL);

	/* Main KB is empty. */
	if (mainkb.nformulas==0){
		if (pgmmode==NONINTERACTIVE){
			printf("%% SZS status InputError : ");
		}
		printf("There are no input formulas to process.\n");
		return;
	}

	#if PRECEDENCE != 0
	/* Compute the final symbol precedence. */
	for (ptr3=kb_hash;ptr3!=NULL;ptr3=ptr3->nextelem){
		ptr3->precedence+=(numsymbols*(maxusecount-ptr3->usecount));
	}
	#endif

	/* Initialize problem used memory and perform pre-processing. */
	ResetStats(pbmstats);
	procctl->solvekb=-1;
	ii=Preprocess(1);
	if (ii){
		if (ii==NOMEMORY){
			if (pgmmode==NONINTERACTIVE){
				printf("%% SZS status MemoryOut : ");
			}
			printf("Not enough memory.\n");
		} else {
			if (pgmmode==NONINTERACTIVE){
				printf("%% SZS status Timeout : ");
			}
			printf("Process timeout\n");
		}
		FreeInfFormulas();
		if ((NOMEMORY==Initialize_KB(&mainkb))&&(pgmmode==INTERACTIVE)){
			printf("No memory available, KB not initialized\n");
		}
		ResetStats(pbmstats);
		procctl->status=STOPPED;
		tot_pbms++;
		procctl->end=clock();
		tot_cpu_time+=((double)(procctl->end-procctl->start))/CLOCKS_PER_SEC;
		return;
	}

	/* Start time measurement after preprocessing. */
	gettimeofday(&mainkb.time,NULL);

	/* Initialize common semaphores. They are initialized in any case because otherwise each */
	/* time a resource in common memory is accessed an additional test would be necessary. */
	sem_init(&procctl->procctl,1,1);
	sem_init(&procctl->statistics,1,1);
	sem_init(&procctl->strategy,1,1);
	sem_init(&procctl->sync,1,0);

	/* Problem was not solved in the pre-processing. */
	jj=active_cores-1;
	*secprc_cpu_time=0;
	if (procctl->status==STOPPED){

		/* Select strategies. */
		procctl->status=RUNNING;
		if (0!=InitStrategies()){
			if (pgmmode==NONINTERACTIVE){
				printf("%% SZS status MemoryOut : Not enough memory available\n");
			} else {
				printf("Cannot process: Not enough memory available\n");
			}
			return;
		}

		/* Single process case. */
		if (active_cores==1){

			/* Set alarm for global timeout detection in the rare case */
			/* in which a timeout is not caught by normal means. */
			gettimeofday(&mainkb.endtime,NULL);
			dd=EXITWAITSECONDS+(TMOWAITMICROSECS/1000000.0)+glblopts.timeout-mainkb.prstats.elapsed_time
					-mainkb.endtime.tv_sec+mainkb.time.tv_sec
					-((mainkb.endtime.tv_usec-mainkb.time.tv_usec)/1000000.0);
			alarmtime.it_interval.tv_sec=alarmtime.it_interval.tv_usec=0;
			alarmtime.it_value.tv_sec=(int32_t)dd;
			alarmtime.it_value.tv_usec=(int32_t)((dd-(int32_t)dd)*1000000.0);
			setitimer(ITIMER_REAL,&alarmtime,NULL);

			/* Initialize maximum process memory and start inference process. */
			avmemory=GetAvailableMemory();
			if (avmemory>=600000000){
				maxmemory=allcdmemory+((avmemory-300000000)/active_cores);
			} else {
				if (avmemory>=5000000){
					maxmemory=allcdmemory+(avmemory/(2*active_cores));
				} else {
					procctl->status=NOMEMORY;
					maxmemory=allcdmemory;
				}
			}
			if (maxmemory>memorylimit){
				maxmemory=memorylimit;
			}
			procnb=0;
			if (procctl->status==RUNNING){
				InferenceProcess();
			}

		/* Multi-process case. */
		} else {

			/* Ensure that there are not pending output so that it is not printed */
			/* several times by the children processes. */
			fflush(stdout);

			/* Initialize process specific semaphores and spawn as many processes */
			/* as number of enabled cores minus one (parent process counts). */
			/* The child process cpu time is set to zero. */
			for (procnb=0;procnb<jj;procnb++){
				sem_init(&procsem[procnb],1,0);
				if (0==(pids[procnb]=fork())){
					cpu_time=0;
					allcdmemory-=mmapsize;
					break;
				} else if (pids[procnb]==-1){
					procctl->status=NOMEMORY;
					break;
				}
			}

			/* Parent process. */
			if (procnb==jj){

				/* Wait until all child processes are ready to run, initialize maximum */
				/* process memory and allow child processes to run. */
				for (ii=0;ii<jj;ii++){
					if (pids[ii]!=-1){
						sem_wait(&procsem[ii]);
						sem_post(&procsem[ii]);
					}
				}
				if (procctl->status==RUNNING){
					avmemory=GetAvailableMemory();
					if (avmemory>=600000000){
						*maxprcmemory=allcdmemory+((avmemory-300000000)/(procnb+1));
					} else {
						if (avmemory>=5000000){
							*maxprcmemory=allcdmemory+(avmemory/(2*(procnb+1)));
						} else {
							procctl->status=NOMEMORY;
							*maxprcmemory=allcdmemory;
						}
					}
					if ((*maxprcmemory)>memorylimit){
						*maxprcmemory=memorylimit;
					}
				}
				sem_post(&procctl->sync);

				/* Set alarm for global timeout detection in the rare case */
				/* in which a timeout is not caught by normal means. */
				gettimeofday(&mainkb.endtime,NULL);
				dd=EXITWAITSECONDS+(TMOWAITMICROSECS/1000000.0)+glblopts.timeout-mainkb.prstats.elapsed_time
						-mainkb.endtime.tv_sec+mainkb.time.tv_sec
						-((mainkb.endtime.tv_usec-mainkb.time.tv_usec)/1000000.0);
				alarmtime.it_interval.tv_sec=alarmtime.it_interval.tv_usec=0;
				alarmtime.it_value.tv_sec=(int32_t)dd;
				alarmtime.it_value.tv_usec=(int32_t)((dd-(int32_t)dd)*1000000.0);
				setitimer(ITIMER_REAL,&alarmtime,NULL);

			/* Child process. Indicate that process is ready to run and wait until parent */
			/* process allows us to run. */
			} else {
				start=clock();
				sem_post(&procsem[procnb]);
				sem_wait(&procctl->sync);
				sem_post(&procctl->sync);
				sem_wait(&procsem[procnb]);
			}

			/* Start inference process. */
			if (procctl->status==RUNNING){
				maxmemory=*maxprcmemory;
				InferenceProcess();
			}

			/* If this is the process that solved the problem then indicate */
			/* that proof building and printing is about to start. */
			if (procctl->solvekb==procnb){
				procctl->prfprntinproc=1;
			}
		}

	/* Problem was solved in the pre-processing. */
	} else {
		procnb=jj;
		pids[0]=0; /* Indicate that there are no children processes now. */
	}

	/* Disable alarm for global timeouts detection. */
	if (procnb==jj){
		alarmtime.it_interval.tv_sec=alarmtime.it_interval.tv_usec=0;
		alarmtime.it_value.tv_sec=alarmtime.it_value.tv_usec=0;
		setitimer(ITIMER_REAL,&alarmtime,NULL);
	}

	/* Print problem output if appropriate. */
	if ((procctl->solvekb==procnb)||((procnb==jj)&&((procctl->solvekb<0)||(procctl->solvekb==active_cores)))){

		/* Display KB data.*/
		if (prtkbdata){
			PrintKBData();
		}

		/* Display statistics. The PrintStats() function initializes the working KB's. */
		PrintStats(0);
	}

	/* Collect memory use information for the problem. */
	sem_wait(&procctl->statistics);
	pbmstats->memory+=mainkb.prstats.memory;
	sem_post(&procctl->statistics);

	/* This is a child process. */
	if (procnb!=jj){

		/* Stop and accumulate child process cpu time measurement. */
		procctl->end=clock();
		cpu_time=((double)(clock()-start))/CLOCKS_PER_SEC;
		sem_wait(&procctl->statistics);
		*secprc_cpu_time+=cpu_time;
		sem_post(&procctl->statistics);

		/* Clean for exit and exit. */
		Clean4Exit();
		exit(0);
	}

	/* IF WE ARE HERE WE ARE THE PARENT PROCESS. */
	/* If this execution was multi-pocess then wait with time limit to have a consolidated */
	/* information from all child processes. */
	if ((active_cores>1)&&(procctl->solvekb!=active_cores)){
		clock_gettime(CLOCK_REALTIME,&starttime);
		gettimeofday(&mainkb.endtime,NULL);
		timeleft=(int32_t)(EXITWAITSECONDS+glblopts.timeout-mainkb.prstats.elapsed_time
				-mainkb.endtime.tv_sec+mainkb.time.tv_sec);
		for (ii=0;ii<jj;ii++){
			endtime=starttime;
			if ((procctl->solvekb==ii)&&(procctl->prfprntinproc)){
				endtime.tv_sec+=timeleft;
			} else {
				endtime.tv_sec+=EXITWAITSECONDS;
			}
			sem_timedwait(&procsem[ii],&endtime);
		}
	}

	/* Stop and accumulate time measurement. */
	procctl->end=clock();
	cpu_time+=*secprc_cpu_time+(((double)(procctl->end-start))/CLOCKS_PER_SEC);
	tot_cpu_time+=cpu_time;
	gettimeofday(&mainkb.endtime,NULL);
	mainkb.prstats.elapsed_time+=(mainkb.endtime.tv_sec-mainkb.time.tv_sec
			+(mainkb.endtime.tv_usec-mainkb.time.tv_usec)/1000000.0);
	pbmstats->elapsed_time=mainkb.prstats.elapsed_time;

	/* Print theorem total statistics. */
	if (pgmmode==NONINTERACTIVE){
		ptr2="% ";
	} else {
		ptr2="";
	}
	if (verbose){
		printf("%s\n%sTheorem statistics:\n",ptr2,ptr2);
		printf("%sNumber of processes %d:\n",ptr2,active_cores);
	}
	PrintStatData(pbmstats,0);
	printf("%sCPU time: %.6f seconds\n",ptr2,cpu_time);

	/* Print theorem memory use information. */
	dd=FormatMemory(pbmstats->memory,&ptr1);
	printf("%sMemory used: %.3f %s\n",ptr2,dd,ptr1);

	/* Accumulate global statistics. After the PrintStats(0) call */
	/* the main KB has the current problem accumulated statistics. */
	tot_pbms++;
	if ((procctl->status==SATISFIABLE)||(procctl->status==UNSATISFIABLE)){
		solved_pbms++;
	}
	AddInferenceStats(pbmstats,&glblstats);
	glblstats.elapsed_time+=mainkb.prstats.elapsed_time;
	if (glblstats.memory<pbmstats->memory){
		glblstats.memory=pbmstats->memory;
	}

	/* Memory problems DEBUG. */
	#ifdef DEBUGMEMORY
	for (ii=0;ii<ndemods;ii++){
		if (memtrack[ii]==1){
			printf("Memory allocation number %d not freed.\n",ii);
		}
	}
	printf("Number of memory allocation calls: %d\n",ndemods);
	#endif

	/* Initialize main KB. */
	FreeInfFormulas();
	if (NOMEMORY==Initialize_KB(&mainkb)){
		printf("No memory available, KB not initialized\n");
	}
	ResetStats(&mainkb.prstats);
	procctl->status=STOPPED;

	/* If this execution was multi-pocess then wait for child processes */
	/* termination and destroy child specific semaphore. */
	if ((active_cores>1)&&(procctl->solvekb!=active_cores)){
		endtime.tv_sec=0;
		endtime.tv_nsec=10000000;
		for (ii=kk=0;ii<jj;ii++){
			if (kk>KILLWAITITERATIONS){
				if (0==waitpid(pids[ii],NULL,WNOHANG)){
					kill(pids[ii],SIGTERM);
					waitpid(pids[ii],NULL,WNOHANG);
				}
			} else {
				for (;kk<=KILLWAITITERATIONS;kk++){
					if (0!=waitpid(pids[ii],NULL,WNOHANG)){
						break;
					}
					nanosleep(&endtime,NULL);
				}
				if (kk>KILLWAITITERATIONS){
					kill(pids[ii],SIGTERM);
					waitpid(pids[ii],NULL,WNOHANG);
				}
			}
			sem_destroy(&procsem[ii]);
		}
		pids[0]=0; /* Indicate that there are no children processes now. */
	}

	/* Destroy common semaphores and return. */
	sem_destroy(&procctl->procctl);
	sem_destroy(&procctl->statistics);
	sem_destroy(&procctl->strategy);
	sem_destroy(&procctl->sync);
	return;
} /* Saturate */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Set strategy specific KB parameters)  OCJ
 *
 *    This function sets strategy specific parameters for a given
 *    strategy. See InitStrategies() general comments for details
 *    on how strategies are selected for each problem type.
 *
 *
 *  ARGUMENTS:
 *
 *    strategyidx: Index to strategies[] set.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void SetMTParams(int32_t strategyidx){
	auto int32_t mm;                                     /* Auxiliary */

	/* Set algorithm. */
	stratflgs=strategies[strategyidx];
	if (OTTER_P==(ALGMASK&stratflgs)){
		kbset.opts.algorithm=OTTER;
	} else {
		kbset.opts.algorithm=DISCOUNT;
	}

	/* Set literal selection options. */
	if (SINGLENEG_P==(SELECTMASK&stratflgs)){
		kbset.opts.select=SINGLENEG;
	} else if (MULTINEG_P==(SELECTMASK&stratflgs)){
		kbset.opts.select=MULTINEG;
	} else if (SINGLEPOS_P==(SELECTMASK&stratflgs)){
		kbset.opts.select=SINGLEPOS;
	} else if (MULTIPOS_P==(SELECTMASK&stratflgs)){
		kbset.opts.select=MULTIPOS;
	} else {
		kbset.opts.select=MAXIMAL;
	}

	/* Set weight factors option for clause weight calculation. */
	if (WFACT1_P==(WFACTSETMASK&stratflgs)){
		kbset.opts.factoridx=1;
	} else {
		kbset.opts.factoridx=2;
	}

	/* Set demodulation option. */
	if (DEMODON_P==(DEMODMASK&stratflgs)){
		kbset.opts.demodulation=DEMODON;
	} else if (DEMODOFF_P==(DEMODMASK&stratflgs)){
		kbset.opts.demodulation=DEMODOFF;
	} else {
		kbset.opts.demodulation=DEMODCPL;
	}

	/* Set factoring inferences enabled/disabled. */
	if (FACTON_P==(FACTORINGMASK&stratflgs)){
		kbset.opts.factoring=1;
	} else {
		kbset.opts.factoring=0;
	}

	/* Set term ordering. */
	if (TRMORDSTD_P==(TERMORDRMASK&stratflgs)){
		kbset.opts.termord=STANDARD;
	} else {
		kbset.opts.termord=NONRECURSIVE;
	}

	/* Set literal and term ordering. */
	if (LITORDSTD_P==(LITORDRMASK&stratflgs)){
		kbset.opts.litord=STANDARD;
	} else if (LITORDNR_P==(LITORDRMASK&stratflgs)){
		kbset.opts.litord=NONRECURSIVE;
	} else {
		kbset.opts.litord=LEXICOGRAPHIC;
	}

	/* Set function weight option. */
	if (UNIFORM_P==(FUNCWMASK&stratflgs)){
		kbset.opts.fweight=UNIFORM;
	} else {
		kbset.opts.fweight=ARITY;
	}

	/* Set maximum weight option. */
	switch (MAXWMASK&stratflgs){
		case LRSOTT_P:
			kbset.opts.mxwghtopt=MXWEIGHTLRSOTT|(glblopts.mxwghtopt&(MXWEIGHTNUM|MXWEIGHTOFF));
			break;
		case LRSOFF_P:
			kbset.opts.mxwghtopt=MXWEIGHTLRSOFF|(glblopts.mxwghtopt&(MXWEIGHTNUM|MXWEIGHTOFF));
			break;
		case LRSALL_P:
			kbset.opts.mxwghtopt=MXWEIGHTLRS|(glblopts.mxwghtopt&(MXWEIGHTNUM|MXWEIGHTOFF));
			break;
		default:
			kbset.opts.mxwghtopt=MXWEIGHTLRSDSC|(glblopts.mxwghtopt&(MXWEIGHTNUM|MXWEIGHTOFF));
			break;
	}
	kbset.opts.maxweight=glblopts.maxweight;

	/* Initialize the learning vector for this process. If no matching record */
	/* is found then take a record matching all strategy parameters except */
	/* the algorithm, if such a record exist. */
	kbset.lrnvector=NULL;
	if ((LEARN_P&stratflgs)||((active_cores==1)&&(timeslices==1)&&(lrnrecords!=NULL))){
		for (mm=0;mm<numlrnrecs;mm++){
			if ((stratflgs&(SELECTMASK|TERMORDRMASK|LITORDRMASK|FUNCWMASK))==
					((SELECTMASK|TERMORDRMASK|LITORDRMASK|FUNCWMASK)&lrnrecords[mm].header)){
				kbset.lrnvector=&lrnrecords[mm].vector[0];
				if ((stratflgs&(ALGMASK|SELECTMASK|TERMORDRMASK|LITORDRMASK|FUNCWMASK))==
						((ALGMASK|SELECTMASK|TERMORDRMASK|LITORDRMASK|FUNCWMASK)&lrnrecords[mm].header)){
					break;
				}
			}
		}
	}

	/* Set remaining options and settings. */
	if ((0==(pbmflags&UNITCLAUSE))&&((glblopts.split==1)||(kbset.opts.algorithm==glblopts.split))){
		kbset.opts.split=1;
	} else {
		kbset.opts.split=0;
	}
	return;
} /* SetMTParams */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Main inference function for the process)  OCJ
 *
 *    This function controls the inference tasks for the process.
 *    There is an instance of this function per executed process.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void InferenceProcess(void){
	auto int32_t slicesleft;                             /* Remaining time slices not yet run */
	auto double totaltimeleft;                           /* Total remaining time at funciton start */
	auto double slicetimeleft;                           /* Slice remaining time */
	auto struct timeval time1,time2;                     /* For time measurements */
	auto cmprefix *ptr1,*ptr2;                           /* Auxiliary pointers */
	auto int32_t ii;                                     /* Auxiliary */

	/* Initialization. */
	gettimeofday(&time1,NULL);
	time2=time1;
	slicesleft=timeslices;
	totaltimeleft=(double)(glblopts.timeout-mainkb.prstats.elapsed_time);
	slicetimeleft=totaltimeleft/slicesleft;

	/* Loop for each time slice. */
	while (1){

		/* Set strategy. Return if no more strategies available. */
		sem_wait(&procctl->strategy);
		ii=*nextstrat;
		(*nextstrat)++;
		sem_post(&procctl->strategy);
		if (ii>=nbstrats){
			kbset.status=UNKNOWN;
			sem_wait(&procctl->procctl);
			if (procctl->status==RUNNING){
				procctl->status=UNKNOWN;
			}
			sem_post(&procctl->procctl);
			break;
		}
		SetMTParams(ii);
		kbset.status=RUNNING;

		/* Initialize the number of formulas in the working KB with */
		/* the number of text formulas in the main KB. This way the */
		/* formula numbering in both kb's will be compatible after */
		/* loading the working KB with passive binary clauses in the */
		/* main KB. */
		kbset.nformulas=mainkb.nformulas;

		/* Load working KB with unprocessed binary clauses in the main KB. */
		for (ptr1=mainkb.frstunproc;ptr1!=NULL;ptr1=ptr1->next){

			/* Allocate storage for the copied clause. */
			if (NULL==(ptr2=mallocmn(sizeof(cmprefix)))){
				kbset.status=NOMEMORY;
				sem_wait(&procctl->procctl);
				if (procctl->status==RUNNING){
					procctl->status=NOMEMORY;
				}
				sem_post(&procctl->procctl);
				return;
			}
			memcpy(ptr2,ptr1,sizeof(cmprefix));
			if (NULL==(ptr2->part2.bin=mallocmn(binpsize+ptr1->part2.bin->size))){
				kbset.status=NOMEMORY;
				sem_wait(&procctl->procctl);
				if (procctl->status==RUNNING){
					procctl->status=NOMEMORY;
				}
				sem_post(&procctl->procctl);
				MYFREE(ptr2);
				return;
			}

			/* Copy binary part of clause and clean equality data. */
			memcpy(ptr2->part2.bin,ptr1->part2.bin,binpsize+ptr1->part2.bin->size);
			ptr2->part2.bin->equflg=0;
			ptr2->part2.bin->clause=ptr2;

			/* Split clause if possible. */
			if (NOMEMORY==(ii=SplitClause(ptr2,0))){
				kbset.status=NOMEMORY;
				sem_wait(&procctl->procctl);
				if (procctl->status==RUNNING){
					procctl->status=NOMEMORY;
				}
				sem_post(&procctl->procctl);
				MYFREE(ptr2->part2.bin);
				MYFREE(ptr2);
				return;
				break;
			}

			/* If clause was not split then add clause to KB as unprocessed. */
			/* In case of an early return there is no need to call FreeEquData() */
			/* because equflg field is zero. */
			if (ii==0){
				if (NOMEMORY==AddBinClause2KB(ptr2,&kbset,SETFNUMBER)){
					kbset.status=NOMEMORY;
					sem_wait(&procctl->procctl);
					if (procctl->status==RUNNING){
						procctl->status=NOMEMORY;
					}
					sem_post(&procctl->procctl);
					MYFREE(ptr2->part2.bin);
					MYFREE(ptr2);
					return;
				}
			}
		}

		/* Get current time for the strategy, update time left for this slice */
		/* and set strategy timeout. */
		gettimeofday(&kbset.time,NULL);
		slicetimeleft-=(kbset.time.tv_sec-time2.tv_sec
				+((kbset.time.tv_usec-time2.tv_usec)/1000000.0));
		kbset.opts.timeout=slicetimeleft;

		/* Start process algorithm. */
		switch (kbset.opts.algorithm){
			case OTTER:
				Otter();
				break;
			case DISCOUNT:
				Discount();
				break;
		}

		/* Set time statistics for this slice. */
		gettimeofday(&kbset.endtime,NULL);
		kbset.prstats.elapsed_time=kbset.endtime.tv_sec-time2.tv_sec
				+(kbset.endtime.tv_usec-time2.tv_usec)/1000000.0;
		time2=kbset.endtime;

		/* Adjust process maximum used memory. */
		if (allcdmemory>mainkb.prstats.memory){
			mainkb.prstats.memory=allcdmemory;
		}

		/* Update problem statistics. */
		sem_wait(&procctl->statistics);
		AddInferenceStats(&kbset.prstats,pbmstats);
		sem_post(&procctl->statistics);

		/* If the problem was not solved or this is not the solving process */
		/* then reinitialize working KB and reset statistics. */
		if (procctl->solvekb!=procnb){
			Initialize_hashcmp();
			Initialize_KB(&kbset);
			ResetStats(&kbset.prstats);
		}

		/* Update slices left and initialize time left for next slice. */
		slicetimeleft=totaltimeleft-(time2.tv_sec-time1.tv_sec+(time2.tv_usec-time1.tv_usec)/1000000.0);
		if (slicetimeleft>0){
			if (slicesleft>0){
				slicesleft--;
			}
			if (slicesleft>0){
				slicetimeleft/=slicesleft;
			}

		/* There is no slice time available. */
		} else {
			sem_wait(&procctl->procctl);
			if ((procctl->status==RUNNING)||(procctl->status==UNKNOWN)){
				procctl->status=TIMEOUT;
			}
			sem_post(&procctl->procctl);
		}

		/* There is a solution or TIMEOUT or NOMEMORY condition. */
		if ((procctl->status!=RUNNING)&&(procctl->status!=UNKNOWN)){
			break;
		}
	}

	return;
} /* InferenceProcess */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Otter saturation algorithm)  OCJ
 *
 *    This function performs the Otter type saturation algorithm.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void Otter(void){
	auto cmprefix *current;                              /* Selected current clause. */
	auto cmprefix *auxcl;                                /* Pointer to auxiliary clauses */
	auto struct timeval time1,time2;                     /* For LRS time measurements and timeout checks */
	auto double timedif;                                 /* Elapsed time in LRS cycle */
	auto double lefttime;                                /* Time available */
	auto float reachcl;                                  /* Number reachable passive clauses */
	auto char *ptr1;                                     /* Auxiliary pointer */
	auto int32_t ii,jj;                                  /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Get initial time and initialize clause selection parameters. */
	/* The initial time for this function is in the time field of the */
	/* kbset structure. */
	time1=kbset.time;
	kbset.selectcnt=0;

	/* Main loop. This loop is held as far as there are passive or unprocessed */
	/* clauses or SAT queue is not empty. */
	kbset.lrscnt=0;
	kbset.flags&=(~RETENTIONUSED);
	while ((kbset.frstpassive!=NULL)||(kbset.frstunproc!=NULL)
			||(kbset.firstsatq!=NULL)){

		/* Check memory limit. */
		if (allcdmemory>maxmemory){
			kbset.status=NOMEMORY;
			return;
		}

		/* SAT queue is not empty. */
		if (kbset.firstsatq!=NULL){

			/* Add SAT clauses in SAT queue to SAT solver and ask SAT solver */
			/* to solve. */
			if (0==(ii=SatQ2SatSolver())){
				ii=SatSolver();
			}
			switch (ii){

				/* TIMEOUT or NOMEMORY condition. */
				case NOMEMORY:
				case TIMEOUT:
					kbset.status=ii;
					return;
					break;

				/* Refutation found. */
				case 0:
					kbset.status=UNSATISFIABLE;
					sem_wait(&procctl->procctl);
					if (procctl->status!=UNSATISFIABLE){
						procctl->status=UNSATISFIABLE;
						procctl->solvekb=procnb;
					} else {
						MYFREE(kbset.frstprfnode);
					}
					sem_post(&procctl->procctl);
					return;
					break;

				/* A SAT model is available. */
				default:
				case 1:
					break;
			}

			/* Apply changes of current interpretation. */
			if (0!=(ii=ApplyModelChanges())){
				kbset.status=ii;
				return;
			}
		}

		/* Debug. */
		#ifdef DEBUGCODE
		#ifdef VERBOSE
		if (active_cores==1){
			printf("--------------------------\nStart process of unprocessed clauses\n");
		}
		#endif
		#endif

		/* Loop through unprocessed clauses. */
		for (auxcl=kbset.frstunproc;auxcl!=NULL;auxcl=kbset.frstunproc){

			/* Unlink clause. Unlink cannot return NOMEMORY because unprocessed */
			/* clauses are not indexed by discrimination trees. */
			Unlink(auxcl,&kbset);

			/* Check empty clause. If the clause is empty then it has assertions so */
			/* add contradiction clause to SAT queue and iterate in the main loop. */
			/* Iterating in the main loop is achieved here by leaving the current */
			/* unprocessed clauses loop. */
			if (UNITEND==auxcl->part2.bin->formula[0]){
				if (0==(NUMBERED&auxcl->flags)){
					kbset.nformulas++;
					auxcl->number=kbset.nformulas;
					auxcl->flags|=NUMBERED;
				}
				if (NOMEMORY==BuildContrClause(auxcl)){
					kbset.status=NOMEMORY;
					MYFREE(auxcl->part2.bin->asserts);
					MYFREE(auxcl->part2.bin);
					MYFREE(auxcl);
					return;
				}
				break;
			}

			/* Split clause if possible. If clause is split then iterate */
			/* in the current loop. In case of NOMEMORY then clause must be */
			/* freed because it is unlinked. */
			if (NOMEMORY==(ii=SplitClause(auxcl,0))){
				kbset.status=NOMEMORY;
				if (auxcl->part2.bin->asserts!=NULL){
					MYFREE(auxcl->part2.bin->asserts);
				}
				MYFREE(auxcl->part2.bin);
				MYFREE(auxcl);
				return;
			}
			if (ii){
				continue;
			}

			/* Do simplifications to and from clause. */
			switch (ii=FwdSimplify(&auxcl,PASSIVE)){

				/* Check timeout and no memory. */
				case NOMEMORY:
				case TIMEOUT:
					kbset.status=ii;
					return;
					break;

				/* Check empty clause without assertions inferred by simplification. */
				case 1:
					sem_wait(&procctl->procctl);
					if (procctl->status!=UNSATISFIABLE){
						procctl->status=UNSATISFIABLE;
						procctl->solvekb=procnb;
					} else {
						MYFREE(kbset.frstprfnode);
					}
					sem_post(&procctl->procctl);
					return;
					break;
			}

			/* Clause has not been locked or deleted by simplification. */
			if ((auxcl!=NULL)&&(0==(LOCKED&auxcl->flags))){

				/* Check empty clause. If the clause is empty then it has assertions so */
				/* add contradiction clause to SAT queue and iterate in the main loop. */
				/* Iterating in the main loop is achieved here by leaving the current */
				/* unprocessed clauses loop. */
				if (UNITEND==auxcl->part2.bin->formula[0]){
					if (0==(NUMBERED&auxcl->flags)){
						kbset.nformulas++;
						auxcl->number=kbset.nformulas;
						auxcl->flags|=NUMBERED;
					}
					if (NOMEMORY==BuildContrClause(auxcl)){
						kbset.status=NOMEMORY;
						MYFREE(auxcl->part2.bin->asserts);
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						return;
					}
					break;
				}

				/* Retention test checking. */
				switch (Retained(auxcl)){

					/* Not enough memory. */
					case NOMEMORY:
						kbset.status=NOMEMORY;
						FreeEquData(auxcl->part2.bin);
						if (auxcl->part2.bin->asserts!=NULL){
							MYFREE(auxcl->part2.bin->asserts);
						}
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						return;
						break;

					/* Clause is retained. */
					case 1:

						/* Perform backward simplification. */
						jj=BckSimplify(auxcl,PASSIVE);
						if ((jj==NOMEMORY)||(jj==TIMEOUT)){
							kbset.status=jj;
							FreeEquData(auxcl->part2.bin);
							if (auxcl->part2.bin->asserts!=NULL){
								MYFREE(auxcl->part2.bin->asserts);
							}
							MYFREE(auxcl->part2.bin);
							MYFREE(auxcl);
							return;
						}

						/* Add clause to passive clauses.*/
						auxcl->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
						auxcl->flags|=PASSIVE;
						if (NOMEMORY==AddBinClause2KB(auxcl,&kbset,SELECT|SETFNUMBER)){
							kbset.status=NOMEMORY;
							FreeEquData(auxcl->part2.bin);
							if (auxcl->part2.bin->asserts!=NULL){
								MYFREE(auxcl->part2.bin->asserts);
							}
							MYFREE(auxcl->part2.bin);
							MYFREE(auxcl);
							return;
						}

						/* If BckSimplify() produced an empty clause without assertions */
						/* then assign a number to the empty clause and return. */
						if (jj){
							kbset.status=UNSATISFIABLE;
							kbset.nformulas++;
							kbset.frstprfnode->ptr.Aclause->number=kbset.nformulas;
							kbset.frstprfnode->ptr.Aclause->flags|=NUMBERED;
							sem_wait(&procctl->procctl);
							if (procctl->status!=UNSATISFIABLE){
								procctl->status=UNSATISFIABLE;
								procctl->solvekb=procnb;
							} else {
								MYFREE(kbset.frstprfnode);
							}
							sem_post(&procctl->procctl);
							return;
						}
						break;
				}
			}
		}

		/* Disposing of clauses in disposal queue. */
		for (auxcl=kbset.lastdisp,ii=0;auxcl!=NULL;auxcl=kbset.lastdisp,ii++){

			/* Clause is not locked. Dispose of clause. */
			if (0==(LOCKED&auxcl->flags)){
				if (NULL==(ptr1=Decompile(auxcl->part2.bin))){
					kbset.status=NOMEMORY;
					return;
				}
				kbset.lastdisp=auxcl->prevdisp;
				auxcl->prevdisp=NULL;
				auxcl->flags&=(~INDISPOSALQUEUE);
				if (NOMEMORY==Unlink(auxcl,&kbset)){
					kbset.status=NOMEMORY;
					return;
				}
				FreeEquData(auxcl->part2.bin);
				if (auxcl->part2.bin->asserts!=NULL){
					MYFREE(auxcl->part2.bin->asserts);
				}
				MYFREE(auxcl->part2.bin);
				auxcl->part2.text=ptr1;
				AddTxtFormula2KB(auxcl,auxcl->inference,&kbset,0);

			/* If the clause is marked as LOCKED then there are locking */
			/* assertions. Unlink the clause from current queue and add */
			/* it to the LOCKED queue. If a NOMEMORY condition arises */
			/* after the clause is unlinked then it is completely deleted */
			/* because it cannot be deleted thereafter as it is unlinked. */
			} else {
				kbset.lastdisp=auxcl->prevdisp;
				auxcl->prevdisp=NULL;
				auxcl->flags&=(~(INDISPOSALQUEUE|LOCKED));
				if (NOMEMORY==Unlink(auxcl,&kbset)){
					kbset.status=NOMEMORY;
					return;
				}
				auxcl->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
				auxcl->flags|=LOCKED;
				if (NOMEMORY==AddBinClause2KB(auxcl,&kbset,0)){
					MYFREE(auxcl->part2.bin->lockasserts);
					if (NULL!=auxcl->part2.bin->asserts){
						MYFREE(auxcl->part2.bin->asserts);
					}
					MYFREE(auxcl->part2.bin);
					MYFREE(auxcl);
					kbset.status=NOMEMORY;
					return;
				}
			}

			/* Check timeout and solution by other process. */
	 		if (ii>=TIMEOUTCCLDSP){
				gettimeofday(&time2,NULL);
				timedif=time2.tv_sec-kbset.time.tv_sec
						+(time2.tv_usec-kbset.time.tv_usec)/1000000.0;
				if (timedif>=kbset.opts.timeout){
					kbset.status=TIMEOUT;
					return;
				}
				if ((procctl->status==UNSATISFIABLE)||(procctl->status==SATISFIABLE)){
					kbset.status=UNKNOWN;
					return;
				}
				if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
					kbset.status=procctl->status;
					return;
				}
				ii=-1;
	 		}
		}

		/* If SAT queue is not empty then iterate in the main loop. */
		/* It is important to place this piece of code after processing */
		/* the disposal queue to ensure that clauses marked as locked are */
		/* effectively locked. */
		if (kbset.firstsatq!=NULL){
			continue;
		}

		/* LRS has been selected. */
		if (kbset.opts.mxwghtopt&(MXWEIGHTLRS|MXWEIGHTLRSOTT)){

			/* Threshold has been reached. */
			kbset.lrscnt++;
			if (kbset.lrscnt>=LRSTHRESHOLD){

				/* Get elapsed time for last cycle and remaining available time. */
				/* Compute number of reachable passive clause. Update current time */
				/* for next cycle*/
				gettimeofday(&time2,NULL);
				timedif=time2.tv_sec-time1.tv_sec+(time2.tv_usec-time1.tv_usec)/1000000.0;
				lefttime=kbset.opts.timeout-(time2.tv_sec-kbset.time.tv_sec
						+(time2.tv_usec-kbset.time.tv_usec)/1000000.0);
				reachcl=LRSCOEF*LRSTHRESHOLD*lefttime/timedif;
				gettimeofday(&time1,NULL);

				/* Not all passive clauses are reachable at the current rate. */
				if (reachcl<kbset.prstats.nbpassive){
					if (NOMEMORY==LRSClean((int32_t)reachcl)){
						kbset.status=NOMEMORY;
						return;
					}
				}

				/* Update LRS count. */
				kbset.lrscnt=0;
			}
		}

		/* Select current clause and unlink it from KB. */
		if (NULL==(current=SelectClause())){
			break;
		}
		if (NOMEMORY==Unlink(current,&kbset)){
			kbset.status=NOMEMORY;
			return;
		}

		/* Debug. */
		#ifdef DEBUGCODE
		if (current->number==dbgnb){
			ptr10=NULL;
		}
		ptr10=Decompile(current->part2.bin);
		#ifdef VERBOSE
		if (active_cores==1){
			printf("--------------------------\nSelected clause [%ld/%ld]:\n%s\n",
					current->number,kbset.nformulas,ptr10);
		}
		#endif
		MYFREE(ptr10);
		#endif

		/* Add current clause to ACTIVE queue. In case of no memory the clause must */
		/* be freed because it is unlinked and will not be freed thereafter. */
		current->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
		current->flags|=ACTIVE;
		if (NOMEMORY==AddBinClause2KB(current,&kbset,SELECT)){
			FreeEquData(current->part2.bin);
			if (current->part2.bin->asserts!=NULL){
				MYFREE(current->part2.bin->asserts);
			}
			MYFREE(current->part2.bin);
			MYFREE(current);
			kbset.status=NOMEMORY;
			return;
		}

		/* Perform deduction inferences. In case of an early return */
		/* current clause is freed by Infere(). */
		Infere(current);
		switch (kbset.status){
			case NOMEMORY:
				return;
				break;
			case UNSATISFIABLE:
				sem_wait(&procctl->procctl);
				if (procctl->status!=UNSATISFIABLE){
					procctl->status=UNSATISFIABLE;
					procctl->solvekb=procnb;
				} else {
					MYFREE(kbset.frstprfnode);
				}
				sem_post(&procctl->procctl);
			/* No break */
			case UNKNOWN:
			case TIMEOUT:
				return;
				break;
		}
	}

	/* If we are here then theory is SATISFIABLE if the algorithm used */
	/* is complete or TIMEOUT/UNKNOWN otherwise. */
	/*if ((0==(kbset.flags&RETENTIONUSED))&&((0==(prproflags&CLFILTERINGPRUNNED))
			||((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))))
			&&((MAXIMAL|SINGLENEG|MULTINEG)&kbset.opts.select)
			&&(0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))){*/
	if ((0==(kbset.flags&RETENTIONUSED))&&((0==(prproflags&CLFILTERINGPRUNNED))
			||((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))))
			&&((SINGLENEG|MULTINEG)&kbset.opts.select)
			&&((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))||(kbset.opts.demodulation==DEMODOFF))){
			//&&((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))||(kbset.opts.demodulation==DEMODOFF)
					//||(kbset.opts.demodulation==DEMODCPL))){
		kbset.status=SATISFIABLE;
		sem_wait(&procctl->procctl);
		if ((procctl->status!=SATISFIABLE)&&(procctl->status!=UNSATISFIABLE)){
			procctl->status=SATISFIABLE;
			procctl->solvekb=procnb;
		}
		sem_post(&procctl->procctl);
	} else {
		gettimeofday(&time2,NULL);
		timedif=time2.tv_sec-kbset.time.tv_sec
				+(time2.tv_usec-kbset.time.tv_usec)/1000000.0;
		if (timedif>=kbset.opts.timeout){
			kbset.status=TIMEOUT;
		} else {
			kbset.status=UNKNOWN;
		}
	}

	return;
} /* Otter */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Discount saturation algorithm)  OCJ
 *
 *    This function performs the Discount type saturation algorithm.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void Discount(void){
	auto cmprefix *current;                              /* Selected current clause. */
	auto cmprefix *auxcl,*auxcl1;                        /* Pointers to auxiliary clause */
	auto struct timeval time1,time2;                     /* For LRS time measurements and timeout checks */
	auto double timedif;                                 /* Elapsed time in LRS cycle */
	auto double lefttime;                                /* Time available */
	auto float reachcl;                                  /* Number reachable passive clauses */
	auto char *ptr1;                                     /* Auxiliary pointer */
	auto int32_t ii,jj,kk;                               /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Get initial time and initialize clause selection parameters. */
	/* The initial time for this function is in the time field of the */
	/* kbset structure. */
	time1=kbset.time;
	kbset.selectcnt=0;

	/* Main loop. */
	kbset.lrscnt=0;
	kbset.flags&=(~RETENTIONUSED);
	while ((kbset.frstpassive!=NULL)||(kbset.frstunproc!=NULL)
			||(kbset.firstsatq!=NULL)){

		/* Check memory limit. */
		if (allcdmemory>maxmemory){
			kbset.status=NOMEMORY;
			return;
		}

		/* SAT queue is not empty. */
		if (kbset.firstsatq!=NULL){

			/* Add SAT clauses in SAT queue to SAT solver and ask SAT solver */
			/* to solve. */
			if (0==(ii=SatQ2SatSolver())){
				ii=SatSolver();
			}
			switch (ii){

				/* TIMEOUT or NOMEMORY condition. */
				case NOMEMORY:
				case TIMEOUT:
					kbset.status=ii;
					return;
					break;

				/* Refutation found. */
				case 0:
					kbset.status=UNSATISFIABLE;
					sem_wait(&procctl->procctl);
					if (procctl->status!=UNSATISFIABLE){
						procctl->status=UNSATISFIABLE;
						procctl->solvekb=procnb;
					} else {
						MYFREE(kbset.frstprfnode);
					}
					sem_post(&procctl->procctl);
					return;
					break;

				/* A SAT model is available. */
				default:
				case 1:
					break;
			}

			/* Apply changes of current interpretation. */
			if (0!=(ii=ApplyModelChanges())){
				kbset.status=ii;
				return;
			}
		}

		/* Debug. */
		#ifdef DEBUGCODE
		#ifdef VERBOSE
		if (active_cores==1){
			printf("--------------------------\nStart process of unprocessed clauses\n");
		}
		#endif
		#endif

		/* Loop through unprocessed clauses. */
		for (auxcl=kbset.frstunproc;auxcl!=NULL;auxcl=kbset.frstunproc){

			/* Unlink clause. Unlink cannot return NOMEMORY because */
			/* unprocessed clauses are not indexed by discrimination trees. */
			Unlink(auxcl,&kbset);

			/* Check empty clause. If the clause is empty then it has assertions so */
			/* add contradiction clause to SAT queue and iterate in the main loop. */
			/* Iterating in the main loop is achieved here by leaving the current */
			/* unprocessed clauses loop. */
			if (UNITEND==auxcl->part2.bin->formula[0]){
				if (0==(NUMBERED&auxcl->flags)){
					kbset.nformulas++;
					auxcl->number=kbset.nformulas;
					auxcl->flags|=NUMBERED;
				}
				if (NOMEMORY==BuildContrClause(auxcl)){
					kbset.status=NOMEMORY;
					MYFREE(auxcl->part2.bin->asserts);
					MYFREE(auxcl->part2.bin);
					MYFREE(auxcl);
					return;
				}
				break;
			}

			/* Split clause if possible. If clause is split then iterate */
			/* in the current loop. In case of NOMEMORY then clause must be */
			/* freed because it is unlinked. */
			if (NOMEMORY==(ii=SplitClause(auxcl,0))){
				kbset.status=NOMEMORY;
				if (auxcl->part2.bin->asserts!=NULL){
					MYFREE(auxcl->part2.bin->asserts);
				}
				MYFREE(auxcl->part2.bin);
				MYFREE(auxcl);
				return;
			}
			if (ii){
				continue;
			}

			/* Do simplifications to and from clause. */
			switch (ii=FwdSimplify(&auxcl,0)){

				/* Check timeout and no memory. */
				case NOMEMORY:
				case TIMEOUT:
					kbset.status=ii;
					return;
					break;

				/* Check empty clause without assertions inferred by simplification. */
				case 1:
					sem_wait(&procctl->procctl);
					if (procctl->status!=UNSATISFIABLE){
						procctl->status=UNSATISFIABLE;
						procctl->solvekb=procnb;
					} else {
						MYFREE(kbset.frstprfnode);
					}
					sem_post(&procctl->procctl);
					return;
					break;
			}

			/* Clause has not been locked or deleted by simplification. */
			if ((auxcl!=NULL)&&(0==(LOCKED&auxcl->flags))){

				/* Check empty clause. If the clause is empty then it has assertions so */
				/* add contradiction clause to SAT queue and iterate in the main loop. */
				/* Iterating in the main loop is achieved here by leaving the current */
				/* unprocessed clauses loop. */
				if (UNITEND==auxcl->part2.bin->formula[0]){
					if (0==(NUMBERED&auxcl->flags)){
						kbset.nformulas++;
						auxcl->number=kbset.nformulas;
						auxcl->flags|=NUMBERED;
					}
					if (NOMEMORY==BuildContrClause(auxcl)){
						kbset.status=NOMEMORY;
						MYFREE(auxcl->part2.bin->asserts);
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						return;
					}
					break;
				}

				/* Retention test checking. */
				switch (Retained(auxcl)){

					/* Not enough memory. */
					case NOMEMORY:
						kbset.status=NOMEMORY;
						FreeEquData(auxcl->part2.bin);
						if (auxcl->part2.bin->asserts!=NULL){
							MYFREE(auxcl->part2.bin->asserts);
						}
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						return;
						break;

					/* Clause is retained. Add clause to passive clauses. */
					case 1:
						auxcl->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
						auxcl->flags|=PASSIVE;
						if (NOMEMORY==AddBinClause2KB(auxcl,&kbset,SELECT|SETFNUMBER)){
							kbset.status=NOMEMORY;
							FreeEquData(auxcl->part2.bin);
							if (auxcl->part2.bin->asserts!=NULL){
								MYFREE(auxcl->part2.bin->asserts);
							}
							MYFREE(auxcl->part2.bin);
							MYFREE(auxcl);
							return;
						}
						break;
				}
			}
		}

		/* If SAT queue is not empty then iterate in the main loop. */
		/* There is no need to process disposal queue before the */
		/* iteration as the previous processing of unprocessed clauses */
		/* has not put any clause  in disposal queue for disposal or */
		/* locking. This is a difference with Otter() algorithm. */
		if (kbset.firstsatq!=NULL){
			continue;
		}

		/* LRS has been selected. */
		if (kbset.opts.mxwghtopt&(MXWEIGHTLRS|MXWEIGHTLRSDSC)){

			/* Threshold has been reached. */
			kbset.lrscnt++;
			if (kbset.lrscnt>=LRSTHRESHOLD){

				/* Get elapsed time for last cycle and remaining available time. */
				/* Compute number of reachable passive clause. Update current time */
				/* for next cycle*/
				gettimeofday(&time2,NULL);
				timedif=time2.tv_sec-time1.tv_sec+(time2.tv_usec-time1.tv_usec)/1000000.0;
				lefttime=kbset.opts.timeout-(time2.tv_sec-kbset.time.tv_sec
						+(time2.tv_usec-kbset.time.tv_usec)/1000000.0);
				reachcl=LRSCOEF*LRSTHRESHOLD*lefttime/timedif;
				gettimeofday(&time1,NULL);

				/* Not all passive clauses are reachable at the current rate. */
				if (reachcl<kbset.prstats.nbpassive){
					if (NOMEMORY==LRSClean((int32_t)reachcl)){
						kbset.status=NOMEMORY;
						return;
					}
				}

				/* Update LRS count. */
				kbset.lrscnt=0;
			}
		}

		/* Select current clause and unlink it from KB. */
		if (NULL==(current=SelectClause())){
			break;
		}
		if (NOMEMORY==Unlink(current,&kbset)){
			kbset.status=NOMEMORY;
			return;
		}

		/* Free equality ordering data. This is necessary because the clause */
		/* may be simplified in the FwdSimplify() process and then added again to */
		/* ACTIVE clauses in the Infere() process. */
		FreeEquData(current->part2.bin);

		/* Debug. */
		#ifdef DEBUGCODE
		if (current->number==dbgnb){
			ptr10=NULL;
		}
		ptr10=Decompile(current->part2.bin);
		#ifdef VERBOSE
		if (active_cores==1){
			printf("--------------------------\nSelected clause [%ld/%ld]:\n%s\n",
					current->number,kbset.nformulas,ptr10);
		}
		#endif
		MYFREE(ptr10);
		#endif

		/* Simplify current clause by and to active clauses */
		/* but of course don't add it to PASSIVE queue. */
		/* Remember origin clause. */
		/* In case of early return the clause is freed by FwdSimplify() */
		/* or one of the functions called by FwdSimplify(). */
		auxcl1=current;
		switch (ii=FwdSimplify(&current,FROMPASSIVE)){

			/* Check timeout and no memory. */
			case NOMEMORY:
			case TIMEOUT:
				kbset.status=ii;
				return;
				break;

				/* Check empty clause without assertions inferred by simplification. */
				case 1:
					sem_wait(&procctl->procctl);
					if (procctl->status!=UNSATISFIABLE){
						procctl->status=UNSATISFIABLE;
						procctl->solvekb=procnb;
					} else {
						MYFREE(kbset.frstprfnode);
					}
					sem_post(&procctl->procctl);
					return;
					break;
		}

		/* Current clause has not been locked or deleted by simplification. */
		if ((current!=NULL)&&(0==(LOCKED&current->flags))){

			/* An empty clause with assertions was inferred in current clause */
			/* simplification. Add contradiction clause to SAT queue and set */
			/* current to NULL so that no additional process is performed with it. */
			if (UNITEND==current->part2.bin->formula[0]){
				if (NOMEMORY==BuildContrClause(current)){
					kbset.status=NOMEMORY;
					MYFREE(current->part2.bin->asserts);
					MYFREE(current->part2.bin);
					MYFREE(current);
					return;
				}
				current=NULL;

			/* No empty clause with assertions was inferred in current clause */
			/* simplification. */
			} else {

				/* Perform backward simplification. */
				jj=BckSimplify(current,0);
				if ((jj==NOMEMORY)||(jj==TIMEOUT)){
					kbset.status=jj;
					FreeEquData(current->part2.bin);
					if (current->part2.bin->asserts!=NULL){
						MYFREE(current->part2.bin->asserts);
					}
					MYFREE(current->part2.bin);
					MYFREE(current);
					return;
				}

				/* Add clause to ACTIVE queue. In case of no memory the clause must be */
				/* freed because it is unlinked and will not be freed thereafter. */
				current->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
				current->flags|=ACTIVE;
				if (NOMEMORY==AddBinClause2KB(current,&kbset,SELECT|(auxcl1==current?0:SETFNUMBER))){
					FreeEquData(current->part2.bin);
					if (current->part2.bin->asserts!=NULL){
						MYFREE(current->part2.bin->asserts);
					}
					MYFREE(current->part2.bin);
					MYFREE(current);
					kbset.status=NOMEMORY;
					return;
				}

				/* If BckSimplify() produced an empty clause without assertions */
				/* then assign a number to the empty clause and return. */
				if (jj){
					kbset.status=UNSATISFIABLE;
					kbset.nformulas++;
					kbset.frstprfnode->ptr.Aclause->number=kbset.nformulas;
					kbset.frstprfnode->ptr.Aclause->flags|=NUMBERED;
					sem_wait(&procctl->procctl);
					if (procctl->status!=UNSATISFIABLE){
						procctl->status=UNSATISFIABLE;
						procctl->solvekb=procnb;
					} else {
						MYFREE(kbset.frstprfnode);
					}
					sem_post(&procctl->procctl);
					return;
				}
			}
		}

		/* Dispose of clauses in disposal queue. */
		for (auxcl=kbset.lastdisp,kk=0;auxcl!=NULL;auxcl=kbset.lastdisp,kk++){

			/* Clause is not locked. Dispose of clause. */
			if (0==(LOCKED&auxcl->flags)){
				if (NULL==(ptr1=Decompile(auxcl->part2.bin))){
					kbset.status=NOMEMORY;
					return;
				}
				kbset.lastdisp=auxcl->prevdisp;
				auxcl->prevdisp=NULL;
				auxcl->flags&=(~INDISPOSALQUEUE);
				if (NOMEMORY==Unlink(auxcl,&kbset)){
					kbset.status=NOMEMORY;
					return;
				}
				FreeEquData(auxcl->part2.bin);
				if (auxcl->part2.bin->asserts!=NULL){
					MYFREE(auxcl->part2.bin->asserts);
				}
				MYFREE(auxcl->part2.bin);
				auxcl->part2.text=ptr1;
				AddTxtFormula2KB(auxcl,auxcl->inference,&kbset,0);

			/* If the clause is marked as LOCKED then there are locking */
			/* assertions. Unlink the clause from current queue and add */
			/* it to the LOCKED queue. If a NOMEMORY condition arises */
			/* after the clause is unlinked then it is completely deleted */
			/* because it cannot be deleted thereafter as it is unlinked. */
			} else {
				kbset.lastdisp=auxcl->prevdisp;
				auxcl->prevdisp=NULL;
				auxcl->flags&=(~(INDISPOSALQUEUE|LOCKED));
				if (NOMEMORY==Unlink(auxcl,&kbset)){
					MYFREE(auxcl->part2.bin->lockasserts);
					kbset.status=NOMEMORY;
					return;
				}
				auxcl->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
				auxcl->flags|=LOCKED;
				if (NOMEMORY==AddBinClause2KB(auxcl,&kbset,0)){
					MYFREE(auxcl->part2.bin->lockasserts);
					if (NULL!=auxcl->part2.bin->asserts){
						MYFREE(auxcl->part2.bin->asserts);
					}
					MYFREE(auxcl->part2.bin);
					MYFREE(auxcl);
					kbset.status=NOMEMORY;
					return;
				}
			}

			/* Check timeout and solution by other process. */
			if (kk>=TIMEOUTCCLDSP){
				gettimeofday(&time2,NULL);
				timedif=time2.tv_sec-kbset.time.tv_sec
						+(time2.tv_usec-kbset.time.tv_usec)/1000000.0;
				if (timedif>=kbset.opts.timeout){
					kbset.status=TIMEOUT;
					return;
				}
				if ((procctl->status==UNSATISFIABLE)||(procctl->status==SATISFIABLE)){
					kbset.status=UNKNOWN;
					return;
				}
				if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
					kbset.status=procctl->status;
					return;
				}
				kk=-1;
			}
		}

		/* Current clause has not been locked or deleted by simplification. */
		/* Perform deduction inferences. In case of an early return */
		/* the current clause is freed in Infere(). */
		if ((current!=NULL)&&(0==(LOCKED&current->flags))){
			Infere(current);
			switch (kbset.status){
				case NOMEMORY:
					return;
					break;
				case UNSATISFIABLE:
					sem_wait(&procctl->procctl);
					if (procctl->status!=UNSATISFIABLE){
						procctl->status=UNSATISFIABLE;
						procctl->solvekb=procnb;
					} else {
						MYFREE(kbset.frstprfnode);
					}
					sem_post(&procctl->procctl);
				/* No break */
				case UNKNOWN:
				case TIMEOUT:
					return;
					break;
			}
		}
	}

	/* If we are here then theory is SATISFIABLE if the algorithm used */
	/* is complete or UNKNOWN otherwise. */
	/*if ((0==(kbset.flags&RETENTIONUSED))&&((0==(prproflags&CLFILTERINGPRUNNED))
			||((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))))
			&&((MAXIMAL|SINGLENEG|MULTINEG)&kbset.opts.select)
			&&(0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))){*/
	if ((0==(kbset.flags&RETENTIONUSED))&&((0==(prproflags&CLFILTERINGPRUNNED))
			||((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))))
			&&((SINGLENEG|MULTINEG)&kbset.opts.select)
			&&((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))||(kbset.opts.demodulation==DEMODOFF))){
			//&&((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))||(kbset.opts.demodulation==DEMODOFF)
					//||(kbset.opts.demodulation==DEMODCPL))){
		kbset.status=SATISFIABLE;
		sem_wait(&procctl->procctl);
		if ((procctl->status!=SATISFIABLE)&&(procctl->status!=UNSATISFIABLE)){
			procctl->status=SATISFIABLE;
			procctl->solvekb=procnb;
		}
		sem_post(&procctl->procctl);
	} else {
		kbset.status=UNKNOWN;
	}

	return;
} /* Discount */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Select current clause from passive clauses)  OCJ
 *
 *    This function selects a "current" clause from the set of
 *    passive clauses in the working KB.
 *
 *    The selection algorithm is based on the "age" and "weight" queues
 *    of the KB and on an age to weight ratio that ensures fairness.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    Pointer to selected "current" clause or NULL if no more passive
 *    clauses available.
 *
 *
 *--------------------------------------------------------------*/
cmprefix *SelectClause(void){
	auto int32_t jj;                                     /* Auxiliary */

	/* There are passive clauses. */
	if (kbset.frstpassive!=NULL){

		/* Get queue index with learning disabled. */
		if (kbset.lrnvector==NULL){
			jj=slctordr[kbset.selectcnt];
			kbset.selectcnt++;
			if (kbset.selectcnt==slctordrsize){
				kbset.selectcnt=0;
			}

		/* Get queue index with learning enabled. */
		} else {
			jj=slctordr2[kbset.selectcnt];
			kbset.selectcnt++;
			if (kbset.selectcnt==slctordrsize2){
				kbset.selectcnt=0;
			}
		}

		/* Select by weight. */
		if (jj<WEIGHTQUEUES){
			#ifdef PREFERPROCESSED
			kbset.frstglweight[jj]->flags|=SELECTEDCLAUSE;
			#endif
			kbset.frstglweight[jj]->flags|=PICKED;
			return(kbset.frstglweight[jj]);

		/* Select by age. Update counter, indicate that clause has been selected and return. */
		} else {
			#ifdef PREFERPROCESSED
			kbset.frstpassive->flags|=SELECTEDCLAUSE;
			#endif
			kbset.frstpassive->flags|=PICKED;
			return(kbset.frstpassive);
		}
	}

	return(NULL);
} /* SelectClause */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if clause is retained)  OCJ
 *
 *    This function checks if a clause is retained. The clause is assumed
 *    to be unlinked from KB.
 *
 *    If the clause is not retained then it is converted to text and the
 *    binary (non text) data is freed.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause.
 *
 *  RETURNS:
 *
 *    0 -> Clause was not retained.
 *    1 -> Clause is retained.
 *
 *
 *--------------------------------------------------------------*/
int32_t Retained(cmprefix *clause){
	auto char *ptr1;                                     /* Auxiliary pointer */

	/* Check clause weight. The weight is only checked for the last */
	/* weight queue. */
	ClauseWeight(clause);
	if (clause->part2.bin->clweight[WEIGHTQUEUES-1]>=kbset.opts.maxweight){

		/* Convert clause to text form. Due to Avatar architecture the same */
		/* clause can go as unprocessed several times, so an unprocessed */
		/* clause is not necessarily new and therefore can be part of the */
		/* theorem proof. */
		if (NULL==(ptr1=Decompile(clause->part2.bin))){
			return(NOMEMORY);
		}
		kbset.flags|=RETENTIONUSED;
		FreeEquData(clause->part2.bin);
		if (clause->part2.bin->asserts!=NULL){
			MYFREE(clause->part2.bin->asserts);
		}
		MYFREE(clause->part2.bin);
		clause->part2.text=ptr1;
		AddTxtFormula2KB(clause,clause->inference,&kbset,SETFNUMBER);
		return(0);
	}

	return(1);
} /* Retained */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Compute clause weight)  OCJ
 *
 *    This function computes the different weights of a clause in compiled
 *    binary format. The following definitions are used:
 *
 *    Distance to conjecture:
 *    ----------------------
 *    - Clauses obtained from conjecture or negated conjecture in the
 *      pre-processing have distance=0.
 *    - Clauses obtained in the prep-rocessing not from conjecture or
 *      negated conjecture have distance=MAXDIST2CONJ.
 *    - If a clause is simplified the resulting clause has the same
 *      distance as the original clause.
 *    - If a clause is obtained by other inferences from parent clauses
 *      with distances m and n then the clause distance is
 *      MIN(MAXDIST2CONJ,1+MIN(m,n)).
 *
 *    Symbol weights:
 *    --------------
 *    There are two types of symbol weight, stored in the symbolw[] set
 *    field of the binprefix structure:
 *    - symbolw[0] is computed assigning weights VARWEIGHT, PREDWEIGHT
 *      or FUNCWEIGHT to each variable, predicate or function symbol.
 *    - symbolw[1] is computed as symbol[0] but if a symbol doesn't
 *      appear in the conjecture then the symbol weight is multiplied
 *      by the appropriate factor, depending on the kind of symbol
 *      and the value of kbset.opts.factoridx, so getting heavier.
 *      See PNOTINCONJFACTOR1, ENOTINCONJFACTOR1 and FNOTINCONJFACTOR1
 *      definitions in global.h.
 *
 *    In addition, for both cases above if a symbol is in a positive
 *    literal then its weights are multiplied by POSLITERALFACTOR1 or
 *    POSLITERALFACTOR2, depending on the value of kbset.opts.factoridx.
 *
 *    The following weights are calculated:
 *
 *    Weight 0:
 *    --------
 *    - This weight is only used if LEARNFROM option is enabled and there is
 *      a learn vector available for the process. The weight is computed as in
 *      weight 1, multiplied by a factor usually less than one and added
 *      BADCLAUSEWEIGHT if the clause is not classified as good by the learning
 *      process.
 *
 *    Weight 1:
 *    --------
 *    - Clauses with distance=0 have weight=0.
 *    - Other clauses obtained in the pre-processing have weight=1. These
 *      clauses have inference=CLAUSIFY and distance=MAXDIST2CONJ.
 *    - For the rest of the clauses weight=symbolw[0]+10. This weight is bigger
 *      than or equal to 20.
 *
 *    Weight 2:
 *    --------
 *    - If the distance of clause to conjecture is less than MAXDIST2CONJ
 *      then then weight=symbolw[0]-10. Otherwise weight=symbolw[0].
 *
 *    Weight 3:
 *    --------
 *    - The weight is computed as weight=symbolw[1]-10.
 *
 *    Weight 4:
 *    --------
 *    - The weight is computed as weight=symbolw[0]-10.
 *
 *   VERY IMPORTANT: lowering of maximum clause weight allowed for clause
 *   retention is done in LRSClean() function based only on the last weight
 *   set (weight 4).
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void ClauseWeight(cmprefix *clause){
	auto double *clvector;                               /* Pointer to clause learning vector */
	auto double dd;                                      /* Auxiliary */
	auto int32_t ii;                                     /* Auxiliary */

	/* Return if clause weight has been already set. */
	if (clause->flags&WEIGHTED){
		return;
	}

	/* Compute the symbolw[] fileds. */
	UpdateClauseSymbolWeights(clause->part2.bin);

	/* Clause weight 1. */
	if (clause->part2.bin->dist2conj==0){
		clause->part2.bin->clweight[1]=0;
	} else if (clause->inference==CLAUSIFY){
		clause->part2.bin->clweight[1]=1;
	} else {
		clause->part2.bin->clweight[1]=(int)(clause->part2.bin->symbolw[0]);
	}

	/* Clause weight 2. */
	if (clause->part2.bin->dist2conj<MAXDIST2CONJ){
		if (clause->part2.bin->symbolw[0]>=10.0){
			clause->part2.bin->clweight[2]=(int)(clause->part2.bin->symbolw[0])-10;
		} else {
			clause->part2.bin->clweight[2]=0;
		}
	} else {
		clause->part2.bin->clweight[2]=(int)(clause->part2.bin->symbolw[0]);
	}

	/* Clause weight 3. */
	if (clause->part2.bin->symbolw[1]>=10.0){
		clause->part2.bin->clweight[3]=(int)(clause->part2.bin->symbolw[1]-10.0);
	} else {
		clause->part2.bin->clweight[3]=0;
	}

	/* Clause weight 4. */
	if (clause->part2.bin->symbolw[0]>=10.0){
		clause->part2.bin->clweight[4]=(int)(clause->part2.bin->symbolw[0]-10.0);
	} else {
		clause->part2.bin->clweight[4]=0;
	}

	/* Clause weight 0. This is done last because because it uses one of the */
	/* previous weights. Currently it uses weight 1. If this is modified then */
	/* please modify also the global comments. */
	if (NULL!=kbset.lrnvector){
		clause->part2.bin->clweight[0]=WEIGHT0FACTOR*clause->part2.bin->clweight[1];
		clvector=GetClauseVector(clause);
		if (clvector!=NULL){
			for (ii=0,dd=0.0;ii<VECTORLENGTH;ii++){
				dd+=(clvector[ii]*kbset.lrnvector[ii]);
			}
			MYFREE(clvector);
			if (dd>0.0){
				clause->flags|=GOODCHOICE;
			}
			if (0==(GOODCHOICE&clause->flags)){
				clause->part2.bin->clweight[0]+=BADCLAUSEWEIGHT;
			}
		}
	}

	/* Indicate that clause has been weighted and return. */
	clause->flags|=WEIGHTED;
	return;
} /* ClauseWeight */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Delete unreachable passive clauses)  OCJ
 *
 *    This function deletes unreachable passive clauses for a KB
 *    given the number of reachable clauses.
 *
 *    VERY IMPORTANT: lowering of maximum clause weight allowed for clause
 *    retention is done based only on the last weight set (weight 4).
 *
 *
 *  ARGUMENTS:
 *
 *    reachcl: Number of reachable passive clauses.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t LRSClean(int32_t reachcl){
	auto cmprefix *wnxtclause[WEIGHTQUEUES];             /* Pointer to next clause in each weight queue to be signed by weight */
	auto cmprefix *anxtclause;                           /* Pointer to next clause to be signed by age */
	auto cmprefix *auxcl1,*auxcl2;                       /* Auxiliary clauses */
	auto int32_t selectcnt;                              /* Current number of selected clauses from queue indexed by selectidx */
	auto char *ptr1;                                     /* Auxiliary pointer */
	auto int32_t ii,jj,kk;                               /* Auxiliary */

	/* Indicate that LRS has been used. */
	kbset.flags|=RETENTIONUSED;

	/* Get first clause in weight and age queues. */
	selectcnt=kbset.selectcnt;
	for (ii=0;ii<WEIGHTQUEUES;ii++){
		wnxtclause[ii]=kbset.frstglweight[ii];
	}
	anxtclause=kbset.frstpassive;

	/* Sign the reachable passive clauses. */
	kbset.signature++;
	for (ii=0;ii<reachcl;ii++){

		/* Select by weight. */
		if (NULL==kbset.lrnvector){
			kk=slctordr[selectcnt];
		} else {
			kk=slctordr2[selectcnt];
		}
		if (kk<WEIGHTQUEUES){

			/* Jump over already signed clauses. */
			for (;wnxtclause[kk]->part2.bin->signature==kbset.signature;
					wnxtclause[kk]=wnxtclause[kk]->part2.bin->nextweight[kk]){
			}

			/* Sign clause and update index and counter. */
			wnxtclause[kk]->part2.bin->signature=kbset.signature;
			wnxtclause[kk]=wnxtclause[kk]->part2.bin->nextweight[kk];
			selectcnt++;
			if (NULL==kbset.lrnvector){
				if (selectcnt==slctordrsize){
					selectcnt=0;
				}
			} else {
				if (selectcnt==slctordrsize2){
					selectcnt=0;
				}
			}

		/* Select by age. */
		} else {

			/* Jump over already signed clauses. */
			for (;anxtclause->part2.bin->signature==kbset.signature;
					anxtclause=anxtclause->next){
			}

			/* Sign clause and update index and counter. */
			anxtclause->part2.bin->signature=kbset.signature;
			anxtclause=anxtclause->next;
			selectcnt++;
			if (NULL==kbset.lrnvector){
				if (selectcnt==slctordrsize){
					selectcnt=0;
				}
			} else {
				if (selectcnt==slctordrsize2){
					selectcnt=0;
				}
			}
		}
	}

	/* Dispose of unsigned clauses. */
	for (auxcl1=wnxtclause[0];auxcl1!=NULL;auxcl1=auxcl2){
		auxcl2=auxcl1->part2.bin->nextweight[0];
		if (auxcl1->part2.bin->signature!=kbset.signature){
			if (NULL==(ptr1=Decompile(auxcl1->part2.bin))){
				return(NOMEMORY);
			}
			if (NOMEMORY==Unlink(auxcl1,&kbset)){
				return(NOMEMORY);
			}
			FreeEquData(auxcl1->part2.bin);
			if (auxcl1->part2.bin->asserts!=NULL){
				MYFREE(auxcl1->part2.bin->asserts);
			}
			MYFREE(auxcl1->part2.bin);
			auxcl1->part2.text=ptr1;
			AddTxtFormula2KB(auxcl1,auxcl1->inference,&kbset,0);
		}
	}

	/* Update current maximum clause weight to the maximum existing weight */
	/* in the last weight queue. */
	for (auxcl1=kbset.lstweight[WEIGHTQUEUES-1][MAXWEIGHT],jj=MAXWEIGHT;auxcl1==NULL;
			auxcl1=kbset.lstweight[WEIGHTQUEUES-1][--jj]){
	}
	if (kbset.opts.maxweight>auxcl1->part2.bin->clweight[WEIGHTQUEUES-1]){
		kbset.opts.maxweight=auxcl1->part2.bin->clweight[WEIGHTQUEUES-1];
	}

	return(0);
} /* LRSClean */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Set select order for clause selection queues)  OCJ
 *
 *    This function initializes the sets pointed by slctordr and
 *    slctord2 pointers. The sets are allocated or reallocated if
 *    necessary. They contain WEIGHTQUEUES+1 elements, one for each
 *    extraction from the age and weight queues during a round-robin
 *    clause selection cycle. Each element indicates from what queue
 *    the extraction must be done.
 *
 *    The slctordrsize and slctordsize2 global variables are also
 *    initialized.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t SetClSelectOrder(void){
	auto int32_t *ptr1;                                  /* Auxiliary pointer */
	auto int32_t ii,jj,kk,mm;                            /* Auxiliary */
	#ifdef PROGCLAUSESELECT
	auto float ff,gg,hh;                                 /* Auxiliary */
	#endif

	/* Count the number of extractions. */
	for (ii=jj=mm=0;ii<=WEIGHTQUEUES;ii++){
		jj+=queuerts[ii];
		mm+=queuerts2[ii];
	}

	/* Allocate or reallocate slctordr and slctord2 set if necessary */
	/* and set slctordrsize and slctordrsize2. */
	if (slctordr==NULL){
		if (NULL==(slctordr=mallocmn(jj*sizeof(*slctordr)))){
			return(NOMEMORY);
		}
		slctordrsize=jj;
	} else if (slctordrsize!=jj){
		if (NULL==(ptr1=reallocmn(slctordr,jj*sizeof(*slctordr)))){
			return(NOMEMORY);
		}
		slctordr=ptr1;
		slctordrsize=jj;
	}
	if (slctordr2==NULL){
		if (NULL==(slctordr2=mallocmn(mm*sizeof(*slctordr2)))){
			MYFREE(slctordr);
			slctordr=NULL;
			return(NOMEMORY);
		}
		slctordrsize2=mm;
	} else if (slctordrsize2!=mm){
		if (NULL==(ptr1=reallocmn(slctordr2,mm*sizeof(*slctordr2)))){
			MYFREE(slctordr);
			slctordr=NULL;
			return(NOMEMORY);
		}
		slctordr2=ptr1;
		slctordrsize2=mm;
	}

	#ifdef PROGCLAUSESELECT
	/* Set slctordr elements by extracting progressively from each queue. */
	/* First allocate a set of WEIGHTQUEUES+1 integers to store the */
	/* number of extractions done so far from each queue. */
	if (NULL==(ptr1=allocmn(WEIGHTQUEUES+1,sizeof(int32_t)))){
		return(NOMEMORY);
	}
	memset(ptr1,0,(WEIGHTQUEUES+1)*sizeof(int32_t));

	/* Loop through all queue extractions */
	/* when learning process disabled. */
	for (kk=0;kk<slctordrsize;kk++){

		/* Loop through all candidate queues to get extraction queue. */
		ff=(float)(kk+1.0)/(float)slctordrsize;
		gg=(ff*queuerts[wqdeplord[0]])-ptr1[wqdeplord[0]];
		for (ii=1,jj=wqdeplord[0];ii<WEIGHTQUEUES;ii++){
			hh=(ff*queuerts[wqdeplord[ii]])-ptr1[wqdeplord[ii]];
			if (gg<hh){
				jj=wqdeplord[ii];
				gg=hh;
			}
		}
		hh=(ff*queuerts[WEIGHTQUEUES])-ptr1[WEIGHTQUEUES];
		if (gg<hh){
			jj=WEIGHTQUEUES;
			gg=hh;
		}
		slctordr[kk]=jj;
		ptr1[jj]++;
	}

	/* Loop through all queue extractions. */
	/* when learning process enabled. */
	memset(ptr1,0,(WEIGHTQUEUES+1)*sizeof(int32_t));
	for (kk=0;kk<slctordrsize2;kk++){

		/* Loop through all candidate queues to get extraction queue. */
		ff=(float)(kk+1.0)/(float)slctordrsize2;
		gg=(ff*queuerts2[wqdeplord2[0]])-ptr1[wqdeplord2[0]];
		for (ii=1,jj=wqdeplord2[0];ii<WEIGHTQUEUES;ii++){
			hh=(ff*queuerts2[wqdeplord2[ii]])-ptr1[wqdeplord2[ii]];
			if (gg<hh){
				jj=wqdeplord2[ii];
				gg=hh;
			}
		}
		hh=(ff*queuerts2[WEIGHTQUEUES])-ptr1[WEIGHTQUEUES];
		if (gg<hh){
			jj=WEIGHTQUEUES;
			gg=hh;
		}
		slctordr2[kk]=jj;
		ptr1[jj]++;
	}
	MYFREE(ptr1);

	#else
	/* Set slctordr elements by depleting each queue one after the other */
	/* when learning process disabled. */
	for (ii=jj=0;ii<WEIGHTQUEUES;ii++){
		for (kk=0;kk<queuerts[wqdeplord[ii]];jj++,kk++){
			slctordr[jj]=wqdeplord[ii];
		}
	}
	for (;jj<slctordrsize;jj++){
		slctordr[jj]=WEIGHTQUEUES;
	}

	/* Set slctordr elements by depleting each queue one after the other */
	/* when learning process enabled. */
	for (ii=jj=0;ii<WEIGHTQUEUES;ii++){
		for (kk=0;kk<queuerts2[wqdeplord2[ii]];jj++,kk++){
			slctordr2[jj]=wqdeplord2[ii];
		}
	}
	for (;jj<slctordrsize2;jj++){
		slctordr2[jj]=WEIGHTQUEUES;
	}
	#endif

	return(0);
} /* SetClSelectOrder */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform problem pre-processing)  OCJ
 *
 *    This function performs the pre-processing needed before the
 *    start of the saturation process or SAT stand alone execution:
 *    - Convert formulas to CNF standard form.
 *    - Perform relevance filtering.
 *    - Update the HORN flag of unprocessed clauses and pbmflags
 *      global variable.
 *
 *
 *  ARGUMENTS:
 *
 *    flag: 1 if called from Saturate, 2 if called from Clausify
 *          function, 0 otherwise.
 *
 *  RETURNS:
 *
 *    NOMEMORY -> Not enough memory
 *    TIMEOUT -> Timeout condition during pre-process
 *    If the function was called from Clausify function (flag==2)
 *    and there was a formula that generated a $false then the
 *    role of that formula is returned, otherwise 0 is returned.
 *
 *
 *--------------------------------------------------------------*/
int32_t Preprocess(int32_t flag){
	auto char *buffer;                               /* Buffer for CNF formula */
	auto int32_t size;                               /* Buffer size in bytes */
	auto cmprefix *ptr1,*ptr2;                       /* Auxiliary pointers */
	auto uint8_t *ptr3;                              /* Auxiliar pointer */
	auto uint64_t ii,jj,kk,nn,mm,pp;                 /* Auxiliary */

	/* Allocate space for CNF conversion. */
	if (NULL==(buffer=mallocmn(2*STR_CHUNK_SIZE))){
		return(NOMEMORY);
	}
	size=2*STR_CHUNK_SIZE;

	/* Loop through text formulas in the main KB. */
	if ((verbose)&&(pgmmode==INTERACTIVE)){
		printf("Converting formulas to CNF...\n");
	}
	pp=ii=0; /* Just to avoid compiler warnings. */
	for (ptr1=mainkb.firsttxt;ptr1!=NULL;ptr1=ptr1->next){

		/* Formula is of type axiom like, NEGCONJ or PREDICATEDEF. */
		if (((PLAIN>=(ii=ptr1->inference))&&(ptr1->inference!=CONJECTURE))||(ptr1->inference==PREDICATEDEF)){

			/* Text formula pre-processing. Pre-processing is repeated in */
			/* case of premature return due to miniscoping time limit */
			/* exceeded. */
			if (NULL==(ptr2=PreprocessFormula(ptr1,&buffer,&size,1))){
				if ((procctl->status!=NOMEMORY)&&(procctl->status!=TIMEOUT)){
					if (NULL==(ptr2=PreprocessFormula(ptr1,&buffer,&size,0))){
						break;
					}
				} else {
					break;
				}
			}

			/* Formula is $false, refutation found. */
			if (0==strcmp("$false",ptr2->part2.text)){
				procctl->status=UNSATISFIABLE;
				procctl->solvekb=active_cores;
				if (flag==1){
					if (NULL!=(mainkb.frstprfnode=mallocmn(sizeof(proofnode)))){
						mainkb.frstprfnode->type=ACLAUSE;
						mainkb.frstprfnode->ptr.Aclause=ptr2;
					}
				}
				pp=ii;
				break;

			/* Formula is not $false or $true, set FROMCONJECTURE flag */
			/* if appropriate and compile CNF formula. */
			} else if (0!=strcmp("$true",ptr2->part2.text)){
				if (ptr1->inference==NEGCONJ){
					ptr2->flags|=FROMCONJECTURE;
				} else {
					ptr2->flags|=(FROMCONJECTURE&ptr1->flags);
				}
				if (NOMEMORY==CompileCNF(buffer,ptr2)){
					procctl->status=NOMEMORY;
					break;;
				}
			}
		}
	}

	/* Free CNF conversion memory. */
	MYFREE(buffer);

	/* Timeout or not enough memory from PreprocessFormula(). */
	if ((procctl->status==TIMEOUT)||(procctl->status==NOMEMORY)){
		return(procctl->status);
	}

	/* Update the HORN flag in each clause and UNITEQUALITY and UNITCLAUSE */
	/* flags of pbmflags global variable. */
	/* Loop through unprocessed clauses. */
	/* ii=1 if there are horn clauses, otherwise ii=0. */
	/* jj=1 if there are non horn clauses, otherwise jj=0. */
	/* nn=1 if there are equalities, otherwise nn=0. */
	jj=nn=0;
	if (procctl->status==UNSATISFIABLE){
		ii=1;
	} else {
		ii=0;
		for (ptr1=mainkb.frstunproc;ptr1!=NULL;ptr1=ptr1->next){

			/* Update UNITCLAUSE and UNITEQUALITY flags. */
			if (ptr1->part2.bin->literals>1){
				pbmflags&=(~UNITCLAUSE);
			} else if (EQUALITY==((NEGATED|EQUALITY)&ptr1->part2.bin->formula[0])){
				pbmflags|=UNITEQUALITY;
			}

			/* Loop through clause items. Set kk to the number of positive */
			/* literals and remove the GROUND flag if clause is not ground. */
			for (ptr3=&ptr1->part2.bin->formula[0],kk=0;(*ptr3)!=UNITEND;
					ptr3=NextItem(ptr3,IMMED)){
				if ((PREDICATE|EQUALITY)&ptr3[0]){
					if (0==(NEGATED&ptr3[0])){
						kk++;
					}
					if (EQUALITY&ptr3[0]){
						nn=1;
					}
				} else if (VARIABLE&ptr3[0]){
					pbmflags&=(~GROUND);
				}
			}

			/* Update ii, jj and clause HORN flag if appropriate. */
			if (kk>1){
				jj=1;
			} else {
				ptr1->flags|=HORN;
				ii=1;
			}
		}
	}

	/* Update NONUNITEQUALITY, HORN and WITHHORNCLAUSES in pbmflags global variable. */
	if (nn&&(0==(pbmflags&UNITEQUALITY))){
		pbmflags|=NONUNITEQUALITY;
	}
	if (ii){
		if (jj){
			pbmflags|=WITHHORNCLAUSES;
		} else {
			pbmflags|=HORN;
		}
	}

	/* If the function was called from Clausify() then accumulate pre-processing elapsed time */
	/* (only if a refutation was found) and return. */
	if (flag==2){
		if (procctl->status==UNSATISFIABLE){
			gettimeofday(&mainkb.endtime,NULL);
			mainkb.prstats.elapsed_time+=(mainkb.endtime.tv_sec-mainkb.time.tv_sec
					+(mainkb.endtime.tv_usec-mainkb.time.tv_usec)/1000000.0);
			pbmstats->elapsed_time=mainkb.prstats.elapsed_time;
			return(pp);
		}
		return(0);
	}

	/* Perform relevance filtering. */
	GetPbmTypeNumber();
	mm=ClauseRelevance();

	/* Accumulate preprocessing elapsed time. */
	gettimeofday(&mainkb.endtime,NULL);
	mainkb.prstats.elapsed_time+=(mainkb.endtime.tv_sec-mainkb.time.tv_sec
			+(mainkb.endtime.tv_usec-mainkb.time.tv_usec)/1000000.0);
	pbmstats->elapsed_time=mainkb.prstats.elapsed_time;

	/* Print preprocessing statistics. */
	if ((verbose)&&(pgmmode==INTERACTIVE)){
		printf("\nPreprocessing statistics:\n");
		PrintStatData(pbmstats,0);
		if (mm){
			printf("%ld clause(s) deleted by relevance filtering\n",mm);
		}
	}
	if (pgmmode==INTERACTIVE){
		if (flag==1){
			printf("Problems processed: %d, problems solved: %d, now working on problem number %d\n",
					tot_pbms,solved_pbms,tot_pbms+1);
		} else {
			printf("Problems processed: %d, now working on problem number %d\n",tot_pbms,tot_pbms+1);
		}
	}

	return(0);
} /* Preprocess */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform relevance filtering)  OCJ
 *
 *    This function select clauses to be used in the saturation process.
 *    Pruning starts with the function and predicate symbols in the goal.
 *    Every axiom that shares such a symbol is considered relevant. Symbols
 *    in relevant axioms become relevant themselves. The process is then
 *    repeated for the time specified in the program options. This relevance
 *    pruning is complete in the non-equational case if allowed to reach a
 *    fixed point. It only provides a relatively coarse measure, however.
 *
 *    If the global variable prproflags has the CLFILTERINGITERLIMIT flag
 *    then the global variable ppthrshld contains the maximun number of
 *    iterations allowed. Otherwise the global variable ppthrshld contains
 *    the maximum allowed time for the algorithm.
 *
 *    A Sine algorithm can be added in the future for a more fine clause
 *    selection.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    Number of deleted clauses by the relevance filter.
 *
 *
 *--------------------------------------------------------------*/
uint64_t ClauseRelevance(void){
	struct timeval sttime,ctime;                         /* Function start and current times */
	auto double timedif;                                 /* Function elapsed time */
	auto cmprefix *ptr1,*ptr2;                           /* Auxiliary pointers */
	auto uint8_t *ptr3;                                  /* Auxiliary pointer */
	auto int32_t ii,jj,mm,nn;                            /* Auxiliary */
	auto uint64_t kk;                                    /* Auxiliary */

	/* Clear CLFILTERINGLIMITREACHED and CLFILTERINGPRUNNED flag from previous jobs. */
	prproflags&=~(CLFILTERINGLIMITREACHED|CLFILTERINGPRUNNED);

	/* Return if no clause filtering must be done. */
	if ((0==(pbmflags&(EXISTCONJ|EXISTNEGCONJ)))||(prproflags&CLFILTERINGOFF)
			||((prproflags&CLFILTERINGNONEQU)&&((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))){
		return(0);
	}

	/* Get start time if necessary. */
	if ((ppthrshld>0.0)&&(0==(prproflags&CLFILTERINGITERLIMIT))){
		gettimeofday(&sttime,NULL);
	}

	/* Set ptr2 to first main KB unprocessed clause and loop until */
	/* no more clauses are flagged as kept or a timeout occurs. */
	ptr2=mainkb.frstunproc;
	mm=0; /* No clause coming from negated conjecture has been processed yet. */
	for (ii=nn=1;ii;nn++){
		ii=0;

		/* Loop through main KB unprocessed clauses. The pointer ptr2 */
		/* is set to first non kept clause in each iteration. */
		for (ptr1=ptr2;ptr1!=NULL;ptr1=ptr1->next){
			if (ptr1==ptr2){
				ptr2=NULL;
			}

			/* Clause is not flagged as kept. */
			if (0==(KEPTCLAUSE&ptr1->flags)){

				/* Clause is coming from a negated conjecture. */
				if (ptr1->part2.bin->dist2conj==0){
					jj=mm=1;

				/* Clause is not coming from a negated conjecture. */
				} else {

					/* At least one clause coming from conjecture has been processed. */
					if (mm){

						/* Loop through clause items and check if any of its */
						/* predicate or function symbols are flagged as RELEVANT. */
						for (ptr3=&ptr1->part2.bin->formula[0],jj=0;(jj==0)&&(ptr3[0]!=UNITEND);ptr3=NextItem(ptr3,IMMED)){
							if (((PREDICATE|FUNCTION)&ptr3[0])&&(RELEVANT&((symbol *)&ptr3[1])->symbol->type)){
								jj=1;
							}
						}

					/* No clause coming from negated conjecture has been processed yet. */
					} else {
						jj=0;
					}
				}

				/* Clause must be flagged as kept. Flag the clause */
				/* and all its predicates and functions. Also indicate */
				/* that main loop must be repeated. */
				if (jj){
					ptr1->flags|=KEPTCLAUSE;
					for (ptr3=&ptr1->part2.bin->formula[0];ptr3[0]!=UNITEND;ptr3=NextItem(ptr3,IMMED)){
						if ((PREDICATE|FUNCTION)&ptr3[0]){
							((symbol *)&ptr3[1])->symbol->type|=RELEVANT;
						}
					}
					ii=1;

				/* Clause must not be flagged as kept. Set ptr2 if not already set. */
				} else {
					if (ptr2==NULL){
						ptr2=ptr1;
					}
				}
			}
		}

		/* Check timeout or number of iterations. */
		if (ppthrshld>0.0){
			if (prproflags&CLFILTERINGITERLIMIT){
				if (nn>=ppthrshld){
					if (ii){
						prproflags|=CLFILTERINGLIMITREACHED;
					}
					break;
				}
			} else {
				gettimeofday(&ctime,NULL);
				timedif=ctime.tv_sec-sttime.tv_sec
						+(ctime.tv_usec-sttime.tv_usec)/1000000.0;
				if (timedif>ppthrshld){
					if (ii){
						prproflags|=CLFILTERINGLIMITREACHED;
					}
					break;
				}
			}
		}
	}

	/* Delete non flagged clauses. Unlink() function cannot */
	/* return NOMEMORY because all the clauses are unprocessed. */
	/* None of the clauses have equality data nor asserts so */
	/* there is no need to free that storage. */
	if ((0==(prproflags&CLFILTERINGLIMITREACHED))||(0==(prproflags&CLFILTERINGRESIGN))){
		for (ptr1=mainkb.frstunproc,kk=0;ptr1!=NULL;ptr1=ptr2){
			ptr2=ptr1->next;
			if (0==(KEPTCLAUSE&ptr1->flags)){
				Unlink(ptr1,&mainkb);
				MYFREE(ptr1->part2.bin);
				MYFREE(ptr1);
				kk++;
				prproflags|=CLFILTERINGPRUNNED;
			} else {
				ptr1->flags&=~(KEPTCLAUSE);
			}
		}
	} else {
		kk=0;
	}

	return(kk);
} /* ClauseRelevance */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Print problem type flags)  OCJ
 *
 *    This function prints the problem type flags in pbmflags global
 *    variable. See global.h for the flags in pbmflags.
 *
 *    In order to do so the problem must be pre-processed.
 *
 *    All main KB (and therefore all KB's data) will be initialized
 *    at the end of the execution of this function.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void GoType(void){
	auto char *uniteq,*unitcl,*horn,*ground,*eq;         /* To store "Y" or "N" for the different simple problem types */
	auto int32_t ii;                                     /* Auxiliary */

	/* Start time measurement. */
	procctl->start=clock();
	gettimeofday(&mainkb.time,NULL);

	/* Main KB is empty. */
	if (mainkb.nformulas==0){
		printf("Main KB has no input formulas to process.\n");
		return;
	}

	/* Perform pre-processing. */
	ii=Preprocess(0);
	if (ii){

		/* Print error information. */
		if (ii==NOMEMORY){
			printf("Not enough memory.\n");
		} else {
			printf("Process timeout\n");
		}
		printf("\n      UNIT_EQUALITY UNIT_CLAUSE HORN GROUND NONUNIT_EQU PBM_TYPE\n");
		printf("====>      ?             ?       ?      ?         ?        ?\n");
		FreeInfFormulas();
		if (NOMEMORY==Initialize_KB(&mainkb)){
			printf("No memory available, KB not initialized\n");
		}
		glblstats.elapsed_time+=mainkb.prstats.elapsed_time;
		ResetStats(pbmstats);
		procctl->status=STOPPED;
		tot_pbms++;

		/* Stop time measurement, collect CPU time and return. */
		procctl->end=clock();
		tot_cpu_time+=((double)(procctl->end-procctl->start))/CLOCKS_PER_SEC;
		return;
	}

	/* Display results. */
	if (pbmflags&UNITEQUALITY){
		uniteq="Y";
	} else {
		uniteq="N";
	}
	if (pbmflags&UNITCLAUSE){
		unitcl="Y";
	} else {
		unitcl="N";
	}
	if (pbmflags&HORN){
		horn="Y";
	} else {
		horn="N";
	}
	if (pbmflags&GROUND){
		ground="Y";
	} else {
		ground="N";
	}
	if (pbmflags&NONUNITEQUALITY){
		eq="Y";
	} else {
		eq="N";
	}
	GetPbmTypeNumber();
	printf("      UNIT_EQUALITY UNIT_CLAUSE HORN GROUND NONUNIT_EQU PBM_TYPE\n");
	printf("====>      %s             %s       %s      %s         %s        %d\n",
			uniteq,unitcl,horn,ground,eq,1+pbmtype);

	/* Collect memory use information. */
	if (glblstats.memory<allcdmemory){
		glblstats.memory=allcdmemory;
	}

	/* Collect inference statistics and initialize main KB. */
	AddInferenceStats(pbmstats,&glblstats);
	FreeInfFormulas();
	if (NOMEMORY==Initialize_KB(&mainkb)){
		printf("No memory available, KB not initialized\n");
	}
	glblstats.elapsed_time+=mainkb.prstats.elapsed_time;
	ResetStats(pbmstats);
	tot_pbms++;

	/* Stop time measurement, collect CPU time and return. */
	procctl->status=STOPPED;
	procctl->end=clock();
	tot_cpu_time+=((double)(procctl->end-procctl->start))/CLOCKS_PER_SEC;

	return;
} /* GoType */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Convert problem formulas to CNF and print them in TPTP format)  OCJ
 *
 *    This function converts problem formulas to CNF and prints them
 *    in TPTP format.
 *
 *    In order to do so the problem must be pre-processed.
 *
 *    All main KB (and therefore all KB's data) will be initialized
 *    at the end of the execution of this function.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void Clausify(void){
	auto int32_t ii;                                     /* Auxiliary */

	/* Start time measurement. */
	procctl->start=clock();
	gettimeofday(&mainkb.time,NULL);
	kbset.time=mainkb.time;

	/* Main KB is empty. */
	if (mainkb.nformulas==0){
		printf("Main KB has no input formulas to process.\n");
		return;
	}

	/* Perform pre-processing. */
	ii=Preprocess(2);
	if ((ii==NOMEMORY)||(ii==TIMEOUT)){

		/* Print error information. */
		if (ii==NOMEMORY){
			printf("Not enough memory.\n");
		} else {
			printf("Process timeout\n");
		}
		FreeInfFormulas();
		if (NOMEMORY==Initialize_KB(&mainkb)){
			printf("No memory available, KB not initialized\n");
		}
		ResetStats(pbmstats);
		procctl->status=STOPPED;
		tot_pbms++;

		/* Stop time measurement, collect CPU time and return. */
		procctl->end=clock();
		tot_cpu_time+=((double)(procctl->end-procctl->start))/CLOCKS_PER_SEC;
		return;
	}

	/* Display results. */
	if (NOMEMORY==PrintClausifiedFormulas(ii)){
		printf("No memory available, printed information is not complete.\n");
	}

	/* Collect memory use information. */
	if (glblstats.memory<allcdmemory){
		glblstats.memory=allcdmemory;
	}

	/* Collect inference statistics and initialize main KB. */
	AddInferenceStats(pbmstats,&glblstats);
	FreeInfFormulas();
	if (NOMEMORY==Initialize_KB(&mainkb)){
		printf("No memory available, KB not initialized\n");
	}
	ResetStats(pbmstats);
	tot_pbms++;

	/* Stop time measurement and collect elapsed and CPU times. */
	procctl->end=clock();
	tot_cpu_time+=((double)(procctl->end-procctl->start))/CLOCKS_PER_SEC;
	gettimeofday(&mainkb.endtime,NULL);
	glblstats.elapsed_time+=(mainkb.endtime.tv_sec-mainkb.time.tv_sec
			+(mainkb.endtime.tv_usec-mainkb.time.tv_usec)/1000000.0);
	procctl->status=STOPPED;

	return;
} /* Clausify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Select strategies for current problem)  OCJ
 *
 *    This function selects the strategies to be used for the current
 *    problem being solved. The selected strategies are set in the
 *    strategies[] global variable. Each strategy selected is a 16 bit
 *    unsigned integer with the values for each type of adjustable parameter.
 *    See definitions of adjustable parameters in global.h.
 *
 *    The best substrategies for each type of problem are stored in
 *    pbmtypestrats[ll][mm][nn] global variable, where:
 *    ll: 0 for strategies without learning, 1 for strategies with learning.
 *    mm: This is problem type index.
 *    nn: This is the index to the strategy itself.
 *
 *    This function selects more strategies than necessary in principle.
 *    This is done to deal with strategy runs that end with no solution
 *    in less time than available due to strategy incompleteness. Of course
 *    the number of possible strategies is limited by the number of active
 *    cores and the number of time slices.
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t InitStrategies(void){
	auto int32_t sel_strats;                             /* Number of currently selected strategies */
	auto int8_t *usedstrats;                             /* Flags for used strategies. */
	auto int32_t ii,jj,kk;                               /* Auxiliary */
	auto uint16_t mm;                                    /* Auxiliary */

	/* Calculate problem type number, read learning if applicable and initialize */
	/* *nextstrat and nbstrats. Set jj=1 if learning is enabled, 0 otherwise. */
	*nextstrat=0;
	if (NULL!=learnfrom){
		if (ReadLearnRecords()){
			jj=0;
		} else {
			jj=1;
		}
	} else {
		jj=0;
	}
	if ((active_cores==1)&&(timeslices==1)){
		nbstrats=1;
	} else {
		nbstrats=active_cores*(timeslices+BKSTRATEGIESPERPROC);
		if (nbstrats>MAXSTRATEGIES){
			nbstrats=MAXSTRATEGIES;
		}
	}

	/* Temporarily reduce the number of time slices if appropriate. It will be */
	/* reset to its original value by Saturate() function. */
	if (MAXSTRATEGIES<=(active_cores*(timeslices-1))){
		timeslices=(MAXSTRATEGIES+active_cores-1)/active_cores;
	}

	/* Get the canonical strategies for the problem. */
	for (sel_strats=0;(sel_strats<TUNEDSTRATEGIES)&&(sel_strats<nbstrats);
			sel_strats++){
		strategies[sel_strats]=StratNb2StratFlags(pbmtypestrats[jj][pbmtype][sel_strats]);
	}

	/* There are user parameter settings. */
	if (usrparam_mask){

		/* Check if user parameter settings match any of the selected strategies. */
		for (kk=0,ii=1;kk<sel_strats;kk++){
			if (usrparam==(strategies[kk]&usrparam_mask)){
				ii=0;
				break;
			}
		}

		/* If none of the selected strategies match the user parameter settings */
		/* then shift one position the selected strategies and combine the */
		/* user parameter settings with the first selected strategy. */
		if (ii){
			for (kk=(sel_strats==nbstrats?sel_strats-1:sel_strats);kk>0;kk--){
				strategies[kk]=strategies[kk-1];
			}
			strategies[0]=usrparam|(strategies[0]&(~usrparam_mask));
			if (sel_strats<nbstrats){
				sel_strats++;
			}
		}
	} else {
		ii=0;
	}

	/* More strategies must be selected. */
	if (sel_strats<nbstrats){

		/* There is enough memory for used strategies flags.*/
		if (NULL!=(usedstrats=mallocmn(MAXSTRATEGIES))){

			/* Set flags of used strategies without user specified strategy. */
			memset(usedstrats,0,MAXSTRATEGIES);
			for (kk=ii;kk<sel_strats;kk++){
				usedstrats[pbmtypestrats[jj][pbmtype][kk]]=1;
			}

			/* Add additional strategies. The additional strategies are selected */
			/* from non used strategies without / with learning alternatively. */
			for (kk=0;sel_strats<nbstrats;kk++){
				if (0==usedstrats[kk]){
					mm=StratNb2StratFlags(kk);
					if (mm!=strategies[0]){
						strategies[sel_strats]=mm;
						sel_strats++;
					}
				}
				if ((NULL!=learnfrom)&&(0==usedstrats[kk+(MAXSTRATEGIES/2)])&&(sel_strats<nbstrats)){
					mm=StratNb2StratFlags(kk+(MAXSTRATEGIES/2));
					if (mm!=strategies[0]){
						strategies[sel_strats]=mm;
						sel_strats++;
					}
				}
			}

			/* Free allocated memory. */
			MYFREE(usedstrats);

		/* Not enough memory. */
		} else {
			return(NOMEMORY);
		}
	}

	return(0);
} /* InitStrategies */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Calculate problem type number)  OCJ
 *
 *    This function calculates a problem type number according to
 *    the following:
 *    LEGEND:
 *    ------
 *    UE: Problem has some positive unit equalities.
 *    NUE: Problem has some equalities none of which are positive unit
 *         equalities.
 *    UC: All clauses in problem are unit clauses.
 *    H: Problem is horn, that is, all clauses contain at most only
 *       one positive literal.
 *    G: Problem is ground, there are no variables in the clauses.
 *
 *    The problem type number is stored in pbmtype global variable.
 *
 *    PROBLEM TYPE NUMBER:
 *    -------------------
 *    0 -> UE&~UC&~H&~G
 *    1 -> UE&~UC&H
 *    2 -> UE&~UC&~H&G
 *    3 -> UE&UC
 *    4 -> NUE&~UC&~H y NUE&UC
 *    5 -> NUE&~UC&H
 *    6 -> ~UE&~NUE&~UC&~H y ~UE&~NUE&UC
 *    7 -> ~UE&~NUE&~UC&H
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    Problem type number.
 *
 *
 *--------------------------------------------------------------*/
void GetPbmTypeNumber(void){

	/* Calculate problem type number. */
	/* 0 -> UE&~UC&~H&~G */
	/* 1 -> UE&~UC&H */
	/* 2 -> UE&~UC&~H&G */
	/* 3 -> UE&UC */
	/* 4 -> NUE&~UC&~H y NUE&UC */
	/* 5 -> NUE&~UC&H */
	/* 6 -> ~UE&~NUE&~UC&~H y ~UE&~NUE&UC */
	/* 7 -> ~UE&~NUE&~UC&H */
	switch (pbmflags&(UNITEQUALITY|NONUNITEQUALITY|UNITCLAUSE|HORN|GROUND)){
		case UNITEQUALITY:
			pbmtype=0;
			break;
		case UNITEQUALITY|HORN|GROUND:
		case UNITEQUALITY|HORN:
			pbmtype=1;
			break;
		case UNITEQUALITY|GROUND:
			pbmtype=2;
			break;
		case UNITEQUALITY|UNITCLAUSE:
		case UNITEQUALITY|UNITCLAUSE|HORN:
		case UNITEQUALITY|UNITCLAUSE|GROUND:
		case UNITEQUALITY|UNITCLAUSE|HORN|GROUND:
			pbmtype=3;
			break;
		case NONUNITEQUALITY:
		case NONUNITEQUALITY|GROUND:
		case NONUNITEQUALITY|UNITCLAUSE:
		case NONUNITEQUALITY|UNITCLAUSE|HORN:
		case NONUNITEQUALITY|UNITCLAUSE|GROUND:
		case NONUNITEQUALITY|UNITCLAUSE|HORN|GROUND:
			pbmtype=4;
			break;
		case NONUNITEQUALITY|HORN:
		case NONUNITEQUALITY|HORN|GROUND:
			pbmtype=5;
			break;
		case 0:
		case GROUND:
		case UNITCLAUSE:
		case UNITCLAUSE|HORN:
		case UNITCLAUSE|GROUND:
		case UNITCLAUSE|HORN|GROUND:
			pbmtype=6;
			break;
		/* The following cases are redundant. They are left just for documentation. */
		/* The compiler optimization will remove the redundance. */
		case HORN:
		case HORN|GROUND:
		default:
			pbmtype=7;
			break;
	}
	/* The following is the same as above written in terms of conditionals. */
	/* It is commented out as documentation and clarification. */
	/*if (pbmflags&UNITEQUALITY){
		if (0==(pbmflags&UNITCLAUSE)){
			if (pbmflags&HORN){
				pbmtype=1;
			} else if (pbmflags&GROUND){
				pbmtype=2;
			} else {
				pbmtype=0;
			}
		} else {
			pbmtype=3;
		}
	} else if (pbmflags&NONUNITEQUALITY){
		if ((0==(pbmflags&UNITCLAUSE))&&(pbmflags&HORN)){
			pbmtype=5;
		} else {
			pbmtype=4;
		}
	} else if ((0==(pbmflags&UNITCLAUSE))&&(pbmflags&HORN)){
		pbmtype=7;
	} else {
		pbmtype=6;
	}*/

	return;
} /* GetPbmTypeNumber */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Convert a strategy number to strategy flags)  OCJ
 *
 *    This function converts an integer with the strategy number to
 *    a 16-bytes unsigned integer with strategy flags.
 *
 *
 *  ARGUMENTS:
 *
 *    stratnb: Strategy number.
 *
 *  RETURNS:
 *
 *    16-bit integer with strategy flags.
 *
 *--------------------------------------------------------------*/
uint16_t StratNb2StratFlags(int32_t stratnb){
	auto uint16_t ii;                                    /* Auxiliary */

	/* Set learning. */
	if (stratnb>=(MAXSTRATEGIES/2)){
		ii=LEARN_P;
		stratnb-=(MAXSTRATEGIES/2);
	} else {
		ii=0;
	}

	/* Set algorithm and maximum weight option. */
	if (stratnb<(MAXSTRATEGIES/3)){
		if (stratnb&1){
			ii|=(DISCOUNT_P|LRSOTT_P);
		} else {
			ii|=(OTTER_P|LRSOTT_P);
		}
		stratnb/=2;
	} else {
		ii|=(OTTER_P|LRSOFF_P);
		stratnb-=(MAXSTRATEGIES/3);
	}

	/* Set literal selection options. */
	switch (stratnb%5){
		case 0:
			ii|=SINGLENEG_P;
			break;
		case 1:
			ii|=MULTINEG_P;
			break;
		case 2:
			ii|=MAXIMAL_P;
			break;
		case 3:
			ii|=SINGLEPOS_P;
			break;
		case 4:
			ii|=MULTIPOS_P;
			break;
	}

	/* Set weight factors option for clause weight calculation. */
	if ((stratnb/5)%2){
		ii|=WFACT2_P;
	} else {
		ii|=WFACT1_P;
	}

	/* Set demodulation option. */
	switch ((stratnb/10)%3){
		case 0:
			ii|=DEMODOFF_P;
			break;
		case 1:
			ii|=DEMODON_P;
			break;
		case 2:
			ii|=DEMODCPL_P;
			break;
	}

	/* Set factoring inferences enabled/disabled. */
	if ((stratnb/30)%2){
		ii|=FACTOFF_P;
	} else {
		ii|=FACTON_P;
	}

	/* Set term ordering. */
	if ((stratnb/60)%2){
		ii|=TRMORDNR_P;
	} else {
		ii|=TRMORDSTD_P;
	}

	/* Set literal ordering. */
	switch ((stratnb/120)%3){
		case 0:
			ii|=LITORDSTD_P;
			break;
		case 1:
			ii|=LITORDNR_P;
			break;
		case 2:
			ii|=LITORDLEX_P;
			break;
	}

	/* Set function weight option. */
	if ((stratnb/360)%2){
		ii|=UNIFORM_P;
	} else {
		ii|=ARITY_P;
	}
	return(ii);
} /* StratNb2StratFlags */
