/*
 ============================================================================
 Name        : ordering.c
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */

/****************************************************************
*
*                 ordering (ordering functions module)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

/*-------------------------------------------------------------
 *
 *  DESCRIPTION:
 *
 *    This source contains the ordering functions of the Drodi package.
 *
 *    Drodi is a multi-platform Open Source Artificial Intelligence application.
 *    See Drodi.c for additional information.
 *
 *
 *  IMPLEMENTATION:
 *
 *    Multi-platform
 *
 *  NOTES:
 *
 *
 *--------------------------------------------------------------*/

/* Include for all C sources */
#include "global.h"

/** Global variables for this module: only those that need initialization in their definition. */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Literals comparison)  OCJ
 *
 *    This function performs a literals comparison. This comparison
 *    is based partially in a term ordering which in turn may be
 *    the standard or the non recursive Knuth-Bendix ordering,
 *    depending on the value of flag. The only addition is that
 *    literals that contain a predicate defined for sub-formula
 *    renaming are always LESSER than other types of literals.
 *
 *    The comparison is performed as indicated in the section 3.1.1
 *    of document "Implementing an efficient theorem prover"
 *    by Alexandre Riazanov. The return conditions of this
 *    function are:
 *
 *    GREATER if literal1 > literal2
 *    LESSER if literal2 > literal1
 *    EQUAL if literal2 = literal1
 *    FAILURE otherwise
 *
 *    See also "The design and implementation of VAMPIRE" section 2.6 by
 *    Alexandre Riazanov and Andrei Voronkov for details about handling
 *    positive and negative literal comparison logic.
 *
 *  ARGUMENTS:
 *
 *    literal1: Pointer to first literal to be compared.
 *    literal2: Pointer to second literal to be compared.
 *    coef1: Pointer to coeff structure for first literal,
 *           only used for STANDARD and NONRECURSIVE atom
 *           reduction orderings.
 *    coef2: Pointer to coeff structure for second literal,
 *           only used for STANDARD and NONRECURSIVE atom
 *           reduction orderings.
 *    flag: STANDARD, NONRECURSIVE or LEXICOGRAPHIC.
 *
 *  RETURNS:
 *
 *    GREATER, LESSER, EQUAL, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t CompareLiterals(uint8_t *literal1,uint8_t *literal2,coeff *coef1,
		coeff *coef2,int32_t flag){

	/* First literal is a predicate definition coming from sub-formula */
	/* renaming. */
	if ((PREDICATE&literal1[0])&&(DEFINEDPRED&((symbol *)&literal1[1])->symbol->type)){

		/* Second literal is not a predicate definition coming from sub-formula */
		/* renaming. */
		if ((0==(PREDICATE&literal2[0]))||(0==(DEFINEDPRED&((symbol *)&literal2[1])->symbol->type))){
			return(LESSER);
		}

	/* First literal is not a predicate definition coming from sub-formula */
	/* renaming. */
	/* Second literal is a predicate definition coming from sub-formula */
	/* renaming. */
	} else if ((PREDICATE&literal1[0])&&(DEFINEDPRED&((symbol *)&literal1[1])->symbol->type)){
		return(GREATER);
	}

	/* First literal is an equality. */
	if (EQUALITY&literal1[0]){

		/* Second literal is an equality. */
		if (EQUALITY&literal2[0]){
			return(CompareEqualities(literal1,literal2,coef1,coef2,flag));
		}

		/* Second literal is not an equality. */
		return(LESSER);

	/* First literal is not an equality, second literal is an equality. */
	} else if (EQUALITY&literal2[0]){
		return(GREATER);
	}

	/* None of the literals is an equality. */
	switch(ComparePredicates(literal1,literal2,coef1,coef2,flag)){
		case GREATER:
			return(GREATER);
			break;
		case LESSER:
			return(LESSER);
			break;
		case EQUAL:
			if ((MAXIMAL|SINGLENEG|MULTINEG)&kbset.opts.select){
				if ((NEGATED&literal1[0])&&(0==(NEGATED&literal2[0]))){
					return(GREATER);
				} else if ((NEGATED&literal2[0])&&(0==(NEGATED&literal1[0]))){
					return(LESSER);
				}
			} else {
				if ((NEGATED&literal1[0])&&(0==(NEGATED&literal2[0]))){
					return(LESSER);
				} else if ((NEGATED&literal2[0])&&(0==(NEGATED&literal1[0]))){
					return(GREATER);
				}
			}
			return(EQUAL);
			break;
		case FAILURE:
			return(FAILURE);
			break;
		case NOMEMORY:
			return(NOMEMORY);
			break;
	}
	return(FAILURE);
 } /* CompareLiterals */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Equality literals comparison)  OCJ
 *
 *    This function performs an equality literals comparison. This
 *    comparison is performed as indicated in the section 3.1.1
 *    of document "Implementing an efficient theorem prover"
 *    by Alexandre Riazanov and is based in the multiset ordering
 *    described in the "Multisets" section of page 26 of the above
 *    document.
 *
 *    Specifically, equalities are compared as multisets, being
 *    {s,t} the multiset assigned to positive equality s=t and
 *    {s,s,t,t} the multiset assigned to negative equality s!=t.
 *    Let |> be the multiset ordering and :> the term reduction
 *    ordering as described in section 3.1.5 of the above document.
 *    Let S1 and S2 be two multisets. Let S1(x) the multiset integer
 *    assigned to element x in multiset S1 and S2(x) the integer
 *    assigned to element x in multiset S2.
 *
 *    Then S1 |> S2 if and only if for all x such that S1(x) < S2(x)
 *    there exists y such that y :> x and S1(y) > S2(y).
 *
 *    The return conditions are:
 *
 *    GREATER if literal1 > literal2
 *    LESSER if literal2 > literal1
 *    EQUAL if literal2 = literal1
 *    FAILURE otherwise
 *
 *    Symmetry axiom is considered for comparisons. So,two literals are
 *    are equal if they are identical or one is identical to the symmetric
 *    of the other.
 *
 *
 *  ARGUMENTS:
 *
 *    literal1: Pointer to first equality literal to be compared.
 *    literal2: Pointer to second equality literal to be compared.
 *    coef1: Pointer to coeff structure for first equality literal,
 *           only used for STANDARD and NONRECURSIVE atom
 *           reduction orderings.
 *    coef2: Pointer to coeff structure for second equality literal,
 *           only used for STANDARD and NONRECURSIVE atom
 *           reduction orderings.
 *    flag: STANDARD, NONRECURSIVE or LEXICOGRAPHIC.
 *
 *  RETURNS:
 *
 *    GREATER, LESSER, EQUAL, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t CompareEqualities(uint8_t *literal1,uint8_t *literal2,coeff *coef1,
		coeff *coef2,int32_t flag){
	auto uint8_t *eqtrm11,*eqtrm21;                 /* Pointers to literal1 equality root terms */
	auto uint8_t *eqtrm12,*eqtrm22;                 /* Pointers to literal2 equality root terms */
	auto int32_t o11,o12,o21,o22;                   /* Results from comparison. */
	auto int32_t ii,jj;                             /* Auxiliary. */


	/* Both equalities are positive or both are negative. */
	ii=flag&STANDARD?STANDARD:NONRECURSIVE;
	if ((literal1[0]&NEGATED)==(literal2[0]&NEGATED)){

		/* Get pointers to equality root terms. */
		eqtrm11=NextItem(literal1,IMMED);
		eqtrm21=NextItem(eqtrm11,OVERSUBTERMS);
		eqtrm12=NextItem(literal2,IMMED);
		eqtrm22=NextItem(eqtrm12,OVERSUBTERMS);

		/* Case EQUAL (two pairs of equality root terms equal). */
		if (NOMEMORY==(o11=CompareItems(eqtrm11,eqtrm12,coef1->eqterm1,coef2->eqterm1,ii))){
			return(NOMEMORY);
		}
		if (NOMEMORY==(o22=CompareItems(eqtrm21,eqtrm22,coef1->eqterm2,coef2->eqterm2,ii))){
			return(NOMEMORY);
		}
		if ((o11==EQUAL)&&(o22==EQUAL)){
			return(EQUAL);
		}
		if (NOMEMORY==(o12=CompareItems(eqtrm11,eqtrm22,coef1->eqterm1,coef2->eqterm2,ii))){
			return(NOMEMORY);
		}
		if (NOMEMORY==(o21=CompareItems(eqtrm21,eqtrm12,coef1->eqterm2,coef2->eqterm1,ii))){
			return(NOMEMORY);
		}
		if ((o12==EQUAL)&&(o21==EQUAL)){
			return(EQUAL);
		}

		/* Case of one pair of equality root terms equal. */
		if (o11==EQUAL){
			if (o22==GREATER){
				return(GREATER);
			} else if (o22==LESSER){
				return(LESSER);
			}
		} else if (o12==EQUAL){
			if (o21==GREATER){
				return(GREATER);
			} else if (o21==LESSER){
				return(LESSER);
			}
		} else if (o21==EQUAL){
			if (o12==GREATER){
				return(GREATER);
			} else if (o12==LESSER){
				return(LESSER);
			}
		} else if (o22==EQUAL){
			if (o11==GREATER){
				return(GREATER);
			} else if (o11==LESSER){
				return(LESSER);
			}
		}

		/* Case of no pairs of equality root terms equal. */
		if (((o11==GREATER)||(21==GREATER))&&((o12==GREATER)||(o22==GREATER))){
			return(GREATER);
		} else if (((o11==LESSER)||(12==LESSER))&&((o21==LESSER)||(o22==LESSER))){
			return(LESSER);
		}

	/* One equality is positive and the other is negative. */
	} else {

		/* Get pointers to equality root terms. The equality sub-index 1 is */
		/* assigned to the equality that produces the bigger multiset. */
		if (((literal1[0]&NEGATED)&&((MAXIMAL|SINGLENEG|MULTINEG)&kbset.opts.select))
				||((0==(literal1[0]&NEGATED))&&((SINGLEPOS|MULTIPOS)&kbset.opts.select))){
			eqtrm11=NextItem(literal1,IMMED);
			eqtrm21=NextItem(eqtrm11,OVERSUBTERMS);
			eqtrm12=NextItem(literal2,IMMED);
			eqtrm22=NextItem(eqtrm12,OVERSUBTERMS);
			jj=1;
		} else {
			eqtrm11=NextItem(literal2,IMMED);
			eqtrm21=NextItem(eqtrm11,OVERSUBTERMS);
			eqtrm12=NextItem(literal1,IMMED);
			eqtrm22=NextItem(eqtrm12,OVERSUBTERMS);
			jj=0;
		}

		/* Case of two pairs of equality root terms equal. */
		if (NOMEMORY==(o11=CompareItems(eqtrm11,eqtrm12,coef1->eqterm1,coef2->eqterm1,ii))){
			return(NOMEMORY);
		}
		if (NOMEMORY==(o22=CompareItems(eqtrm21,eqtrm22,coef1->eqterm2,coef2->eqterm2,ii))){
			return(NOMEMORY);
		}
		if ((o11==EQUAL)&&(o22==EQUAL)){
			if (jj){
				return(GREATER);
			} else {
				return(LESSER);
			}
		}
		if (NOMEMORY==(o12=CompareItems(eqtrm11,eqtrm22,coef1->eqterm1,coef2->eqterm2,ii))){
			return(NOMEMORY);
		}
		if (NOMEMORY==(o21=CompareItems(eqtrm21,eqtrm12,coef1->eqterm2,coef2->eqterm1,ii))){
			return(NOMEMORY);
		}
		if ((o12==EQUAL)&&(o21==EQUAL)){
			if (jj){
				return(GREATER);
			} else {
				return(LESSER);
			}
		}

		/* Case of one pair of equality root terms equal. */
		if (o11==EQUAL){
			if (o22==GREATER){
				if (jj){
					return(GREATER);
				} else {
					return(LESSER);
				}
			} else if ((o22==LESSER)&&(o12==LESSER)){
				if (jj){
					return(LESSER);
				} else {
					return(GREATER);
				}
			}
		} else if (o12==EQUAL){
			if (o21==GREATER){
				if (jj){
					return(GREATER);
				} else {
					return(LESSER);
				}
			} else if ((o21==LESSER)&&(o11==LESSER)){
				if (jj){
					return(LESSER);
				} else {
					return(GREATER);
				}
			}
		} else if (o21==EQUAL){
			if (o12==GREATER){
				if (jj){
					return(GREATER);
				} else {
					return(LESSER);
				}
			} else if ((o12==LESSER)&&(o22==LESSER)){
				if (jj){
					return(LESSER);
				} else {
					return(GREATER);
				}
			}
		} else if (o22==EQUAL){
			if (o11==GREATER){
				if (jj){
					return(GREATER);
				} else {
					return(LESSER);
				}
			} else if ((o11==LESSER)&&(o21==LESSER)){
				if (jj){
					return(LESSER);
				} else {
					return(GREATER);
				}
			}
		}

		/* Case of no pairs of equality root terms equal. */
		if (((o11==GREATER)||(21==GREATER))&&((o12==GREATER)||(o22==GREATER))){
			if (jj){
				return(GREATER);
			} else {
				return(LESSER);
			}
		} else if (((o11==LESSER)||(12==LESSER))&&((o21==LESSER)||(o22==LESSER))){
			if (jj){
				return(LESSER);
			} else {
				return(GREATER);
			}
		}
	}

	return(FAILURE);
 } /* CompareEqualities */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Predicates comparison)  OCJ
 *
 *    This function performs a predicates comparison. The comparison
 *    is performed just between predicates without taking into account
 *    positivity or negativity. That must be managed by the calling
 *    function.
 *
 *    The comparison is performed as indicated in the section 3.1.1
 *    of document "Implementing an efficient theorem prover"
 *    by Alexandre Riazanov. See CompareItems() function for details
 *    on the standard and non recursive KBO ordering.
 *
 *    The return conditions of this function are:
 *
 *    GREATER if pred1 > pred2
 *    LESSER if pred2 > pred1
 *    EQUAL if pred2 = pred1
 *    FAILURE otherwise
 *
 *
 *  ARGUMENTS:
 *
 *    pred1: Pointer to first predicate to be compared.
 *    pred2: Pointer to second predicate to be compared.
 *    coef1: Pointer to coeff structure for first predicate,
 *           only used for STANDARD and NONRECURSIVE atom
 *           reduction orderings.
 *    coef2: Pointer to coeff structure for second predicate,
 *           only used for STANDARD and NONRECURSIVE atom
 *           reduction orderings.
 *    flag: STANDARD, NONRECURSIVE or LEXICOGRAPHIC.
 *
 *  RETURNS:
 *
 *    GREATER, LESSER, EQUAL, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t ComparePredicates(uint8_t *pred1,uint8_t *pred2,coeff *coef1,
		coeff *coef2,int32_t flag){

	/* Comparison by precedence. */
	if (flag&LEXICOGRAPHIC){
		return(SecondKBOCheck(pred1,pred2,NONRECURSIVE));

	/* Standard or non recursive KBO comparison. */
	} else {
		return(CompareItems(pred1,pred2,coef1,coef2,flag));
	}

	return(FAILURE);
 } /* ComparePredicates */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Terms and predicates comparison)  OCJ
 *
 *    This function performs a standard or non recursive Knuth-Bendix
 *    comparison of terms and predicates, depending on the value of flags.
 *    Equalities are not allowed.
 *
 *    Let :> be the KBO ordering relation (standard or non recursive,
 *    depending on the value of flag) of definition 3.14 or definition 3.15
 *    as described in document "Implementing an efficient theorem prover"
 *    by Alexandre Riazanov. Then the return conditions of this function are:
 *
 *    GREATER if term1 :> term2
 *    LESSER if term2 :> term1
 *    EQUAL if term1 = term2
 *    FAILURE otherwise
 *
 *
 *  ARGUMENTS:
 *
 *    term1: Pointer to first term to be compared.
 *    term2: Pointer to second term to be compared.
 *    coef1: Pointer to coeff structure for first item.
 *           This will be ignored if NULL.
 *    coef2: Pointer to coeff structure for second item.
 *           This will be ignored if NULL.
 *    flag: STANDARD or NONRECURSIVE
 *
 *  RETURNS:
 *
 *    EQUAL, GREATER, LESSER, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t CompareItems(uint8_t *term1,uint8_t *term2,coeff *coef1,
		coeff *coef2,int32_t flag){
	auto int32_t ii;                                /* Auxiliary. */

	/* Second item is a variable. */
	if (term2[0]&VARIABLE){

		/* First item is a variable. */
		if (term1[0]&VARIABLE){

			/* Both items are the same variable. */
			if ((*((int16_t *)&term1[1]))==(*((int16_t *)&term2[1]))){
				return(EQUAL);

			/* Both items are not the same variable. */
			} else {
				return(FAILURE);
			}

		/* Variable in second item exists in first item. */
		} else if (CheckVarInTerm2(term2,term1)){
			return(GREATER);

		/* Variable in second item doesn't exist in first item. */
		} else {
			return(FAILURE);
		}

	/* First item is a variable. */
	} else if (term1[0]&VARIABLE){

		/* The variable in first item exists in the second item. */
		if (CheckVarInTerm2(term1,term2)){
			return(LESSER);

		/* The variable in first item doesn't exist in the second item. */
		} else {
			return(FAILURE);
		}
	}

	/* Compare term weights. */
	switch (CompareItemWeight(term1,term2,coef1,coef2)){
		case GREATER:
			return(GREATER);
			break;
		case LESSER:
			return(LESSER);
			break;
		case EQUAL:
			return(SecondKBOCheck(term1,term2,flag));
			break;
		case NOGREATER:
			switch (SecondKBOCheck(term2,term1,flag)){
				/* Case EQUAL cannot happen as in order for CompareItems() to */
				/* return EQUAL the call to CompareItemWeight() above must */
				/* return EQUAL. */
				case GREATER:
					return(LESSER);
					break;
				case LESSER:
				case FAILURE:
					return(FAILURE);
					break;
				case NOMEMORY:
					return(NOMEMORY);
					break;
			}
			break;
		case NOLESSER:
			if (LESSER==(ii=SecondKBOCheck(term1,term2,flag))){
				return(FAILURE);
			} else {
				return(ii);
			}
			break;
		case NOMEMORY:
			return(NOMEMORY);
			break;
	}

	return(FAILURE);
 } /* CompareItems */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Equality root terms comparison)  OCJ
 *
 *    This function performs a standard or non recursive Knuth-Bendix
 *    comparison of equality terms and subterms, depending on the value
 *    of flags. This function also updates the result field of the
 *    equpnodes structures of terms being compared.
 *
 *    This function works in a similar way as CompareItems() with the following
 *    differences:
 *    - It uses CompareItemWeight2() function instead of CompareItemWeight().
 *      The CompareItemWeight2() function uses the compact form of linear
 *      polynomials.
 *    - It uses the equality polynomials tree structure instead of the
 *      long polynomial structure.
 *    - It stores the result in the result field of the appropriate
 *      equpnode structures.
 *    - It uses specialized code to improve speed as described in section 9 of
 *      the document "Implementing an efficient theorem prover" by Alexandre
 *      Riazanov.
 *
 *    See CompareItems() function comments for additional details.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of clause to which the
 *          terms belong.
 *    term1: Pointer to first term to be compared.
 *    term2: Pointer to second term to be compared.
 *    pnode1: Address of pointer to equpnode node structure for first term.
 *    pnode2: Address of pointer to equpnode node structure for second term.
 *    flag: STANDARD or NONRECURSIVE
 *
 *  RETURNS:
 *
 *    EQUAL, GREATER, LESSER, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t CompareEquTerms(binprefix *binp,uint8_t *term1,uint8_t *term2,equpnode **pnode1,
		equpnode **pnode2,int32_t flag){

	/* Setup nodes data if nodes are empty. */
	if ((*pnode1)==NULL){
		if (NOMEMORY==SetEquChildNode(binp,term1,pnode1)){
			return(NOMEMORY);
		}
		if (NOMEMORY==SetEquChildNode(binp,term2,pnode2)){
			FreeEquNode(term1,*pnode1);
			*pnode1=NULL;
			return(NOMEMORY);
		}
		if (NOMEMORY==SetLftRhtPols(binp,&((*pnode1)->coefs),&((*pnode2)->coefs))){
			FreeEquNode(term1,*pnode1);
			FreeEquNode(term2,*pnode2);
			*pnode1=*pnode2=NULL;
			return(NOMEMORY);
		}
	}

	/* If the coefs field of *pnode1 is NULL then there is already */
	/* a definite comparison result. */
	if ((*pnode1)->coefs==NULL){
		return((EQUAL|GREATER|LESSER)&(*pnode1)->result);
	}

	/* Second item is a variable. */
	if (term2[0]&VARIABLE){

		/* First item is a variable. */
		if (term1[0]&VARIABLE){

			/* Both items are the same variable. */
			if ((*((int16_t *)&term1[1]))==(*((int16_t *)&term2[1]))){
				(*pnode1)->result=(*pnode2)->result=EQUAL|EQUALW;
				return(EQUAL);

			/* Both items are not the same variable. */
			} else {
				(*pnode1)->result=(*pnode2)->result=FAILURE|FAILUREW;
				return(FAILURE);
			}

		/* Variable in second item exists in first item. */
		} else if (CheckVarInTerm2(term2,term1)){
			(*pnode1)->result=GREATER;
			(*pnode2)->result=LESSER;
			return(GREATER);

		/* Variable in second item doesn't exist in first item. */
		} else {
			(*pnode1)->result=(*pnode2)->result=FAILURE|FAILUREW;
			return(FAILURE);
		}

	/* First item is a variable. */
	} else if (term1[0]&VARIABLE){

		/* The variable in first item exists in the second item. */
		if (CheckVarInTerm2(term1,term2)){
			(*pnode1)->result=LESSER;
			(*pnode2)->result=GREATER;
			return(LESSER);

		/* The variable in first item doesn't exist in the second item. */
		} else {
			(*pnode1)->result=(*pnode2)->result=FAILURE|FAILUREW;
			return(FAILURE);
		}
	}

	/* Compare term weights. */
	switch (CompareItemWeight2((*pnode1)->coefs,(*pnode2)->coefs)){
		case GREATER:
			FreeEquNodeData(term1,*pnode1);
			FreeEquNodeData(term2,*pnode2);
			(*pnode1)->result=GREATER;
			(*pnode2)->result=LESSER;
			return(GREATER);
			break;
		case LESSER:
			FreeEquNodeData(term1,*pnode1);
			FreeEquNodeData(term2,*pnode2);
			(*pnode1)->result=LESSER;
			(*pnode2)->result=GREATER;
			return(LESSER);
			break;
		case EQUAL:
			switch (SecondKBOCheck2(binp,term1,term2,*pnode1,*pnode2,flag)){
				case GREATER:
					FreeEquNodeData(term1,*pnode1);
					FreeEquNodeData(term2,*pnode2);
					(*pnode1)->result=GREATER|EQUALW;
					(*pnode2)->result=LESSER|EQUALW;
					return(GREATER);
					break;
				case LESSER:
					FreeEquNodeData(term1,*pnode1);
					FreeEquNodeData(term2,*pnode2);
					(*pnode1)->result=LESSER|EQUALW;
					(*pnode2)->result=GREATER|EQUALW;
					return(LESSER);
					break;
				case EQUAL:
					FreeEquNodeData(term1,*pnode1);
					FreeEquNodeData(term2,*pnode2);
					(*pnode1)->result=(*pnode2)->result=EQUAL;
					return(EQUAL);
					break;
				case FAILURE:
					(*pnode1)->result=(*pnode2)->result=FAILURE|EQUALW;
					return(FAILURE);
					break;
				case NOMEMORY:
					FreeEquNodeData(term1,*pnode1);
					FreeEquNodeData(term2,*pnode2);
					return(NOMEMORY);
					break;
			}
			break;
		case NOGREATER:
			switch (SecondKBOCheck2(binp,term2,term1,*pnode2,*pnode1,flag)){
				/* Case EQUAL cannot happen as in order for CompareEquTerms() to */
				/* return EQUAL the call to CompareItemWeight2() above must */
				/* return EQUAL. */
				case GREATER:
					FreeEquNodeData(term1,*pnode1);
					FreeEquNodeData(term2,*pnode2);
					(*pnode1)->result=LESSER|NOGREATER;
					(*pnode2)->result=GREATER|NOLESSER;
					return(LESSER);
					break;
				case LESSER:
				case FAILURE:
					(*pnode1)->result=FAILURE|NOGREATER;
					(*pnode2)->result=FAILURE|NOLESSER;
					return(FAILURE);
					break;
				case NOMEMORY:
					FreeEquNodeData(term1,*pnode1);
					FreeEquNodeData(term2,*pnode2);
					return(NOMEMORY);
					break;
			}
			break;
		case NOLESSER:
			switch (SecondKBOCheck2(binp,term1,term2,*pnode1,*pnode2,flag)){
				/* Case EQUAL cannot happen as in order for CompareEquTerms() to */
				/* return EQUAL the call to CompareItemWeight2() above must */
				/* return EQUAL. */
				case GREATER:
					FreeEquNodeData(term1,*pnode1);
					FreeEquNodeData(term2,*pnode2);
					(*pnode1)->result=GREATER|NOLESSER;
					(*pnode2)->result=LESSER|NOGREATER;
					return(GREATER);
					break;
				case LESSER:
				case FAILURE:
					(*pnode1)->result=FAILURE|NOLESSER;
					(*pnode1)->result=FAILURE|NOGREATER;
					return(FAILURE);
					break;
				case NOMEMORY:
					return(NOMEMORY);
					break;
			}
			break;
		case NOMEMORY:
			return(NOMEMORY);
			break;
	}

	/* Set result to both failures and return. */
	(*pnode1)->result=FAILURE|FAILUREW;
	(*pnode2)->result=FAILURE|FAILUREW;
	return(FAILURE);
 } /* CompareEquTerms */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Equality root terms comparison with substitution)  OCJ
 *
 *    This function performs a standard or non recursive Knuth-Bendix
 *    comparison of equality terms and subterms, depending on the value
 *    of flags, given a substituion to be applied to the terms.
 *
 *    This function works in a similar way as CompareEquTerms() but the
 *    comparison is done after applying a substitution to the terms. The
 *    substitution is no applied directly to the terms but to their linear
 *    polynomials. See section 9 of the document "Implementing an efficient
 *    theorem prover" by Alexandre Riazanov for details.
 *
 *    See CompareEquTerms() function comments for additional details.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of clause to which the
 *          terms belong.
 *    term1: Pointer to first term to be compared.
 *    term2: Pointer to second term to be compared.
 *    pnode1: Address of pointer to equpnode node structure for first term.
 *    pnode2: Address of pointer to equpnode node structure for second term.
 *    Subst: Pointer to substitution to be applied to term1 and term2
 *           before checking the ordering constraint.
 *    varshift: Shift for variable numbers belonging to a secondary
 *              merged clause. Variable numbers coming from secondary clause
 *              increased by varshift match variable numbers in the substitution.
 *    flag: STANDARD or NONRECURSIVE depending on the ordering type.
 *          MAXIMUM: If this flag is set then the first term is checked
 *          for being greater than second term. If this flag is not set
 *          then the first term is checked for being not lesser nor equal
 *          than the second term.
 *          ITEM1SEC: If this flag is set then both terms belong to
 *          a secondary item with respect to the substitution. If this
 *          flag is not set then the terms belong to a primary item.
 *
 *  RETURNS:
 *
 *    EQUAL, GREATER, LESSER, FAILURE or NOMEMORY.
 *    If MAXIMUM flag is set the return value may be FAILURE even if a comparison
 *    with MAXIMUM flag not set gives a different value.
 *
 *--------------------------------------------------------------*/
int32_t CompSubstEquTrms(binprefix *binp,uint8_t *term1,uint8_t *term2,equpnode **pnode1,
		equpnode **pnode2,subst *Subst,int32_t varshift,int32_t flag){
	auto int32_t ii;                                /* Auxiliary. */

	/* If comparison without substitution is not available then do it. */
	if (((*pnode1)==NULL)||(0==(*pnode1)->result)){
		if (FAILURE!=(ii=CompareEquTerms(binp,term1,term2,pnode1,pnode2,flag))){
			return(ii);
		}
	}

	/* Check if results without substitution are conclusive for a return. */
	ii=(*pnode1)->result&(EQUAL|GREATER|LESSER);
	if (ii){
		return(ii);
	}

	/* At least one of the terms is a variable. */
	if ((term1[0]&VARIABLE)||(term2[0]&VARIABLE)){

		/* Compare term weights with substitution. */
		if (NOMEMORY==(ii=CompItmWeightSubst(*pnode1,*pnode2,Subst,varshift,flag))){
			return(NOMEMORY);
		}

		/* Result suffices for an early return. */
		if (ii&(GREATER|LESSER|FAILURE|NOMEMORY)){
			return(ii);

		/* Result doesn't suffice for an early return. */
		} else {
			return(CompareSubstItems(term1,term2,ii,Subst,varshift,flag));
		}
	}

	/* If we are here then term1=f(s1,s2,...,sm) and term2=g(t1,t2,...,tn). */
	/* Cases of weight comparison without substitution results. */
	switch ((NOGREATER|NOLESSER|EQUALW|FAILUREW)&(*pnode1)->result){

		/* If weight comparison without substitution is EQUAL then weight */
		/* comparison with substitution is also EQUAL. As both terms */
		/* are incomparable then we must have f=g (see above). */
		case EQUALW:
			return(CompareLexic(binp,term1,term2,*pnode1,*pnode2,Subst,varshift,flag));
			break;

		/* Other cases. */
		case NOLESSER:
		case FAILUREW:
			switch(ii=CompItmWeightSubst(*pnode1,*pnode2,Subst,varshift,flag)){
				case LESSER:
				case GREATER:
				case NOMEMORY:
					return(ii);
					break;
				case EQUAL:
					return(SecondKBOCheck3(binp,term1,term2,*pnode1,*pnode2,Subst,varshift,flag));
					break;
				case NOLESSER:
					if (LESSER!=(ii=SecondKBOCheck3(binp,term1,term2,*pnode1,*pnode2,Subst,varshift,flag))){
						return(ii);
					}
					break;
				case NOGREATER:
					if (flag&MAXIMUM){
						return(FAILURE);
					}
					if (GREATER!=(ii=SecondKBOCheck3(binp,term1,term2,*pnode1,*pnode2,Subst,varshift,flag))){
						return(ii);
					}
					break;
			}
			break;
		case NOGREATER:
			if (flag&MAXIMUM){
				return(FAILURE);
			}
			switch(ii=CompItmWeightSubst(*pnode1,*pnode2,Subst,varshift,flag)){
				/* Case NOLESSER cannot happen because if weight is NOLESSER */
				/* and NOGREATER at the same time then it is EQUAL. */
				case GREATER:
				case LESSER:
				case NOMEMORY:
					return(ii);
					break;
				case EQUAL:
					return(SecondKBOCheck3(binp,term1,term2,*pnode1,*pnode2,Subst,varshift,flag));
					break;
				case NOGREATER:
					if (GREATER!=(ii=SecondKBOCheck3(binp,term1,term2,*pnode1,*pnode2,Subst,varshift,flag))){
						return(ii);
					}
					break;
			}
			break;
	}
	return(FAILURE);
 } /* CompSubstEquTrms */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Terms comparison with substitution)  OCJ
 *
 *    This function performs a comparison between two terms after applying
 *    a substitution to them. The weight comparison result must be known
 *    and is passed as a parameter to this function. It is supposed that
 *    the weight comparison is NOGREATER, NOLESSER or EQUAL, otherwise
 *    the term comparison is fully defined and there is no need to call
 *    this function.
 *
 *    The substitution is applied by building a dummy clause with just
 *    an equality in which the two root terms are the terms passed as
 *    arguments to this function. Then ApplySubst2Clause() function
 *    is called.
 *
 *    The final result is obtained by taking into account the weight
 *    comparison results and calling SecondKBOCheck() function.
 *
 *
 *  ARGUMENTS:
 *
 *    term1: Pointer to first term to be compared after substitution.
 *    term2: Pointer to second term to be compared after substitution.
 *    cwresult: Weight comparison result for substituted terms.
 *    Subst: Pointer to substitution to be applied to term1 and term2
 *           before checking the ordering constraint.
 *    varshift: Shift for variable numbers belonging to a secondary
 *              merged clause. Variable numbers coming from secondary clause
 *              increased by varshift match variable numbers in the substitution.
 *    flag: STANDARD or NONRECURSIVE depending on the ordering type.
 *          ITEM1SEC: If this flag is set then both terms belong to
 *          a secondary item with respect to the substitution. If this
 *          flag is not set then the terms belong to a primary item.
 *
 *  RETURNS:
 *
 *    EQUAL, GREATER, LESSER, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t CompareSubstItems(uint8_t *term1,uint8_t *term2,int32_t cwresult,
		subst *Subst,int32_t varshift,int32_t flag){
	auto cmprefix *dmyclause;                       /* Dummy clause cmprefix structure pointer */
	auto uint8_t *sterm1,*sterm2;                   /* Pointers to substituted terms */
	auto int32_t ii;                                /* Auxiliary. */

	/* Apply substitution to terms. */
	if (NOMEMORY==ApplySbst2TwoItm(term1,term2,&sterm1,&sterm2,&dmyclause,Subst,varshift,flag)){
		return(NOMEMORY);
	}

	/* Second item is a variable. */
	if (sterm2[0]&VARIABLE){

		/* First item is a variable. */
		if (sterm1[0]&VARIABLE){

			/* Both items are the same variable. */
			if ((*((int16_t *)&sterm1[1]))==(*((int16_t *)&sterm2[1]))){
				if (dmyclause!=NULL){
					MYFREE(dmyclause->part2.bin);
					MYFREE(dmyclause);
				}
				return(EQUAL);

			/* Both items are not the same variable. */
			} else {
				if (dmyclause!=NULL){
					MYFREE(dmyclause->part2.bin);
					MYFREE(dmyclause);
				}
				return(FAILURE);
			}

		/* Variable in second item exists in first item. */
		} else if (CheckVarInTerm2(sterm2,sterm1)){
			if (dmyclause!=NULL){
				MYFREE(dmyclause->part2.bin);
				MYFREE(dmyclause);
			}
			return(GREATER);

		/* Variable in second item doesn't exist in first item. */
		} else {
			if (dmyclause!=NULL){
				MYFREE(dmyclause->part2.bin);
				MYFREE(dmyclause);
			}
			return(FAILURE);
		}

	/* First item is a variable. */
	} else if (sterm1[0]&VARIABLE){

		/* The variable in first item exists in the second item. */
		if (CheckVarInTerm2(sterm1,sterm2)){
			if (dmyclause!=NULL){
				MYFREE(dmyclause->part2.bin);
				MYFREE(dmyclause);
			}
			return(LESSER);

		/* The variable in first item doesn't exist in the second item. */
		} else {
			if (dmyclause!=NULL){
				MYFREE(dmyclause->part2.bin);
				MYFREE(dmyclause);
			}
			return(FAILURE);
		}
	}

	/* Call SecondKBOCheck() for final result. Free memory and */
	/* return as appropriate */
	switch (cwresult){
		case EQUAL:
			ii=SecondKBOCheck(sterm1,sterm2,flag);
			if (dmyclause!=NULL){
				MYFREE(dmyclause->part2.bin);
				MYFREE(dmyclause);
			}
			return(ii);
			break;
		case NOGREATER:
			ii=SecondKBOCheck(sterm2,sterm1,flag);
			if (dmyclause!=NULL){
				MYFREE(dmyclause->part2.bin);
				MYFREE(dmyclause);
			}
			switch (ii){
				case GREATER:
					return(LESSER);
					break;
				case LESSER:
				case FAILURE:
					return(FAILURE);
					break;
				case NOMEMORY:
					return(NOMEMORY);
					break;
				/* Case EQUAL may happen even if cwresult is not EQUAL, for instance */
				/* paramodulating from a clause with a(b,b)=c to a clause with a(X1,X0)=a(X0,X1). */
				case EQUAL:
					return(EQUAL);
					break;
			}
			break;
		case NOLESSER:
			ii=SecondKBOCheck(sterm1,sterm2,flag);
			if (dmyclause!=NULL){
				MYFREE(dmyclause->part2.bin);
				MYFREE(dmyclause);
			}
			if (ii==LESSER){
				return(FAILURE);
			}
			return(ii);
			break;
		case NOMEMORY:
			if (dmyclause!=NULL){
				MYFREE(dmyclause->part2.bin);
				MYFREE(dmyclause);
			}
			return(NOMEMORY);
			break;
	}
	if (dmyclause!=NULL){
		MYFREE(dmyclause->part2.bin);
		MYFREE(dmyclause);
	}
	return(FAILURE);
 } /* CompareSubstItems */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform second KBO check)  OCJ
 *
 *    This function performs the second Knuth-Bendix ordering check
 *    for the case of first check failure, for both standard and
 *    non recursive KBO ordering depending on the value of flags.
 *
 *    This function can be used for terms that are not variables
 *    as well as for predicates, but not for equalities. In the case
 *    of predicates it just compares the atoms, it doesn't take into
 *    account any possible negations.
 *
 *    Let :> be the KBO ordering relation (standard or non recursive,
 *    depending on the value of flag) of definition 3.14 or definition 3.15
 *    as described in document "Implementing an efficient theorem prover"
 *    by Alexandre Riazanov. Let > and ·> be the binary relations of
 *    definition 3.13 of above document. As indicated in the document,
 *    when the relation > fails a second check is performed. This is
 *    the check performed by this function to verify that item1 ·> item2.
 *
 *    The return conditions are:
 *
 *    GREATER if item1 :> item2
 *    LESSER if item2 :> item2
 *    EQUAL if item1 = item2
 *    FAILURE otherwise
 *
 *
 *  ARGUMENTS:
 *
 *    item1: Pointer to first item to be compared.
 *    item2: Pointer to second item to be compared.
 *    flag: STANDARD or NONRECURSIVE
 *
 *
 *  RETURNS:
 *
 *    GREATER, LESSER, EQUAL, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t SecondKBOCheck(uint8_t *item1,uint8_t *item2,int32_t flag){
	auto uint8_t *ptr1,*ptr2,*ptr3;                 /* Auxiliary pointers */
	auto int32_t ii;                                /* Auxiliary. */

	/* Item 1 is not a "distinct object" and item 1 is a "distinct object". */
	/* Symbol precedence relation (>>) holds. */
	if ((DISTINCTOBJECT&((symbol *)&item1[1])->symbol->type)
			&&(0==(DISTINCTOBJECT&((symbol *)&item2[1])->symbol->type))){
		return(GREATER);
	}

	/* Item 1 is a "distinct object" and item 1 is not a "distinct object". */
	/* Symbol precedence relation (<<) holds. */
	if ((DISTINCTOBJECT&((symbol *)&item2[1])->symbol->type)
			&&(0==(DISTINCTOBJECT&((symbol *)&item1[1])->symbol->type))){
		return(LESSER);
	}

	/* Symbol precedence relation (>>) holds. */
	if (((symbol *)&item1[1])->symbol->precedence>((symbol *)&item2[1])->symbol->precedence){
		return(GREATER);
	}

	/* Symbol precedence relation (<<) holds. */
	if (((symbol *)&item1[1])->symbol->precedence<((symbol *)&item2[1])->symbol->precedence){
		return(LESSER);
	}

	/* Both symbols have the same precedence. */
	/* Standard ordering. Check standard KBO ordering relation recursively. */
	if (flag&STANDARD){

		/* Loop through the arguments of both items. */
		for (ptr1=NextItem(item1,IMMED),ptr2=NextItem(item2,IMMED),ptr3=NextItem(item1,OVERSUBTERMS);
				ptr1<ptr3;ptr1=NextItem(ptr1,OVERSUBTERMS),ptr2=NextItem(ptr2,OVERSUBTERMS)){
			if (EQUAL!=(ii=CompareItems(ptr1,ptr2,NULL,NULL,flag))){
				return(ii);
			}
		}

	/* Non recursive ordering. Check >>> ordering relation (slightly modified). */
	} else {

		/* Loop through the arguments of both items. */
		for (ptr1=NextItem(item1,IMMED),ptr2=NextItem(item2,IMMED),ptr3=NextItem(item1,OVERSUBTERMS);
				ptr1<ptr3;ptr1=NextItem(ptr1,OVERSUBTERMS),ptr2=NextItem(ptr2,OVERSUBTERMS)){

			/* None of the items is a variable. */
			if ((0==(ptr1[0]&VARIABLE))&&(0==(ptr2[0]&VARIABLE))){
				if (EQUAL!=(ii=SecondKBOCheck(ptr1,ptr2,flag))){
					return(ii);
				}

			/* Both items are a variable. */
			} else if ((ptr1[0]&VARIABLE)&&(ptr2[0]&VARIABLE)){

				/* The variables are not the same. */
				if ((*((int16_t *)&ptr1[1]))!=(*((int16_t *)&ptr2[1]))){
					return(FAILURE);
				}

			/* One item is a variable and the other is not a variable. */
			} else {
				return(FAILURE);
			}
		}
	}

	return(EQUAL);
 } /* SecondKBOCheck */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform second KBO check for polynomial trees)  OCJ
 *
 *    This function is analogous to the SecondKBOCheck() function
 *    except that it calls CompareEquTerms() and itself instead
 *    of calling CompareItems() and SecondKBOCheck(). It performs
 *    the same function but works with polynomial trees instead
 *    of long polynomials.
 *
 *    See SecondKBOCheck() function comments for additional details.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of clause to which the
 *          terms belong.
 *    item1: Pointer to first item to be compared.
 *    item2: Pointer to second item to be compared.
 *    node1: Pointer to equpnode node structure for first term. This
 *           is not used in case of non recursive ordering so it is
 *           NULL on input in this case.
 *    node2: Pointer to equpnode node structure for second term. This
 *           is not used in case of non recursive ordering so it is
 *           NULL on input in this case.
 *    flag: STANDARD or NONRECURSIVE
 *
 *  RETURNS:
 *
 *    GREATER, LESSER, EQUAL, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t SecondKBOCheck2(binprefix *binp,uint8_t *item1,uint8_t *item2,equpnode *node1,
		equpnode *node2,int32_t flag){
	auto uint8_t *ptr1,*ptr2,*ptr3;                 /* Auxiliary pointers */
	auto int32_t ii,jj;                             /* Auxiliary. */

	/* Item 1 is not a "distinct object" and item 1 is a "distinct object". */
	/* Symbol precedence relation (>>) holds. */
	if ((DISTINCTOBJECT&((symbol *)&item1[1])->symbol->type)
			&&(0==(DISTINCTOBJECT&((symbol *)&item2[1])->symbol->type))){
		return(GREATER);
	}

	/* Item 1 is a "distinct object" and item 1 is not a "distinct object". */
	/* Symbol precedence relation (<<) holds. */
	if ((DISTINCTOBJECT&((symbol *)&item2[1])->symbol->type)
			&&(0==(DISTINCTOBJECT&((symbol *)&item1[1])->symbol->type))){
		return(LESSER);
	}

	/* Symbol precedence relation (>>) holds. */
	if (((symbol *)&item1[1])->symbol->precedence>((symbol *)&item2[1])->symbol->precedence){
		return(GREATER);
	}

	/* Symbol precedence relation (<<) holds. */
	if (((symbol *)&item1[1])->symbol->precedence<((symbol *)&item2[1])->symbol->precedence){
		return(LESSER);
	}

	/* Both symbols have the same precedence. */
	/* Standard ordering. Check standard KBO ordering relation recursively. */
	if (flag&STANDARD){

		/* Loop through the arguments of both items. */
		for (ptr1=NextItem(item1,IMMED),ptr2=NextItem(item2,IMMED),ptr3=NextItem(item1,OVERSUBTERMS),jj=0;
				ptr1<ptr3;ptr1=NextItem(ptr1,OVERSUBTERMS),ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){
			if (EQUAL!=(ii=CompareEquTerms(binp,ptr1,ptr2,&node1->subterms[jj],
					&node2->subterms[jj],flag))){
				return(ii);
			}
		}

	/* Non recursive ordering. Check >>> ordering relation (slightly modified). */
	} else {

		/* Loop through the arguments of both items. */
		for (ptr1=NextItem(item1,IMMED),ptr2=NextItem(item2,IMMED),ptr3=NextItem(item1,OVERSUBTERMS),jj=0;
				ptr1<ptr3;ptr1=NextItem(ptr1,OVERSUBTERMS),ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){

			/* None of the items is a variable. */
			if ((0==(ptr1[0]&VARIABLE))&&(0==(ptr2[0]&VARIABLE))){
				if (EQUAL!=(ii=SecondKBOCheck2(binp,ptr1,ptr2,NULL,NULL,flag))){
					return(ii);
				}

			/* Both items are a variable. */
			} else if ((ptr1[0]&VARIABLE)&&(ptr2[0]&VARIABLE)){

				/* The variables are not the same. */
				if ((*((int16_t *)&ptr1[1]))!=(*((int16_t *)&ptr2[1]))){
					return(FAILURE);
				}

			/* One item is a variable and the other is not a variable. */
			} else {
				return(FAILURE);
			}
		}
	}

	return(EQUAL);
 } /* SecondKBOCheck2 */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform second KBO check with substitution)  OCJ
 *
 *    This function is analogous to the SecondKBOCheck2() function
 *    except that applies a given substitution to terms being compared
 *    if necessary.
 *
 *    See SecondKBOCheck2() function comments for additional details.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of clause to which the
 *          terms belong.
 *    item1: Pointer to first item to be compared.
 *    item2: Pointer to second item to be compared.
 *    node1: Pointer to equpnode node structure for first term. This
 *           is not used in case of non recursive ordering so it is
 *           NULL on input in this case.
 *    node2: Pointer to equpnode node structure for second term. This
 *           is not used in case of non recursive ordering so it is
 *           NULL on input in this case.
 *    Subst: Pointer to substitution to be applied before checking
 *           the ordering constraint.
 *    varshift: Shift for variable numbers belonging to a secondary
 *              merged clause. Variable numbers coming from secondary clause
 *              increased by varshift match variable numbers in the substitution.
 *    flag: STANDARD or NONRECURSIVE depending on the ordering type.
 *          MAXIMUM: If this flag is set then the first term is checked
 *          for being greater than second term. If this flag is not set
 *          then the first term is checked for being not lesser nor equal
 *          than the second term.
 *          ITEM1SEC: If this flag is set then both terms belong to
 *          a secondary item with respect to the substitution. If this
 *          flag is not set then the terms belong to a primary item.
 *
 *  RETURNS:
 *
 *    GREATER, LESSER, EQUAL, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t SecondKBOCheck3(binprefix *binp,uint8_t *item1,uint8_t *item2,equpnode *node1,
		equpnode *node2,subst *Subst,int32_t varshift,int32_t flag){

	/* Item 1 is not a "distinct object" and item 1 is a "distinct object". */
	/* Symbol precedence relation (>>) holds. */
	if ((DISTINCTOBJECT&((symbol *)&item1[1])->symbol->type)
			&&(0==(DISTINCTOBJECT&((symbol *)&item2[1])->symbol->type))){
		return(GREATER);
	}

	/* Item 1 is a "distinct object" and item 1 is not a "distinct object". */
	/* Symbol precedence relation (<<) holds. */
	if ((DISTINCTOBJECT&((symbol *)&item2[1])->symbol->type)
			&&(0==(DISTINCTOBJECT&((symbol *)&item1[1])->symbol->type))){
		return(LESSER);
	}

	/* Symbol precedence relation (>>) holds. */
	if (((symbol *)&item1[1])->symbol->precedence>((symbol *)&item2[1])->symbol->precedence){
		return(GREATER);
	}

	/* Symbol precedence relation (<<) holds. */
	if (((symbol *)&item1[1])->symbol->precedence<((symbol *)&item2[1])->symbol->precedence){
		return(LESSER);
	}

	/* Both symbols have the same precedence. */
	return(CompareLexic(binp,item1,item2,node1,node2,Subst,varshift,flag));
 } /* SecondKBOCheck3 */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform lexicographical comparison with substitution)  OCJ
 *
 *    This function performs a lexicographical comparison of two terms
 *    of the form item1=f(s1,s2,...,sn) and item2=f(t1,t2,...,tn) after
 *    applying a substitution to the terms.
 *
 *    Pre-computed comparison information without substitution must
 *    be available with the arguments node1 and node2.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of clause to which the
 *          terms belong.
 *    item1: Pointer to first item to be compared.
 *    item2: Pointer to second item to be compared.
 *    node1: Pointer to equpnode node structure for first term. This
 *           is not used in case of non recursive ordering so it may
 *           be NULL on input in this case.
 *    node2: Pointer to equpnode node structure for second term. This
 *           is not used in case of non recursive ordering so it may
 *           be NULL on input in this case.
 *    Subst: Pointer to substitution to be applied before checking
 *           the ordering constraint.
 *    varshift: Shift for variable numbers belonging to a secondary
 *              merged clause. Variable numbers coming from secondary clause
 *              increased by varshift match variable numbers in the substitution.
 *    flag: STANDARD or NONRECURSIVE depending on the ordering type.
 *          MAXIMUM: If this flag is set then the first term is checked
 *          for being greater than second term. If this flag is not set
 *          then the first term is checked for being not lesser nor equal
 *          than the second term.
 *          ITEM1SEC: If this flag is set then both terms belong to
 *          a secondary item with respect to the substitution. If this
 *          flag is not set then the terms belong to a primary item.
 *
 *  RETURNS:
 *
 *    GREATER, LESSER, EQUAL, FAILURE or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t CompareLexic(binprefix *binp,uint8_t *item1,uint8_t *item2,equpnode *node1,
		equpnode *node2,subst *Subst,int32_t varshift,int32_t flag){
	auto cmprefix *dmyclause;                       /* Dummy clause cmprefix structure pointer */
	auto uint8_t *sterm1,*sterm2;                   /* Pointers to substituted terms */
	auto uint8_t *ptr1,*ptr2,*ptr3,*ptr4;           /* Auxiliary pointers */
	auto int32_t ii,jj;                             /* Auxiliary. */

	/* Standard ordering. Check standard KBO ordering relation recursively. */
	if (flag&STANDARD){

		/* Loop through the arguments of both items. */
		for (ptr1=NextItem(item1,IMMED),ptr2=NextItem(item2,IMMED),ptr3=NextItem(item1,OVERSUBTERMS),jj=0;
				ptr1<ptr3;ptr1=NextItem(ptr1,OVERSUBTERMS),ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){
			if (EQUAL!=(ii=CompSubstEquTrms(binp,ptr1,ptr2,&node1->subterms[jj],
					&node2->subterms[jj],Subst,varshift,flag))){
				return(ii);
			}
		}

	/* Non recursive ordering. Check >>> ordering relation (slightly modified). */
	} else {

		/* Loop through the arguments of both items. */
		for (ptr1=NextItem(item1,IMMED),ptr2=NextItem(item2,IMMED),ptr3=NextItem(item1,OVERSUBTERMS),jj=0;
				ptr1<ptr3;ptr1=NextItem(ptr1,OVERSUBTERMS),ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){

			/* None of the items is a variable. */
			if ((0==(ptr1[0]&VARIABLE))&&(0==(ptr2[0]&VARIABLE))){
				if (EQUAL!=(ii=SecondKBOCheck3(binp,ptr1,ptr2,NULL,NULL,Subst,varshift,flag))){
					return(ii);
				}

			/* Both items are a variable. */
			} else if ((ptr1[0]&VARIABLE)&&(ptr2[0]&VARIABLE)){

				/* The variables are not the same. Apply substitution to items and compare them. */
				if ((*((int16_t *)&ptr1[1]))!=(*((int16_t *)&ptr2[1]))){

					/* Apply substitution to items. */
					if (NOMEMORY==ApplySbst2TwoItm(ptr1,ptr2,&sterm1,&sterm2,&dmyclause,Subst,varshift,flag)){
						return(NOMEMORY);
					}

					/* None of the substituted items is a variable. The substitution is */
					/* temporarily disabled and the ITEM1SEC flag is disabled in the call. */
					if ((0==(sterm1[0]&VARIABLE))&&(0==(sterm2[0]&VARIABLE))){
						ptr4=Subst->buffer;
						Subst->buffer=NULL;
						ii=SecondKBOCheck3(binp,sterm1,sterm2,NULL,NULL,Subst,varshift,flag&(~ITEM1SEC));
						Subst->buffer=ptr4;
						if (EQUAL!=ii){
							if (dmyclause!=NULL){
								MYFREE(dmyclause->part2.bin);
								MYFREE(dmyclause);
							}
							return(ii);
						}

					/* Both substituted items are a variable. */
					} else if ((sterm1[0]&VARIABLE)&&(sterm2[0]&VARIABLE)){

						/* The variables are not the same. */
						if ((*((int16_t *)&sterm1[1]))!=(*((int16_t *)&sterm2[1]))){
							if (dmyclause!=NULL){
								MYFREE(dmyclause->part2.bin);
								MYFREE(dmyclause);
							}
							return(FAILURE);
						}

					/* One item is a variable and the other is not a variable. */
					} else {
						if (dmyclause!=NULL){
							MYFREE(dmyclause->part2.bin);
							MYFREE(dmyclause);
						}
						return(FAILURE);
					}
					if (dmyclause!=NULL){
						MYFREE(dmyclause->part2.bin);
						MYFREE(dmyclause);
					}
				}

			/* One item is a variable and the other is not a variable. */
			} else {

				/* Apply substitution to items. */
				if (NOMEMORY==ApplySbst2TwoItm(ptr1,ptr2,&sterm1,&sterm2,&dmyclause,Subst,varshift,flag)){
					return(NOMEMORY);
				}

				/* None of the items is a variable. The substitution is temporarily */
				/* disabled and the ITEM1SEC flag is disabled in the call. */
				if ((0==(sterm1[0]&VARIABLE))&&(0==(sterm2[0]&VARIABLE))){
					ptr4=Subst->buffer;
					Subst->buffer=NULL;
					ii=SecondKBOCheck3(binp,sterm1,sterm2,NULL,NULL,Subst,varshift,flag&(~ITEM1SEC));
					Subst->buffer=ptr4;
					if (EQUAL!=ii){
						if (dmyclause!=NULL){
							MYFREE(dmyclause->part2.bin);
							MYFREE(dmyclause);
						}
						return(ii);
					}

				/* One item is a variable and the other is not a variable. */
				} else {
					if (dmyclause!=NULL){
						MYFREE(dmyclause->part2.bin);
						MYFREE(dmyclause);
					}
					return(FAILURE);
				}
				if (dmyclause!=NULL){
					MYFREE(dmyclause->part2.bin);
					MYFREE(dmyclause);
				}
			}
		}
	}
	return(EQUAL);
 } /* CompareLexic */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Terms and predicates weight comparison)  OCJ
 *
 *    This function performs a weight comparison of two items
 *    that can be terms or predicates.
 *
 *    Let > and ·> the binary relations of definition 3.13 as
 *    described in the document "Implementing an efficient
 *    theorem prover" by Alexandre Riazanov. Then the return
 *    conditions of this function are:
 *
 *    EQUAL if weight(item1) = weight(item2), that is,
 *          if weight(item1) ·> weight(item2)
 *          and weight(item2) ·> weight(item1)
 *    GREATER if weight(item1) > weight(item2)
 *    LESSER if weight(item2) > weight(item1)
 *    NOGREATER if weight(item2) ·> weight(item1)
 *              but neither weight(item2) = weight(item1) (see EQUAL
 *              condition) nor weight(item2) > weight(item1)
 *    NOLESSER if weight(item1) ·> weight(item2)
 *             but neither weight(item1) = weight(item2) (see EQUAL
 *             condition) nor weight(item1) > weight(item2)
 *    FAILURE otherwise
 *
 *    Both items must belong to the same clause.
 *
 *
 *  ARGUMENTS:
 *
 *    item1: Pointer to first item to be compared.
 *    item2: Pointer to second item to be compared.
 *    coef1: Pointer to coeff structure for first item.
 *           This will be ignored if NULL.
 *    coef2: Pointer to coeff structure for second item.
 *           This will be ignored if coef1 is NULL.
 *
 *  RETURNS:
 *
 *    EQUAL, GREATER, LESSER, NOGREATER, NOLESSER, FAILURE
 *    or NOMEMORY (see above).
 *
 *--------------------------------------------------------------*/
int32_t CompareItemWeight(uint8_t *item1,uint8_t *item2,coeff *coef1,coeff *coef2){
	auto binprefix *binp;                           /* Pointer to binprefix structure */
	auto coeff lclcoef1,lclcoef2;                   /* Local coeff structures. */
	auto int32_t *alpha,*beta;                      /* Pointers to polynomial coefficients */
	auto int32_t alpha0,beta0;                      /* Zero order polynomial coefficients */
	auto int32_t jj,kk,mm,nn,pp;                    /* Auxiliary */

	/* Case of two variables. */
	if ((VARIABLE&item1[0])&&(VARIABLE&item2[0])){
		if ((*((int16_t *)&item1[1]))==(*((int16_t *)&item2[1]))){
			return(EQUAL);
		}
		return(FAILURE);
	}

	/* Get pointer to clause binprefix structure. */
	if (VARIABLE&item2[0]){
		binp=(binprefix *)(item1-(binpsize+((symbol *)&item1[1])->offset));
	} else {
		binp=(binprefix *)(item2-(binpsize+((symbol *)&item2[1])->offset));
	}

	/* Coefficients data is available. */
	if (coef1!=NULL){
		alpha0=coef1->alpha0;
		alpha=coef1->alpha;
		beta0=coef2->alpha0;
		beta=coef2->alpha;

	/* Coefficients data is not available. */
	/* Allocate memory for counting variable occurrences and */
	/* Compute polynomial coefficients. */
	} else {
		if (NOMEMORY==CalcCoeffs(item1,&lclcoef1,binp->maxvarnb)){
			return(NOMEMORY);
		}
		if (NOMEMORY==CalcCoeffs(item2,&lclcoef2,binp->maxvarnb)){
			MYFREE(lclcoef1.alpha);
			return(NOMEMORY);
		}
		alpha0=lclcoef1.alpha0;
		alpha=lclcoef1.alpha;
		beta0=lclcoef2.alpha0;
		beta=lclcoef2.alpha;
	}

	/* Check > and ·> binary relations. */
	/* At the end of the loop we get: */
	/* kk=2 if alpha[i] - beta[i] >= 0 for all i and */
	/*         alpha[i] - beta[i] > 0 for some i. */
	/* kk=0 if alpha[i] - beta[i] <= 0 for all i and */
	/*         alpha[i] - beta[i] < 0 for some i. */
	/* kk=1 if alpha[i] - beta[i] = 0 for all i. */
	/* kk=3 if alpha[i] - beta[i] > 0 for some i and */
	/*         alpha[j] - beta[j] < 0 for some j. */
	/* For the first three cases we get nn=sum(alpha[i]-beta[i]. */
	mm=alpha0-beta0;
	for (jj=nn=0,kk=1;jj<=binp->maxvarnb;jj++){
		pp=alpha[jj]-beta[jj];
		if (pp<0){
			if (kk==2){
				kk=3;
				break;
			}
			kk=0;
		} else if (pp>0){
			if (kk==0){
				kk=3;
				break;
			}
			kk=2;
		}
		nn+=pp;
	}

	/* Free memory and return as appropriate. */
	/* IMPORTANT: The commented code corresponds to what it is indicated as */
	/* "in practice..." in section 3.1.5 of the document "Implementing an */
	/* efficient theorem prover" by Alexandre Riazanov for weight comparison. */
	/* The used code is what I interpret corresponds to the definition 3.13 */
	/* in the above document. */
	if (coef1==NULL){
		MYFREE(lclcoef1.alpha);
		MYFREE(lclcoef2.alpha);
	}
	switch (kk){
		case 0:
			/*if (mm<=0){
				return(LESSER);
			}
			if ((-nn)>=mm){
				return(NOGREATER);
			}*/
			if ((mm<=0)||((-nn)>mm)){
				return(LESSER);
			}
			if ((-nn)==mm){
				return(NOGREATER);
			}
			break;
		case 1:
			if (mm<0){
				return(LESSER);
			}
			if (mm>0){
				return(GREATER);
			}
			return(EQUAL);
			break;
		case 2:
			/*if (mm>=0){
				return(GREATER);
			}
			if (nn>=(-mm)){
				return(NOLESSER);
			}*/
			if ((mm>=0)||(nn>(-mm))){
				return(GREATER);
			}
			if (nn==(-mm)){
				return(NOLESSER);
			}
			break;
	}
 	return(FAILURE);
 } /* CompareItemWeight */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Equality subterms weight comparison)  OCJ
 *
 *    This function performs a weight comparison of two subterms
 *    that are in different sides of an equality or inequality.
 *    It gives the same output than CompareItemWeight() function.
 *    The difference is that this function takes as input the
 *    polynomials of the subterms in compact form.
 *
 *    This function uses some faster algorithms described in section
 *    9.2.1 of the document "Implementing an efficient theorem prover"
 *    by Alexandre Riazanov.
 *
 *    See CompareItemWeight() function comments for a description
 *    of weight comparison of two linear polynomials.
 *
 *
 *  ARGUMENTS:
 *
 *    coef1: Pointer to coeffcpt structure for first subterm.
 *    coef2: Pointer to coeffcpt structure for second subterm.
 *
 *  RETURNS:
 *
 *    EQUAL, GREATER, LESSER, NOGREATER, NOLESSER or FAILURE
 *    (see CompareItemWeight() function comments).
 *
 *--------------------------------------------------------------*/
int32_t CompareItemWeight2(coeffcpt *coef1,coeffcpt *coef2){
	auto int32_t ii,jj,mm,nn,pp;                    /* Auxiliary */

	/* Check > and ·> binary relations. In the following alpha[i] and beta[i] */
	/* are the coefficients of the linear polynomials. The values of the */
	/* variables for the coefficients checked so far after a loop iteration are: */
	/* mm=0 if alpha[i]-beta[i]=0 for all i. */
	/* mm=1 if alpha[i]-beta[i]=0 for all i>0 and alpha[0]-beta[0]>0. */
	/* mm=2 if alpha[i]-beta[i]=0 for all i>0 and alpha[0]-beta[0]<0. */
	/* mm=3 if alpha[i]-beta[i]>=0 for all i, alpha[i]-beta[i]>0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]=0. */
	/* mm=4 if alpha[i]-beta[i]<=0 for all i, alpha[i]-beta[i]<0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]=0. */
	/* mm=5 if alpha[i]-beta[i]>=0 for all i, alpha[i]-beta[i]>0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]>0. */
	/* mm=6 if alpha[i]-beta[i]<=0 for all i, alpha[i]-beta[i]<0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]<0. */
	/* mm=7 if alpha[i]-beta[i]>=0 for all i>0, alpha[i]-beta[i]>0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]<0. */
	/* mm=8 if alpha[i]-beta[i]<=0 for all i>0, alpha[i]-beta[i]<0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]>0. */
	/* nn=sum(alpha[i]-beta[i]) for all i>0 at the end of the loop. */
	/* pp=-(alpha[0] - beta[0]) at the end of the loop. */
	ii=jj=mm=nn=pp=0;
	while ((coef1[ii].coefnb>=0)||(coef2[jj].coefnb>=0)) {

		/* Beta is zero for this alpha or there are no more alpha's. */
		if (coef1[ii].coefnb<coef2[jj].coefnb){

			/* Beta is zero for this alpha. */
			if (coef1[ii].coefnb>=0){
				switch (mm){
					case 0:
						if (coef1[ii].coefnb==0){
							pp=-coef1[ii].coef;
							mm=1;
						} else {
							mm=3;
						}
						break;
					case 1:
						mm=5;
						break;
					case 2:
						mm=7;
						break;
					case 4:
					case 6:
					case 8:
						return(FAILURE);
						break;
				}
				if (coef1[ii].coefnb!=0){
					nn+=coef1[ii].coef;
				}
				ii++;

			/* There are no more alpha's. */
			} else {
				switch (mm){
					case 0:
					case 2:
					case 4:
					case 6:
						return(LESSER);
						break;
					case 1:
						mm=8;
						break;
					case 3:
					case 5:
					case 7:
						return(FAILURE);
						break;
				}
				for (;coef2[jj].coefnb>=0;jj++){
					nn-=coef2[jj].coef;
				}
				break;
			}

		/* Neither alpha nor beta are zero. */
		} else if (coef1[ii].coefnb==coef2[jj].coefnb){

			/* alpha>beta. */
			if (coef1[ii].coef>coef2[jj].coef){
				switch (mm){
					case 0:
						if (coef1[ii].coefnb==0){
							mm=1;
						} else {
							mm=3;
						}
						break;
					case 1:
						mm=5;
						break;
					case 2:
						mm=7;
						break;
					case 4:
					case 6:
					case 8:
						return(FAILURE);
						break;
				}

			/* beta>alpha. */
			} else if (coef1[ii].coef<coef2[jj].coef){
				switch (mm){
					case 0:
						if (coef1[ii].coefnb==0){
							mm=2;
						} else {
							mm=4;
						}
						break;
					case 1:
						mm=8;
						break;
					case 2:
						mm=6;
						break;
					case 3:
					case 5:
					case 7:
						return(FAILURE);
						break;
				}
			}

			/* Prepare next iteration. */
			if (coef1[ii].coefnb==0){
				pp=coef2[jj].coef-coef1[ii].coef;
			} else {
				nn+=(coef1[ii].coef-coef2[jj].coef);
			}
			ii++;
			jj++;

		/* Alpha is zero for this beta or there are no more beta's. */
		} else {

			/* Alpha is zero for this beta. */
			if (coef2[jj].coefnb>=0){
				switch (mm){
					case 0:
						if (coef2[jj].coefnb==0){
							pp=coef2[jj].coef;
							mm=2;
						} else {
							mm=4;
						}
						break;
					case 1:
						mm=8;
						break;
					case 2:
						mm=6;
						break;
					case 3:
					case 5:
					case 7:
						return(FAILURE);
						break;
				}
				if (coef2[jj].coefnb!=0){
					nn-=coef2[jj].coef;
				}
				jj++;

			/* There are no more beta's. */
			} else {
				switch (mm){
					case 0:
					case 1:
					case 3:
					case 5:
						return(GREATER);
						break;
					case 2:
						mm=7;
						break;
					case 4:
					case 6:
					case 8:
						return(FAILURE);
						break;
				}
				for (;coef1[ii].coefnb>=0;ii++){
					nn+=coef1[ii].coef;
				}
				break;
			}
		}
	}

	/* Return as appropriate. */
	/* IMPORTANT: The commented code corresponds to what it is indicated as */
	/* "in practice..." in section 3.1.5 of the document "Implementing an */
	/* efficient theorem prover" by Alexandre Riazanov for weight comparison. */
	/* The used code is what I interpret corresponds to the definition 3.13 */
	/* in the above document. */
	switch (mm){
		case 0:
			return(EQUAL);
			break;
		case 1:
		case 3:
		case 5:
			return(GREATER);
			break;
		case 2:
		case 4:
		case 6:
			return(LESSER);
			break;
		case 7:
			/*if (nn>=pp){
				return(NOLESSER);
			}*/
			if (nn==pp){
				return(NOLESSER);
			} else if (nn>pp){
				return(GREATER);
			}
			break;
		case 8:
			/*if (nn<=pp){
				return(NOGREATER);
			}*/
			if (nn==pp){
				return(NOGREATER);
			} else if (nn<pp){
				return(LESSER);
			}
			break;
	}
 	return(FAILURE);
 } /* CompareItemWeight2 */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Equality subterms weight comparison with substitution)  OCJ
 *
 *    This function performs a weight comparison of two subterms
 *    that are in different sides of an equality or inequality
 *    after a substitution is applied to both subterms. The substitution
 *    is not applied to the subterms themselves. It is applied to
 *    the precomputed left and right linear polynomials stored
 *    in the equality ordering data tree.
 *
 *    It is assumed that the subterm weights have been compared
 *    without a substitution and the results are available in
 *    the result field of the corresponding equpnode structure of the
 *    equality ordering tree.
 *
 *    This function uses some faster algorithms described in section
 *    9.2.1 of the document "Implementing an efficient theorem prover"
 *    by Alexandre Riazanov.
 *
 *    See CompareItemWeight() function comments for a description
 *    of weight comparison of two linear polynomials.
 *
 *
 *  ARGUMENTS:
 *
 *    node1: Pointer to equpnode structure for first subterm.
 *    node2: Pointer to equpnode structure for secod subterm.
 *    Subst: Pointer to substitution to be applied to the subterm
 *           polynomials.
 *    varshift: Shift for variable numbers belonging to a secondary
 *              merged clause. Variable numbers coming from secondary clause
 *              increased by varshift match variable numbers in the substitution.
 *    flag: MAXIMUM: If this flag is set then the first term is checked
 *          for being greater than second term. If this flag is not set
 *          then the first term is checked for being not lesser nor equal
 *          than the second term.
 *          ITEM1SEC: If this flag is set then both terms belong to
 *          a secondary item with respect to the substitution. If this
 *          flag is not set then the terms belong to a primary item.
 *
 *  RETURNS:
 *
 *    EQUAL, GREATER, LESSER, NOGREATER, NOLESSER, FAILURE or NOMEMORY.
 *    (see CompareItemWeight() function comments).
 *
 *--------------------------------------------------------------*/
int32_t CompItmWeightSubst(equpnode *node1,equpnode *node2,subst *Subst,int32_t varshift,int32_t flag){
	auto lkedcoef *lclcoef1,*lclcoef2;              /* Pointers to temporary polynomials */
	auto lkedcoef *auxcoef1,*auxcoef2;              /* Pointers to more temporary polynomials */
	auto int32_t ii,jj,kk,mm,nn,pp;                 /* Auxiliary */

	/* Substitution is NULL or there are conclusive weight comparison */
	/* results without a substitution applied. */
	if ((Subst->buffer==NULL)||(Subst->substsize==1)||(EQUALW&node1->result)){
		switch ((EQUALW|NOGREATER|NOLESSER)|node1->result){
			case EQUALW:
				return(EQUAL);
				break;
			case NOGREATER:
				return(NOGREATER);
				break;
			case NOLESSER:
				return(NOLESSER);
				break;
		}
	}

	/* First special case: The first term has no alpha[0] coefficient and the */
	/* second term has only a beta[0] coefficient. */
	if ((node1->coefs->coefnb!=0)&&(node2->coefs->coefnb==0)&&(node2->coefs[1].coefnb==-1)){

		/* Compute the minimal weight of a ground instance for all coefficients */
		/* in first term polynomial. */
		mm=0;
		for (ii=0,jj=0;node1->coefs[ii].coefnb!=-1;ii++){
			jj+=node1->coefs[ii].coef;
		}

		/* Loop through coefficients of first term polynomial. */
		for (ii=0;node1->coefs[ii].coefnb!=-1;ii++){

			/* Get polynomial obtained after applying the substitution */
			/* to the variable that corresponds to the coefficient. */
			if (NULL==(lclcoef1=mallocmn(sizeof(lkedcoef)))){
				return(NOMEMORY);
			}
			lclcoef1->coefnb=node1->coefs[ii].coefnb;
			lclcoef1->coef=node1->coefs[ii].coef;
			lclcoef1->next=NULL;
			if (NOMEMORY==ApplySubst2Pol(&lclcoef1,Subst,varshift,flag)){
				for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
					auxcoef2=auxcoef1->next;
					MYFREE(auxcoef1);
				}
				return(NOMEMORY);
			}

			/* Compute minimal weight of ground instance of the new substitution polynomial. */
			for (auxcoef1=lclcoef1,kk=0;auxcoef1!=NULL;auxcoef1=auxcoef1->next){
				kk+=auxcoef1->coef;

				/* If the new substitution polynomial is non ground then indicate so. */
				if (auxcoef1->coefnb!=0){
					mm=1;
				}
			}

			/* Free temporary substitution polynomial. */
			for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
				auxcoef2=auxcoef1->next;
				MYFREE(auxcoef1);
			}

			/* Compute the minimum weight guaranteed so far. */
			jj+=(kk-node1->coefs[ii].coef);

			/* Check if this suffices for a return. */
			if (jj>node2->coefs->coef){
				return(GREATER);
			}
		}

		/* Return as appropriate. */
		if (jj==node2->coefs->coef){
			if (mm){
				return(NOLESSER);
			}
			return(EQUAL);
		}
		if (mm==0){
			return(LESSER);
		}
		return(FAILURE);
	}

	/* Second special case: The second term has no beta[0] coefficient and the */
	/* first term has only an alpha[0] coefficient. */
	if ((node1->coefs->coefnb!=0)&&(node2->coefs->coefnb==0)&&(node2->coefs[1].coefnb==-1)){


		/* Compute the minimal weight of a ground instance for all coefficients */
		/* in second term polynomial. */
		mm=0;
		for (ii=0,jj=0;node2->coefs[ii].coefnb!=-1;ii++){
			jj+=node2->coefs[ii].coef;
		}

		/* Loop through coefficients of second term polynomial. */
		for (ii=0;node2->coefs[ii].coefnb!=-1;ii++){

			/* Get polynomial obtained after applying the substitution */
			/* to the variable that corresponds to the coefficient. */
			if (NULL==(lclcoef1=mallocmn(sizeof(lkedcoef)))){
				return(NOMEMORY);
			}
			lclcoef1->coefnb=node2->coefs[ii].coefnb;
			lclcoef1->coef=node2->coefs[ii].coef;
			lclcoef1->next=NULL;
			if (NOMEMORY==ApplySubst2Pol(&lclcoef1,Subst,varshift,flag)){
				for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
					auxcoef2=auxcoef1->next;
					MYFREE(auxcoef1);
				}
				return(NOMEMORY);
			}

			/* The new substitution polynomial is non-ground. */
			if ((lclcoef1->coefnb>0)||(lclcoef1->next!=NULL)){

				/* The flag MAXIMUM  is set. Return FAILURE as this suffices although */
				/* the relationship may also be LESSER or NOGREATER. */
				if (flag&MAXIMUM){
					for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					return(FAILURE);

				/* The flag MAXIMUM is not set. Set mm=1 to indicate that the */
				/* new substitution polynomial is non-ground. */
				} else {
					mm=1;
				}
			}

			/* Compute minimal weight of ground instance of the new substitution polynomial. */
			for (auxcoef1=lclcoef1,kk=0;auxcoef1!=NULL;auxcoef1=auxcoef1->next){
				kk+=auxcoef1->coef;
			}

			/* Free temporary substitution polynomial. */
			for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
				auxcoef2=auxcoef1->next;
				MYFREE(auxcoef1);
			}

			/* Compute the minimum weight guaranteed so far. */
			jj+=(kk-node2->coefs[ii].coef);

			/* Check if this suffices for a return. */
			if (jj>node1->coefs->coef){
				return(LESSER);
			}
		}

		/* Return as appropriate. */
		if (jj==node1->coefs->coef){
			if (mm){
				return(NOGREATER);
			}
			return(EQUAL);
		}
		if (mm==0){
			return(GREATER);
		}
		return(FAILURE);
	}

	/* GENERAL CASE. */
	/* Build linked list polynomials for first term. */
	for (ii=0,lclcoef1=auxcoef1=NULL;node1->coefs[ii].coefnb!=-1;ii++){
		auxcoef2=auxcoef1;
		if (NULL==(auxcoef1=mallocmn(sizeof(lkedcoef)))){
			for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
				auxcoef2=auxcoef1->next;
				MYFREE(auxcoef1);
			}
			return(NOMEMORY);
		}
		auxcoef1->coef=node1->coefs[ii].coef;
		auxcoef1->coefnb=node1->coefs[ii].coefnb;
		if (auxcoef2!=NULL){
			auxcoef2->next=auxcoef1;
		} else {
			lclcoef1=auxcoef1;
		}
	}
	if (auxcoef1!=NULL){
		auxcoef1->next=NULL;
	}

	/* Build linked list polynomials for second term. */
	for (ii=0,lclcoef2=auxcoef1=NULL;node2->coefs[ii].coefnb!=-1;ii++){
		auxcoef2=auxcoef1;
		if (NULL==(auxcoef1=mallocmn(sizeof(lkedcoef)))){
			for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
				auxcoef2=auxcoef1->next;
				MYFREE(auxcoef1);
			}
			for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
				auxcoef2=auxcoef1->next;
				MYFREE(auxcoef1);
			}
			return(NOMEMORY);
		}
		auxcoef1->coef=node2->coefs[ii].coef;
		auxcoef1->coefnb=node2->coefs[ii].coefnb;
		if (auxcoef2!=NULL){
			auxcoef2->next=auxcoef1;
		} else {
			lclcoef2=auxcoef1;
		}
	}
	if (auxcoef1!=NULL){
		auxcoef1->next=NULL;
	}

	/* Apply substitution to linked list polynomials for first and second terms. */
	if (NOMEMORY==ApplySubst2Pol(&lclcoef1,Subst,varshift,flag)){
		for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
			auxcoef2=auxcoef1->next;
			MYFREE(auxcoef1);
		}
		for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
			auxcoef2=auxcoef1->next;
			MYFREE(auxcoef1);
		}
		return(NOMEMORY);
	}
	if (NOMEMORY==ApplySubst2Pol(&lclcoef2,Subst,varshift,flag)){
		for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
			auxcoef2=auxcoef1->next;
			MYFREE(auxcoef1);
		}
		for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
			auxcoef2=auxcoef1->next;
			MYFREE(auxcoef1);
		}
		return(NOMEMORY);
	}

	/* Check > and ·> binary relations for linked list polynomials of first and */
	/* second terms. In the following alpha[i] and beta[i] are the coefficients */
	/* of the linear polynomials. The values of the variables for the coefficients */
	/* checked so far after a loop iteration are: */
	/* mm=0 if alpha[i]-beta[i]=0 for all i. */
	/* mm=1 if alpha[i]-beta[i]=0 for all i>0 and alpha[0]-beta[0]>0. */
	/* mm=2 if alpha[i]-beta[i]=0 for all i>0 and alpha[0]-beta[0]<0. */
	/* mm=3 if alpha[i]-beta[i]>=0 for all i, alpha[i]-beta[i]>0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]=0. */
	/* mm=4 if alpha[i]-beta[i]<=0 for all i, alpha[i]-beta[i]<0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]=0. */
	/* mm=5 if alpha[i]-beta[i]>=0 for all i, alpha[i]-beta[i]>0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]>0. */
	/* mm=6 if alpha[i]-beta[i]<=0 for all i, alpha[i]-beta[i]<0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]<0. */
	/* mm=7 if alpha[i]-beta[i]>=0 for all i>0, alpha[i]-beta[i]>0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]<0. */
	/* mm=8 if alpha[i]-beta[i]<=0 for all i>0, alpha[i]-beta[i]<0 for at */
	/*      least one i!=0 and alpha[0]-beta[0]>0. */
	/* nn=sum(alpha[i]-beta[i]) for all i>0 at the end of the loop. */
	/* pp=-(alpha[0] - beta[0]) at the end of the loop. */
	mm=nn=pp=0;
	auxcoef1=lclcoef1;auxcoef2=lclcoef2;
	while ((auxcoef1!=NULL)||(auxcoef2!=NULL)){

		/* There are no more alpha's. */
		if (auxcoef1==NULL){
			switch (mm){
				case 0:
				case 2:
				case 4:
				case 6:
					for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					return(LESSER);
					break;
				case 1:
					mm=8;
					break;
				case 3:
				case 5:
				case 7:
					for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					return(FAILURE);
					break;
			}
			for (;auxcoef2!=NULL;auxcoef2=auxcoef2->next){
				nn-=auxcoef2->coef;
			}
			break;

		/* There are no more betas's. */
		} else if (auxcoef2==NULL){
			switch (mm){
				case 0:
				case 1:
				case 3:
				case 5:
					for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					return(GREATER);
					break;
				case 2:
					mm=7;
					break;
				case 4:
				case 6:
				case 8:
					for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					return(FAILURE);
					break;
			}
			for (;auxcoef1!=NULL;auxcoef1=auxcoef1->next){
				nn+=auxcoef1->coef;
			}
			break;
		}

		/* Beta is zero for this alpha. */
		if (auxcoef1->coefnb<auxcoef2->coefnb){
			switch (mm){
				case 0:
					if (auxcoef1->coefnb==0){
						pp=-auxcoef1->coef;
						mm=1;
					} else {
						mm=3;
					}
					break;
				case 1:
					mm=5;
					break;
				case 2:
					mm=7;
					break;
				case 4:
				case 6:
				case 8:
					for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					return(FAILURE);
					break;
			}
			if (auxcoef1->coefnb!=0){
				nn+=auxcoef1->coef;
			}
			auxcoef1=auxcoef1->next;

		/* Alpha is zero for this beta. */
		} else if (auxcoef2->coefnb<auxcoef1->coefnb){
			switch (mm){
				case 0:
					if (auxcoef2->coefnb==0){
						pp=auxcoef2->coef;
						mm=2;
					} else {
						mm=4;
					}
					break;
				case 1:
					mm=8;
					break;
				case 2:
					mm=6;
					break;
				case 3:
				case 5:
				case 7:
					for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
						auxcoef2=auxcoef1->next;
						MYFREE(auxcoef1);
					}
					return(FAILURE);
					break;
			}
			if (auxcoef2->coefnb!=0){
				nn-=auxcoef2->coef;
			}
			auxcoef2=auxcoef2->next;

		/* Neither alpha nor beta are zero. */
		} else {

			/* alpha>beta. */
			if (auxcoef1->coef>auxcoef2->coef){
				switch (mm){
					case 0:
						if (auxcoef1->coefnb==0){
							mm=1;
						} else {
							mm=3;
						}
						break;
					case 1:
						mm=5;
						break;
					case 2:
						mm=7;
						break;
					case 4:
					case 6:
					case 8:
						for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
							auxcoef2=auxcoef1->next;
							MYFREE(auxcoef1);
						}
						for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
							auxcoef2=auxcoef1->next;
							MYFREE(auxcoef1);
						}
						return(FAILURE);
						break;
				}

			/* beta>alpha. */
			} else if (auxcoef1->coef<auxcoef2->coef){
				switch (mm){
					case 0:
						if (auxcoef1->coefnb==0){
							mm=2;
						} else {
							mm=4;
						}
						break;
					case 1:
						mm=8;
						break;
					case 2:
						mm=6;
						break;
					case 3:
					case 5:
					case 7:
						for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
							auxcoef2=auxcoef1->next;
							MYFREE(auxcoef1);
						}
						for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
							auxcoef2=auxcoef1->next;
							MYFREE(auxcoef1);
						}
						return(FAILURE);
						break;
				}
			}

			/* Prepare next iteration. */
			if (auxcoef1->coefnb==0){
				pp=auxcoef2->coef-auxcoef1->coef;
			} else {
				nn+=(auxcoef1->coef-auxcoef2->coef);
			}
			auxcoef1=auxcoef1->next;
			auxcoef2=auxcoef2->next;
		}
	}

	/* Free memory. */
	for (auxcoef1=lclcoef1;auxcoef1!=NULL;auxcoef1=auxcoef2){
		auxcoef2=auxcoef1->next;
		MYFREE(auxcoef1);
	}
	for (auxcoef1=lclcoef2;auxcoef1!=NULL;auxcoef1=auxcoef2){
		auxcoef2=auxcoef1->next;
		MYFREE(auxcoef1);
	}

	/* Return as appropriate. */
	/* IMPORTANT: The commented code corresponds to what it is indicated as */
	/* "in practice..." in section 3.1.5 of the document "Implementing an */
	/* efficient theorem prover" by Alexandre Riazanov for weight comparison. */
	/* The used code is what I interpret corresponds to the definition 3.13 */
	/* in the above document. */
	switch (mm){
		case 0:
			return(EQUAL);
			break;
		case 1:
		case 3:
		case 5:
			return(GREATER);
			break;
		case 2:
		case 4:
		case 6:
			return(LESSER);
			break;
		case 7:
			/*if (nn>=pp){
				return(NOLESSER);
			}*/
			if (nn==pp){
				return(NOLESSER);
			} else if (nn>pp){
				return(GREATER);
			}
			break;
		case 8:
			/*if (nn<=pp){
				return(NOGREATER);
			}*/
			if (nn==pp){
				return(NOGREATER);
			} else if (nn<pp){
				return(LESSER);
			}
			break;
	}
 	return(FAILURE);
 } /* CompItmWeightSubst */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Apply substitution to polynomial)  OCJ
 *
 *    This function applies a substitution to a polynomial in linked
 *    list compact format.
 *
 *
 *  ARGUMENTS:
 *
 *    coef: Address of pointer to lkedcoef structure. On input it points
 *          to the first term of the input linear polynomial. On output
 *          it will point to the first term of the polynomial with the
 *          substitution applied.
 *    Subst: Pointer to substitution to be applied.
 *    varshift: Shift for variable numbers belonging to a secondary
 *              merged clause. Variable numbers coming from secondary clause
 *              increased by varshift match variable numbers in the substitution.
 *    flag: ITEM1SEC: If this flag is set then both terms belong to
 *          a secondary item with respect to the substitution. If this
 *          flag is not set then the terms belong to a primary item.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *--------------------------------------------------------------*/
int32_t ApplySubst2Pol(lkedcoef **coef,subst *Subst,int32_t varshift,int32_t flag){
	auto uint8_t *ptr1,*ptr2,*ptr3;                 /* Auxiliary pointers */
	auto lkedcoef *ptr4,*ptr5,*ptr6;                /* Auxiliary pointers */
	auto lkedcoef *prevcoef;                        /* Pointer to previous term in polynomial */
	auto lkedcoef *auxcoef;                         /* Pointer to current term in polynomial */
	auto int32_t ii,jj,kk,mm;                       /* Auxiliary */

	/* Shift variable numbers in polynomial if necessary. */
	if ((flag&ITEM1SEC)&&(varshift>0)){
		for (auxcoef=*coef;auxcoef!=NULL;auxcoef=auxcoef->next){
			if (auxcoef->coefnb>0){
				auxcoef->coefnb+=varshift;
			}
		}
	}

	/* Return if substitution is empty. */
	if ((Subst->buffer==NULL)||(Subst->substsize==1)){
		return(0);
	}

	/* Loop until no substitutions are applied. */
	do {

		/* Loop through coefficients of variables in polynomial. */
		prevcoef=NULL;
		for (ii=0,auxcoef=*coef;auxcoef!=NULL;auxcoef=(auxcoef==NULL?*coef:auxcoef->next)){

			/* This coefficient is not alpha[0]. */
			if (auxcoef->coefnb!=0){

				/* Loop through substitutions. */
				for (ptr1=Subst->buffer;ptr1[0]!=UNITEND;ptr1=NextItem(&ptr1[3],OVERSUBTERMS)){

					/* The corresponding variable must be substituted. */
					if ((*((int16_t *)&ptr1[1])+(ptr1[0]&ITEM1SEC?1+varshift:1))==auxcoef->coefnb){

						/* Unlink variable from polynomial but remember its coefficient. */
						jj=auxcoef->coef;
						if (prevcoef!=NULL){
							prevcoef->next=auxcoef->next;
						} else {
							*coef=auxcoef->next;
						}
						MYFREE(auxcoef);
						auxcoef=prevcoef;

						/* Loop through substitution term items. */
						ptr3=NextItem(&ptr1[3],OVERSUBTERMS);
						for (ptr2=&ptr1[3];ptr2<ptr3;ptr2=NextItem(ptr2,IMMED)){

							/* Find the polynomial term number for this substitution term item. */
							/* If the substitution term is secondary then all variables in int */
							/* must be shifted. */
							if (VARIABLE&ptr2[0]){
								if (ptr1[0]&ITEM2SEC){
									mm=1+varshift+*(int16_t *)&ptr2[1];
								} else {
									mm=1+*(int16_t *)&ptr2[1];
								}
							} else {
								mm=0;
							}

							/* Loop through current polynomial to find the slot for */
							/* the variable item. */
							for (ptr4=*coef,ptr5=NULL,kk=0;ptr4!=NULL;ptr4=(ptr5=ptr4)->next){
								if (ptr4->coefnb==mm){
									kk=1;
									break;
								} else if (ptr4->coefnb>mm){
									break;
								}
							}

							/* New slot at the end of the chain. */
							if (ptr4==NULL){
								if (NULL==(ptr6=mallocmn(sizeof(lkedcoef)))){
									return(NOMEMORY);
								}
								if (ptr5!=NULL){
									ptr5->next=ptr6;
								} else {
									*coef=ptr6;
								}
								ptr6->next=NULL;
								ptr6->coefnb=mm;
								if (mm){
									ptr6->coef=jj;
								} else {
									if (kbset.opts.fweight==ARITY){
										ptr6->coef=jj*(1+10*(((symbol *)&ptr2[1])->symbol->arity));
									} else {
										ptr6->coef=jj;
									}
								}

							/* Slot not at the end of the chain. */
							} else {

								/* Existing slot. */
								if (kk){
									if (mm){
										ptr4->coef+=jj;
									} else {
										if (kbset.opts.fweight==ARITY){
											ptr4->coef+=jj*(1+10*(((symbol *)&ptr2[1])->symbol->arity));
										} else {
											ptr4->coef+=jj;
										}
									}

								/* New slot. */
								} else {
									if (NULL==(ptr6=mallocmn(sizeof(lkedcoef)))){
										return(NOMEMORY);
									}
									if (ptr5!=NULL){
										ptr6->next=ptr5->next;
										ptr5->next=ptr6;
									} else {
										ptr6->next=*coef;
										*coef=ptr6;
									}
									ptr6->coefnb=mm;
									if (mm){
										ptr6->coef=jj;
									} else {
										if (kbset.opts.fweight==ARITY){
											ptr6->coef=jj*(1+10*(((symbol *)&ptr2[1])->symbol->arity));
										} else {
											ptr6->coef=jj;
										}
									}
								}
							}
						}

						/* Signal substitution applied and leave inner loop. */
						ii=1;
						break;
					}
				}
			}

			/* Update prevcoef. */
			prevcoef=auxcoef;
		}
	} while (ii);
 	return(0);
 } /* ApplySubst2Pol */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Apply substitution to two terms)  OCJ
 *
 *    This function applies a substitution to two items that belong to
 *    the same clause. The substitution is applied by building a dummy
 *    clause with just an equality in which the two root terms are the
 *    terms passed as arguments to this function. Then ApplySubst2Clause()
 *    function is called.
 *
 *
 *  ARGUMENTS:
 *
 *    term1: Pointer to first term.
 *    term2: Pointer to second term.
 *    pterm1s: Address of pointer to first term with substitution applied.
 *    pterm2s: Address of pointer to second term with substitution applied.
 *    pclause: Address of pointer to cmprefix structure of dummy clause.
 *    Subst: Pointer to substitution to be applied to term1 and term2
 *           before checking the ordering constraint.
 *    varshift: Shift for variable numbers belonging to a secondary
 *              merged clause. Variable numbers in substitution
 *              coming from secondary clause increased by varshift
 *              match variable numbers in the substitution.
 *    flag: ITEM1SEC: If this flag is set then both terms belong to
 *          a secondary item with respect to the substitution. If this
 *          flag is not set then the terms belong to a primary item.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *--------------------------------------------------------------*/
int32_t ApplySbst2TwoItm(uint8_t *term1,uint8_t *term2,uint8_t **pterm1s,uint8_t **pterm2s,
		cmprefix **pclause,subst *Subst,int32_t varshift,int32_t flag){
	auto uint8_t *ptr1;                             /* Auxiliary pointer */
	auto int32_t ii,jj,kk;                          /* Auxiliary. */

	/* Return if substitution is empty. */
	if ((Subst->buffer==NULL)||(Subst->substsize==1)){
		*pclause=NULL;
		*pterm1s=term1;
		*pterm2s=term2;
		return(0);
	}

	/* Allocate space for dummy clause. */
	if (NULL==((*pclause)=mallocmn(sizeof(cmprefix)))){
		return(NOMEMORY);
	}
	jj=NextItem(term1,OVERSUBTERMS)-term1;
	kk=NextItem(term2,OVERSUBTERMS)-term2;
	ii=sizeof(binprefix)+2*jj+2*kk+1+sizeof(symbol)+CLAUSE_CHUNK_SIZE;
	if (NULL==(((*pclause)->part2.bin)=mallocmn(ii))){
		MYFREE(*pclause);
		return(NOMEMORY);
	}
	(*pclause)->part2.bin->clause=*pclause;
	(*pclause)->part2.bin->equflg=0;

	/* Build dummy equality clause formula. */
	(*pclause)->part2.bin->formula[0]=EQUALITY;
	((symbol *)&(*pclause)->part2.bin->formula[1])->symbol=&equkey;
	(*pterm1s)=&(*pclause)->part2.bin->formula[1]+sizeof(symbol);
	memcpy(*pterm1s,term1,jj);
	memcpy(*pterm1s+jj,term2,kk);
	(*pterm1s)[jj+kk]=UNITEND;

	/* If terms belong to a secondary item then shift variable numbers. */
	if (ITEM1SEC&flag){
		for (ptr1=*pterm1s;ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,IMMED)){
			if (VARIABLE&ptr1[0]){
				*((uint16_t *)&ptr1[1])+=varshift;
			}
		}
	}

	/* Apply substitution to dummy clause and update clause parameters. */
	if (NOMEMORY==ApplySubst2Clause(*pclause,&ii,Subst,varshift)){
		MYFREE((*pclause)->part2.bin);
		MYFREE(*pclause);
		return(NOMEMORY);
	}
	(*pclause)->part2.bin->maxvarnb=-1;
	UpdateOffsets((*pclause)->part2.bin,&(*pclause)->part2.bin->formula[0]);

	/* Get pointers to substituted terms. */
	(*pterm1s)=&(*pclause)->part2.bin->formula[1]+sizeof(symbol);
	(*pterm2s)=NextItem(*pterm1s,OVERSUBTERMS);
	return(0);
 } /* ApplySbst2TwoItm */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Calculate weight polynomial coefficients)  OCJ
 *
 *    This function calculates the weight polynomial coefficients
 *    of a literal or term.
 *
 *    This function allocates memory that must be freed by the caller.
 *
 *
 *  ARGUMENTS:
 *
 *    item: Pointer to literal or term.
 *    coef: Pointer to coeff structure where the coefficient data
 *          will be placed.
 *    maxvarnb: Máximum variable number.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *--------------------------------------------------------------*/
int32_t CalcCoeffs(uint8_t *item,coeff *coef,int32_t maxvarnb){
	auto uint8_t *ptr1,*ptr2;                       /* Auxiliary pointers */
	auto int32_t ii;                                /* Auxiliary */

	/* Allocate memory for counting variable occurrences. */
	ii=sizeof(*coef->alpha)*(maxvarnb+1);
	if (ii>0){
		if (NULL==(coef->alpha=mallocmn(ii))){
			return(NOMEMORY);
		}
		memset(coef->alpha,0,ii);
	} else {
		coef->alpha=NULL;
	}
	coef->alpha0=0;

	/* Compute polynomial coefficients. */
	for (ptr1=item,ptr2=NextItem(item,OVERSUBTERMS);ptr1<ptr2;ptr1=NextItem(ptr1,IMMED)){
		if (ptr1[0]&VARIABLE){
			coef->alpha[*((int16_t *)&ptr1[1])]++;
		} else {
			if (kbset.opts.fweight==ARITY){
				coef->alpha0+=(1+10*(((symbol *)&ptr1[1])->symbol->arity));
			} else {
				coef->alpha0++;
			}
		}
	}

 	return(0);
 } /* CalcCoeffs */

#ifdef SINGLEANY
/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Select items in a clause)  OCJ
 *
 *    This function selects the following items in a clause:
 *    - Appropriate literals in active clauses (see below).
 *    - Maximal equality root terms in selected equalities of active
 *      clauses.
 *    - Maximum equality root terms in single literal active and passive
 *      clauses are flagged as MAXIMUM.
 *
 *    Literals in active clauses are selected as follows:
 *    - If selection is set to MAXIMAL then all maximal literals
 *      are selected. Of course if there is a maximum literal it will be
 *      the only literal selected because it is the only maximal literal.
 *    - If selection is set to SINGLENEG then:
 *      - If there are negative literals then the first one is selected
 *        and the other literals are not selected.
 *      - Otherwise all maximal literals are selected.
 *    - If selection is set to MULTINEG then:
 *      - If there are negative literals then the negative literals that
 *        are maximal in the set of negative literals are selected and
 *        and the other literals are not selected. Of course if there
 *        is a maximum negative literal that will be the only negative
 *        literal selected because the maximum is the only maximal
 *        literal.
 *      - Otherwise all maximal literals are selected.
 *    - If selection is set to SINGLEPOS then:
 *      - If there are positive literals then the first one is selected
 *        and the other literals are not selected.
 *      - Otherwise all maximal literals are selected.
 *    - If selection is set to MULTIPOS then:
 *      - If there are positive literals then the positive literals that
 *        are maximal in the set of positive literals are selected and
 *        and the other literals are not selected. Of course if there
 *        is a maximum negative literal that will be the only positive
 *        literal selected because the maximum is the only maximal
 *        literal.
 *      - Otherwise all maximal literals are selected.
 *
 *    See "The design and implementation of VAMPIRE" section 2.6 by Alexandre
 *    Riazanov and Andrei Voronkov for details about this logic.
 *
 *    Also SELECTED positive equalities in clauses without negative literals
 *    are flagged as PMSELECTED to indicate that they are selected for
 *    paramodulating from them.
 *
 *    IMPORTANT: This function must never be called for main KB.
 *    The clause must belong to the working KB.
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in binary form.
 *    flag: If not zero then the literal selection loop will be
 *          executed not only for ACTIVE clauses but for any type
 *          of clauses.
 *
 *  RETURNS:
 *
 *    0 if successful, NOMEMORY if not enough memory.
 *
 *--------------------------------------------------------------*/
int32_t SelectItems(cmprefix *clause,int32_t flag){
	auto uint8_t *ptr1,*ptr2,*ptr3,*ptr4;           /* Auxiliary pointers */
	auto uint8_t *firstneg,*firstpos;               /* Pointer to first negative and positive literals */
	auto coeff *coefs;                              /* Pointer to coefficient structures */
	auto int32_t eqs;                               /* Number of positive and negative equality literals. */
	auto uint8_t *demodequ;                         /* Pointer to equality suitable for demodulating from if */
	auto int32_t ii,jj,kk,nn;                       /* Auxiliary */

	/* Loop through clause literals. */
	for (ptr1=&clause->part2.bin->formula[0],eqs=0,demodequ=firstneg=firstpos=NULL;
			(*ptr1)!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Deselect literal and equality root terms. Count number of */
		/* equality literals. Equality root terms are deselected only */
		/* if clause has not been used for backward demodulation, as in */
		/* this case root terms are correctly selected. */
		ptr1[0]&=(~(SELECTED|PMSELECTED|MAXIMUM));
		if (ptr1[0]&EQUALITY){
			eqs++;
			if (clause->part2.bin->equflg!=2){
				ptr2=NextItem(ptr1,IMMED);
				ptr2[0]&=(~(SELECTED|MAXIMUM));
				ptr2=NextItem(ptr2,OVERSUBTERMS);
				ptr2[0]&=(~(SELECTED|MAXIMUM));
				if (clause->part2.bin->equflg==0){
					((symbol *)&ptr1[1])->leftnode=NULL;
				}
			}
		}

		/* Remember first negative and positive literals. */
		if ((firstneg==NULL)&&(ptr1[0]&NEGATED)){
			firstneg=ptr1;
		}
		if ((firstpos==NULL)&&(0==(ptr1[0]&NEGATED))){
			firstpos=ptr1;
		}
	}

	/* Check if clause is suitable for performing demodulation from it. */
	if ((clause->part2.bin->literals==1)&&(((EQUALITY|NEGATED)&clause->part2.bin->formula[0])==EQUALITY)){

		/* Remember position of equality suitable for demodulating from it. */
		demodequ=&clause->part2.bin->formula[0];

		/* If clause has been used for backward demodulation then equality */
		/* data has been set, so just select the equality literal if */
		/* appropriate and return. */
		if (clause->part2.bin->equflg==2){
			if (clause->flags&ACTIVE){
				demodequ[0]|=(SELECTED|PMSELECTED);
			}
			return(0);
		}
	} else {
		demodequ=NULL;
	}

	/* Clause is suitable for demodulation from it. */
	if (demodequ!=NULL){

		/* Get pointers to equality root terms. */
		ptr2=NextItem(demodequ,IMMED);
		ptr3=NextItem(ptr2,OVERSUBTERMS);

		/* Calculate equality ordering data and remember that */
		/* this was done for a potentially demodulating clause. */
		if (NOMEMORY==SetEquData(clause->part2.bin,demodequ,ptr2,ptr3)){
			return(NOMEMORY);
		}
		clause->part2.bin->equflg=2;

		/* Select maximum equality root terms. */
		/* Release equality tree data if the root terms ordering */
		/* is completely defined. */
		switch (CompareEquTerms(clause->part2.bin,ptr2,ptr3,&((symbol *)&demodequ[1])->leftnode,
				&((symbol *)&demodequ[1])->rightnode,kbset.opts.termord)){
			case NOMEMORY:
				return(NOMEMORY);
				break;
			case GREATER:
				ptr2[0]|=MAXIMUM|SELECTED;
				FreeEquNode(ptr2,((symbol *)&demodequ[1])->leftnode);
				FreeEquNode(ptr3,((symbol *)&demodequ[1])->rightnode);
				((symbol *)&demodequ[1])->leftnode=NULL;
				break;
			case LESSER:
				ptr3[0]|=MAXIMUM|SELECTED;
				FreeEquNode(ptr2,((symbol *)&demodequ[1])->leftnode);
				FreeEquNode(ptr3,((symbol *)&demodequ[1])->rightnode);
				((symbol *)&demodequ[1])->leftnode=NULL;
				break;
			case EQUAL:
				FreeEquNode(ptr2,((symbol *)&demodequ[1])->leftnode);
				FreeEquNode(ptr3,((symbol *)&demodequ[1])->rightnode);
				((symbol *)&demodequ[1])->leftnode=NULL;
				break;
			case FAILURE:
				ptr2[0]|=SELECTED;
				ptr3[0]|=SELECTED;
				break;
		}

		/* Select literal if clause is active. */
		if (clause->flags&ACTIVE){
			demodequ[0]|=(SELECTED|PMSELECTED);
		}

		/* Return. */
		return(0);
	}

	/* Active clause or flag is set. */
	if ((clause->flags&ACTIVE)||flag){

		/* More than one literal may be potentially selected. */
		if ((clause->part2.bin->literals>1)&&(((kbset.opts.select!=SINGLENEG)||(firstneg==NULL))
				&&((kbset.opts.select!=SINGLEPOS)||(firstpos==NULL)))){

			/* Allocate memory for coefficient structures. */
			if (NULL==(coefs=mallocmn((clause->part2.bin->literals+(2*eqs))*sizeof(coeff)))){
				return(NOMEMORY);
			}

			/* Initialize pointers to equality root term coefficients. */
			for (ptr1=&clause->part2.bin->formula[0],ii=clause->part2.bin->literals,jj=0;
					(*ptr1)!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS),jj++){
				if (ptr1[0]&EQUALITY){
					coefs[jj].eqterm1=&coefs[ii];
					coefs[jj].eqterm2=&coefs[ii+1];
					ii+=2;
				} else {
					coefs[jj].eqterm1=NULL;
				}
			}

			/* Allocate and initialize coefficients data. */
			for (ptr1=&clause->part2.bin->formula[0],ii=0;(*ptr1)!=UNITEND;
					ptr1=NextItem(ptr1,OVERSUBTERMS),ii++){

				/* Literal is an equality. Allocate and initialize */
				/* data for equality root terms. */
				if (ptr1[0]&EQUALITY){
					ptr2=NextItem(ptr1,IMMED);
					if (NOMEMORY==CalcCoeffs(ptr2,coefs[ii].eqterm1,clause->part2.bin->maxvarnb)){
						for (jj=0;jj<ii;jj++){
							if (clause->part2.bin->maxvarnb>=0){
								if (coefs[jj].eqterm1!=NULL){
									MYFREE(coefs[jj].eqterm1->alpha);
									MYFREE(coefs[jj].eqterm2->alpha);
								} else {
									MYFREE(coefs[jj].alpha);
								}
							}
						}
						MYFREE(coefs);
						return(NOMEMORY);
					}
					if (NOMEMORY==CalcCoeffs(NextItem(ptr2,OVERSUBTERMS),coefs[ii].eqterm2,
							clause->part2.bin->maxvarnb)){
						if (clause->part2.bin->maxvarnb>=0){
							for (jj=0;jj<ii;jj++){
								if (coefs[jj].eqterm1!=NULL){
									MYFREE(coefs[jj].eqterm1->alpha);
									MYFREE(coefs[jj].eqterm2->alpha);
								} else {
									MYFREE(coefs[jj].alpha);
								}
							}
							MYFREE(coefs[ii].eqterm1->alpha);
						}
						MYFREE(coefs);
						return(NOMEMORY);
					}

				/* Literal is a predicate. Allocate and initialize data for predicate. */
				} else {
					if (NOMEMORY==CalcCoeffs(ptr1,&coefs[ii],clause->part2.bin->maxvarnb)){
						if (clause->part2.bin->maxvarnb>=0){
							for (jj=0;jj<ii;jj++){
								if (coefs[jj].eqterm1!=NULL){
									MYFREE(coefs[jj].eqterm1->alpha);
									MYFREE(coefs[jj].eqterm2->alpha);
								} else {
									MYFREE(coefs[jj].alpha);
								}
							}
						}
						MYFREE(coefs);
						return(NOMEMORY);
					}
				}
			}

			/* Loop through all literals. Flag MAXIMUM is temporarily */
			/* used here as an auxiliary flag to indicate that a literal */
			/* is lesser than a previous literal. */
			nn=clause->part2.bin->equflg;
			for (ptr1=&clause->part2.bin->formula[0],ii=0;(*ptr1)!=UNITEND;
					ptr1=NextItem(ptr1,OVERSUBTERMS),ii++){

				/* Literal is a candidate for selection. */
				if (((ptr1[0]&NEGATED)||(kbset.opts.select!=MULTINEG)||(firstneg==NULL))
						&&((0==(ptr1[0]&NEGATED))||(kbset.opts.select!=MULTIPOS)||(firstpos==NULL))){

					/* Literal flagged as being lesser than a previous one. Reset MAXIMUM flag. */
					if (ptr1[0]&MAXIMUM){
						ptr1[0]&=(~MAXIMUM);

					/* Literal not flagged as being lesser than a previous one. */
					} else {

						/* Loop through literals following current literal. */
						for (ptr2=NextItem(ptr1,OVERSUBTERMS),jj=ii+1,kk=1;((*ptr2)!=UNITEND)&&(kk);
								ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){

							/* Second loop literal is a candidate for selection. */
							if ((0==(ptr2[0]&MAXIMUM))
									&&((ptr2[0]&NEGATED)||(kbset.opts.select!=MULTINEG)||(firstneg==NULL))
									&&((0==(ptr2[0]&NEGATED))||(kbset.opts.select!=MULTIPOS)||(firstpos==NULL))){
								switch (CompareLiterals(ptr1,ptr2,&coefs[ii],&coefs[jj],kbset.opts.litord)){
									case NOMEMORY:
										if (clause->part2.bin->maxvarnb>=0){
											for (jj=0;jj<clause->part2.bin->literals;jj++){
												if (coefs[jj].eqterm1!=NULL){
													MYFREE(coefs[jj].eqterm1->alpha);
													MYFREE(coefs[jj].eqterm2->alpha);
												} else {
													MYFREE(coefs[jj].alpha);
												}
											}
										}
										MYFREE(coefs);
										return(NOMEMORY);
										break;
									case GREATER:
										ptr2[0]|=MAXIMUM;
										break;
									case LESSER:
										kk=0;
										break;
								}
							}
						}

						/* Select literal of first loop and equality root terms */
						/* if appropriate. */
						if (kk){
							ptr1[0]|=SELECTED;

							/* Literal is an equality or inequality. */
							if (ptr1[0]&EQUALITY){

								/* If there are not negative literals then this is a positive equality */
								/* eligible for paramodulating from it. Set flag PMSELECTED. */
								/* The condition SELECTED!=PMSELECTED forces the compiler to */
								/* ignore this code if SELECTED==PMSELECTED. */
								if ((SELECTED!=PMSELECTED)&&(firstneg==NULL)){
									ptr1[0]|=PMSELECTED;
								}

								/* Get pointers to equality root terms. */
								ptr2=NextItem(ptr1,IMMED);
								ptr3=NextItem(ptr2,OVERSUBTERMS);

								/* Calculate equality ordering data. */
								if (nn==0){
									if (NOMEMORY==SetEquData(clause->part2.bin,ptr1,ptr2,ptr3)){
										if (clause->part2.bin->maxvarnb>=0){
											for (jj=0;jj<clause->part2.bin->literals;jj++){
												if (coefs[jj].eqterm1!=NULL){
													MYFREE(coefs[jj].eqterm1->alpha);
													MYFREE(coefs[jj].eqterm2->alpha);
												} else {
													MYFREE(coefs[jj].alpha);
												}
											}
										}
										MYFREE(coefs);
										return(NOMEMORY);
									}
								}

								/* Set equality root terms as appropriate. */
								/* Release equality tree data if the root terms ordering */
								/* is completely defined. */
								switch (CompareEquTerms(clause->part2.bin,ptr2,ptr3,&((symbol *)&ptr1[1])->leftnode,
										&((symbol *)&ptr1[1])->rightnode,kbset.opts.termord)){
									case NOMEMORY:
										if (clause->part2.bin->maxvarnb>=0){
											for (jj=0;jj<clause->part2.bin->literals;jj++){
												if (coefs[jj].eqterm1!=NULL){
													MYFREE(coefs[jj].eqterm1->alpha);
													MYFREE(coefs[jj].eqterm2->alpha);
												} else {
													MYFREE(coefs[jj].alpha);
												}
											}
										}
										MYFREE(coefs);
										return(NOMEMORY);
										break;
									case GREATER:
										ptr2[0]|=(SELECTED|MAXIMUM);
										FreeEquNode(ptr2,((symbol *)&ptr1[1])->leftnode);
										FreeEquNode(ptr3,((symbol *)&ptr1[1])->rightnode);
										((symbol *)&ptr1[1])->leftnode=NULL;
										break;
									case LESSER:
										ptr3[0]|=(SELECTED|MAXIMUM);
										FreeEquNode(ptr2,((symbol *)&ptr1[1])->leftnode);
										FreeEquNode(ptr3,((symbol *)&ptr1[1])->rightnode);
										((symbol *)&ptr1[1])->leftnode=NULL;
										break;
									case EQUAL:
										FreeEquNode(ptr2,((symbol *)&ptr1[1])->leftnode);
										FreeEquNode(ptr3,((symbol *)&ptr1[1])->rightnode);
										((symbol *)&ptr1[1])->leftnode=NULL;
										break;
									case FAILURE:
										ptr2[0]|=SELECTED;
										ptr3[0]|=SELECTED;
										break;
								}
							}
						}
					}
				}
			}

			/* Free coefficients memory. */
			if (clause->part2.bin->maxvarnb>=0){
				for (jj=0;jj<clause->part2.bin->literals;jj++){
					if (coefs[jj].eqterm1!=NULL){
						MYFREE(coefs[jj].eqterm1->alpha);
						MYFREE(coefs[jj].eqterm2->alpha);
					} else {
						MYFREE(coefs[jj].alpha);
					}
				}
			}
			MYFREE(coefs);

		/* Only one literal can be selected. */
		} else {

			/* Clause has only one literal. */
			if (clause->part2.bin->literals==1){
				ptr4=&clause->part2.bin->formula[0];
				ptr4[0]|=SELECTED;

			/* Clause has more than one literal. */
			} else {

				/* Select first negative or positive literal as appropriate. */
				if (kbset.opts.select==SINGLENEG){
					firstneg[0]|=SELECTED;
					ptr4=firstneg;
				} else {
					firstpos[0]|=SELECTED;
					ptr4=firstpos;
				}
			}

			/* If literal is an equality select the appropriate root terms. */
			if (ptr4[0]&EQUALITY){

				/* If there are not negative literals then this is a positive equality */
				/* eligible for paramodulating from it. Set flag PMSELECTED. */
				/* The condition SELECTED!=PMSELECTED forces the compiler to */
				/* ignore this code if SELECTED==PMSELECTED. */
				if ((SELECTED!=PMSELECTED)&&(firstneg==NULL)){
					ptr4[0]|=PMSELECTED;
				}

				/* Get pointers to equality root terms. */
				ptr2=NextItem(ptr4,IMMED);
				ptr3=NextItem(ptr2,OVERSUBTERMS);

				/* Calculate equality ordering data. */
				if (clause->part2.bin->equflg==0){
					if (NOMEMORY==SetEquData(clause->part2.bin,ptr4,ptr2,ptr3)){
						return(NOMEMORY);
					}
				}

				/* Set equality root terms as appropriate. */
				/* Release equality tree data if the root terms ordering */
				/* is completely defined. */
				switch (CompareEquTerms(clause->part2.bin,ptr2,ptr3,&((symbol *)&ptr4[1])->leftnode,
						&((symbol *)&ptr4[1])->rightnode,kbset.opts.termord)){
					case NOMEMORY:
						return(NOMEMORY);
						break;
					case GREATER:
						ptr2[0]|=(SELECTED|MAXIMUM);
						FreeEquNode(ptr2,((symbol *)&ptr4[1])->leftnode);
						FreeEquNode(ptr3,((symbol *)&ptr4[1])->rightnode);
						((symbol *)&ptr4[1])->leftnode=NULL;
						break;
					case LESSER:
						ptr3[0]|=(SELECTED|MAXIMUM);
						FreeEquNode(ptr2,((symbol *)&ptr4[1])->leftnode);
						FreeEquNode(ptr3,((symbol *)&ptr4[1])->rightnode);
						((symbol *)&ptr4[1])->leftnode=NULL;
						break;
					case EQUAL:
						FreeEquNode(ptr2,((symbol *)&ptr4[1])->leftnode);
						FreeEquNode(ptr3,((symbol *)&ptr4[1])->rightnode);
						((symbol *)&ptr4[1])->leftnode=NULL;
						break;
					case FAILURE:
						ptr2[0]|=SELECTED;
						ptr3[0]|=SELECTED;
						break;
				}
			}
		}
	}

	return(0);
} /* SelectItems */
#else
/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Select items in a clause)  OCJ
 *
 *    This function selects the following items in a clause:
 *    - Appropriate literals in active clauses (see below).
 *    - Maximal equality root terms in selected equalities of active
 *      clauses.
 *    - Maximum equality root terms in single literal active and passive
 *      clauses are flagged as MAXIMUM.
 *
 *    Literals in active clauses are selected as follows:
 *    - If selection is set to MAXIMAL then all maximal literals
 *      are selected. Of course if there is a maximum literal it will be
 *      the only literal selected because it is the only maximal literal.
 *    - If selection is set to SINGLENEG then:
 *      - If there are negative literals then the first negative literal
 *        that is maximal with respect to the set of negative literals
 *        will be selected. Of course if there is a negative literal
 *        that is maximum with respect to the set of negative literals that
 *        will be selected as it is the only maximal negative literal.
 *      - Otherwise all maximal literals are selected.
 *    - If selection is set to MULTIENEG then:
 *      - If there are negative literals then the negative literals that
 *        are maximal with respect to the set of negative literals are
 *        selected and the other literals are not selected. Of course if
 *        there is a maximum negative literal it will be the only literal
 *        selected because it is the only maximal negative literal.
 *      - Otherwise all maximal literals are selected.
 *    - If selection is set to SINGLEPOS then:
 *      - If there are positive literals then the first positive literal
 *        that is maximal with respect to the set of positive literals
 *        will be selected. Of course if there is a positive literal
 *        that is maximum with respect to the set of positive literals that
 *        will be selected as it is the only maximal positive literal.
 *      - Otherwise all maximal literals are selected.
 *    - If selection is set to MULTIEPOS then:
 *      - If there are positive literals then the positive literals that
 *        are maximal with respect to the set of positive literals are
 *        selected and the other literals are not selected. Of course if
 *        there is a maximum positive literal it will be the only literal
 *        selected because it is the only maximal positive literal.
 *      - Otherwise all maximal literals are selected.
 *
 *    See "The design and implementation of VAMPIRE" section 2.6 by Alexandre
 *    Riazanov and Andrei Voronkov for details about this logic.
 *
 *    Also SELECTED positive equalities in clauses without negative literals
 *    are flagged as PMSELECTED to indicate that they are selected for
 *    paramodulating from them.
 *
 *    IMPORTANT: This function must never be called for main KB.
 *    The clause must belong to the working KB.
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in binary form.
 *    flag: If not zero then the literal selection loop will be
 *          executed not only for ACTIVE clauses but for any type
 *          of clauses.
 *
 *  RETURNS:
 *
 *    0 if successful, NOMEMORY if not enough memory.
 *
 *--------------------------------------------------------------*/
int32_t SelectItems(cmprefix *clause,int32_t flag){
	auto uint8_t *ptr1,*ptr2,*ptr3;                 /* Auxiliary pointers */
	auto uint8_t *lastselected;                     /* Pointer to last selected literal */
	auto coeff *coefs;                              /* Pointer to coefficient structures */
	auto int32_t eqs;                               /* Number of positive and negative equality literals. */
	auto int32_t existneg;                          /* Set to 1 if there are negative literals, 0 otherwise */
	auto int32_t existpos;                          /* Set to 1 if there are positive literals, 0 otherwise */
	auto uint8_t *demodequ;                         /* Pointer to equality suitable for demodulating from if */
	auto int32_t ii,jj,kk,nn;                       /* Auxiliary */

	/* Loop through clause literals. */
	for (ptr1=&clause->part2.bin->formula[0],eqs=existneg=existpos=0,demodequ=NULL;
			(*ptr1)!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Deselect literal and equality root terms. Count number of */
		/* equality literals. Equality root terms are deselected only */
		/* if clause has not been used for backward demodulation, as in */
		/* this case root terms are correctly selected. */
		ptr1[0]&=(~(SELECTED|PMSELECTED|MAXIMUM));
		if (ptr1[0]&EQUALITY){
			eqs++;
			if (clause->part2.bin->equflg!=2){
				ptr2=NextItem(ptr1,IMMED);
				ptr2[0]&=(~(SELECTED|MAXIMUM));
				ptr2=NextItem(ptr2,OVERSUBTERMS);
				ptr2[0]&=(~(SELECTED|MAXIMUM));
				if (clause->part2.bin->equflg==0){
					((symbol *)&ptr1[1])->leftnode=NULL;
				}
			}
		}

		/* Remember that there are negative literals. */
		if (ptr1[0]&NEGATED){
			existneg=1;
		} else {
			existpos=1;
		}
	}

	/* Check if clause is suitable for performing demodulation from it. */
	if ((clause->part2.bin->literals==1)&&(((EQUALITY|NEGATED)&clause->part2.bin->formula[0])==EQUALITY)){

		/* Remember position of equality suitable for demodulating from it. */
		demodequ=&clause->part2.bin->formula[0];

		/* If clause has been used for backward demodulation then equality */
		/* data has been set, so just select the equality literal if */
		/* appropriate and return. */
		if (clause->part2.bin->equflg==2){
			if (clause->flags&ACTIVE){
				demodequ[0]|=(SELECTED|PMSELECTED);
			}
			return(0);
		}
	} else {
		demodequ=NULL;
	}

	/* Clause is suitable for demodulation from it. */
	if (demodequ!=NULL){

		/* Get pointers to equality root terms. */
		ptr2=NextItem(demodequ,IMMED);
		ptr3=NextItem(ptr2,OVERSUBTERMS);

		/* Calculate equality ordering data and remember that */
		/* this was done for a potentially demodulating clause. */
		if (NOMEMORY==SetEquData(clause->part2.bin,demodequ,ptr2,ptr3)){
			return(NOMEMORY);
		}
		clause->part2.bin->equflg=2;

		/* Select maximum equality root terms. */
		/* Release equality tree data if the root terms ordering */
		/* is completely defined. */
		switch (CompareEquTerms(clause->part2.bin,ptr2,ptr3,&((symbol *)&demodequ[1])->leftnode,
				&((symbol *)&demodequ[1])->rightnode,kbset.opts.termord)){
			case NOMEMORY:
				return(NOMEMORY);
				break;
			case GREATER:
				ptr2[0]|=MAXIMUM|SELECTED;
				FreeEquNode(ptr2,((symbol *)&demodequ[1])->leftnode);
				FreeEquNode(ptr3,((symbol *)&demodequ[1])->rightnode);
				((symbol *)&demodequ[1])->leftnode=NULL;
				break;
			case LESSER:
				ptr3[0]|=MAXIMUM|SELECTED;
				FreeEquNode(ptr2,((symbol *)&demodequ[1])->leftnode);
				FreeEquNode(ptr3,((symbol *)&demodequ[1])->rightnode);
				((symbol *)&demodequ[1])->leftnode=NULL;
				break;
			case EQUAL:
				FreeEquNode(ptr2,((symbol *)&demodequ[1])->leftnode);
				FreeEquNode(ptr3,((symbol *)&demodequ[1])->rightnode);
				((symbol *)&demodequ[1])->leftnode=NULL;
				break;
			case FAILURE:
				ptr2[0]|=SELECTED;
				ptr3[0]|=SELECTED;
				break;
		}

		/* Select literal if clause is active. */
		if (clause->flags&ACTIVE){
			demodequ[0]|=(SELECTED|PMSELECTED);
		}

		/* Return. */
		return(0);
	}

	/* Active clause or flag is set. */
	if ((clause->flags&ACTIVE)||flag){

		/* Clause has more than one literal. */
		if (clause->part2.bin->literals>1){

			/* Allocate memory for coefficient structures. */
			if (NULL==(coefs=mallocmn((clause->part2.bin->literals+(2*eqs))*sizeof(coeff)))){
				return(NOMEMORY);
			}

			/* Initialize pointers to equality root term coefficients. */
			for (ptr1=&clause->part2.bin->formula[0],ii=clause->part2.bin->literals,jj=0;
					(*ptr1)!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS),jj++){
				if (ptr1[0]&EQUALITY){
					coefs[jj].eqterm1=&coefs[ii];
					coefs[jj].eqterm2=&coefs[ii+1];
					ii+=2;
				} else {
					coefs[jj].eqterm1=NULL;
				}
			}

			/* Allocate and initialize coefficients data. */
			for (ptr1=&clause->part2.bin->formula[0],ii=0;(*ptr1)!=UNITEND;
					ptr1=NextItem(ptr1,OVERSUBTERMS),ii++){

				/* Literal is an equality. Allocate and initialize */
				/* data for equality root terms. */
				if (ptr1[0]&EQUALITY){
					ptr2=NextItem(ptr1,IMMED);
					if (NOMEMORY==CalcCoeffs(ptr2,coefs[ii].eqterm1,clause->part2.bin->maxvarnb)){
						for (jj=0;jj<ii;jj++){
							if (clause->part2.bin->maxvarnb>=0){
								if (coefs[jj].eqterm1!=NULL){
									MYFREE(coefs[jj].eqterm1->alpha);
									MYFREE(coefs[jj].eqterm2->alpha);
								} else {
									MYFREE(coefs[jj].alpha);
								}
							}
						}
						MYFREE(coefs);
						return(NOMEMORY);
					}
					if (NOMEMORY==CalcCoeffs(NextItem(ptr2,OVERSUBTERMS),coefs[ii].eqterm2,
							clause->part2.bin->maxvarnb)){
						if (clause->part2.bin->maxvarnb>=0){
							for (jj=0;jj<ii;jj++){
								if (coefs[jj].eqterm1!=NULL){
									MYFREE(coefs[jj].eqterm1->alpha);
									MYFREE(coefs[jj].eqterm2->alpha);
								} else {
									MYFREE(coefs[jj].alpha);
								}
							}
							MYFREE(coefs[ii].eqterm1->alpha);
						}
						MYFREE(coefs);
						return(NOMEMORY);
					}

				/* Literal is a predicate. Allocate and initialize data for predicate. */
				} else {
					if (NOMEMORY==CalcCoeffs(ptr1,&coefs[ii],clause->part2.bin->maxvarnb)){
						if (clause->part2.bin->maxvarnb>=0){
							for (jj=0;jj<ii;jj++){
								if (coefs[jj].eqterm1!=NULL){
									MYFREE(coefs[jj].eqterm1->alpha);
									MYFREE(coefs[jj].eqterm2->alpha);
								} else {
									MYFREE(coefs[jj].alpha);
								}
							}
						}
						MYFREE(coefs);
						return(NOMEMORY);
					}
				}
			}

		/* Clause has only one literal. This is just to prevent compiler warning messages. */
		} else {
			coefs=NULL;
		}

		/* Loop through all literals. Flag MAXIMUM is temporarily */
		/* used here as an auxiliary flag to indicate that a literal */
		/* is lesser than a previous literal. */
		nn=clause->part2.bin->equflg;
		lastselected=NULL;
		for (ptr1=&clause->part2.bin->formula[0],ii=0;(*ptr1)!=UNITEND;
				ptr1=NextItem(ptr1,OVERSUBTERMS),ii++){

			/* Literal is a candidate for selection. */
			if (((ptr1[0]&NEGATED)||((MAXIMAL|SINGLEPOS|MULTIPOS)&kbset.opts.select)||(existneg==0))
					&&((0==(ptr1[0]&NEGATED))||((MAXIMAL|SINGLENEG|MULTINEG)&kbset.opts.select)||(existpos==0))){

				/* Literal flagged as being lesser than a previous one. Reset MAXIMUM flag. */
				if (ptr1[0]&MAXIMUM){
					ptr1[0]&=(~MAXIMUM);

				/* Literal is not flagged as being lesser than a previous one */
				/* and search for selectable literals must be continued. */
				} else if ((0==((SINGLENEG|SINGLEPOS)&kbset.opts.select))
						||((SINGLENEG&kbset.opts.select)&&(existneg==0))
						||((SINGLEPOS&kbset.opts.select)&&(existpos==0))||(lastselected==NULL)){

					/* Clause has only one literal. */
					if (clause->part2.bin->literals==1){
						kk=1;

					/* Clause has more than one literal. */
					} else {

						/* Loop through literals following current literal. */
						for (ptr2=NextItem(ptr1,OVERSUBTERMS),jj=ii+1,kk=1;((*ptr2)!=UNITEND)&&(kk);
								ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){

							/* Second loop literal is a candidate for selection. */
							if ((0==(ptr2[0]&MAXIMUM))
									&&((ptr2[0]&NEGATED)||((MAXIMAL|SINGLEPOS|MULTIPOS)&kbset.opts.select)||(existneg==0))
									&&((0==(ptr2[0]&NEGATED))||((MAXIMAL|SINGLENEG|MULTINEG)&kbset.opts.select)||(existpos==0))){
								switch (CompareLiterals(ptr1,ptr2,&coefs[ii],&coefs[jj],kbset.opts.litord)){
									case NOMEMORY:
										if (clause->part2.bin->maxvarnb>=0){
											for (jj=0;jj<clause->part2.bin->literals;jj++){
												if (coefs[jj].eqterm1!=NULL){
													MYFREE(coefs[jj].eqterm1->alpha);
													MYFREE(coefs[jj].eqterm2->alpha);
												} else {
													MYFREE(coefs[jj].alpha);
												}
											}
										}
										MYFREE(coefs);
										return(NOMEMORY);
										break;
									case GREATER:
										ptr2[0]|=MAXIMUM;
										break;
									case LESSER:
										kk=0;
										break;
								}
							}
						}
					}

					/* Literal is maximal. */
					if (kk){

						/* Select the literal. */
						ptr1[0]|=SELECTED;
						lastselected=ptr1;

						/* Literal is an equality or inequality. */
						if (ptr1[0]&EQUALITY){

							/* If there are not negative literals then this is a positive equality */
							/* eligible for paramodulating from it. Set flag PMSELECTED. */
							/* The condition SELECTED!=PMSELECTED forces the compiler to */
							/* ignore this code if SELECTED==PMSELECTED. */
							if ((SELECTED!=PMSELECTED)&&(existneg==0)){
								ptr1[0]|=PMSELECTED;
							}

							/* Get pointers to equality root terms. */
							ptr2=NextItem(ptr1,IMMED);
							ptr3=NextItem(ptr2,OVERSUBTERMS);

							/* Calculate equality ordering data. */
							if (nn==0){
								if (NOMEMORY==SetEquData(clause->part2.bin,ptr1,ptr2,ptr3)){
									if ((clause->part2.bin->literals>1)){
										if (clause->part2.bin->maxvarnb>=0){
											for (jj=0;jj<clause->part2.bin->literals;jj++){
												if (coefs[jj].eqterm1!=NULL){
													MYFREE(coefs[jj].eqterm1->alpha);
													MYFREE(coefs[jj].eqterm2->alpha);
												} else {
													MYFREE(coefs[jj].alpha);
												}
											}
										}
										MYFREE(coefs);
									}
									return(NOMEMORY);
								}
							}

							/* Set equality root terms as appropriate. */
							/* Release equality tree data if the root terms ordering */
							/* is completely defined. */
							switch (CompareEquTerms(clause->part2.bin,ptr2,ptr3,&((symbol *)&ptr1[1])->leftnode,
									&((symbol *)&ptr1[1])->rightnode,kbset.opts.termord)){
								case NOMEMORY:
									if ((clause->part2.bin->literals>1)){
										if (clause->part2.bin->maxvarnb>=0){
											for (jj=0;jj<clause->part2.bin->literals;jj++){
												if (coefs[jj].eqterm1!=NULL){
													MYFREE(coefs[jj].eqterm1->alpha);
													MYFREE(coefs[jj].eqterm2->alpha);
												} else {
													MYFREE(coefs[jj].alpha);
												}
											}
										}
										MYFREE(coefs);
									}
									return(NOMEMORY);
									break;
								case GREATER:
									ptr2[0]|=(SELECTED|MAXIMUM);
									FreeEquNode(ptr2,((symbol *)&ptr1[1])->leftnode);
									FreeEquNode(ptr3,((symbol *)&ptr1[1])->rightnode);
									((symbol *)&ptr1[1])->leftnode=NULL;
									break;
								case LESSER:
									ptr3[0]|=(SELECTED|MAXIMUM);
									FreeEquNode(ptr2,((symbol *)&ptr1[1])->leftnode);
									FreeEquNode(ptr3,((symbol *)&ptr1[1])->rightnode);
									((symbol *)&ptr1[1])->leftnode=NULL;
									break;
								case EQUAL:
									FreeEquNode(ptr2,((symbol *)&ptr1[1])->leftnode);
									FreeEquNode(ptr3,((symbol *)&ptr1[1])->rightnode);
									((symbol *)&ptr1[1])->leftnode=NULL;
									break;
								case FAILURE:
									ptr2[0]|=SELECTED;
									ptr3[0]|=SELECTED;
									break;
							}
						}
					}
				}
			}
		}

		/* Free coefficients memory. */
		if ((clause->part2.bin->literals>1)){
			if (clause->part2.bin->maxvarnb>=0){
				for (jj=0;jj<clause->part2.bin->literals;jj++){
					if (coefs[jj].eqterm1!=NULL){
						MYFREE(coefs[jj].eqterm1->alpha);
						MYFREE(coefs[jj].eqterm2->alpha);
					} else {
						MYFREE(coefs[jj].alpha);
					}
				}
			}
			MYFREE(coefs);
		}
	}

	return(0);
} /* SelectItems */
#endif

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check ordering constraints in equality root terms)  OCJ
 *
 *    This function checks ordering constraints in equality root terms.
 *    of a literal or term after applying a substitution to the terms.
 *
 *    It is assumed that term1 and term2 are root terms of the same equality,
 *    they has been compared with no substitution applied and have the following
 *    flags set:
 *      SELECTED if the term is not lesser nor equal than the other term.
 *      MAXIMUM if the term is greater than the other term.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of clause to which the
 *          terms belong.
 *    equality: Pointer to the equality literal in binp in which
 *              term1 and term2 are the root terms.
 *    term1: Equality root term that is checked for being greater than
 *           (MAXIMUM flag set) or not lesser (MAXIMUM flag no set) than
 *           term2.
 *    term2: The other equality root terms that is checked against term1.
 *    Subst: Pointer to substitution to be applied to term1 and term2
 *           before checking the ordering constraint.
 *    varshift: Shift for variable numbers belonging to a secondary
 *              merged clause. Variable numbers in substitution
 *              coming from secondary clause increased by varshift
 *              match variable numbers in the substitution.
 *    flag: MAXIMUM: If this flag is set then term1 is checked for being
 *          greater than term2. If this flag is not set then term1 is
 *          checked for being not lesser nor equal than term2.
 *          ITEM1SEC: If this flag is set then term1 and term2 belong to
 *          a secondary item with respect to the substitution. If this
 *          flag is not set then the terms belong to a primary item.
 *
 *  RETURNS:
 *
 *    1 if terms comply with the ordering constraint.
 *    0 if terms don't comply with the ordering constraint.
 *    NOMEMORY if there is not enough memory.
 *
 *--------------------------------------------------------------*/
int32_t CheckEqRootOrder(binprefix *binp,uint8_t *equality,uint8_t *term1,uint8_t *term2,subst *Subst,
		int32_t varshift,int32_t flag){
	auto equpnode *node1,*node2;                         /* Linear polynomial node pointers */
	auto int32_t mm;                                     /* Auxiliary */

	/* If term1 is greater than term2 before applying the substitution */
	/* then the ordering constraint is met. */
	if (term1[0]&MAXIMUM){
		return(1);
	}

	/* Return failure if both equality root terms are EQUAL. This is */
	/* detected because leftnode equpnode pointer is NULL. */
	if (((symbol *)&equality[1])->leftnode==NULL){
		return(0);
	}

	/* Substitution is empty. */
	if ((Subst->buffer==NULL)||(Subst->substsize==1)){
		if (0==(flag&MAXIMUM)){
			return(1);
		} else {
			return(0);
		}
	}

	/* Compare equality root terms with substition. */
	if (term1<term2){
		node1=((symbol *)&equality[1])->leftnode;
		node2=((symbol *)&equality[1])->rightnode;
	} else {
		node1=((symbol *)&equality[1])->rightnode;
		node2=((symbol *)&equality[1])->leftnode;
	}
	if (NOMEMORY==(mm=CompSubstEquTrms(binp,term1,term2,&node1,&node2,
			Subst,varshift,flag|kbset.opts.termord))){
		return(NOMEMORY);
	}

	/* Return as appropriate. */
	switch (mm){
		case NOMEMORY:
			return(NOMEMORY);
			break;
		case GREATER:
			return(1);
			break;
		case FAILURE:
			if (0==(flag&MAXIMUM)){
				return(1);
			}
			break;
	}

 	return(0);
 } /* CheckEqRootOrder */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Update item offsets)  OCJ
 *
 *    This function updates the nextoff and offset parameters of
 *    an item and all its subterms such that NextItem() function
 *    will work properly.
 *
 *    The maximum variable number in the binprefix structure is
 *    also updated for use by CompareItemWeight() function when
 *    called by CheckEqRootOrder().
 *
 *    This function is recursively called.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of item.
 *    item: Pointer to item.
 *
 *  RETURNS:
 *
 *    Pointer to next byte after item and all its subterms.
 *
 *
 *--------------------------------------------------------------*/
uint8_t *UpdateOffsets(binprefix *binp,uint8_t *item){
	auto uint8_t *ptr1;                                  /* Auxiliary pointer */
	auto int32_t ii,jj;                                  /* Auxiliary */

	/* Item is a variable. */
	if (item[0]&VARIABLE){
		ptr1=&item[3];
		if ((*(int16_t *)&item[1])>binp->maxvarnb){
			binp->maxvarnb=*(int16_t *)&item[1];
		}

	/* Item is not a variable. */
	} else {

		/* Compute symbol arity and set offset field. */
		((symbol *)&item[1])->offset=item-&binp->formula[0];
		if (item[0]&EQUALITY){
			ii=2;
		} else {
			ii=((symbol *)&item[1])->symbol->arity;
		}

		/* Loop through item subterms. */
		for (ptr1=NextItem(item,IMMED),jj=0;jj<ii;jj++){
			ptr1=UpdateOffsets(binp,ptr1);
		}

		/* Update symbol nextoff field. */
		((symbol *)&item[1])->nextoff=((symbol *)&item[1])->offset+(ptr1-item);
	}

	return(ptr1);
} /* UpdateOffsets */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Initialize equality for ordering constraints)  OCJ
 *
 *    This function initializes and sets the equality ordering data
 *    for the structures pointed by the leftnode and rightnode fields
 *    of symbol structure in equality and inequality literals and for
 *    the two child nodes of the root node. Before calling this function
 *    the caller function must ensure that all leftnode fields of
 *    equality symbol structures with no equality data are set to NULL
 *    if the affected clause has equflg==0.
 *
 *    The child nodes of equality data ordering trees are set by calling
 *    SetEquChildNode() function.
 *
 *    SetEquChildNode() and other functions called subsequently allocate
 *    storage for the equality ordering trees. This storage must be freed
 *    by calling FreeEquData() function from the appropriate code points.
 *
 *    All this stuff is for accelerating fast ordering constraint checks
 *    of equality root terms that are dependent on substitutions.
 *
 *    This function sets the equflg field in the binprefix structure to 1.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of the equality or inequality.
 *    equality: Pointer to equality literal.
 *    term1: Pointer to first (left) equality root term.
 *    term2: Pointer to second (right) equality root term.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t SetEquData(binprefix *binp,uint8_t *equality,uint8_t *term1,uint8_t *term2){

	/* Initialize and set child nodes data. Set equflg field. */
	if (NOMEMORY==SetEquChildNode(binp,term1,&((symbol *)&equality[1])->leftnode)){
		return(NOMEMORY);
	}
	if (NOMEMORY==SetEquChildNode(binp,term2,&((symbol *)&equality[1])->rightnode)){
		FreeEquNode(term1,((symbol *)&equality[1])->leftnode);
		return(NOMEMORY);
	}
	if (NOMEMORY==SetLftRhtPols(binp,&((symbol *)&equality[1])->leftnode->coefs,
			&((symbol *)&equality[1])->rightnode->coefs)){
		FreeEquNode(term1,((symbol *)&equality[1])->leftnode);
		FreeEquNode(term2,((symbol *)&equality[1])->rightnode);
		return(NOMEMORY);
	}
	binp->equflg=1;

	return(0);
} /* SetEquData */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Initialize equality tree node for ordering constraints)  OCJ
 *
 *    This function allocates storage for a equpnode tree equality node
 *    structure and initializes and sets its data.
 *
 *    This function allocate storage that must be freed by calling FreeEquData()
 *    function by other functions,
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of the equality or inequality.
 *    term: Pointer to equality (sub)term.
 *    pnode: Address of pointer to node structure.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t SetEquChildNode(binprefix *binp,uint8_t *term,equpnode **pnode){
	auto int32_t arity;                                  /* Term arity */
	auto int32_t ii;                                     /* Auxiliary */

	/* Allocate memory for child node. If the term is a VARIABLE */
	/* or a symbol of zero arity then the storage allocated is the */
	/* size of equpnode structure minus the size of the last element */
	/* of the structure. Otherwise the storage allocates is the size */
	/* of the structure plus the size of the last element of the */
	/* structure times (arity-1). */
	if (VARIABLE&term[0]){
		arity=0;
	} else {
		arity=((symbol *)&term[1])->symbol->arity;
	}
	if (NULL==((*pnode)=mallocmn(sizeof(equpnode)+((arity-1)*sizeof(void *))))){
		return(NOMEMORY);
	}

	/* Allocate and set the storage for the coefficients structure in compact form. */
	if (NOMEMORY==SetEqTrmPol(binp,term,&((*pnode)->coefs))){
		return(NOMEMORY);
	}

	/* Initialize results field to zero and subterms field pointers to NULL. */
	(*pnode)->result=0;
	for (ii=0;ii<arity;ii++){
		(*pnode)->subterms[ii]=NULL;
	}

	return(0);
} /* SetEquChildNode */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Set equality subterm compact polynomial)  OCJ
 *
 *    This function sets an equality or inequality term or subterm linear
 *    polynomial in compact form. The memory for the polynomial is
 *    allocated here. It must be freed by FreeEquData() function.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of the equality or inequality.
 *    term: Pointer to (sub)term of equality or inequality.
 *    pcoeff: Address of pointer to coeffcpt structure where the linear
 *            coefficients list in compact form will be placed.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t SetEqTrmPol(binprefix *binp,uint8_t *term,coeffcpt **pcoeff){
	auto coeff coefflng;                                 /* Coefficients of linear polynomials, long format */
	auto int32_t jj,kk;                                  /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */

	ptr10=Decompile(binp);
	MYFREE(ptr10);
	#endif

	/* Get coefficients of polynomial in long format. */
	if (NOMEMORY==CalcCoeffs(term,&coefflng,binp->maxvarnb)){
		return(NOMEMORY);
	}

	/* Allocate memory for coefficients in compact format. */
	if (NULL==((*pcoeff)=mallocmn((3+binp->maxvarnb)*sizeof(coeffcpt)))){
		MYFREE(coefflng.alpha);
		return(NOMEMORY);
	}

	/* Set coefficient zero (for symbols). */
	if (coefflng.alpha0>0){
		(*pcoeff)[0].coefnb=0;
		(*pcoeff)[0].coef=coefflng.alpha0;
		jj=1;
	} else {
		jj=0;
	}

	/* Set coefficients for variables. */
	for (kk=0;kk<=binp->maxvarnb;kk++){
		if (coefflng.alpha[kk]>0){
			(*pcoeff)[jj].coefnb=kk+1;
			(*pcoeff)[jj].coef=coefflng.alpha[kk];
			jj++;
		}
	}

	/* Mark the end of list, free the long format coefficients memory */
	/* and return. */
	(*pcoeff)[jj].coefnb=-1;
	MYFREE(coefflng.alpha);
	return(0);
} /* SetEqTrmPol */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Convert compact polynomials to left and right format)  OCJ
 *
 *    This function takes two compact form polynomials and converts them to
 *    left and right polynomials, also in compact form but without the redundant
 *    part. See "Implementing an efficient theorem prover" document section
 *    by Alexandre Riazanov for details.
 *
 *    The polynomials don't need to be the left and right equality root
 *    terms. They may be any couple of (sub)terms polynomials that must be
 *    compared.
 *
 *    Memory for polynomials is reallocated as needed.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of clause where (sub)terms belong.
 *    pcoefflft: Address of pointer to coeffcpt structure with the linear
 *               polynomial of "left" (sub)term. This serves both as input
 *               and output (see above).
 *    pcoeffrht: Address of pointer to coeffcpt structure with the linear
 *               polynomial of "right" (sub)term. This serves both as input
 *               and output (see above).
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t SetLftRhtPols(binprefix *binp,coeffcpt **pcoefflft,coeffcpt **pcoeffrht){
	auto coeffcpt *coefflft;                             /* Pointer to coefficients of left hand linear polynomial, compact format */
	auto coeffcpt *coeffrht;                             /* Pointer to coefficients of right hand linear polynomial, compact format */
	auto int32_t ii,jj,kk,mm;                            /* Auxiliary */

	/* Allocate memory for left hand and right hand format polynomials. */
	if (NULL==(coefflft=mallocmn((3+binp->maxvarnb)*sizeof(coeffcpt)))){
		return(NOMEMORY);
	}
	if (NULL==(coeffrht=mallocmn((3+binp->maxvarnb)*sizeof(coeffcpt)))){
		MYFREE(coefflft);
		return(NOMEMORY);
	}

	/* Convert compact linear polynomials to left hand and right hand format. */
	/* The "alpha" and "beta" in the following comments refer to the compact */
	/* input polynomials, not to the left hand or right hand polynomials. */
	ii=jj=kk=mm=0;
	do {

		/* Beta is zero for this alpha or there are no more alpha's. */
		if ((*pcoefflft)[ii].coefnb<(*pcoeffrht)[kk].coefnb){

			/* Beta is zero for this alpha. */
			if ((*pcoefflft)[ii].coefnb>=0){
				coefflft[jj].coefnb=(*pcoefflft)[ii].coefnb;
				coefflft[jj].coef=(*pcoefflft)[ii].coef;
				ii++;jj++;
				if ((*pcoefflft)[ii].coefnb<0){
					coefflft[jj].coefnb=-1;
					jj++;
				}

			/* There are no more alpha's. */
			} else {
				do {
					coeffrht[mm].coefnb=(*pcoeffrht)[kk].coefnb;
					coeffrht[mm].coef=(*pcoeffrht)[kk].coef;
					kk++;mm++;
				} while ((*pcoeffrht)[kk-1].coefnb>=0);
				kk--;
			}

		/* Neither alpha nor beta are zero. */
		} else if ((*pcoefflft)[ii].coefnb==(*pcoeffrht)[kk].coefnb){
			if ((*pcoefflft)[ii].coef>(*pcoeffrht)[kk].coef){
				coefflft[jj].coef=(*pcoefflft)[ii].coef-(*pcoeffrht)[kk].coef;
				coefflft[jj].coefnb=(*pcoefflft)[ii].coefnb;
				ii++;jj++;kk++;
				if ((*pcoefflft)[ii].coefnb<0){
					coefflft[jj].coefnb=-1;
					jj++;
				}
				if ((*pcoeffrht)[kk].coefnb<0){
					coeffrht[mm].coefnb=-1;
					mm++;
				}
			} else if ((*pcoefflft)[ii].coef<(*pcoeffrht)[kk].coef){
				coeffrht[mm].coef=(*pcoeffrht)[kk].coef-(*pcoefflft)[ii].coef;
				coeffrht[mm].coefnb=(*pcoeffrht)[kk].coefnb;
				ii++;kk++;mm++;
				if ((*pcoefflft)[ii].coefnb<0){
					coefflft[jj].coefnb=-1;
					jj++;
				}
				if ((*pcoeffrht)[kk].coefnb<0){
					coeffrht[mm].coefnb=-1;
					mm++;
				}
			} else {
				ii++;kk++;
				if ((*pcoefflft)[ii].coefnb<0){
					coefflft[jj].coefnb=-1;
					jj++;
				}
				if ((*pcoeffrht)[kk].coefnb<0){
					coeffrht[mm].coefnb=-1;
					mm++;
				}
			}

		/* Alpha is zero for this beta or there are no more beta's. */
		} else {

			/* Alpha is zero for this beta. */
			if ((*pcoeffrht)[kk].coefnb>=0){
				coeffrht[mm].coefnb=(*pcoeffrht)[kk].coefnb;
				coeffrht[mm].coef=(*pcoeffrht)[kk].coef;
				kk++;mm++;
				if ((*pcoeffrht)[kk].coefnb<0){
					coeffrht[mm].coefnb=-1;
					mm++;
				}

			/* There are no more beta's. */
			} else {
				do {
					coefflft[jj].coefnb=(*pcoefflft)[ii].coefnb;
					coefflft[jj].coef=(*pcoefflft)[ii].coef;
					ii++;jj++;
				} while ((*pcoefflft)[ii-1].coefnb>=0);
				ii--;
			}
		}
	} while (((*pcoefflft)[ii].coefnb>=0)||((*pcoeffrht)[kk].coefnb>=0));

	/* Replace old coefficients with new ones. */
	MYFREE(*pcoefflft);
	MYFREE(*pcoeffrht);
	(*pcoefflft)=coefflft;
	(*pcoeffrht)=coeffrht;

	return(0);
} /* SetLftRhtPols */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Frees equality ordering data of a compiled clause)  OCJ
 *
 *    This function frees equality ordering data storage allocated for
 *    a compiled clause.
 *
 *    The storage is freed only if the equflg field of clause binprefix
 *    structure is NOT zero and only for those equalities or inequalities
 *    in the clause that are SELECTED and have a NON NULL pointer in the
 *    leftnode field of the equality symbol structure. Otherwise it is
 *    assumed that there is no allocated storage to free.
 *
 *    In order to ensure that storage is properly allocated the following
 *    procedure must be followed:
 *    - Functions allocating binary clauses that will persist after the
 *      function returns must set the equflg field of clause binprefix
 *      to 0 if equality ordering data is not generated by calling
 *      SetEquData() function.
 *    - Equality ordering data is generated by calling SetEquData() function
 *      which sets equflg field of clause binprefix to 1.
 *    - The function SetEquData() is called from SelectItems() function.
 *      If the clause may be potentially used for demodulation from it
 *      then the equflg field of clause binprefix must be reset to 2 by
 *      SelectItems() after SetEquData() return.
 *    - The function SetEquData() is also called from BckDemodulation()
 *      function, in which case the equflg field of clause binprefix
 *      must be reset to 2 by BckDemodulation() after SetEquData()
 *      return.
 *    - Functions allocating binary clauses that will NOT persist after
 *      the function returns don't need to set the equflg field.
 *    - Functions freeing binary clause storage for clauses that were
 *      NOT created in the function must call FreeEquData() function
 *      to free equality data storage before freeing binary clause
 *      storage.
 *    - Functions freeing binary clause storage for clauses that were
 *      created in the function don't need to call FreeEquData() function.
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure of clause for which the
 *          equality data must be freed.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void FreeEquData(binprefix *binp){
	auto uint8_t *ptr1,*ptr2;                             /* Auxiliary pointers */

	/* The equflg flag is set. */
	if (binp->equflg){

		/* Loop through clause literals. */
		for (ptr1=&binp->formula[0];ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

			/* Literal is an equality or inequality with existing equality data */
			/* that must be freed. */
			if ((ptr1[0]&EQUALITY)&&(((symbol *)&ptr1[1])->leftnode!=NULL)){
				ptr2=NextItem(ptr1,IMMED);
				FreeEquNode(ptr2,((symbol *)&ptr1[1])->leftnode);
				FreeEquNode(NextItem(ptr2,OVERSUBTERMS),((symbol *)&ptr1[1])->rightnode);
				((symbol *)&ptr1[1])->leftnode=NULL;
			}
		}

		/* Reset binp->equflg because the clause will not necessarily */
		/* be deleted. */
		binp->equflg=0;
	}

	return;
} /* FreeEquData */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Frees equality ordering node)  OCJ
 *
 *    This function frees an equality ordering node storage allocated in
 *    the equality ordering tree structure. It is called recursively.
 *
 *
 *  ARGUMENTS:
 *
 *    term: Pointer to term owning the node.
 *    node: Pointer to equpnode structure of node to be freed.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void FreeEquNode(uint8_t *term,equpnode *node){

	/* Free node data. */
	FreeEquNodeData(term,node);

	/* Free node and return. */
	MYFREE(node);
	return;
} /* FreeEquNode */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Frees equality ordering node)  OCJ
 *
 *    This function frees the data allocated for an equality ordering
 *    node storage in the equality ordering tree structure. The storage
 *    allocated for the node structure itself is not freed.
 *
 *
 *  ARGUMENTS:
 *
 *    term: Pointer to term owning the node.
 *    node: Pointer to equpnode structure of node to be freed.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void FreeEquNodeData(uint8_t *term,equpnode *node){
	auto int arity;                                      /* Term arity */
	auto uint8_t *ptr1;                                  /* Auxiliary pointer */
	auto int ii;                                         /* Auxiliary */

	/* Term is not a variable. */
	if (0==(term[0]&VARIABLE)){

		/* Get arity of term and free child nodes. */
		arity=((symbol *)&term[1])->symbol->arity;
		for (ii=0,ptr1=NextItem(term,IMMED);ii<arity;ii++,ptr1=NextItem(ptr1,OVERSUBTERMS)){
			if (NULL!=node->subterms[ii]){
				FreeEquNode(ptr1,node->subterms[ii]);
				node->subterms[ii]=NULL;
			}
		}
	}

	/* Free linear polynomial. */
	if (node->coefs!=NULL){
		MYFREE(node->coefs);
		node->coefs=NULL;
	}
	return;
} /* FreeEquNodeData */
