/*
 ============================================================================
 Name        : parsing.c
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */
/****************************************************************
*
*                 parsing (parse program options and interactive commands)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

/*-------------------------------------------------------------
 *
 *  DESCRIPTION:
 *
 *    This module include routines necessary to parse program options
 *    and commands entered in interactivemode.
 *
 *    Drodi is a multi-platform Open Source Artificial Intelligence application.
 *    Current functionality is just a first order logic knowledge base with
 *    a saturation algorithms based mainly in ordered resolution, paramodulation
 *    and factoring.
 *
 *    See Drodi.c for additional details.
 *
 *
 *  IMPLEMENTATION:
 *
 *    Multi-platform
 *
 *  NOTES:
 *
 *
 *--------------------------------------------------------------*/

/* Include for all C sources */
#include "global.h"

/** Global variables for this module: only those that need initialization in their definition. */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Command processor)  OCJ
 *
 *    This function process the commands in interactive mode.
 *
 *    Commands are not case insensitive. However the command
 *    arguments may be case sensitive (for instance in the case
 *    of formulas or file names).
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void CommandProc(void){
	auto char *line;                                /* Input line buffer */
	auto int32_t size;                              /* Size of input line buffer */
	auto char *ptr1;                                /* Auxiliary pointer */
	auto int32_t ii,jj;                             /* Auxiliary */

	/* Initial memory allocation for input line. */
	printf("%s\n",VERSION);
	if (NULL==(line=mallocmn(INPUT_CHUNK_SIZE))){
		printf("No memory available, exiting program\n");
		return;
	}
	size=INPUT_CHUNK_SIZE;

	/* Read command loop. */
	while (1){

		/* Read line. */
		printf("Enter command: ");
		ii=0;
		do {
			fgets(&line[ii],size-ii,stdin);
			if (line[strlen(line)-1]!='\n'){
				ii=size-1;
				jj=1;
				size+=INPUT_CHUNK_SIZE;
				if (NULL==(ptr1=reallocmn(line,size))){
					printf("No memory available, exiting program\n");
					MYFREE(line);
					return;
				}
				line=ptr1;
			} else {
				jj=0;
			}
		} while (jj);
		line[strlen(line)-1]=0;

		/* Parse and execute command in line. */
		ii=ParseCommand(line,&ptr1);
		if (ptr1!=NULL){
			printf("%s\n",ptr1);
		}
		if (ii){
			break;
		}
	}

	/* Free memory, perform cleaning for exit and return. */
	MYFREE(line);
	return;
} /* CommandProc */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Parse program arguments)  OCJ
 *
 *    This function process the arguments passed to the program
 *    when executing it in non-interactive mode.
 *
 *    If more than one of the options -GOSAT, -PBMTYPE, -CLAUSIFY
 *    and -PROCLEARN is specified only the last one will be processed
 *    and the process will be done after processing all the other
 *    options. For other options they will be processed one after the
 *    other even if some of them are repeated.
 *
 *    Options are NOT case sensitive and must be preceded by a "-".
 *    If the option has arguments then they must be enclosed in
 *    parenthesis and separated by commas with no spaces in between
 *    except for program names enclosed in single quotes. For instance:
 *      -relevance(off)
 *      -proclearn('my file name',10)
 *
 *
 *  ARGUMENTS:
 *
 *    argc: Number of arguments passed, including the program filename.
 *    argv: Pointer to set of pointers to each argument. The first
 *          argument argument is the program filename. The last argument
 *          must be a valid problem filename or if -proclearn option
 *          is specified then it must be a valid learnfrom filename.
 *          The remaining arguments must be valid options. It is assumed
 *          than there are some arguments (that is,argc>1) but this is
 *          not checked
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void ParseArgs(int argc,char **argv){
	auto char *argline;                             /* Pointer to copy of a single argument */
	auto int32_t size;                              /* Size of argline buffer */
	auto int32_t exectype;                          /* Type of executing option */
	auto int32_t prclrn;                            /* Set to 1 if PROCLEARN option executed, otherwise 0 */
	auto int32_t help;                              /* Set to 1 if HELP option executed, otherwise 0 */
	auto char *ptr1;                                /* Auxiliary pointer */
	auto int32_t ii,jj,kk;                          /* Auxiliary */

	/* Initial memory allocation for input line. */
	printf("%% %s\n",VERSION);
	if (NULL==(argline=mallocmn(INPUT_CHUNK_SIZE))){
		printf("No memory available, exiting program\n");
		return;
	}
	size=INPUT_CHUNK_SIZE;

	/* Loop through program parameters. */
	exectype=prclrn=help=0;
	for (ii=1;ii<argc;ii++){

		/* Option not preceded by a "-" character. Leave the loop if last option */
		/* as it could be the target filename, return otherwise. */
		if (*argv[ii]!='-'){
			if (ii<(argc-1)){
				printf("Invalid option %s, exiting program\n",argv[ii]);
				MYFREE(argline);
				return;
			} else {
				break;
			}
		}

		/* Copy option to argline not including the initial dash. */
		jj=strlen(argv[ii]);
		if (jj>size){
			size+=jj;
			if (NULL==(ptr1=reallocmn(argline,size))){
				printf("No memory available, exiting program\n");
				MYFREE(argline);
				return;
			}
			argline=ptr1;
		}
		strcpy(argline,&argv[ii][1]);

		/* First option validity check and conversion to interactive command format. */
		/* Last parameter may be a valid filename if not a valid option. */
		if (ChkConvertOption(argline)){
			if (ii<(argc-1)){
				printf("Invalid option %s, exiting program\n",argv[ii]);
				MYFREE(argline);
				return;
			} else {
				break;
			}
		}

		/* Convert first token to upper case. */
		for (jj=0;(argline[jj]!=0)&&(argline[jj]!=' ')&&(argline[jj]!='\t');jj++){
			argline[jj]=toupper(argline[jj]);
		}

		/* Process option.  */
		/* Action options. Remember the last one. Last parameter may be */
		/* a valid filename if not a valid option. */
		kk=0;
		if ((0==strncmp("GOSAT",argline,jj))&&(jj==5)){
			if (argline[jj]!=0){
				if (ii<(argc-1)){
					printf("Invalid option %s, exiting program\n",argv[ii]);
					MYFREE(argline);
					return;
				} else {
					kk=1;
				}
			} else {
				exectype=1;
			}
		} else if ((0==strncmp("PBMTYPE",argline,jj))&&(jj==7)){
			if (argline[jj]!=0){
				if (ii<(argc-1)){
					printf("Invalid option %s, exiting program\n",argv[ii]);
					MYFREE(argline);
					return;
				} else {
					kk=1;
				}
			} else {
				exectype=2;
			}
		} else if ((0==strncmp("CLAUSIFY",argline,jj))&&(jj==5)){
			if (argline[jj]!=0){
				if (ii<(argc-1)){
					printf("Invalid option %s, exiting program\n",argv[ii]);
					MYFREE(argline);
					return;
				} else {
					kk=1;
				}
			} else {
				exectype=3;
			}

		/* Non action option. Pass it to ParseCommand(). Last parameter */
		/* may be a valid filename if not a valid option. */
		} else {
			if ((0==strncmp("PROCLEARN",argline,jj))&&(jj==5)){
				prclrn=1;
			} else if (0==strncmp("HELP",argline,jj)){
				help=1;
			}
			kk=ParseCommand(argline,&ptr1);
			if (kk){
				if (ii<(argc-1)){
					if (ptr1!=NULL){
						printf("%s in %s\n",ptr1,argv[ii]);
					}
					MYFREE(argline);
					return;
				}
			} else if (help){
				MYFREE(argline);
				return;
			}
		}

		/* If last parameter has been correctly processed and */
		/* there is no PROCLEARN option then there is nothing */
		/* to do. Report status and return. */
		if (ii==(argc-1)){
			if ((prclrn==0)&&(kk==0)){
				printf("No filename specified, nothing to do\n");
				MYFREE(argline);
				return;
			} else {
				if (prclrn==0){
					if (ptr1!=NULL){
						printf("%s in %s\n",ptr1,argv[ii]);
					}
					MYFREE(argline);
					return;
				}
				exectype=4;
			}
		}
	}

	/* Perform the requested action and return. */
	if (exectype!=4){
		syntaxmode=TPTP;
		importfile=argv[argc-1];
		if (0!=Import(importfile,NULL,0,1)){
			Initialize_KB(&mainkb);
			ResetStats(pbmstats);
		} else {
			switch (exectype){
				case 0:
					Saturate();
					break;
				case 1:
					GoSat();
					break;
				case 2:
					GoType();
					break;
				case 3:
					Clausify();
					break;
			}
		}
	}
	MYFREE(argline);
	return;
} /* ParseArgs */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check and convert option to interactive command)  OCJ
 *
 *    This function makes an initial validity check to an option included
 *    in program invocation and converts it to an interactive command.
 *
 *
 *  ARGUMENTS:
 *
 *    argline: Pointer to option as entered included in program invocation
 *             on entry, conversion result on exit.
 *
 *  RETURNS:
 *
 *    0 if OK or 1 if option not valid.
 *
 *--------------------------------------------------------------*/
int32_t ChkConvertOption(char *argline){
	auto int32_t jj,kk,mm;                          /* Auxiliary */


	/* Loop through option characters. */
	for (jj=0,kk=mm=0;argline[jj]!=0;jj++){
		if (argline[jj]=='\''){
			kk^=1;
		} else if (kk==0){
			switch (argline[jj]){
				case '(':
					if (mm==0){
						if ((argline[strlen(argline)-1]!=')')||(argline[jj+1]==')')||(argline[jj+1]==',')){
							return(1);
						}
						mm=1;
						argline[jj]=' ';
					} else {
						return(1);
					}
					break;
				case ')':
					if ((mm==0)||(argline[jj+1]!=0)){
						return(1);
					}
					argline[jj]=0;
					break;
				case ',':
					if ((argline[jj+1]==')')||(argline[jj+1]==',')){
						return(1);
					}
					argline[jj]=' ';
					break;
				case ' ':
					return(1);
					break;
			}
		}
	}
	if (kk){
		return(1);
	}
	return(0);
} /* ChkConvertOption */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Command parser)  OCJ
 *
 *    This function parses the commands entered in interactive
 *    mode or built from options passed as arguments at program
 *    invocation.
 *
 *
 *  ARGUMENTS:
 *
 *    command: Pointer to string with the command.
 *    msg: Address of pointer where an output message will be
 *         stored. It is the calles responsibility to display
 *         this message when appropriate.
 *
 *  RETURNS:
 *
 *    0 if OK or 1 if program must terminate due to an error
 *    or an user request.
 *
 *--------------------------------------------------------------*/
int32_t ParseCommand(char *command,char **msg){
	auto clock_t start,end;                         /* For CPU time measurements */
	auto char *ltkn;                                /* Pointer to token */
	auto int32_t ratios[WEIGHTQUEUES];              /* Temporary weight queue ratios for SELECTRATIO command */
	auto char *ptr1,*ptr2,*ptr3,*ptr4;              /* Auxiliary pointers */
	auto int32_t ii,jj,nn,cc;                       /* Auxiliary */
	auto uint64_t kk;                               /* Auxiliary */
	auto float ff;                                  /* Auxiliary */

	/* Check number of occurrences of ':' char for SELECTRATIO parsing. */
	*msg=NULL;
	for (ptr1=command-1,cc=-1;ptr1!=NULL;ptr1=strchr(ptr1+1,':'),cc++){
	}

	/* Parse first token and convert to upper case. */
	if (NULL==(ltkn=strtok(command," \t"))){
		return(0);
	}
	for (ii=0;ii<strlen(ltkn);ii++){
		ltkn[ii]=toupper(ltkn[ii]);
	}

	/* EXIT command. */
	if (0==strcmp(ltkn,"EXIT")){
		if (pgmmode==NONINTERACTIVE){
			*msg="Invalid option";
			return(1);
		}
		if (NULL==(ltkn=strtok(NULL," \t"))){
			if (CheckAnswer("Exit program? [y/N]: ","")){
				return(1);
			}
		} else {
			printf("Invalid EXIT option: %s\n",ltkn);
			printf("This command has no options\n");
		}

	/* HELP command. */
	} else if (0==strcmp(ltkn,"HELP")){
		Help(ltkn);
		if (pgmmode==NONINTERACTIVE){
			return(1);
		}

	/* NEW command. */
	} else if (0==strcmp(ltkn,"NEW")){
		if (pgmmode==NONINTERACTIVE){
			*msg="Invalid option";
			return(1);
		}
		if (NULL==(ltkn=strtok(NULL," \t"))){
			if (CheckAnswer("Initialize all data? [y/N]: ","KB not initialized")){
				if (NOMEMORY==Initialize_KB(&mainkb)){
					printf("No memory available, exiting program\n");
					return(1);
				}
				ResetStats(pbmstats);
				printf("Knowledge bases have been initialized\n");
			}
		} else {
			printf("Invalid NEW option: %s\n",ltkn);
			printf("This command has no options\n");
		}

	/* TELL command. */
	} else if (0==strcmp(ltkn,"TELL")){
		if (pgmmode==NONINTERACTIVE){
			*msg="Invalid option";
			return(1);
		}
		ltkn=strtok(NULL,"");
		gettimeofday(&mainkb.time,NULL);
		start=clock();
		ii=Tell(ltkn,AXIOM,FROMMAINFILE);
		gettimeofday(&mainkb.endtime,NULL);
		mainkb.prstats.elapsed_time+=(mainkb.endtime.tv_sec-mainkb.time.tv_sec
				+(mainkb.endtime.tv_usec-mainkb.time.tv_usec)/1000000.0);
		end=clock();
		cpu_time+=(((double)(end-start))/CLOCKS_PER_SEC);
		switch (ii){
			case NOMEMORY:
				return(1);
				break;
			case TIMEOUT:
				if (NOMEMORY==Initialize_KB(&mainkb)){
					printf("No memory available, exiting program\n");
					return(1);
				}
				ResetStats(pbmstats);
				printf("Knowledge bases have been initialized\n");
				break;
		}

	/* ASK command. */
	} else if (0==strcmp(ltkn,"ASK")){
		if (pgmmode==NONINTERACTIVE){
			*msg="Invalid option";
			return(1);
		}
		ltkn=strtok(NULL,"");
		gettimeofday(&mainkb.time,NULL);
		start=clock();
		ii=Ask(ltkn);
		gettimeofday(&mainkb.endtime,NULL);
		mainkb.prstats.elapsed_time+=(mainkb.endtime.tv_sec-mainkb.time.tv_sec
				+(mainkb.endtime.tv_usec-mainkb.time.tv_usec)/1000000.0);
		end=clock();
		cpu_time+=(((double)(end-start))/CLOCKS_PER_SEC);
		switch (ii){
			case NOMEMORY:
				return(1);
				break;
			case TIMEOUT:
				if (NOMEMORY==Initialize_KB(&mainkb)){
					printf("No memory available, exiting program\n");
					return(1);
				}
				ResetStats(pbmstats);
				printf("Knowledge bases have been initialized\n");
				break;
		}

	/* GO command. */
	} else if (0==strcmp(ltkn,"GO")){
		if (pgmmode==NONINTERACTIVE){
			*msg="Invalid option";
			return(1);
		}
		if (NULL==(ltkn=strtok(NULL," \t"))){
			Saturate();
		} else {
			printf("Invalid GO option: %s\n",ltkn);
			printf("This command has no options\n");
		}

	/* GOSAT command. */
	} else if (0==strcmp(ltkn,"GOSAT")){
		if (NULL!=(ltkn=strtok(NULL,""))){
			if (NULL==(ptr1=ParseFilename(ltkn,&ptr2))){
				*msg="Invalid filename";
				return(pgmmode==INTERACTIVE?0:1);
			}
			ii=syntaxmode;
			syntaxmode=TPTP;
			importfile=ptr1;
			if (0==Import(ptr1,NULL,0,1)){
				GoSat();
			} else {
				if (glblstats.memory<allcdmemory){
					glblstats.memory=allcdmemory;
				}
				tot_cpu_time+=cpu_time;
				tot_pbms++;
			}
			syntaxmode=ii;
			importfile=NULL;
		} else {
			GoSat();
		}
		if (NOMEMORY==Initialize_KB(&mainkb)){
			printf("No memory available, exiting program\n");
			return(1);
		}
		ResetStats(pbmstats);
		printf("Knowledge bases have been initialized\n");

	/* PBMTYPE command. */
	} else if (0==strcmp(ltkn,"PBMTYPE")){
		if (NULL!=(ltkn=strtok(NULL,""))){
			if (NULL==(ptr1=ParseFilename(ltkn,&ptr2))){
				*msg="Invalid filename";
				return(pgmmode==INTERACTIVE?0:1);
			}
			printf("Reading file %s...\n",ptr1);
			ii=syntaxmode;
			syntaxmode=TPTP;
			if (Import(ptr1,NULL,0,1)){
				printf("      UNIT_EQUALITY UNIT_CLAUSE HORN GROUND NONUNIT_EQU PBM_TYPE\n");
				printf("====>      ?             ?       ?      ?         ?        ?\n");
				if (glblstats.memory<allcdmemory){
					glblstats.memory=allcdmemory;
				}
				tot_cpu_time+=cpu_time;
				if (NOMEMORY==Initialize_KB(&mainkb)){
					printf("No memory available, exiting program\n");
					return(1);
				}
				ResetStats(pbmstats);
				printf("Knowledge bases have been initialized\n");
				syntaxmode=ii;
				tot_pbms++;
				return(0);
			}
			syntaxmode=ii;
		} else if (pgmmode==NONINTERACTIVE){
			*msg="Missing PBMTYPE parameter";
			return(1);
		}
		GoType();

	/* CLAUSIFY command. */
	} else if (0==strcmp(ltkn,"CLAUSIFY")){
		if (NULL!=(ltkn=strtok(NULL,""))){
			if (NULL==(ptr1=ParseFilename(ltkn,&ptr2))){
				*msg="Invalid filename";
				return(pgmmode==INTERACTIVE?0:1);
			}
			printf("Converting formulas in file %s to CNF format...\n",ptr1);
			ii=syntaxmode;
			syntaxmode=TPTP;
			importfile=ptr1;
			if (0==Import(ptr1,NULL,0,1)){
				Clausify();
			} else {
				if (glblstats.memory<allcdmemory){
					glblstats.memory=allcdmemory;
				}
				tot_cpu_time+=cpu_time;
				tot_pbms++;
			}
			syntaxmode=ii;
			importfile=NULL;
		} else {
			printf("Converting formulas to CNF format...\n");
			Clausify();
		}
		if (NOMEMORY==Initialize_KB(&mainkb)){
			printf("No memory available, exiting program\n");
			return(1);
		}
		ResetStats(pbmstats);
		printf("Knowledge bases have been initialized\n");

	/* TERMORDER command. */
	} else if (0==strcmp(ltkn,"TERMORDER")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"STANDARD")){
				glblopts.termord=STANDARD;
				usrparam&=~TERMORDRMASK;
				usrparam|=TRMORDSTD_P;
				usrparam_mask|=TERMORDRMASK;
				*msg="Term ordering has been set to STANDARD";
			} else if (0==strcmp(ltkn,"NONRECURSIVE")){
				glblopts.termord=NONRECURSIVE;
				usrparam&=~TERMORDRMASK;
				usrparam|=TRMORDNR_P;
				usrparam_mask|=TERMORDRMASK;
				*msg="Term ordering has been set to NONRECURSIVE";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.termord=STANDARD;
				usrparam&=~TERMORDRMASK;
				usrparam_mask&=~TERMORDRMASK;
				*msg="Term ordering has been set to DEFAULT";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing TERMORDER parameter";
					return(1);
				}
				if (0==(usrparam_mask&TERMORDRMASK)){
					*msg="Term ordering is set to DEFAULT";
				} else {
					if (glblopts.termord==STANDARD){
						*msg="Term ordering is STANDARD";
					} else {
						*msg="Term ordering is NON RECURSIVE";
					}
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid TERMORDER option, valid options are STANDARD, NONRECURSIVE or DEFAULT";
					return(1);
				}
				printf("Invalid TERMORDER option: %s\n",ltkn);
				printf("Valid options are STANDARD, NONRECURSIVE or DEFAULT\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing TERMORDER parameter";
				return(1);
			}
			if (0==(usrparam_mask&TERMORDRMASK)){
				printf("Term ordering is set to DEFAULT\n");
			} else {
				if (glblopts.termord==STANDARD){
					printf("Term ordering is STANDARD\n");
				} else {
					printf("Term ordering is NON RECURSIVE\n");
				}
			}
		}

	/* LITORDER command. */
	} else if (0==strcmp(ltkn,"LITORDER")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"STANDARD")){
				glblopts.litord=STANDARD;
				usrparam&=~LITORDRMASK;
				usrparam|=LITORDSTD_P;
				usrparam_mask|=LITORDRMASK;
				*msg="Literal ordering has been set to STANDARD";
			} else if (0==strcmp(ltkn,"NONRECURSIVE")){
				glblopts.litord=NONRECURSIVE;
				usrparam&=~LITORDRMASK;
				usrparam|=LITORDNR_P;
				usrparam_mask|=LITORDRMASK;
				*msg="Literal ordering has been set to NONRECURSIVE";
			} else if (0==strcmp(ltkn,"LEXICOGRAPHIC")){
				glblopts.litord=LEXICOGRAPHIC;
				usrparam&=~LITORDRMASK;
				usrparam|=LITORDLEX_P;
				usrparam_mask|=LITORDRMASK;
				*msg="Literal ordering has been set to LEXICOGRAPHIC";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.litord=STANDARD;
				usrparam&=~LITORDRMASK;
				usrparam_mask&=~LITORDRMASK;
				*msg="Literal ordering has been set to DEFAULT";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing LITORDER parameter";
					return(1);
				}
				if (0==(usrparam_mask&LITORDRMASK)){
					printf("Literal ordering is set to DEFAULT\n");
				} else {
					switch (glblopts.litord){
						case STANDARD:
							printf("Literal ordering is STANDARD\n");
							break;
						case NONRECURSIVE:
							printf("Literal ordering is NON RECURSIVE\n");
							break;
						case LEXICOGRAPHIC:
							printf("Literal ordering is LEXICOGRAPHIC\n");
							break;
					}
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid LITORDER option, valid options are STANDARD, NONRECURSIVE, LEXICOGRAPHIC or DEFAULT";
					return(1);
				}
				printf("Invalid LITORDER option: %s\n",ltkn);
				printf("Valid options are STANDARD, NONRECURSIVE, LEXICOGRAPHIC or DEFAULT\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing LITORDER parameter";
				return(1);
			}
			if (0==(usrparam_mask&LITORDRMASK)){
				printf("Literal ordering is set to DEFAULT\n");
			} else {
				switch (glblopts.litord){
					case STANDARD:
						printf("Literal ordering is STANDARD\n");
						break;
					case NONRECURSIVE:
						printf("Literal ordering is NON RECURSIVE\n");
						break;
					case LEXICOGRAPHIC:
						printf("Literal ordering is LEXICOGRAPHIC\n");
						break;
				}
			}
		}

	/* FUNCTIONW command. */
	} else if (0==strcmp(ltkn,"FUNCTIONW")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"UNIFORM")){
				glblopts.fweight=UNIFORM;
				usrparam&=~FUNCWMASK;
				usrparam|=UNIFORM_P;
				usrparam_mask|=FUNCWMASK;
				*msg="Function weight has been set to UNIFORM";
			} else if (0==strcmp(ltkn,"ARITY")){
				glblopts.fweight=ARITY;
				usrparam&=~FUNCWMASK;
				usrparam|=ARITY_P;
				usrparam_mask|=FUNCWMASK;
				*msg="Function weight has been set to ARITY";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.fweight=ARITY;
				usrparam&=~FUNCWMASK;
				usrparam_mask&=~FUNCWMASK;
				*msg="Function weight has been set to DEFAULT";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing FUNCTIONW parameter";
					return(1);
				}
				if (0==(usrparam_mask&FUNCWMASK)){
					printf("Function weight is set to DEFAULT\n");;
				} else {
					if (glblopts.fweight==UNIFORM){
						printf("Function weight UNIFORM\n");
					} else {
						printf("Function weight by ARITY\n");
					}
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid FUNCTIONW parameter, valid options are UNIFORM, ARITY or DEFAULT";
					return(1);
				}
				printf("Invalid FUNCTIONW parameter: %s\n",ltkn);
				printf("Valid options are UNIFORM, ARITY or DEFAULT\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing FUNCTIONW parameter";
				return(1);
			}
			if (0==(usrparam_mask&FUNCWMASK)){
				printf("Function weight is set to DEFAULT\n");;
			} else {
				if (glblopts.fweight==UNIFORM){
					printf("Function weight UNIFORM\n");
				} else {
					printf("Function weight by ARITY\n");
				}
			}
		}

	/* SELECTRATIO command. */
	} else if (0==strcmp(ltkn,"SELECTRATIO")){
		if (cc==4){
			for (ii=1;ii<WEIGHTQUEUES;ii++){
				if (NULL!=(ltkn=strtok(NULL,": \t"))){
					jj=strtol(ltkn,&ptr1,10);
					if ((jj>=0)&&(ptr1[0]==0)){
						ratios[ii]=jj;
					} else {
						break;
					}
				} else {
					break;
				}
			}
			if ((ii==WEIGHTQUEUES)&&(NULL!=(ltkn=strtok(NULL,": \t")))){
				jj=strtol(ltkn,&ptr1,10);
				if ((jj>0)&&(ptr1[0]==0)){
					for (ii=1;ii<WEIGHTQUEUES;ii++){
						queuerts[ii]=ratios[ii];
					}
					queuerts[WEIGHTQUEUES]=jj;
					if (NOMEMORY==SetClSelectOrder()){
						*msg="No memory available, exiting program";
						return(1);
					}
					if (pgmmode==INTERACTIVE){
						printf("SELECTRATIO value has been set to ");
						for (ii=1;ii<WEIGHTQUEUES;ii++){
							printf("%d:",queuerts[ii]);
						}
						printf("%d\n",queuerts[WEIGHTQUEUES]);
					}
				} else {
					ii=0;
				}
			} else {
				ii=0;
			}
			if (ii==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="SELECTRATIO parameter missing or invalid.";
					return(1);
				}
				printf("SELECTRATIO parameter missing or invalid.\n");
				printf("Parameter must be a set of five positive integers of the form:\n");
				printf("  weight1:weight2:weight3:weight4:age\n");
				printf("The fifth integer, corresponding to the age queue, must be strictly positive.\n");
			}
		} else if ((cc==0)&&(NULL==strtok(NULL," \t"))){
			if (pgmmode==NONINTERACTIVE){
				*msg="SELECTRATIO parameter missing or invalid.";
				return(1);
			}
			printf("Current SELECTRATIO value is ");
			for (ii=1;ii<WEIGHTQUEUES;ii++){
				printf("%d:",queuerts[ii]);
			}
			printf("%d\n",queuerts[WEIGHTQUEUES]);
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="SELECTRATIO parameter missing or invalid.";
				return(1);
			}
			printf("SELECTRATIO parameter missing or invalid.\n");
			printf("Parameter must be a set of five positive integers of the form:\n");
			printf("  weight1:weight2:weight3:weight4:age\n");
			printf("The fifth integer, corresponding to the age queue, must be strictly positive.\n");
		}

	/* SELECTRATIO2 command. */
	} else if (0==strcmp(ltkn,"SELECTRATIO2")){
		if (cc==5){
			for (ii=0;ii<WEIGHTQUEUES;ii++){
				if (NULL!=(ltkn=strtok(NULL,": \t"))){
					jj=strtol(ltkn,&ptr1,10);
					if ((jj>=0)&&(ptr1[0]==0)){
						ratios[ii]=jj;
					} else {
						break;
					}
				} else {
					break;
				}
			}
			if ((ii==WEIGHTQUEUES)&&(NULL!=(ltkn=strtok(NULL,": \t")))){
				jj=strtol(ltkn,&ptr1,10);
				if ((jj>0)&&(ptr1[0]==0)){
					for (ii=0;ii<WEIGHTQUEUES;ii++){
						queuerts2[ii]=ratios[ii];
					}
					queuerts2[WEIGHTQUEUES]=jj;
					if (NOMEMORY==SetClSelectOrder()){
						*msg="No memory available, exiting program";
						return(1);
					}
					if (pgmmode==INTERACTIVE){
						printf("SELECTRATIO2 value has been set to ");
						for (ii=0;ii<WEIGHTQUEUES;ii++){
							printf("%d:",queuerts2[ii]);
						}
						printf("%d\n",queuerts2[WEIGHTQUEUES]);
					}
				} else {
					ii=0;
				}
			} else {
				ii=0;
			}
			if (ii==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="SELECTRATIO2 parameter missing or invalid.";
					return(1);
				}
				printf("SELECTRATIO2 parameter missing or invalid.\n");
				printf("Parameter must be a set of six positive integers of the form:\n");
				printf("  weight1:weight2:weight3:weight4:weight5:age\n");
				printf("The sixth integer, corresponding to the age queue, must be strictly positive.\n");
			}
		} else if ((cc==0)&&(NULL==strtok(NULL," \t"))){
			if (pgmmode==NONINTERACTIVE){
				*msg="SELECTRATIO2 parameter missing or invalid.";
				return(1);
			}
			printf("Current SELECTRATIO2 value is ");
			for (ii=0;ii<WEIGHTQUEUES;ii++){
				printf("%d:",queuerts2[ii]);
			}
			printf("%d\n",queuerts2[WEIGHTQUEUES]);
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="SELECTRATIO2 parameter missing or invalid.";
				return(1);
			}
			printf("SELECTRATIO2 parameter missing or invalid.\n");
			printf("Parameter must be a set of six positive integers of the form:\n");
			printf("  weight1:weight2:weight3:weight4:weight5:age\n");
			printf("The sixth integer, corresponding to the age queue, must be strictly positive.\n");
		}

	/* TIMEOUT command. */
	} else if (0==strcmp(ltkn,"TIMEOUT")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			ii=strtol(ltkn,&ptr1,10);
			if ((ii>0)&&(ptr1[0]==0)){
				glblopts.timeout=ii;
				if (pgmmode==INTERACTIVE){
					printf("Timeout value has been set to %.1f seconds\n",glblopts.timeout);
				}
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing timeout parameter";
					return(1);
				}
				printf("Current timeout value is %.1f seconds\n",glblopts.timeout);
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid timeout parameter";
					return(1);
				}
				printf("Invalid timeout parameter: %s\n",ltkn);
				printf("Parameter must be the timeout in seconds given as a strictly positive integer.\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing timeout parameter";
				return(1);
			}
			printf("Current timeout value is %.1f\n",glblopts.timeout);
		}

	/* MAXMEMORY command. */
	} else if (0==strcmp(ltkn,"MAXMEMORY")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			ii=strtoll(ltkn,&ptr1,10);
			kk=ii*1000000L;
			if ((kk>0)&&(ptr1[0]==0)){
				memorylimit=kk;
				if (pgmmode==INTERACTIVE){
					printf("MAXMEMORY limit has been set to %d MB\n",ii);
				}
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing MAXMEMORY parameter";
					return(1);
				}
				printf("Current MAXMEMORY value is %lu MB\n",memorylimit/1000000);
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid MAXMEMORY parameter";
					return(1);
				}
				printf("Invalid MAXMEMORY parameter: %s\n",ltkn);
				printf("Parameter must be the memory limit in megabytes given as a strictly positive integer.\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing MAXMEMORY parameter";
				return(1);
			}
			printf("Current MAXMEMORY value is %lu MB\n",memorylimit/1000000);
		}

	/* RELEVANCE command. */
	} else if (0==strcmp(ltkn,"RELEVANCE")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"OFF")){
				prproflags=CLFILTERINGOFF;
				jj=0;
				*msg="Relevance filtering has been set to OFF";
			} else if (0==strcmp(ltkn,"ON")){
				jj=CLFILTERINGON;
			} else if (0==strcmp(ltkn,"NONEQU")){
				jj=CLFILTERINGNONEQU;
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid RELEVANCE parameter";
					return(1);
				}
				printf("Invalid RELEVANCE parameter: %s\n",ltkn);
				printf("The first parameter must be OFF, ON or NONEQU.\n");
				jj=0;
			}
			if (jj){
				if (NULL!=(ltkn=strtok(NULL," \t"))){
					ff=strtol(ltkn,&ptr1,10);
					if (ptr1[0]==0){
						if (ff>=0){
							jj|=CLFILTERINGITERLIMIT;
						} else {
							if (pgmmode==NONINTERACTIVE){
								*msg="Invalid RELEVANCE threshold parameter";
								return(1);
							}
							printf("Invalid threshold parameter: %s\n",ltkn);
							printf("Threshold must be a positive number in float or integer format.\n");
							jj=0;
						}
					} else {
						ff=strtof(ltkn,&ptr1);
						if (ptr1[0]==0){
							if (ff<0.0){
								if (pgmmode==NONINTERACTIVE){
									*msg="Invalid RELEVANCE threshold parameter";
									return(1);
								}
								printf("Invalid threshold parameter: %s\n",ltkn);
								printf("Threshold must be a positive number in float or integer format.\n");
								jj=0;
							} else if (ff==0){
								jj|=CLFILTERINGITERLIMIT;
							}
						} else {
							if (pgmmode==NONINTERACTIVE){
								if (ltkn[0]==0){
									*msg="Missing threshold value";
								} else {
									*msg="Invalid RELEVANCE threshold parameter";
								}
								return(1);
							}
							jj=0;
							if (ltkn[0]==0){
								printf("Missing threshold value\n");
							} else {
								printf("Invalid threshold parameter: %s\n",ltkn);
								printf("Threshold must be a positive number in float or integer format.\n");
							}
						}
					}
					if (jj){
						if (NULL!=(ltkn=strtok(NULL," \t"))){
							for (ii=0;ii<strlen(ltkn);ii++){
								ltkn[ii]=toupper(ltkn[ii]);
							}
							if (0==strcmp(ltkn,"RESIGN")){
								ppthrshld=ff;
								prproflags=jj|CLFILTERINGRESIGN;
							} else {
								if (pgmmode==NONINTERACTIVE){
									*msg="Invalid parameter. The only optional final parameter is RESIGN";
									return(1);
								}
								printf("Invalid parameter: %s\n",ltkn);
								printf("The only optional final parameter is RESIGN.\n");
								jj=0;
							}
						} else {
							ppthrshld=ff;
							prproflags=jj;
						}
						if ((jj)&&(pgmmode==INTERACTIVE)){
							printf("Relevance filtering has been set to ");
							if (prproflags&CLFILTERINGON){
								printf("ON ");
							} else {
								printf("NONEQU ");
							}
							if (prproflags&CLFILTERINGITERLIMIT){
								printf("%ld",(int64_t)ppthrshld);
							} else {
								printf("%f",ppthrshld);
							}
							if (prproflags&CLFILTERINGRESIGN){
								printf(" RESIGN");
							}
							printf("\n");
						}
					}
				} else {
					*msg="Missing RELEVANCE threshold parameter";
					if (pgmmode==NONINTERACTIVE){
						return(1);
					}
				}
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing RELEVANCE parameters";
				return(1);
			}
			printf("Current relevance filtering setting is ");
			if (prproflags&CLFILTERINGOFF){
				printf("OFF\n");
			} else {
				if (prproflags&CLFILTERINGON){
					printf("ON ");
				} else {
					printf("NONEQU ");
				}
				if (prproflags&CLFILTERINGITERLIMIT){
					printf("%ld",(int64_t)ppthrshld);
				} else {
					printf("%f",ppthrshld);
				}
				if (prproflags&CLFILTERINGRESIGN){
					printf(" RESIGN");
				}
				printf("\n");
			}
		}

	/* TIMESLICES command. */
	} else if (0==strcmp(ltkn,"TIMESLICES")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			ii=strtol(ltkn,&ptr1,10);
			if ((ii>0)&&(ii<64)&&(ptr1[0]==0)){
				timeslices=ii;
				if (pgmmode==INTERACTIVE){
					printf("Number of time slices has been set to %d\n",timeslices);
				}
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing UNIFSLICES parameter";
					return(1);
				}
				printf("Current number of time slices is %d\n",timeslices);
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid number of time slices";
					return(1);
				}
				printf("Invalid number of time slices: %s\n",ltkn);
				printf("Number of time slices must be an integer in the range 1-63.\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing UNIFSLICES parameter";
				return(1);
			}
			printf("Current number of time slices is %d\n",timeslices);
		}

	/* CORES command. */
	} else if (0==strcmp(ltkn,"CORES")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			ii=strtol(ltkn,&ptr1,10);
			if ((ii>=0)&&(ptr1[0]==0)){
				if ((ii<=num_cores)&&(ii<=MAXSTRATEGIES)){
					if (ii==0){
						ii=num_cores;
					}
					nn=0;
				} else {
					if (MAXSTRATEGIES>num_cores){
						ii=num_cores;
						nn=1;
					} else {
						ii=MAXSTRATEGIES;
						nn=2;
					}
				}
				active_cores=ii;
				procnb=active_cores-1;
				if (pgmmode==INTERACTIVE){
					switch (nn){
						case 0:
						default:
							printf("Now using %d core(s). %d core(s) available.\n",active_cores,num_cores);
							break;
						case 1:
							printf("Only %d core(s) are available.\nOnly %d core(s) enabled\n",num_cores,num_cores);
							break;
						case 2:
							printf("Only %d tuned strategies are available.\nOnly %d core(s) enabled\n",
									MAXSTRATEGIES,MAXSTRATEGIES);
							break;
					}
				}

			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing CORES parameter";
					return(1);
				}
				printf("Using %d cores. %d cores available.\n",active_cores,num_cores);
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid CORES parameter, parameter must be a positive integer";
					return(1);
				}
				printf("Invalid CORES parameter: %s\n",ltkn);
				printf("Parameter must be a positive integer.\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing CORES parameter";
				return(1);
			}
			printf("Using %d cores. %d cores available.\n",active_cores,num_cores);
		}

	/* MAXWEIGHT command. */
	} else if (0==strcmp(ltkn,"MAXWEIGHT")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"OFF")){
				glblopts.mxwghtopt|=MXWEIGHTOFF;
				glblopts.mxwghtopt&=(~MXWEIGHTNUM);
				*msg="MAXWEIGHT has been set to OFF";
				glblopts.maxweight=ABSMAXWEIGHT;
			} else if (0==strcmp(ltkn,"LRS")){
				glblopts.mxwghtopt|=MXWEIGHTLRS;
				glblopts.mxwghtopt&=(~(MXWEIGHTLRSOTT|MXWEIGHTLRSDSC|MXWEIGHTLRSOFF));
				usrparam&=~MAXWMASK;
				usrparam|=LRSALL_P;
				usrparam_mask|=MAXWMASK;
				*msg="MAXWEIGHT has been set to LRS";
			} else if (0==strcmp(ltkn,"LRSOTTER")){
				glblopts.mxwghtopt|=MXWEIGHTLRSOTT;
				glblopts.mxwghtopt&=(~(MXWEIGHTLRS|MXWEIGHTLRSDSC|MXWEIGHTLRSOFF));
				usrparam&=~MAXWMASK;
				usrparam|=LRSOTT_P;
				usrparam_mask|=MAXWMASK;
				*msg="MAXWEIGHT has been set to LRSOTTER";
			} else if (0==strcmp(ltkn,"LRSDISCOUNT")){
				glblopts.mxwghtopt|=MXWEIGHTLRSDSC;
				glblopts.mxwghtopt&=(~(MXWEIGHTLRS|MXWEIGHTLRSOTT|MXWEIGHTLRSOFF));
				usrparam&=~MAXWMASK;
				usrparam|=LRSDSC_P;
				usrparam_mask|=MAXWMASK;
				*msg="MAXWEIGHT has been set to LRSDISCOUNT";
			} else if (0==strcmp(ltkn,"LRSOFF")){
				glblopts.mxwghtopt|=MXWEIGHTLRSOFF;
				glblopts.mxwghtopt&=(~(MXWEIGHTLRS|MXWEIGHTLRSOTT|MXWEIGHTLRSDSC));
				usrparam&=~MAXWMASK;
				usrparam|=LRSOFF_P;
				usrparam_mask|=MAXWMASK;
				*msg="MAXWEIGHT has been set to LRSOFF";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.mxwghtopt|=MXWEIGHTLRSOTT;
				glblopts.mxwghtopt&=(~(MXWEIGHTLRS|MXWEIGHTLRSDSC|MXWEIGHTLRSOFF));
				usrparam&=~MAXWMASK;
				usrparam_mask&=~MAXWMASK;
				*msg="MAXWEIGHT has been set to DEFAULT";
			} else {
				ii=strtol(ltkn,&ptr1,10);
				if ((ii>0)&&(ptr1[0]==0)){
					glblopts.mxwghtopt|=MXWEIGHTNUM;
					glblopts.mxwghtopt&=(~MXWEIGHTOFF);
					glblopts.maxweight=ii*10;
					if (pgmmode==INTERACTIVE){
						printf("MAXWEIGHT has been set to %d.\n",ii);
					}
				} else if (ltkn[0]==0){
					if (pgmmode==NONINTERACTIVE){
						*msg="Missing MAXWEIGHT parameter";
						return(1);
					}
					if (glblopts.mxwghtopt&MXWEIGHTOFF){
						printf("MAXWEIGHT is set to OFF and ");
					} else {
						printf("MAXWEIGHT is set to %d and ",glblopts.maxweight/10);
					}
					if (0==(usrparam_mask&MAXWMASK)){
						printf("LRS DEFAULT\n");
					} else {
						switch (glblopts.mxwghtopt&(MXWEIGHTLRS|MXWEIGHTLRSOTT|MXWEIGHTLRSDSC|MXWEIGHTLRSOFF)){
							case MXWEIGHTLRS:
								printf("LRS.\n");
								break;
							case MXWEIGHTLRSOFF:
								printf("LRSOFF.\n");
								break;
							case MXWEIGHTLRSOTT:
								printf("LRSOTTER.\n");
								break;
							default:
								printf("LRSDISCOUNT.\n");
								break;
						}
					}
				} else {
					if (pgmmode==NONINTERACTIVE){
						*msg="Invalid MAXWEIGHT parameter";
						return(1);
					}
					printf("Invalid MAXWEIGHT parameter: %s\n",ltkn);
					printf("Parameter must be OFF, LRS, LRSOTTER, LRSDISCOUNT, an unsigned integer or DEFAULT.\n");
				}
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing MAXWEIGHT parameter";
				return(1);
			}
			if (glblopts.mxwghtopt&MXWEIGHTOFF){
				printf("MAXWEIGHT is set to OFF and ");
			} else {
				printf("MAXWEIGHT is set to %d and ",glblopts.maxweight/10);
			}
			if (0==(usrparam_mask&MAXWMASK)){
				printf("LRS DEFAULT\n");
			} else {
				switch (glblopts.mxwghtopt&(MXWEIGHTLRS|MXWEIGHTLRSOTT|MXWEIGHTLRSDSC|MXWEIGHTLRSOFF)){
					case MXWEIGHTLRS:
						printf("LRS.\n");
						break;
					case MXWEIGHTLRSOFF:
						printf("LRSOFF.\n");
						break;
					case MXWEIGHTLRSOTT:
						printf("LRSOTTER.\n");
						break;
					default:
						printf("LRSDISCOUNT.\n");
						break;
				}
			}
		}

	/* DEMODULATION command. */
	} else if (0==strcmp(ltkn,"DEMODULATION")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"ON")){
				glblopts.demodulation=DEMODON;
				usrparam&=~DEMODMASK;
				usrparam|=DEMODON_P;
				usrparam_mask|=DEMODMASK;
				*msg="Demodulation has been enabled";
			} else if (0==strcmp(ltkn,"OFF")){
				glblopts.demodulation=DEMODOFF;
				usrparam&=~DEMODMASK;
				usrparam|=DEMODOFF_P;
				usrparam_mask|=DEMODMASK;
				*msg="Demodulation has been disabled";
			} else if (0==strcmp(ltkn,"COMPLETE")){
				glblopts.demodulation=DEMODCPL;
				usrparam&=~DEMODMASK;
				usrparam|=DEMODCPL_P;
				usrparam_mask|=DEMODMASK;
				*msg="Demodulation is enabled and will preserve completeness";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.demodulation=DEMODCPL;
				usrparam&=~DEMODMASK;
				usrparam_mask&=~DEMODMASK;
				*msg="Demodulation has been set to DEFAULT";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing DEMODULATION parameter";
					return(1);
				}
				if (0==(usrparam_mask&DEMODMASK)){
					printf("Demodulation check is set to DEFAULT\n");
				} else {
					if (glblopts.demodulation==DEMODON){
						printf("Demodulation is set to ON\n");
					} else if (glblopts.demodulation==DEMODCPL){
						printf("Demodulation is set to COMPLETE\n");
					} else {
						printf("Demodulation is set to OFF\n");
					}
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid DEMODULATION parameter";
					return(1);
				}
				printf("Invalid DEMODULATION option: %s\n",ltkn);
				printf("Valid options are ON, OFF, COMPLETE or DEFAULT\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing DEMODULATION parameter";
				return(1);
			}
			if (0==(usrparam_mask&DEMODMASK)){
				printf("Demodulation check is set to DEFAULT\n");
			} else {
				if (glblopts.demodulation==DEMODON){
					printf("Demodulation is set to ON\n");
				} else if (glblopts.demodulation==DEMODCPL){
					printf("Demodulation is set to COMPLETE\n");
				} else {
					printf("Demodulation is set to OFF\n");
				}
			}
		}

	/* FACTORSET command. */
	} else if (0==strcmp(ltkn,"FACTORSET")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"1")){
				glblopts.factoridx=1;
				usrparam&=~WFACTSETMASK;
				usrparam|=WFACT1_P;
				usrparam_mask|=WFACTSETMASK;
				*msg="Weight clause factor set has been set to 1";
			} else if (0==strcmp(ltkn,"2")){
				glblopts.factoridx=2;
				usrparam&=~WFACTSETMASK;
				usrparam|=WFACT2_P;
				usrparam_mask|=WFACTSETMASK;
				*msg="Weight clause factor set has been set to 2";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.factoridx=1;
				usrparam&=~WFACTSETMASK;
				usrparam_mask&=~WFACTSETMASK;
				*msg="Weight clause factor set has been set to DEFAULT";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing FACTORSET parameter";
					return(1);
				}
				if (0==(usrparam_mask&WFACTSETMASK)){
					printf("Weight clause factor set is set to DEFAULT\n");
				} else {
					if (glblopts.factoridx==1){
						printf("Weight clause factor set is set to 1\n");
					} else {
						printf("Weight clause factor set is set to 2\n");
					}
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid FACTORSET parameter";
					return(1);
				}
				printf("Invalid FACTORSET option: %s\n",ltkn);
				printf("Valid options are 1, 2 or DEFAULT\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing FACTORSET parameter";
				return(1);
			}
			if (0==(usrparam_mask&WFACTSETMASK)){
				printf("Weight clause factor set is set to DEFAULT\n");
			} else {
				if (glblopts.factoridx==1){
					printf("Weight clause factor set is set to 1\n");
				} else {
					printf("Weight clause factor set is set to 2\n");
				}
			}
		}

	/* LOADREDUNDANT command. */
	} else if (0==strcmp(ltkn,"LOADREDUNDANT")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"ON")){
				loadredundant=1;
				*msg="LOADREDUNDANT has been set to ON";
			} else if (0==strcmp(ltkn,"OFF")){
				loadredundant=0;
				*msg="LOADREDUNDANT has been set to OFF";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing LOADREDUNDANT parameter";
					return(1);
				}
				if (loadredundant==1){
					printf("LOADREDUNDANT is ON\n");
				} else {
					printf("LOADREDUNDANT is OFF\n");
				}
			} else {
				printf("Invalid LOADREDUNDANT option: %s\n",ltkn);
				printf("Valid options are ON or OFF\n");
				return(pgmmode==INTERACTIVE?0:1);
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing LOADREDUNDANT parameter";
				return(1);
			}
			if (loadredundant==1){
				printf("LOADREDUNDANT is ON\n");
			} else {
				printf("Equivalent factoring is OFF\n");
			}
		}

	/* VERBOSE command. */
	} else if (0==strcmp(ltkn,"VERBOSE")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"ON")){
				verbose=1;
				*msg="VERBOSE has been set to ON";
			} else if (0==strcmp(ltkn,"OFF")){
				verbose=0;
				*msg="VERBOSE has been set to OFF";
			} else if (ltkn[0]==0){
				if (verbose==1){
					printf("VERBOSE is ON\n");
				} else {
					printf("VERBOSE is OFF\n");
				}
				return(pgmmode==INTERACTIVE?0:1);
			} else {
				printf("Invalid VERBOSE option: %s\n",ltkn);
				printf("Valid options are ON or OFF\n");
				return(pgmmode==INTERACTIVE?0:1);
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing VERBOSE parameter";
				return(1);
			}
			if (verbose==1){
				printf("VERBOSE is ON\n");
			} else {
				printf("VERBOSE is OFF\n");
			}
		}

	/* LEARNTO command. */
	} else if (0==strcmp(ltkn,"LEARNTO")){
		if (NULL!=(ltkn=strtok(NULL,""))){
			if (strlen(ltkn)==3){
				if (('O'==toupper(ltkn[0]))&&('F'==toupper(ltkn[1]))&&('F'==toupper(ltkn[2]))){
					if (learnto!=NULL){
						MYFREE(learnto);
						learnto=NULL;
					}
					*msg="LEARNTO has been set to OFF";
					return(0);
				}
			}
			if (NULL==(ptr1=ParseFilename(ltkn,&ptr2))){
				*msg="Invalid LEARNTO parameter. It must be OFF or a valid filename";
				return(pgmmode==INTERACTIVE?0:1);
			}
			if (NULL!=(ltkn=strtok(ptr2," \t"))){
				if (strlen(ltkn)==3){
					if (('N'==toupper(ltkn[0]))&&('E'==toupper(ltkn[1]))&&('W'==toupper(ltkn[2]))){
						learnnew=1;
					} else {
						printf("Invalid LEARNTO parameter %s\n",ltkn);
						return(pgmmode==INTERACTIVE?0:1);
					}
				} else {
					printf("Invalid LEARNTO parameter %s\n",ltkn);
					return(pgmmode==INTERACTIVE?0:1);
				}
			} else {
				learnnew=0;
			}
			if (NULL==(ptr3=strdupmn(ptr1))){
				*msg="No memory available, exiting program";
				return(1);
			}
			if (learnto!=NULL){
				MYFREE(learnto);
			}
			learnto=ptr3;
			if (pgmmode==INTERACTIVE){
				if (learnnew){
					printf("LEARNTO has been set to %s NEW\n",learnto);
				} else {
					printf("LEARNTO has been set to %s\n",learnto);
				}
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing LEARNTO parameter";
				return(1);
			}
			if (learnto==NULL){
				printf("LEARNTO is OFF\n");
			} else {
				if (learnnew){
					printf("LEARNTO is set to %s NEW\n",learnto);
				} else {
					printf("LEARNTO is set to %s\n",learnto);
				}
			}
		}

	/* LEARNFROM command. */
	} else if (0==strcmp(ltkn,"LEARNFROM")){
		if (NULL!=(ltkn=strtok(NULL,""))){
			if (strlen(ltkn)==3){
				if (('O'==toupper(ltkn[0]))&&('F'==toupper(ltkn[1]))&&('F'==toupper(ltkn[2]))){
					if (learnfrom!=NULL){
						MYFREE(learnfrom);
						MYFREE(lrnrecords);
						learnfrom=NULL;
						lrnrecords=NULL;
					}
					*msg="LEARNFROM has been set to OFF";
					return(0);
				}
			}
			if (NULL==(ptr1=ParseFilename(ltkn,&ptr2))){
				*msg="Invalid LEARNFROM parameter. It must be OFF or a valid filename.";
				return(pgmmode==INTERACTIVE?0:1);
			}
			if (NULL==(ptr3=strdupmn(ptr1))){
				*msg="No memory available, exiting program";
				return(1);
			}
			if (learnfrom!=NULL){
				MYFREE(learnfrom);
				MYFREE(lrnrecords);
				lrnrecords=NULL;
			}
			learnfrom=ptr3;
			if (pgmmode==INTERACTIVE){
				printf("LEARNFROM has been set to %s\n",learnfrom);
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing LEARNFROM parameter";
				return(1);
			}
			if (learnfrom==NULL){
				printf("LEARNFROM is OFF\n");
			} else {
				printf("LEARNFROM is set to %s\n",learnfrom);
			}
		}

	/* PROCLEARN command. */
	} else if (0==strcmp(ltkn,"PROCLEARN")){
		if (NULL!=(ltkn=strtok(NULL,""))){
			if (NULL==(ptr1=ParseFilename(ltkn,&ptr3))){
				printf("Invalid PROCLEARN input file name: %s\n",ltkn);
				return(pgmmode==INTERACTIVE?0:1);
			}
			if (NULL==(ptr2=ParseFilename(ptr3,&ptr4))){
				printf("Invalid PROCLEARN output file name: %s\n",ptr3);
				return(pgmmode==INTERACTIVE?0:1);
			}
			if (NULL!=(ltkn=strtok(ptr4," \t"))){
				ii=strtol(ltkn,&ptr3,10);
				if ((ii<=0)||(ptr3[0]!=0)){
					printf("Invalid boostfactor parameter: %s\n",ltkn);
					printf("The boostfactor parameter must be an strictly positive integer.\n");
					return(pgmmode==INTERACTIVE?0:1);
				}
			} else {
				ii=10;
			}
			if (ProcessLearnData(ptr1,ptr2,ii)){
				if (pgmmode==NONINTERACTIVE){
					return(1);
				}
			}
		} else {
			printf("Missing PROCLEARN filename parameters, no process done.\n");
			return(pgmmode==INTERACTIVE?0:1);
		}

	/* SPLIT command. */
	} else if (0==strcmp(ltkn,"SPLIT")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"ON")){
				glblopts.split=1;
				*msg="Split has been set to ON";
			} else if (0==strcmp(ltkn,"OFF")){
				glblopts.split=0;
				*msg="Split has been set to OFF";
			} else if (0==strcmp(ltkn,"OTTER")){
				glblopts.split=OTTER;
				*msg="Split has been set to ON only for Otter algorithm";
			} else if (0==strcmp(ltkn,"DISCOUNT")){
				glblopts.split=DISCOUNT;
				*msg="Split has been set to ON only for Discount algorithm";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing SPLIT parameter";
					return(1);
				}
				switch (glblopts.split){
					case 0:
						printf("Split is OFF, interface between first order prover\n");
						printf("and SAT solver is disabled.\n");
						break;
					case 1:
						printf("Split is ON, interface between first order prover\n");
						printf("and SAT solver is enabled.\n");
						break;
					case OTTER:
						printf("Split is OTTER, interface between first order prover\n");
						printf("and SAT solver is enabled only for Otter algorithm.\n");
						break;
					case DISCOUNT:
						printf("Split is DISCOUNT, interface between first order prover\n");
						printf("and SAT solver is enabled only for Discount algorithm.\n");
						break;
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid SPLIT parameter";
					return(1);
				}
				printf("Invalid SPLIT option: %s\n",ltkn);
				printf("Valid options are ON, OFF, Otter or Discount.\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing SPLIT parameter";
				return(1);
			}
			switch (glblopts.split){
				case 0:
					printf("Split is OFF, interface between first order prover\n");
					printf("and SAT solver is disabled.\n");
					break;
				case 1:
					printf("Split is ON, interface between first order prover\n");
					printf("and SAT solver is enabled.\n");
					break;
				case OTTER:
					printf("Split is OTTER, interface between first order prover\n");
					printf("and SAT solver is enabled only for Otter algorithm.\n");
					break;
				case DISCOUNT:
					printf("Split is DISCOUNT, interface between first order prover\n");
					printf("and SAT solver is enabled only for Discount algorithm.\n");
					break;
			}
		}

	/* FACTORING command. */
	} else if (0==strcmp(ltkn,"FACTORING")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"ON")){
				glblopts.factoring=1;
				usrparam&=~FACTORINGMASK;
				usrparam|=FACTON_P;
				usrparam_mask|=FACTORINGMASK;
				*msg="Factoring has been set to ON";
			} else if (0==strcmp(ltkn,"OFF")){
				glblopts.factoring=0;
				usrparam&=~FACTORINGMASK;
				usrparam|=FACTOFF_P;
				usrparam_mask|=FACTORINGMASK;
				*msg="Factoring has been set to OFF";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.factoring=1;
				usrparam&=~FACTORINGMASK;
				usrparam_mask&=~FACTORINGMASK;
				*msg="Factoring has been set to DEFAULT";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing FACTORING parameter";
					return(1);
				}
				if (0==(usrparam_mask&FACTORINGMASK)){
					printf("Factoring is set to DEFAULT\n");
				} else {
					if (glblopts.factoring==1){
						printf("Factoring is ON\n");
					} else {
						printf("Factoring is OFF\n");
					}
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid FACTORING parameter";
					return(1);
				}
				printf("Invalid FACTORING option: %s\n",ltkn);
				printf("Valid options are ON, OFF or DEFAULT\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing FACTORING parameter";
				return(1);
			}
			if (0==(usrparam_mask&FACTORINGMASK)){
				printf("Factoring is set to DEFAULT\n");
			} else {
				if (glblopts.factoring==1){
					printf("Factoring is ON\n");
				} else {
					printf("Factoring is OFF\n");
				}
			}
		}


	/* SELECT command. */
	} else if (0==strcmp(ltkn,"SELECT")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"MAXIMAL")){
				glblopts.select=MAXIMAL;
				usrparam&=~SELECTMASK;
				usrparam|=MAXIMAL_P;
				usrparam_mask|=SELECTMASK;
				*msg="Select has been set to MAXIMAL";
			} else if (0==strcmp(ltkn,"SINGLENEG")){
				glblopts.select=SINGLENEG;
				usrparam&=~SELECTMASK;
				usrparam|=SINGLENEG_P;
				usrparam_mask|=SELECTMASK;
				*msg="Select has been set to SINGLENEG";
			} else if (0==strcmp(ltkn,"MULTINEG")){
				glblopts.select=MULTINEG;
				usrparam&=~SELECTMASK;
				usrparam|=MULTINEG_P;
				usrparam_mask|=SELECTMASK;
				*msg="Select has been set to MULTINEG";
			} else if (0==strcmp(ltkn,"SINGLEPOS")){
				glblopts.select=SINGLEPOS;
				usrparam&=~SELECTMASK;
				usrparam|=SINGLEPOS_P;
				usrparam_mask|=SELECTMASK;
				*msg="Select has been set to SINGLEPOS";
			} else if (0==strcmp(ltkn,"MULTIPOS")){
				glblopts.select=MULTIPOS;
				usrparam&=~SELECTMASK;
				usrparam|=MULTIPOS_P;
				usrparam_mask|=SELECTMASK;
				*msg="Select has been set to MULTIPOS";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.select=SINGLENEG;
				usrparam&=~SELECTMASK;
				usrparam_mask&=~SELECTMASK;
				*msg="Select has been set to DEFAULT";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing SELECT parameter";
					return(1);
				}
				if (0==(usrparam_mask&SELECTMASK)){
					printf("Select option is set to DEFAULT\n");
				} else {
					switch (glblopts.select){
						case MAXIMAL:
							printf("Select is MAXIMAL\n");
							break;
						case SINGLENEG:
							printf("Select is SINGLENEG\n");
							break;
						default:
						case MULTINEG:
							printf("Select is MULTINEG\n");
							break;
					}
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid SELECT parameter";
					return(1);
				}
				printf("Invalid SELECT option: %s\n",ltkn);
				printf("Valid options are MAXIMAL, SINGLENEG, MULTINEG or DEFAULT\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing SELECT parameter";
				return(1);
			}
			if (0==(usrparam_mask&SELECTMASK)){
				printf("Select option is set to DEFAULT\n");
			} else {
				switch (glblopts.select){
					case MAXIMAL:
						printf("Select is MAXIMAL\n");
						break;
					case SINGLENEG:
						printf("Select is SINGLENEG\n");
						break;
					default:
					case MULTINEG:
						printf("Select is MULTINEG\n");
						break;
				}
			}
		}

	/* SYNTAX command. */
	} else if (0==strcmp(ltkn,"SYNTAX")){
		if (pgmmode==NONINTERACTIVE){
			*msg="Invalid option";
			return(1);
		}
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"TPTP")){
				syntaxmode=TPTP;
				*msg="Syntax mode has been set to TPTP";
			} else if (0==strcmp(ltkn,"DRODI")){
				if (proofmode==TPTP){
					printf("WARNING: If SYNTAX is set to DRODI and the problem is introduced via TELL and ASK commands the proof\n");
					printf("will always be displayed in DRODI format, because TPTP format is not compatible with DRODI syntax.\n\n");
				}
				syntaxmode=DRODI;
				*msg="Syntax mode has been set to DRODI";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing SYNTAX parameter";
					return(1);
				}
				if (syntaxmode==TPTP){
					printf("Syntax mode is TPTP\n");
				} else {
					printf("Syntax mode is DRODI\n");
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid SYNTAX parameter";
					return(1);
				}
				printf("Invalid SYNTAX option: %s\n",ltkn);
				printf("Valid options are TPTP or DRODI\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing SYNTAX parameter";
				return(1);
			}
			if (syntaxmode==TPTP){
				printf("Syntax mode is set to TPTP\n");
			} else {
				printf("Syntax mode is set to DRODI\n");
			}
		}

	/* PROOF command. */
	} else if (0==strcmp(ltkn,"PROOF")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"TPTP")){
				if ((pgmmode==INTERACTIVE)&&(syntaxmode==DRODI)){
					printf("WARNING: SYNTAX is set to DRODI. If the problem is introduced via TELL and ASK commands the proof\n");
					printf("will always be displayed in DRODI format, because TPTP format is not compatible with DRODI syntax.\n");
				}
				proofmode=TPTP;
				*msg="Proof mode has been set to TPTP";
			} else if (0==strcmp(ltkn,"DRODI")){
				proofmode=DRODI;
				*msg="Proof mode has been set to DRODI";
			} else if (0==strcmp(ltkn,"OFF")){
				proofmode=PROOFOFF;
				*msg="Proof mode has been set to OFF";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing PROOF parameter";
					return(1);
				}
				if (proofmode==TPTP){
					printf("Proof mode is TPTP\n");
				} else if (proofmode==DRODI){
					printf("Proof mode is DRODI\n");
				} else {
					printf("Proof mode is OFF\n");
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid PROOF parameter";
					return(1);
				}
				printf("Invalid PROOF option: %s\n",ltkn);
				printf("Valid options are TPTP or DRODI\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing PROOF parameter";
				return(1);
			}
			if (proofmode==TPTP){
				printf("Proof mode is TPTP\n");
			} else if (proofmode==DRODI){
				printf("Proof mode is DRODI\n");
			} else {
				printf("Proof mode is OFF\n");
			}
		}

	/* PRINTKBDATA command. */
	} else if (0==strcmp(ltkn,"PRINTKBDATA")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"ON")){
				prtkbdata=1;
				*msg="Printkbdata mode has been set to ON";
			} else if (0==strcmp(ltkn,"OFF")){
				prtkbdata=0;
				*msg="Printkbdata mode has been set to OFF";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing PRINTKBDATA parameter";
					return(1);
				}
				if (prtkbdata){
					printf("Printkbdata mode is ON\n");
				} else {
					printf("Printkbdata mode is OFF\n");
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid PRINTKBDATA parameter";
					return(1);
				}
				printf("Invalid PRINTKBDATA option: %s\n",ltkn);
				printf("Valid options are WINNER, ALL or OFF\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing PRINTKBDATA parameter";
				return(1);
			}
			if (prtkbdata){
				printf("Printkbdata mode is ON\n");
			} else {
				printf("Printkbdata mode is OFF\n");
			}
		}

	/* ALGORITHM command. */
	} else if (0==strcmp(ltkn,"ALGORITHM")){
		if (NULL!=(ltkn=strtok(NULL," \t"))){
			for (ii=0;ii<strlen(ltkn);ii++){
				ltkn[ii]=toupper(ltkn[ii]);
			}
			if (0==strcmp(ltkn,"OTTER")){
				glblopts.algorithm=OTTER;
				usrparam&=~ALGMASK;
				usrparam|=OTTER_P;
				usrparam_mask|=ALGMASK;
				*msg="Selected algorithm has been set to OTTER";
			} else if (0==strcmp(ltkn,"DISCOUNT")){
				glblopts.algorithm=DISCOUNT;
				usrparam&=~ALGMASK;
				usrparam|=DISCOUNT_P;
				usrparam_mask|=ALGMASK;
				*msg="Selected algorithm has been set to DISCOUNT";
			} else if (0==strcmp(ltkn,"DEFAULT")){
				glblopts.algorithm=OTTER;
				usrparam&=~ALGMASK;
				usrparam_mask&=~ALGMASK;
				*msg="Selected algorithm has been set to DEFAULT";
			} else if (ltkn[0]==0){
				if (pgmmode==NONINTERACTIVE){
					*msg="Missing ALGORITHM parameter";
					return(1);
				}
				if (0==(usrparam_mask&ALGMASK)){
					printf("Selected algorithm is set to DEFAULT\n");
				} else {
					if (glblopts.algorithm==OTTER){
						printf("Selected algorithm is OTTER\n");
					} else {
						printf("Selected algorithm is DISCOUNT\n");
					}
				}
			} else {
				if (pgmmode==NONINTERACTIVE){
					*msg="Invalid ALGORITHM parameter";
					return(1);
				}
				printf("Invalid ALGORITHM option: %s\n",ltkn);
				printf("Valid options are OTTER, DISCOUNT or DEFAULT\n");
			}
		} else {
			if (pgmmode==NONINTERACTIVE){
				*msg="Missing ALGORITHM parameter";
				return(1);
			}
			if (0==(usrparam_mask&ALGMASK)){
				printf("Selected algorithm is set to DEFAULT\n");
			} else {
				if (glblopts.algorithm==OTTER){
					printf("Selected algorithm is OTTER\n");
				} else {
					printf("Selected algorithm is DISCOUNT\n");
				}
			}
		}

	/* IMPORT command. */
	} else if (0==strcmp(ltkn,"IMPORT")){
		if (pgmmode==NONINTERACTIVE){
			*msg="Invalid option";
			return(1);
		}
		if (NULL!=(ltkn=strtok(NULL,""))){
			if (NULL==(ptr1=ParseFilename(ltkn,&ptr2))){
				*msg="Invalid IMPORT parameter. It must be a valid filename";
				return(pgmmode==INTERACTIVE?0:1);
			}
			printf("Reading file %s...\n",ptr1);
			ii=syntaxmode;
			syntaxmode=TPTP;
			importfile=ptr1;
			if (0!=(ii=Import(ptr1,NULL,0,1))){
				if (ii!=2){
					if (NOMEMORY==Initialize_KB(&mainkb)){
						printf("No memory available, exiting program\n");
						return(1);
					}
					ResetStats(pbmstats);
					printf("Knowledge bases have been initialized\n");
				}
			} else {
				Saturate();
			}
			syntaxmode=ii;
			importfile=NULL;
		} else {
			printf("Missing file name\n");
		}

	/* LOG command. */
	} else if (0==strcmp(ltkn,"LOG")){
		if (NULL!=(ltkn=strtok(NULL,""))){
			MYFREE(logpath);
			if (strlen(ltkn)==3){
				if (('O'==toupper(ltkn[0]))&&('F'==toupper(ltkn[1]))&&('F'==toupper(ltkn[2]))){
					logpath=NULL;
					*msg="LOG has been set to OFF";
					return(0);
				}
			}
			if (strlen(ltkn)==2){
				if (('O'==toupper(ltkn[0]))&&('N'==toupper(ltkn[1]))){
					if (NULL==(logpath=mallocmn(1))){
						*msg="No memory available, exiting program";
						return(1);
					}
					logpath[0]=0;
					*msg="LOG has been set to ON";
					return(0);
				}
			}
			if (NULL==(logpath=strdupmn(ltkn))){
				*msg="No memory available, exiting program";
				return(1);
			}
			if (pgmmode==NONINTERACTIVE){
				*msg="LOG has been set to file\n";
			} else {
				printf("LOG has been set to %s\n",ltkn);
			}
		} else {
			*msg="Missing log parameter";
			if (pgmmode==NONINTERACTIVE){
				return(1);
			}
		}

	/* Comment. */
	} else if (ltkn[0]=='#'){
		if (pgmmode==NONINTERACTIVE){
			*msg="Invalid option #";
			return(1);
		}
		printf("\n");

	#ifdef BENCHMARKING
	/* BENCHFILE command. */
	} else if (0==strcmp(ltkn,"BENCHFILE")){
		if (NULL!=(ltkn=strtok(NULL,""))){
			if (NULL==(benchfname=mallocmn(1+strlen(ltkn)))){
				*msg="Not enough memory for benchmarking output file name";
				return(pgmmode==INTERACTIVE?0:1);
			} else {
				strcpy(benchfname,ltkn);
				if (pgmmode==INTERACTIVE){
					printf("BENCHFILE has been set to %s\n",benchfname);
				}
			}
		} else {
			*msg="Missing file name";
			if (pgmmode==NONINTERACTIVE){
				return(1);
			}
		}
	#endif

	/* Invalid option/command. */
	} else {
		printf("Invalid option %s\n",ltkn);
		return(pgmmode==INTERACTIVE?0:1);
	}
	return(0);
} /* ParseOption */
