/*
 ============================================================================
 Name        : simplify.c
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */

/****************************************************************
*
*               simplify (Simplification inference functions module)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

/*-------------------------------------------------------------
 *
 *  DESCRIPTION:
 *
 *    This source contains the simplification inferences of the Drodi package.
 *
 *    Drodi is a multi-platform Open Source Artificial Intelligence application.
 *    See Drodi.c for additional information.
 *
 *
 *  IMPLEMENTATION:
 *
 *    Multi-platform
 *
 *  NOTES:
 *
 *
 *--------------------------------------------------------------*/

/* Include for all C sources */
#include "global.h"

/** Global variables for this module: only those that need initialization in their definition. */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if a clause is a tautology)  OCJ
 *
 *    This function checks if a clause in internal compiled format
 *    is a tautology.
 *
 *    Two checks are performed to check for a tautology:
 *
 *    The first check is positive if the clause contains a literal
 *    and its negation, that is, for clauses of the form "C|P|~P"
 *    where C is a sub clause and P is an atom. The literals must
 *    be one exactly the negation of the other, not just unify with
 *    each other. For instance, consider the following clause:
 *      A(B,x)|!A(y,C)
 *    where x and y are universally quantified variables. This
 *    clause is not a tautology because let A(B,D) be false and
 *    A(E,C) be true. Then the clause is false, so it is not
 *    necessarily true.
 *    Another example, the clause:
 *      A(x)|!A(B)
 *    where x is an universally quantified variable. This is not
 *    a tautology because let A(C) be false and A(B) be true.
 *    Then the clause is false.
 *
 *    The second check is for equational tautologies, that is,
 *    for clauses of the form C|t=t or C|"a"="a" or C|"a"!="b" where C
 *    is a sub clause, t is a term and "a" and "b" are "distinct objects".
 *
 *
 *  ARGUMENTS:
 *
 *    clause: pointer to clause in binary format.
 *
 *  RETURNS:
 *
 *    0 -> Clause is not a tautology.
 *    1 -> Clause is a tautology.
 *
 *
 *--------------------------------------------------------------*/
int32_t IsTautology(cmprefix *clause){
	auto hashchain *rtterm1s,*rtterm2s;                  /* Pointers to root term symbols in equalities */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */

	/* Loop through literals in clause to detect equalities */
	/* of identical terms or "distinct objects" tautologies. */
	if ((UNITEQUALITY|NONUNITEQUALITY)&pbmflags){
		for (ptr1=&clause->part2.bin->formula[0];(*ptr1)!=UNITEND;
				ptr1=NextItem(ptr1,OVERSUBTERMS)){
			if (EQUALITY&ptr1[0]){
				ptr2=NextItem(ptr1,IMMED);
				ptr3=NextItem(ptr2,OVERSUBTERMS);
				if (0==(NEGATED&ptr1[0])){
					if (IsEqual(ptr2,ptr3,0)){
						return(1);
					}
				} else {
					if ((FUNCTION&ptr2[0])&&(FUNCTION&ptr3[0])){
						rtterm1s=((symbol *)&ptr2[1])->symbol;
						rtterm2s=((symbol *)&ptr3[1])->symbol;
						if ((rtterm1s!=rtterm2s)&&(DISTINCTOBJECT&rtterm1s->type)
								&&(DISTINCTOBJECT&rtterm2s->type)){
							return(1);
						}
					}
				}
			}
		}
	}

	/* Loop through pairs of literals in clause. */
	for (ptr1=&clause->part2.bin->formula[0];(*ptr1)!=UNITEND;
			ptr1=NextItem(ptr1,OVERSUBTERMS)){
		for (ptr2=NextItem(ptr1,OVERSUBTERMS);(*ptr2)!=UNITEND;
				ptr2=NextItem(ptr2,OVERSUBTERMS)){
			if (IsEqual(ptr1,ptr2,NEGATED)){
				return(1);
			}
		}
	}

	return(0);
} /* IsTautology */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Remove literal duplicates from clause)  OCJ
 *
 *    This function removes literal duplicates from a clause in
 *    internal compiled binary format. The simplified clause
 *    keeps the original assertions.
 *
 *    If the clause is successfully simplified then it is placed in
 *    new allocated storage and flagged as PASSIVE. If buildtxt
 *    parameter is not zero then the original clause is converted
 *    to text form and also placed in a new allocated storage that
 *    is linked to the original cmprefix structure. No other modifications
 *    are done to the cmprefix structure. The original binprefix storage
 *    is freed.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *    See EquivalentFactor() function for comments regarding program
 *    performance.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by factoring.
 *    pnewcl: Address of pointer to cmprefix structure of new clause
 *            to be allocated if there are literal removals.
 *    buildtxt: If not zero and a simplification results then the
 *              original clause is converted to text form and placed
 *              in a new allocated storage that is linked to the original
 *              cmprefix structure.
 *
 *  RETURNS:
 *
 *    0 -> process was OK and no duplicate literals were found.
 *    1 -> Some duplicate literals were removed.
 *    NOMEMORY -> not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t RemoveDuplicates(cmprefix *clause,cmprefix **pnewcl,int32_t buildtxt){
	auto cmprefix *auxcl;                               /* Pointer to auxiliary clause */
	auto int32_t size;                                  /* Size of auxiliary clause binary buffer in bytes */
	auto uint8_t *ptr1,*ptr2;                           /* Auxiliary pointers */
	auto char *ptr10;                                   /* Buffer for clause in text form. */
	auto binprefix *ptr4;                               /* Auxiliary pointer */
	auto int32_t ii,pp;                                 /* Auxiliary */

	#ifdef DEBUGCODE
	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Allocate storage for auxiliary clause and copy */
	/* clause into the first auxiliary clause. */
	size=clause->part2.bin->size+binpsize;
	if ((auxcl=mallocmn(sizeof(cmprefix)))==NULL){
		return(NOMEMORY);
	}
	if ((auxcl->part2.bin=mallocmn(size))==NULL){
		MYFREE(auxcl);
		return(NOMEMORY);
	}
	memcpy(auxcl->part2.bin,clause->part2.bin,size);
	auxcl->part2.bin->clause=auxcl;

	/* Loop through literals in auxiliary clause. */
	for (ptr1=&auxcl->part2.bin->formula[0],pp=0;(*ptr1)!=UNITEND;
			ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Loop through pairs of literals in auxiliary clause. */
		for (ptr2=NextItem(ptr1,OVERSUBTERMS);(*ptr2)!=UNITEND;
				ptr2=(ii?NextItem(ptr2,OVERSUBTERMS):ptr2)){
			ii=1;

			/* Identical literals. Delete the ptr2 literal. */
			if (IsEqual(ptr1,ptr2,0)){
					DeleteLiteral(auxcl->part2.bin,ptr2);
					pp=1; /* Indicate there was a removal. */
					ii=0; /* Keep in the same literal position in next iteration for ptr2. */
			}
		}
	}

	/* There is a removal. */
	if (pp){

		/* Reallocate simplification result to exact size */
		/* and set clause pointer. The original assertions */
		/* are kept. but their asymbol's must be adjusted */
		/* to the new clause. */
		if (NULL==(ptr4=reallocmn(auxcl->part2.bin,auxcl->part2.bin->size+binpsize))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->part2.bin=ptr4;
		auxcl->part2.bin->equflg=0;
		auxcl->inference=REMOVEDUPS;
		auxcl->parent1=clause;
		auxcl->parent2=NULL;
		auxcl->flags=UNPROC;
		if (auxcl->part2.bin->asserts!=NULL){
			for (ptr1=&auxcl->part2.bin->asserts[0];ptr1[0]!=UNITEND;ptr1+=(1+sizeof(asymbol))){
				((asymbol *)&ptr1[1])->part2.binp=auxcl->part2.bin;
			}
		}
		*pnewcl=auxcl;

		/* Convert original clause to text form if needed. */
		if (buildtxt){
			if (NULL==(ptr10=Decompile(clause->part2.bin))){
				MYFREE(auxcl->part2.bin);
				MYFREE(auxcl);
				return(NOMEMORY);
			}
			MYFREE(clause->part2.bin);
			clause->part2.text=ptr10;
		}

	/* If there was not a removal then free auxiliary clause storage. */
	} else {
		MYFREE(auxcl->part2.bin);
		MYFREE(auxcl);
	}

	return(pp);
} /* RemoveDuplicates */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Trivial equality resolution inference)  OCJ
 *
 *    This function performs trivial equality resolution inference
 *    simplification of a clause in internal compiled binary format.
 *    The simplified clause keeps the original assertions.
 *
 *    If the clause is successfully simplified then it is placed in
 *    new allocated storage and flagged as PASSIVE. The original clause
 *    is converted to text form and also placed in a new allocated
 *    storage that is linked to the original cmprefix structure.
 *    No other modifications are done to the cmprefix structure.
 *    The original binprefix storage is freed.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by equivalent factoring.
 *    pnewcl: Address of pointer to cmprefix structure of new clause
 *            to be allocated if a simplification is performed.
 *
 *  RETURNS:
 *
 *    0 -> process was OK and no simplification was performed.
 *    1 -> Clause has been successfully simplified.
 *    NOMEMORY -> not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t TrivEqResolution(cmprefix *clause,cmprefix **pnewcl){
	auto cmprefix *auxcl;                               /* Pointer to auxiliary clause */
	auto int32_t size;                                  /* Size of auxiliary clause binary buffer in bytes */
	auto hashchain *rtterm1s,*rtterm2s;                 /* Pointers to root term symbols in equalities */
	auto uint8_t *ptr1,*ptr2,*ptr3;                     /* Auxiliary pointers */
	auto char *ptr10;                                   /* Buffer for clause in text form. */
	auto binprefix *ptr4;                               /* Auxiliary pointer */
	auto int32_t ii,pp;                                 /* Auxiliary */

	#ifdef DEBUGCODE
	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Allocate storage for auxiliary clause and copy */
	/* clause into the first auxiliary clause. */
	size=clause->part2.bin->size+binpsize;
	if ((auxcl=mallocmn(sizeof(cmprefix)))==NULL){
		return(NOMEMORY);
	}
	if ((auxcl->part2.bin=mallocmn(size))==NULL){
		MYFREE(auxcl);
		return(NOMEMORY);
	}
	memcpy(auxcl->part2.bin,clause->part2.bin,size);
	auxcl->part2.bin->clause=auxcl;

	/* Loop through literals in auxiliary clause. */
	for (ptr1=&auxcl->part2.bin->formula[0],pp=0;(*ptr1)!=UNITEND;
			ptr1=(ii?NextItem(ptr1,OVERSUBTERMS):ptr1)){
		ii=1;

		/* Literal is an equality. */
		if (EQUALITY&ptr1[0]){
			ptr2=NextItem(ptr1,IMMED);
			ptr3=NextItem(ptr2,OVERSUBTERMS);

			/* Negated equality and root terms are equal. */
			if (NEGATED&ptr1[0]){
				if (IsEqual(ptr2,ptr3,0)){
					DeleteLiteral(auxcl->part2.bin,ptr1);
					pp=1; /* Indicate there was a simplification. */
					ii=0; /* Keep in the same literal position in next iteration for ptr2. */
				}

			/* Positive equality and root terms are unequal "distinct objects". */
			} else if ((FUNCTION&ptr2[0])&&(FUNCTION&ptr3[0])){
				rtterm1s=((symbol *)&ptr2[1])->symbol;
				rtterm2s=((symbol *)&ptr3[1])->symbol;
				if ((rtterm1s!=rtterm2s)&&(DISTINCTOBJECT&rtterm1s->type)
									&&(DISTINCTOBJECT&rtterm2s->type)){
					DeleteLiteral(auxcl->part2.bin,ptr1);
					pp=1; /* Indicate there was a simplification. */
					ii=0; /* Keep in the same literal position in next iteration for ptr2. */
				}
			}
		}
	}

	/* If there is a simplification then set clause pointer */
	/* and convert original clause to text form. The original */
	/* assertions are kept but their asymbol's must be */
	/* adjusted to the new clause. */
	if (pp){
		if (NULL==(ptr4=reallocmn(auxcl->part2.bin,auxcl->part2.bin->size+binpsize))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->part2.bin=ptr4;
		if (NULL==(ptr10=Decompile(clause->part2.bin))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->inference=TRIVEQRES;
		auxcl->parent1=clause;
		auxcl->parent2=NULL;
		auxcl->flags=UNPROC;
		auxcl->part2.bin->equflg=0;
		if (auxcl->part2.bin->asserts!=NULL){
			for (ptr1=&auxcl->part2.bin->asserts[0];ptr1[0]!=UNITEND;ptr1+=(1+sizeof(asymbol))){
				((asymbol *)&ptr1[1])->part2.binp=auxcl->part2.bin;
			}
		}
		*pnewcl=auxcl;
		FreeEquData(clause->part2.bin);
		MYFREE(clause->part2.bin);
		clause->part2.text=ptr10;

	/* If there was not a removal then free auxiliary clause storage. */
	} else {
		MYFREE(auxcl->part2.bin);
		MYFREE(auxcl);
	}

	return(pp);
} /* TrivEqResolution */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Destructive equality resolution inference)  OCJ
 *
 *    This function performs destructive equality resolution inference
 *    simplification of a clause in internal compiled binary format.
 *    The simplified clause keeps the original assertions.
 *
 *    If the clause is successfully simplified then it is placed in
 *    new allocated storage and flagged as PASSIVE. The original clause
 *    is converted to text form and also placed in a new allocated
 *    storage that is linked to the original cmprefix structure.
 *    No other modifications are done to the cmprefix structure.
 *    The original binprefix storage is freed.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by equivalent factoring.
 *    pnewcl: Address of pointer to cmprefix structure of new clause
 *            to be allocated if a simplification is performed.
 *
 *  RETURNS:
 *
 *    0 -> process was OK and no simplification was performed.
 *    1 -> Clause has been successfully simplified.
 *    NOMEMORY -> not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t DestrEqResolution(cmprefix *clause,cmprefix **pnewcl){
	auto cmprefix *auxcl;                               /* Pointer to auxiliary clause */
	auto int32_t size;                                  /* Size of auxiliary clause binary buffer in bytes */
	auto uint8_t *ptr1,*ptr2,*ptr5;                     /* Auxiliary pointers */
	auto char *ptr10;                                   /* Buffer for clause in text form. */
	auto binprefix *ptr4;                               /* Auxiliary pointer */
	auto subst Subst;                                   /* Auxiliary substitution */
	auto int32_t ii,jj,pp;                              /* Auxiliary */

	#ifdef DEBUGCODE
	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Allocate storage for auxiliary clause and copy */
	/* clause into the first auxiliary clause. */
	size=clause->part2.bin->size+binpsize;
	if ((auxcl=mallocmn(sizeof(cmprefix)))==NULL){
		return(NOMEMORY);
	}
	if ((auxcl->part2.bin=mallocmn(size))==NULL){
		MYFREE(auxcl);
		return(NOMEMORY);
	}
	memcpy(auxcl->part2.bin,clause->part2.bin,size);
	auxcl->part2.bin->clause=auxcl;

	/* Initialize substitution. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop through literals in auxiliary clause. */
	for (ptr1=&auxcl->part2.bin->formula[0],pp=jj=0;(*ptr1)!=UNITEND;
			ptr1=(ii?NextItem(ptr1,OVERSUBTERMS):ptr1),jj=(ii?jj+1:jj)){
		ii=1;

		/* Literal is a negated equality. */
		if ((NEGATED|EQUALITY)==((NEGATED|EQUALITY)&ptr1[0])){

			/* Check if one of the root equality terms is a variable. */
			/* If completeness is enabled then both root equality */
			/* root terms must be variables. */
			ptr2=NextItem(ptr1,IMMED);
			if (DEMODCPL!=kbset.opts.demodulation){
				if (VARIABLE&ptr2[0]){
					ptr5=NextItem(ptr2,OVERSUBTERMS);
				} else {
					ptr2=NextItem(ptr5=ptr2,OVERSUBTERMS);
					if (0==(VARIABLE&ptr2[0])){
						ptr5=NULL;
					}
				}
			} else {
				ptr5=NextItem(ptr2,OVERSUBTERMS);
				if ((0==(VARIABLE&ptr2[0]))||(0==(VARIABLE&ptr5[0]))){
					ptr5=NULL;
				}
			}

			/* One of the root equality terms is a variable. */
			if (ptr5!=NULL){

				/* Empty substitution. */
				if (Subst.buffer!=NULL){
					Subst.substsize=1;
					Subst.buffer[0]=UNITEND;
				}

				/* Check unification and build substitution. */
				switch (UnifyVar(ptr2,ptr5,&Subst,0)){

					/* Not enough memory. */
					case NOMEMORY:
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						if (auxcl->part2.bin->asserts!=NULL){
							MYFREE(auxcl->part2.bin->asserts);
						}
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						return(NOMEMORY);
						break;

					/* Unification is possible and we have a substitution. */
					case 1:

						/* Delete equality. */
						DeleteLiteral(auxcl->part2.bin,ptr1);

						/* Apply substitution. */
						if (NOMEMORY==ApplySubst2Clause(auxcl,&size,&Subst,0)){
							if (Subst.buffer!=NULL){
								MYFREE(Subst.buffer);
							}
							if (auxcl->part2.bin->asserts!=NULL){
								MYFREE(auxcl->part2.bin->asserts);
							}
							MYFREE(auxcl->part2.bin);
							MYFREE(auxcl);
							return(NOMEMORY);
						}

						/* Update ptr1. This is necessary because auxiliary clause */
						/* binprefix may have been reallocated by ApplySubst2Clause() */
						/* and/or the positions of the literals may have changed by */
						/* substitution application. */
						ptr1=GetLiteralPtr(auxcl,jj);

						/* Indicate there was a simplification and keep in the */
						/* same literal position in next iteration for ptr2. */
						pp=1;
						ii=0;
						break;
				}
			}
		}
	}

	/* Free substitution storage. */
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}

	/* There is a simplification. */
	if (pp){

		/* Set clause pointer and convert original clause to text form. */
		/* The original assertions are kept but their asymbol's must be */
		/* adjusted to the new clause. */
		if (NULL==(ptr4=reallocmn(auxcl->part2.bin,auxcl->part2.bin->size+binpsize))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->part2.bin=ptr4;
		ptr4->equflg=0;
		if (NULL==(ptr10=Decompile(clause->part2.bin))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->inference=DESTEQRES;
		auxcl->parent1=clause;
		auxcl->parent2=NULL;
		auxcl->flags=UNPROC;
		if (auxcl->part2.bin->asserts!=NULL){
			for (ptr1=&auxcl->part2.bin->asserts[0];ptr1[0]!=UNITEND;ptr1+=(1+sizeof(asymbol))){
				((asymbol *)&ptr1[1])->part2.binp=auxcl->part2.bin;
			}
		}
		*pnewcl=auxcl;
		MYFREE(clause->part2.bin);
		clause->part2.text=ptr10;

		/* Debug. */
		#ifdef DEBUGCODE
		ptr10=Decompile(auxcl->part2.bin);
		#ifdef VERBOSE
		if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
			printf("Destructive equality resolution simplification result [%ld]:\n%s\n",kbset.nformulas,ptr10);
		}
		#endif
		MYFREE(ptr10);
		#endif

	/* If there was not a removal then free auxiliary clause storage. */
	} else {
		MYFREE(auxcl->part2.bin);
		MYFREE(auxcl);
	}

	return(pp);
} /* DestrEqResolution */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if clause is subsumed by a KB)  OCJ
 *
 *    This function checks if the given clause in compiled or binary
 *    format is subsumed by some clause that exists in the indicated KB
 *    by calling IsSubsumedByClause() function for every candidate clause
 *    in the requested KB. It must also be specified the type of subsuming
 *    clauses that will be considered, either active or passive clauses.
 *
 *    Candidate clauses are selected by inspecting every literal in the
 *    argument clause and selecting clauses with a generalization literal.
 *    To prevent selection of the same candidate subsuming clauses by two
 *    different literals once a clause is selected it is signed with an
 *    8 bytes integer. If a clause is signed then it is not selected.
 *
 *    Once all candidate clauses corresponding to a given literal have
 *    been unsuccessfully tested then that literal is excluded from the
 *    IsSubsumedByClause() function call to speed the algorithm.
 *
 *    If the clause is subsumed then its locking assertions are set and
 *    the clause is locked if appropriate. However the clause is not
 *    deleted even if it is not locked. This must be done by the calling
 *    function.
 *
 *    See IsSubsumedByClause() function for details of clause subsumption
 *    condition.
 *
 *    This function must only be called for the working KB. It must NOT
 *    be called for main KB.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    type: PASSIVE if passive clauses are to be checked as subsuming
 *          clauses. Active clauses are always checked. Any other
 *          flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Clause is not subsumed by any clause of the indicated type
 *         in the indicated KB.
 *    1 -> Clause is subsumed by a clause of the indicated type in the
 *         indicated KB.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FwSubsuming(cmprefix *clause,int32_t type){
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto uint8_t *equality;                              /* Pointer to equality with root terms in reverse order */
	auto uint8_t *eqterms;                               /* Pointer to equality root terms in reverse order */
	auto uint8_t *queryitm;                              /* Pointer to QueryDscTree() query item */
	auto int32_t size;                                   /* Size of eqterms buffer */
	auto subst Subst;                                    /* Substitution for calls to IsSubsumedByClause() */
	auto uint64_t signature;                             /* KB signature. */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3,*ptr4,*ptr5;          /* Auxiliary pointers */
	auto int32_t ii,jj,kk,mm,pp,qq,rr,ss,nn;             /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Remember KB signature and update it. */
	signature=kbset.signature;
	kbset.signature++;

	/* Initialize substitution for IsSubsumedByClause() calls. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop through clause literals. */
	eqterms=equality=NULL;
	size=0;
	for (ptr1=&clause->part2.bin->formula[0],mm=0;ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS),mm++){

		/* Loop for equality symmetry management. */
		ptr4=NextItem(ptr1,IMMED);
		pp=(ptr1[0]&EQUALITY?2:1);
		for (qq=0;qq<pp;qq++){

			/* This is the first iteration. */
			if (qq==0){
				queryitm=ptr1;

			/* This is the second iteration for an equality literal. */
			/* Build the set of root terms in reverse order. */
			} else {
				ptr3=NextItem(ptr4,OVERSUBTERMS);
				rr=ptr3-ptr4;
				ss=NextItem(ptr3,OVERSUBTERMS)-ptr3;
				if (size==0){
					if (NULL==(equality=mallocmn(size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
						MYFREE(Subst.buffer);
						return(NOMEMORY);
					}
					eqterms=&equality[1+sizeof(symbol)];
				} else if (size<(2+sizeof(symbol)+rr+ss)){
					if (NULL==(ptr5=reallocmn(equality,size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
						MYFREE(Subst.buffer);
						MYFREE(equality);
						return(NOMEMORY);
					}
					equality=ptr5;
					eqterms=&equality[1+sizeof(symbol)];
				}
				memcpy(equality,ptr1,1+sizeof(symbol));
				memcpy(eqterms,ptr3,ss);
				memcpy(&eqterms[ss],ptr4,rr);
				eqterms[rr+ss]=UNITEND;
				queryitm=equality;
			}

			/* Loop through ACTIVE and PASSIVE clauses. ACTIVE clauses have */
			/* selected and non selected literals. */
			for (kk=ACTIVE;kk!=-1;
					kk=(kk==ACTIVE?ACTIVE|SELECTED:(kk==PASSIVE?-1:(type&PASSIVE?PASSIVE:-1)))){

				/* Get pointer to top tree node. */
				switch (kk){
					case ACTIVE:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[3];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[2];
						}
						break;
					case PASSIVE:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[5];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[4];
						}
						break;
					default:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[1];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[0];
						}
						break;
				}

				/* Loop through generalization candidates. */
				nn=0;
				do {

					/* Check timeout and solution by other process. */
					gettimeofday(&timestr,NULL);
					time=timestr.tv_sec-kbset.time.tv_sec
							+(timestr.tv_usec-kbset.time.tv_usec)/1000000.0;
					if ((time>=kbset.opts.timeout)||(procctl->status==UNSATISFIABLE)
							||(procctl->status==SATISFIABLE)){
						MYFREE(Subst.buffer);
						MYFREE(equality);
						return(TIMEOUT);
					}
					if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
						MYFREE(Subst.buffer);
						MYFREE(equality);
						return(procctl->status);
					}

					/* Get a candidate subsuming literal. */
					if (0!=(ii=QueryDscTree(queryitm,&ptr2,ptr3,GENERALIZATION,nn))){

						/* Get pointer to candidate binprefix structure. */
						nn=1;
						binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

						/* Debug. */
						#ifdef DEBUGCODE
						ptr10=Decompile(binp);
						MYFREE(ptr10);
						#endif

						/* Check if candidate clause has been already processed. */
						if (binp->signature!=signature){

							/* Check subsumption. */
							switch (jj=IsSubsumedByClause(ptr1,clause->part2.bin->literals-mm,
									binp,&Subst,NULL)){

								/* No memory. */
								case NOMEMORY:
									MYFREE(Subst.buffer);
									MYFREE(equality);
									return(NOMEMORY);
									break;

								/* Clause is not subsumed. Sign candidate to prevent checking */
								/* it again. */
								case 0:
									binp->signature=signature;
									break;

								/* Clause is subsumed. */
								default:

									/* Get locking assertions. */
									clause->flags&=~(TEXTFORMULA|UNPROC|PASSIVE|ACTIVE|LOCKED);
									if (NOMEMORY==GetLocksAsserts(binp,clause->part2.bin,NULL)){
										MYFREE(Subst.buffer);
										MYFREE(equality);
										return(NOMEMORY);
									}

									/* Free memory and return. */
									MYFREE(Subst.buffer);
									MYFREE(equality);
									return(jj);
									break;
							}
						}
					}
				} while(ii);
			}
		}
	}

	/* Free memory and return clause not subsumed. */
	MYFREE(Subst.buffer);
	MYFREE(equality);
	return(0);
} /* FwSubsuming */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform backward subsumption inference)  OCJ
 *
 *    This function checks if the given clause in compiled or binary
 *    format subsumes one or more clauses that exist in the indicated KB
 *    by calling IsSubsumedByClause() function for every candidate clause
 *    in the requested KB. It must also be specified the type of subsuming
 *    clauses that will be considered, either active or passive clauses.
 *
 *    Candidate clauses are selected by inspecting the first literal in the
 *    argument clause and selecting clauses with an instantiation literal.
 *
 *    Subsumed clauses are either locked and have their locking assertions
 *    set or they are added to the disposal queue but not removed from
 *    their current queue.
 *
 *    See IsSubsumedByClause() function for details of clause subsumption
 *    condition.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    type: PASSIVE if passive clauses are to be checked as subsumed
 *          clauses. Active clauses are always checked. Any other
 *          flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Process was ok.
 *    1 -> Process performed and an empty clause without assertions was
 *         generated. This function never returns 1 but this is
 *         documented here as a possible return code from functions
 *         called by BckSimplify() function.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t BckSubsuming(cmprefix *clause,int32_t type){
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto uint8_t *equality;                              /* Pointer to equality with root terms in reverse order */
	auto uint8_t *eqterms;                               /* Pointer to equality root terms in reverse order */
	auto uint8_t *queryitm;                              /* Pointer to QueryDscTree() query item */
	auto int32_t size;                                   /* Size of eqterms buffer */
	auto subst Subst;                                    /* Substitution for calls to IsSubsumedByClause() */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3,*ptr4,*ptr5;          /* Auxiliary pointers */
	auto int32_t ii,jj,kk,nn,pp,qq,rr,ss;                /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Initialize substitution for IsSubsumedByClause() calls. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop for equality symmetry management. */
	ptr1=&clause->part2.bin->formula[0];
	ptr4=NextItem(ptr1,IMMED);
	pp=(ptr1[0]&EQUALITY?2:1);
	eqterms=equality=NULL;
	size=0;
	for (qq=0;qq<pp;qq++){

		/* This is the first iteration. */
		if (qq==0){
			queryitm=ptr1;

		/* This is the second iteration for an equality literal. */
		/* Build the set of root terms in reverse order. */
		} else {
			ptr3=NextItem(ptr4,OVERSUBTERMS);
			rr=ptr3-ptr4;
			ss=NextItem(ptr3,OVERSUBTERMS)-ptr3;
			if (size==0){
				if (NULL==(equality=mallocmn(size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
					MYFREE(Subst.buffer);
					return(NOMEMORY);
				}
				eqterms=&equality[1+sizeof(symbol)];
			} else if (size<(2+sizeof(symbol)+rr+ss)){
				if (NULL==(ptr5=reallocmn(equality,size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
					MYFREE(Subst.buffer);
					MYFREE(equality);
					return(NOMEMORY);
				}
				equality=ptr5;
				eqterms=&equality[1+sizeof(symbol)];
			}
			memcpy(equality,ptr1,1+sizeof(symbol));
			memcpy(eqterms,ptr3,ss);
			memcpy(&eqterms[ss],ptr4,rr);
			eqterms[rr+ss]=UNITEND;
			queryitm=equality;
		}

		/* Loop through ACTIVE and PASSIVE clauses. ACTIVE clauses have */
		/* selected and non selected literals. */
		for (kk=ACTIVE;kk!=-1;
				kk=(kk==ACTIVE?ACTIVE|SELECTED:(kk==PASSIVE?-1:(type&PASSIVE?PASSIVE:-1)))){

			/* Get pointer to top tree node. */
			switch (kk){
				case ACTIVE:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr[3];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr[2];
					}
					break;
				case PASSIVE:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr[5];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr[4];
					}
					break;
				default:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr[1];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr[0];
					}
					break;
			}

			/* Loop through instantiation candidates. */
			nn=0;
			do {

				/* Check timeout and solution by other process. */
				gettimeofday(&timestr,NULL);
				time=timestr.tv_sec-kbset.time.tv_sec
						+(timestr.tv_usec-kbset.time.tv_usec)/1000000.0;
				if ((time>=kbset.opts.timeout)||(procctl->status==UNSATISFIABLE)
						||(procctl->status==SATISFIABLE)){
					MYFREE(Subst.buffer);
					MYFREE(equality);
					return(TIMEOUT);
				}
				if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
					MYFREE(Subst.buffer);
					MYFREE(equality);
					return(procctl->status);
				}

				/* Get a candidate subsumed literal. */
				if (0!=(ii=QueryDscTree(queryitm,&ptr2,ptr3,INSTANCE,nn))){

					/* Get pointer to candidate binprefix structure. */
					nn=1;
					binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

					/* Debug. */
					#ifdef DEBUGCODE
					ptr10=Decompile(binp);
					MYFREE(ptr10);
					#endif

					/* Clause is not in disposal queue therefore */
					/* it has not been already simplified. */
					if (0==(binp->clause->flags&INDISPOSALQUEUE)){

						/* Check subsumption. */
						switch (jj=IsSubsumedByClause(&binp->formula[0],binp->literals,
								clause->part2.bin,&Subst,NULL)){

							/* Not enough memory. */
							case NOMEMORY:
								MYFREE(Subst.buffer);
								MYFREE(equality);
								return(NOMEMORY);
								break;

							/* Candidate clause is subsumed. */
							case 1:

								/* Get locking assertions. */
								if (NOMEMORY==GetLocksAsserts(clause->part2.bin,binp,NULL)){
									MYFREE(Subst.buffer);
									MYFREE(equality);
									return(NOMEMORY);
								}

								/* Add subsumed clause to disposal queue for later deletion or */
								/* locking. This queue interconnects clauses from other queues. */
								/* The clause must not be removed from any queue to which it is */
								/* linked. */
								binp->clause->flags|=INDISPOSALQUEUE;
								binp->clause->prevdisp=kbset.lastdisp;
								kbset.lastdisp=binp->clause;
								kbset.prstats.bcksubsum++;
								kbset.lrscnt++;

								/* Debug. */
								#ifdef DEBUGCODE
								ptr10=Decompile(binp);
								#ifdef VERBOSE
								if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
									printf("Clause backward subsumed:\n%s\n",ptr10);
								}
								#endif
								MYFREE(ptr10);
								#endif
								break;
						}
					}
				}
			} while(ii);
		}
	}

	/* Free memory and return. */
	MYFREE(Subst.buffer);
	MYFREE(equality);
	return(0);
} /* BckSubsuming */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if clause is subsumed by another clause)  OCJ
 *
 *    This function checks if clause of formula1 in compiled binary format
 *    is subsumed by clause in formula2 also in compiled binary format.
 *    The clause in formula1 is subsumed by the clause in binp2 if there
 *    exist a substitution such that applied to binp2 and adding 0 or more
 *    literals then formula1 is obtained. This means that clause in binp2 is
 *    both more general and simpler than formula1 because it is true whenever
 *    formula in binp2 is true and it has the same or fewer literals. The
 *    substitution is only applied to clause in binp2 so the variables substituted
 *    must all belong to clause in binp2.
 *
 *    The substitution may be a complex one formed by a set
 *    of single substitutions.
 *
 *    This function doesn't consider assertions at all. This must be done
 *    by the calling function when applicable.
 *
 *    The process is as follows:
 *    - This function computes data related to which literals of the candidate
 *      subsumed clause are instances of each literal of the candidate subsuming
 *      clause.
 *    - Any of the following conditions produce an early return as
 *      with result of no subsuming:
 *      - There are more literals in formula2 than in formula1.
 *      - Two different literals in binp2 have the same single instance literal
 *        in formula1.
 *      - There is a literal in binp2 without any assignable candidate
 *        instance literals.
 *      - There are more literals binp2 than instance literals in formula1.
 *    - The function calls LiteralSubsuming() function and returns the value
 *      obtained.
 *
 *    This algorithm is potentially exponential in execution time. To avoid
 *    hang ups a timeout condition checking is introduced. In case of timeout
 *    it is assumed that the clause is not subsumed. This will affect a little
 *    bit the program performance but it will prevent hang ups and I think that
 *    it preserves completeness. This timeout condition must not be confused
 *    with the global process timeout condition.
 *
 *    IMPORTANT: To ensure that this function works properly the candidate
 *    subsuming clause must not have duplicate literals because if there is a
 *    single instance literal for both duplicate literals this function will be
 *    unable to assign it to both duplicate literals. This is not a problem
 *    if the clause is in the KB as it has been properly simplified already.
 *    However, for cases like equivalent factor simplifications the removal of
 *    duplicate literals must be done in the calling function. For instance,
 *    when equivalent factoring is applied to the clause:
 *      A(x,y)|A(u,v)|B(C,D)|B(x,y)|B(u,v)
 *    where x, y, u, v are variables then a candidate equivalent factor is
 *      A(u,v)|B(C,D)|B(u,v)|B(u,v)
 *    This function cannot assign both duplicate literals B(u,v) to the
 *    instance literal B(x,y) or B(u,v), therefore failing to identify
 *    a subsumption.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    formula1: Pointer to formula in compiled binary form as it is in the
 *              formula field of binprefix structure. This formula will be
 *              checked for being subsumed by binp2. It points to a literal
 *              in the clause that is being checked for being subsumed, but
 *              not necessarily to the first literal. This is useful for
 *              forward subsumption, when the first literals of candidate
 *              subsumed clause have been already checked and found that
 *              they cannot participate in the subsumption.
 *    lits1: Number of literals starting at formula1 pointer.
 *    binp2: Pointer to binprefix structure of clause in compiled binary. This
 *           clause will be checked for subsuming formula1.
 *    Subst: Pointer to a substitution structure. This structure is
 *           kept between calls to this function in order to save
 *           allocation and freeing operations of substitution buffer.
 *    plit: On entry this is either NULL or the address of a pointer to a
 *          literal in binp2. If it is not NULL on entry and subsumption is
 *          successful then on exit it will contain the address of a pointer to
 *          the literal in formula1 that is an instance of the literal in binp2.
 *          If subsumption is not successful then the content on exit is
 *          unpredictable.
 *
 *  RETURNS:
 *
 *    0 -> Clause is not subsumed or a function timeout condition occurred.
 *    1 -> Clause is subsumed.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t IsSubsumedByClause(uint8_t *formula1,int32_t lits1,binprefix *binp2,
		subst *Subst,uint8_t **plit){
	auto subsumdata ssdata;                              /* Subsuming data structure. */
	struct timeval sttime;                               /* Function starting time */
	auto uint8_t *ptr1,*ptr2;                            /* Auxiliary pointers */
	auto int32_t ii,jj,nn;                               /* Auxiliary */

	/* Check the number of literals. This is not strictly necessary */
	/* unless factoring is on, as then the factor would always be */
	/* subsumed by the parent. However it helps simplify clauses */
	/* and speeds up the program. */
	if (binp2->literals>lits1){
		return(0);
	}

	/* There is only one literal of each clause involved in the subsumption. */
	/* Just check for a valid generalization. */
	if (lits1==1){
		if (plit!=NULL){
			*plit=formula1;
		}
		if (EQUALITY&binp2->formula[0]){
			if (0!=(ii=Generalize(&binp2->formula[0],formula1,Subst,INITSUBST))){
				return(ii);
			}
			return(SymEquGeneralize(&binp2->formula[0],formula1,Subst,INITSUBST));
		} else {
			return(Generalize(&binp2->formula[0],formula1,Subst,INITSUBST));
		}
	}

	/* Allocate subsumdata buffers. */
	if (NULL==(ssdata.pinst=mallocmn((lits1*sizeof(int32_t))+(binp2->literals*((sizeof(int32_t)*(1+lits1))
			+(sizeof(void *)*(lits1+2))))))){
		return(NOMEMORY);
	}

	/* Get start time. */
	gettimeofday(&sttime,NULL);

	/* Initialize basic data in subsumdata structure. */
	ssdata.lits1=lits1;
	ssdata.lits2=binp2->literals;
	ssdata.pinst[0]=(uint8_t **)&ssdata.pinst[binp2->literals];
	for (ii=1;ii<binp2->literals;ii++){
		ssdata.pinst[ii]=&ssdata.pinst[ii-1][lits1];
	}
	ssdata.instnb=(int32_t **)&ssdata.pinst[binp2->literals-1][lits1];
	ssdata.instnb[0]=(int32_t *)&ssdata.instnb[binp2->literals];
	for (ii=1;ii<binp2->literals;ii++){
		ssdata.instnb[ii]=&ssdata.instnb[ii-1][lits1];
	}
	ssdata.nbofinst=(int32_t *)&ssdata.instnb[binp2->literals-1][lits1];
	ssdata.usedinst=&ssdata.nbofinst[binp2->literals];

	/* Set to not assigned all literals in the candidate subsumed clause. */
	memset(ssdata.usedinst,0,lits1*sizeof(int32_t));

	/* Fill data in buffers of subsumdata structure. */
	/* Loop through literals of candidate subsuming clause. */
	for (ptr1=&binp2->formula[0],ii=0;ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS),ii++){

		/* Initialize number of instance literals for this candidate subsuming literal. */
		ssdata.nbofinst[ii]=0;

		/* Loop through literals of candidate subsumed formula. */
		for (ptr2=formula1,jj=0;ptr2[0]!=UNITEND;ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){

			/* Check if ptr1 is a generalization of ptr2. */
			if (0==(nn=Generalize(ptr1,ptr2,Subst,INITSUBST))){
				if (ptr1[0]&EQUALITY){
					nn=SymEquGeneralize(ptr1,ptr2,Subst,INITSUBST);
				}
			}
			switch (nn){

				/* Not enough memory. */
				case NOMEMORY:
					MYFREE(ssdata.pinst);
					return(NOMEMORY);
					break;

				/* Valid generalization. Set subsuming data. */
				case 1:
					ssdata.pinst[ii][ssdata.nbofinst[ii]]=ptr2;
					ssdata.instnb[ii][ssdata.nbofinst[ii]]=jj;
					ssdata.nbofinst[ii]++;
					break;
			}
		}

		/* Return no subsumption if there are no candidate instance literals */
		/* for this literal. */
		if (ssdata.nbofinst[ii]==0){
			MYFREE(ssdata.pinst);
			return(0);
		}
	}

	/* Initialize substitution if necessary. */
	if (Subst->buffer!=NULL){
		Subst->substsize=1;
		Subst->buffer[0]=UNITEND;
	} else {
		Subst->substsize=0;
	}

	/* Call LiteralSubsuming() function. */
	ii=LiteralSubsuming(&binp2->formula[0],&binp2->formula[0],0,&ssdata,Subst,&sttime,plit);
	if (ii==TIMEOUT){
		ii=0;
	}

	/* Free memory and return. */
	MYFREE(ssdata.pinst);
	return(ii);
} /* IsSubsumedByClause */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check literal subsuming)  OCJ
 *
 *    This function checks if there is a valid corresponding literal in
 *    candidate subsumed formula that can be subsumed by the passed literal
 *    in candidate subsuming formula that allows candidate subsuming formula
 *    to subsume candidate subsumed formula.
 *
 *    This function is called recursively.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    formula2: Pointer to formula in compiled binary form as it is in the
 *              formula field of binprefix structure. This formula will be
 *              checked for subsuming clause1.
 *    literal: Pointer to literal in candidate subsuming formula to be checked.
 *    litnb: Literal number of literal in candidate subsuming formula to be
 *           checked.
 *    pssdata: Pointer to structure with subsuming data.
 *    Subst: Pointer to a substitution structure. This structure is
 *           kept between calls to this function in order to save
 *           allocation and freeing operations of substitution buffer.
 *    sttime: Structure with the function starting time to check function
 *            timeouts. See IsSubsumedByClause() global comments for details.
 *    plit: On entry this is either NULL or the address of a pointer to a
 *          literal in formula2. If it is not NULL on entry and subsumption
 *          is successful then on exit it will contain the address of a pointer
 *          to the literal in subsumed clause that is an instance of the literal
 *          in binp2. If subsumption is not successful then the content on exit
 *          is unpredictable.
 *
 *  RETURNS:
 *
 *    0 -> There is no valid corresponding literal that allows subsuming.
 *    1 -> There is a valid corresponding literal that allows subsuming.
 *    NOMEMORY -> Not enough memory.
 *    TIMEOUT -> Function timeout condition
 *
 *
 *--------------------------------------------------------------*/
int32_t LiteralSubsuming(uint8_t *formula2,uint8_t *literal,int32_t litnb,subsumdata *pssdata,
		subst *Subst,struct timeval *sttime,uint8_t **plit){
	auto struct timeval endtime;                         /* Function current time */
	auto double timedif;                                 /* Function elapsed time */
	static uint8_t *literal2;                            /* Value of *plit on entry */
	auto int32_t ii,jj,kk,mm,nn,pp;                      /* Auxiliary */

	/* If no more literals then return valid subsumption. */
	if (literal[0]==UNITEND){
		return(1);
	}

	/* If this is the first call then remember the value of *plit. */
	if ((formula2==literal)){
		if (plit!=NULL){
			literal2=*plit;
		} else {
			literal2=NULL;
		}
	}

	/* Remember substitution status. */
	if (Subst->buffer!=NULL){
		kk=Subst->substsize;
	} else {
		kk=-1;
	}

	/* Loop through candidate instance literals for the given literal. */
	for (ii=0;ii<pssdata->nbofinst[litnb];ii++){

		/* Candidate instance literal is not yet assigned. */
		if (0==pssdata->usedinst[pssdata->instnb[litnb][ii]]){

			/* Equality symmetry loop. */
			mm=(EQUALITY&literal[0]?2:1);
			for (pp=0;pp<mm;pp++){

				/* Check valid generalization. */
				if (pp==1){
					nn=SymEquGeneralize(literal,pssdata->pinst[litnb][ii],Subst,0);
				} else {
					nn=Generalize(literal,pssdata->pinst[litnb][ii],Subst,0);
				}
				switch (nn){

					/* Not enough memory. */
					case NOMEMORY:
						return(NOMEMORY);
						break;

					/* Valid generalization. */
					case 1:

						/* Set instance literal as assigned. */
						pssdata->usedinst[pssdata->instnb[litnb][ii]]=1;

						/* If literal is literal2 then update *plit with the pairing literal */
						/* in the subsumed clause. */
						if (literal==literal2){
							*plit=pssdata->pinst[litnb][ii];
						}

						/* Make recursive call. */
						jj=LiteralSubsuming(formula2,NextItem(literal,OVERSUBTERMS),
								litnb+1,pssdata,Subst,sttime,plit);
						if (0!=jj){
							return(jj);
						}

						/* Set instance literal as unassigned. */
						pssdata->usedinst[pssdata->instnb[litnb][ii]]=0;
						break;
				}

				/* Reset substitution to its previous state. */
				if (kk>0){
					Subst->substsize=kk;
					Subst->buffer[kk-1]=UNITEND;
				} else if (Subst->buffer!=NULL){
					Subst->substsize=1;
					Subst->buffer[0]=UNITEND;
				} else {
					Subst->substsize=0;
				}

				/* Check timeout. */
				gettimeofday(&endtime,NULL);
				timedif=endtime.tv_sec-sttime->tv_sec
						+(endtime.tv_usec-sttime->tv_usec)/1000000.0;
				if (timedif>0.0185){
					return(TIMEOUT);
				}
			}
		}
	}

	/* Return no subsumption. */
	return(0);
} /* LiteralSubsuming */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Forward and backward simplifications)  OCJ
 *
 *    This function performs the forward simplifications of a clause
 *    in the working KB. This function also checks for getting an empty
 *    clause without assertions. If empty clause without assertions is
 *    detected then it is added to the UNPROC queue.
 *
 *    The input clause must be unlinked form the owner KB.
 *
 *    This function just calls the appropriate functions that perform
 *    each simplification inference. Any required memory freeing is
 *    performed by this function or the functions called by it, even
 *    in the case of error and also for the memory allocated to *pclause
 *    by the caller.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    pclause: Address of pointer to cmprefix structure of clause.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          FROMPASSIVE if the clause is coming from PASSIVE queue and
 *          therefore some simplifications must not be performed.
 *
 *  RETURNS:
 *
 *    0 -> Process performed and no empty clause without assertions inferred.
 *    1 -> Process performed and an empty clause without assertions was inferred.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FwdSimplify(cmprefix **pclause,int32_t type){
	auto int32_t ii;                                     /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;

	/* Debug. */
	ptr10=Decompile((*pclause)->part2.bin);
	#ifdef VERBOSE
	if (active_cores==1){
		printf("===> Starting forward simplifications to clause:\n%s\n",ptr10);
	}
	#endif
	MYFREE(ptr10);
	#endif

	#if DEMODPOSITION == 1
	/* Perform demodulations if there are equalities. */
	/* An initial tautology detection is done before */
	/* the demodulation loop. */
	if (((UNITEQUALITY|NONUNITEQUALITY)&pbmflags)&&(kbset.opts.demodulation!=DEMODOFF)){

		/* Check tautology. */
		if ((0==(SIMPLIFIED&(*pclause)->flags))&&(IsTautology(*pclause))){
			kbset.prstats.tautologies++;
			FreeEquData((*pclause)->part2.bin);
			if ((*pclause)->part2.bin->asserts!=NULL){
				MYFREE((*pclause)->part2.bin->asserts);
			}
			MYFREE((*pclause)->part2.bin);
			MYFREE(*pclause);
			*pclause=NULL;
			return(0);
		}

		/* Demodulation loop. */
		for (ii=1;ii!=0;){
			switch (ii=FwdDemodulation(pclause,type)){
				case TIMEOUT:
				case NOMEMORY:
					return(ii);
					break;
				case 1:
					type&=(~FROMPASSIVE);
					break;
				case 2:
					*pclause=NULL;
					return(0);
					break;
			}
		}
	}
	#endif

	/* Perform initial forward simplifications. If there is a NOMEMORY */
	/* error or if the clause is deleted then memory freeing is */
	/* performed by FirstSimplify() function, not here. */
	switch (ii=FirstSimplify(pclause,type)){

		/* Not enough memory or timeout. */
		case NOMEMORY:
		case TIMEOUT:
			return(ii);
			break;

		/* Clause has been deleted or subsumed. */
		case 2:
		case 3:
			*pclause=NULL;
			return(0);
			break;

		/* Clause has been locked. */
		case 4:
			return(0);
			break;
	}

	/* Check empty clause. This can be achieved by trivial equality resolution */
	/* and destructive equality resolution in FirstSimplify() function. */
	if ((*pclause)->part2.bin->formula[0]==UNITEND){

		/* Empty clause has not assertions. */
		if ((*pclause)->part2.bin->asserts==NULL){
			(*pclause)->flags=UNPROC;
			if (NOMEMORY==AddBinClause2KB(*pclause,&kbset,SETFNUMBER)){
				FreeEquData((*pclause)->part2.bin);
				if ((*pclause)->part2.bin->asserts!=NULL){
					MYFREE((*pclause)->part2.bin->asserts);
				}
				MYFREE((*pclause)->part2.bin);
				MYFREE(*pclause);
				return(NOMEMORY);
			}
			if (NULL==(kbset.frstprfnode=mallocmn(sizeof(proofnode)))){
				return(NOMEMORY);
			}
			kbset.frstprfnode->type=ACLAUSE;
			kbset.frstprfnode->ptr.Aclause=*pclause;
			kbset.status=UNSATISFIABLE;
			return(1);
		}

		/* Empty clause has assertions. Return as an empty clause cannot */
		/* be further simplified. */
		return(0);
	}

	#ifdef UNITCLRESOLUTION
	/* Perform unit clause resolution simplification. This is done only */
	/* for unit clauses without assertions. */
	if (((*pclause)->part2.bin->literals==1)&&((*pclause)->part2.bin->asserts==NULL)){

		/* Perform call and check NOMEMORY and TIMEOUT. */
		switch (ii=UCResolution(*pclause)){
			case NOMEMORY:
			case TIMEOUT:
				return(ii);
				break;
		}

		/* Unit clause resolution was successful. */
		if (ii){
			return(1);
		}
	}
	#endif

	#if DEMODPOSITION == 2
	/* Perform demodulations if there are equalities. */
	if (((UNITEQUALITY|NONUNITEQUALITY)&pbmflags)&&(kbset.opts.demodulation!=DEMODOFF)){
		for (ii=1;ii!=0;){
			switch (ii=FwdDemodulation(pclause,type)){
				case TIMEOUT:
				case NOMEMORY:
					return(ii);
					break;
				case 2:
					*pclause=NULL;
					return(0);
					break;
			}
		}
	}
	#endif

	/* Perform additional forward simplification. If there is a NOMEMORY */
	/* error or if the clause is deleted then memory freeing is */
	/* performed by FwdSimplify() function, not here. */
	switch (ii=SecondSimplify(pclause,type)){
		case NOMEMORY:
		case TIMEOUT:
			return(ii);
			break;
	}

	/* Check empty clause without assertions. This can be achieved by subsumption */
	/* resolution in SecondSimplify() function. */
	if ((*pclause)->part2.bin->formula[0]==UNITEND){

		/* Empty clause has not assertions. */
		if ((*pclause)->part2.bin->asserts==NULL){
			(*pclause)->flags=UNPROC;
			if (NOMEMORY==AddBinClause2KB(*pclause,&kbset,SETFNUMBER)){
				FreeEquData((*pclause)->part2.bin);
				if ((*pclause)->part2.bin->asserts!=NULL){
					MYFREE((*pclause)->part2.bin->asserts);
				}
				MYFREE((*pclause)->part2.bin);
				MYFREE(*pclause);
				return(NOMEMORY);
			}
			if (NULL==(kbset.frstprfnode=mallocmn(sizeof(proofnode)))){
				return(NOMEMORY);
			}
			kbset.frstprfnode->type=ACLAUSE;
			kbset.frstprfnode->ptr.Aclause=*pclause;
			kbset.status=UNSATISFIABLE;
			return(1);
		}

		/* Empty clause has assertions. Return as an empty clause cannot */
		/* be further simplified. */
		return(0);
	}

	#if DEMODPOSITION == 3
	/* Perform demodulations if there are equalities. */
	if (((UNITEQUALITY|NONUNITEQUALITY)&pbmflags)&&(kbset.opts.demodulation!=DEMODOFF)){
		for (ii=1;ii!=0;){
			switch (ii=FwdDemodulation(pclause,type)){
				case TIMEOUT:
				case NOMEMORY:
					return(ii);
					break;
				case 2:
					*pclause=NULL;
					return(0);
					break;
			}
		}
	}
	#endif

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile((*pclause)->part2.bin);
	MYFREE(ptr10);
	#endif

	return(0);
} /* FwdSimplify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (First clause simplification)  OCJ
 *
 *    This function performs the following simplifications in a clause:
 *    - Simplifications that detect clause redundancy, leading to
 *      clause deletion: tautology detection and forward subsumption.
 *    - Simplifications that don't involve any other external clause:
 *      remove duplicate literals, trivial equality resolution and
 *      destructive equality resolution.
 *
 *    If the clause is deleted then its storage is freed. The clause
 *    storage is also freed in case of a NOMEMORY error condition.
 *
 *    This function just calls the appropriate functions that perform
 *    each simplification inference. Any required memory freeing is
 *    performed by this function, even in the case of error and also
 *    for the memory allocated to *pclause by the caller.
 *
 *    With AVATAR a clause may go through the simplification process
 *    more than once. Except for clause subsumptions all simplifications
 *    called from this function except FwSubsuming() don't need to be
 *    performed for clauses that have been simplified in the past. See
 *    SIMPLIFIED flag in global.h file that is used for this.
 *
 *    IMPORTANT: This function must never be called for main KB.
 *
 *
 *  ARGUMENTS:
 *
 *    pclause: Address of pointer to cmprefix structure of clause.
 *    flags: PASSIVE if passive clauses are to be checked for
 *           simplifications. Active clauses are always checked.
 *          FROMPASSIVE if the clause is coming from PASSIVE queue and
 *          therefore some simplifications must not be performed.
 *
 *  RETURNS:
 *
 *    0 -> No simplifications were done.
 *    1 -> Some simplifications were done.
 *    2 -> Clause was deleted because it is a tautology.
 *    3 -> Clause was subsumed.
 *    4 -> Clause was locked because it was subsumed.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FirstSimplify(cmprefix **pclause,int32_t flags){
	auto cmprefix *smpclauses[4];                        /* Pointers to simplified clauses */
	auto int32_t smpsize;                                /* Number of elements in smpclauses set */
	auto char *ptr10;                                    /* Auxiliary pointer */
	auto int32_t ii,jj;                                  /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr11;                                    /* Auxiliary pointer */
	#endif

	/* Initialize the set of simplified clauses. The last element with */
	/* index [smpsize-1] contains the last result of a simplification */
	/* in binary format. The other elements contain simplified clauses */
	/* in text format. */
	smpsize=1;
	smpclauses[0]=*pclause;

	/* Simplifications done only if clause is not coming from PASSIVE */
	/* and clause has not been previously simplified. */
	if ((0==(SIMPLIFIED&(*pclause)->flags))&&(0==(flags&FROMPASSIVE))){

		/* Remove duplicate literals. */
		switch (RemoveDuplicates(*pclause,&smpclauses[1],1)){
			case NOMEMORY:
				FreeEquData((*pclause)->part2.bin);
				if ((*pclause)->part2.bin->asserts!=NULL){
					MYFREE((*pclause)->part2.bin->asserts);
				}
				MYFREE((*pclause)->part2.bin);
				MYFREE(*pclause);
				return(NOMEMORY);
				break;
			case 1:
				smpsize=2;
				kbset.prstats.removedups++;
				break;
		}

		/* Debug. */
		#ifdef DEBUGCODE
		ptr11=Decompile(smpclauses[smpsize-1]->part2.bin);
		MYFREE(ptr11);
		#endif

		/* There are not equalities, check if clause is a tautology. */
		/* This is done here only if there are not equalities. Otherwise */
		/* it is done later (see below).*/
		if (0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags)){
			if (IsTautology(smpclauses[smpsize-1])){
				kbset.prstats.tautologies++;
				for (ii=0;ii<(smpsize-1);ii++){
					MYFREE(smpclauses[ii]->part2.text);
					MYFREE(smpclauses[ii]);
				}
				FreeEquData(smpclauses[smpsize-1]->part2.bin);
				if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
					MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
				}
				MYFREE(smpclauses[smpsize-1]->part2.bin);
				MYFREE(smpclauses[smpsize-1]);
				return(2);
			}

		/* There are equalities. */
		} else {

			/* Trivial equality resolution. */
			switch (TrivEqResolution(smpclauses[smpsize-1],&smpclauses[smpsize])){

				/* Not enough memory. */
				case NOMEMORY:
					for (ii=0;ii<(smpsize-1);ii++){
						MYFREE(smpclauses[ii]->part2.text);
						MYFREE(smpclauses[ii]);
					}
					FreeEquData(smpclauses[smpsize-1]->part2.bin);
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(NOMEMORY);
					break;

				/* Clause was simplified. If the result is an empty clause */
				/* then add text clauses to KB, set pclause to simplified */
				/* clause and return. */
				case 1:
					smpsize++;
					kbset.prstats.triveqres++;
					if (smpclauses[smpsize-1]->part2.bin->formula[0]==UNITEND){
						for (ii=0;ii<(smpsize-1);ii++){
							AddTxtFormula2KB(smpclauses[ii],smpclauses[ii]->inference,&kbset,SETFNUMBER);
						}
						*pclause=smpclauses[smpsize-1];
						return(1);
					}
					break;
			}

			/* Destructive equality resolution. */
			switch (jj=DestrEqResolution(smpclauses[smpsize-1],&smpclauses[smpsize])){

				/* Not enough memory. */
				case NOMEMORY:
					for (ii=0;ii<(smpsize-1);ii++){
						MYFREE(smpclauses[ii]->part2.text);
						MYFREE(smpclauses[ii]);
					}
					FreeEquData(smpclauses[smpsize-1]->part2.bin);
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(NOMEMORY);
					break;

				/* Clause was simplified. If the result is an empty clause */
				/* then add text clauses to KB, set pclause to simplified */
				/* clause and return. */
				case 1:
					smpsize++;
					kbset.prstats.desteqres++;
					if (smpclauses[smpsize-1]->part2.bin->formula[0]==UNITEND){
						for (ii=0;ii<(smpsize-1);ii++){
							AddTxtFormula2KB(smpclauses[ii],smpclauses[ii]->inference,&kbset,SETFNUMBER);
						}
						*pclause=smpclauses[smpsize-1];
						return(1);
					}
					break;
			}

			/* If clause was simplified by destructive equality resolution then check */
			/* if clause is a tautology. This is done here because destructive */
			/* equality resolution may produce a tautology. For instance, the clause */
			/* ~X0=X1|~X2=X3|vapp(X0,X2)=vapp(X1,X3) produces the clause */
			/* vapp(X0,X2)=vapp(X0,X2) that is a tautology. */
			if ((jj)||(DEMODPOSITION!=1)){
				if (IsTautology(smpclauses[smpsize-1])){
					kbset.prstats.tautologies++;
					for (ii=0;ii<(smpsize-1);ii++){
						MYFREE(smpclauses[ii]->part2.text);
						MYFREE(smpclauses[ii]);
					}
					FreeEquData(smpclauses[smpsize-1]->part2.bin);
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(2);
				}
			}
		}
	}

	/* Check clause subsumption. This is done even if the clause has been */
	/* previously simplified because there may be new causes that could */
	/* subsume the current clause. */
	switch (jj=FwSubsuming(smpclauses[smpsize-1],flags)){

		/* Not enough memory. */
		case NOMEMORY:
		case TIMEOUT:
			for (ii=0;ii<(smpsize-1);ii++){
				MYFREE(smpclauses[ii]->part2.text);
				MYFREE(smpclauses[ii]);
			}
			FreeEquData(smpclauses[smpsize-1]->part2.bin);
			if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
				MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
			}
			MYFREE(smpclauses[smpsize-1]->part2.bin);
			MYFREE(smpclauses[smpsize-1]);
			return(jj);
			break;

		/* Clause is not subsumed. */
		case 0:

			/* Debug. */
			#ifdef DEBUGCODE
			#ifdef VERBOSE
			if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
				for (ii=1;ii<(smpsize-1);ii++){
					printf("First simplification result:\n%s\n",smpclauses[ii]->part2.text);
				}
			}
			#endif
			if (smpsize>1){
				ptr10=Decompile(smpclauses[smpsize-1]->part2.bin);
				#ifdef VERBOSE
				if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
					printf("First simplification result:\n%s\n",ptr10);
				}
				#endif
				MYFREE(ptr10);
			}
			#endif

			/* Add text clauses to KB and set pclause to simplified clause. */
			for (ii=0;ii<(smpsize-1);ii++){
				AddTxtFormula2KB(smpclauses[ii],smpclauses[ii]->inference,&kbset,SETFNUMBER);
			}
			*pclause=smpclauses[smpsize-1];
			break;

		/* Clause is subsumed. */
		case 1:
			#if defined(DEBUGCODE) && defined(VERBOSE)
			if ((active_cores==1)&&(procctl->status==RUNNING)&&(verbfrom<=kbset.nformulas)
					&&(verbto>=kbset.nformulas)){
				printf("Inferred clause was subsumed\n");
			}
			#endif

			/* Count subsumption. */
			kbset.prstats.fwdsubsum++;

			/* Add text clauses to KB. This is always done because due to AVATAR */
			/* successive lock and unlock cycles a clause can be subsumed after */
			/* generating a simplified successor an so be part of the proof queue. */
			for (ii=0;ii<(smpsize-1);ii++){
				AddTxtFormula2KB(smpclauses[ii],smpclauses[ii]->inference,&kbset,SETFNUMBER);
			}

			/* Clause is not locked. Convert it to text form. This is always done */
			/* because due to AVATAR successive lock and unlock cycles a clause can */
			/* be subsumed after it generates a simplified successor an so be part of */
			/* the proof queue. */
			if (0==(smpclauses[smpsize-1]->flags&LOCKED)){
				if (NULL==(ptr10=Decompile(smpclauses[smpsize-1]->part2.bin))){
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(NOMEMORY);
				}
				FreeEquData(smpclauses[smpsize-1]->part2.bin);
				if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
					MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
				}
				MYFREE(smpclauses[smpsize-1]->part2.bin);
				smpclauses[smpsize-1]->part2.text=ptr10;
				AddTxtFormula2KB(smpclauses[smpsize-1],smpclauses[smpsize-1]->inference,&kbset,SETFNUMBER);
				return(3);

			/* Clause is locked. */
			} else {

				/* Set pclause to simplified clause. */
				*pclause=smpclauses[smpsize-1];

				/* Add clause to LOCKED queue and return. */
				/* If a NOMEMORY condition arises then delete it. */
				smpclauses[smpsize-1]->flags&=(~(UNPROC|PASSIVE|ACTIVE));
				smpclauses[smpsize-1]->flags|=LOCKED;
				if (NOMEMORY==AddBinClause2KB(smpclauses[smpsize-1],&kbset,0)){
					MYFREE(smpclauses[smpsize-1]->part2.bin->lockasserts);
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(NOMEMORY);
				}
				return(4);
			}
			break;
	}

	/* Set SIMPLIFIED flag and return. */
	(*pclause)->flags|=SIMPLIFIED;
	if (smpsize>1){
		return(1);
	}
	return(0);
} /* FirstSimplify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Forward clause simplification)  OCJ
 *
 *    This function performs the forward simplifications in a clause
 *    that are not performed by FirstSimplify() function:
 *    - Subsumption resolution
 *    - Other possible simplifications to be added in the future
 *
 *    The clause storage is freed in case of a NOMEMORY error condition.
 *
 *    This function just calls the appropriate functions that perform
 *    each simplification inference. Any required memory freeing is
 *    performed by this function, even in the case of error and also
 *    for the memory allocated to *pclause by the caller.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    pclause: Address of pointer to cmprefix structure of clause.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> No simplifications were done.
 *    1 -> Simplifications were done.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t SecondSimplify(cmprefix **pclause,int32_t type){
	auto cmprefix *ptr1;                                 /* Auxiliary pointer */
	auto int32_t ii;                                     /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Auxiliary */
	#endif

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile((*pclause)->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Perform forward subsumption resolution. */
	switch (ii=FwSubsResltn(*pclause,&ptr1,type)){

		/* Not enough memory or timeout. */
		case NOMEMORY:
		case TIMEOUT:
			FreeEquData((*pclause)->part2.bin);
			if ((*pclause)->part2.bin->asserts!=NULL){
				MYFREE((*pclause)->part2.bin->asserts);
			}
			MYFREE((*pclause)->part2.bin);
			MYFREE(*pclause);
			return(ii);
			break;

		/* The clause was simplified. */
		case 1:

			/* If original clause is not locked then add it in text form to KB. */
			if (0==((*pclause)->flags&LOCKED)){
				AddTxtFormula2KB(*pclause,(*pclause)->inference,&kbset,SETFNUMBER);

			/* If original clause is locked then add it to LOCKED queue. */
			} else {
				(*pclause)->flags&=(~(UNPROC|PASSIVE|ACTIVE));
				if (NOMEMORY==AddBinClause2KB(*pclause,&kbset,SETFNUMBER)){
					MYFREE((*pclause)->part2.bin->lockasserts);
					FreeEquData((*pclause)->part2.bin);
					if ((*pclause)->part2.bin->asserts!=NULL){
						MYFREE((*pclause)->part2.bin->asserts);
					}
					MYFREE((*pclause)->part2.bin);
					MYFREE(*pclause);
					return(NOMEMORY);
				}
			}

			/* Get simplified clause and account for this inference. */
			*pclause=ptr1;
			kbset.prstats.fwsubsresltns++;

			/* If an empty clause has been generated then return as an empty */
			/* clause cannot be further simplified. */
			if (ptr1->part2.bin->formula[0]==UNITEND){
				return(0);
			}
			break;
	}

	return(ii);
} /* SecondSimplify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Backward simplification)  OCJ
 *
 *    This function performs backward simplifications by a clause.
 *
 *    Also, if the function called gets an empty clause without asertions
 *    (as may be the case of subsumption resolution) then it must add
 *    the empty clause to the KB as UNPROC, set the frstprfnode field
 *    of the KB and return with an return code 1 such that this
 *    circumstance is recognized by this function.
 *
 *    This function just calls the appropriate functions that perform
 *    each simplification inference.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Process performed and no empty clause without assertions generated.
 *    1 -> Process performed and an empty clause without assertions was generated.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t BckSimplify(cmprefix *clause,int32_t type){
	auto int32_t ii;                                     /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Auxiliary */
	#endif

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile(clause->part2.bin);
	#ifdef VERBOSE
	if (active_cores==1){
		printf("===> Starting backward simplifications by clause:\n%s\n",ptr10);
	}
	#endif
	MYFREE(ptr10);
	#endif

	/* WARNING: Make sure the clause parameters are correctly updated */
	/* as appropriate by the functions that generate them before adding */
	/* them to the KB. */

	/* Perform backward demodulation. */
	if (((UNITEQUALITY|NONUNITEQUALITY)&pbmflags)&&(kbset.opts.demodulation!=DEMODOFF)){
		switch (ii=((UNITEQUALITY|NONUNITEQUALITY)&pbmflags?BckDemodulation(clause,type):0)){

			/* Not enough memory or timeout. */
			case NOMEMORY:
			case TIMEOUT:
				return(ii);
				break;
		}
	}

	/* Perform backward subsumption. */
	switch (ii=BckSubsuming(clause,type)){

		/* Not enough memory or timeout. */
		case NOMEMORY:
		case TIMEOUT:
			return(ii);
			break;
	}

	/* Perform backward subsumption resolution. */
	switch (ii=BkSubsResltn(clause,type)){

		/* Not enough memory, timeout or empty clause without assertions generated. */
		case NOMEMORY:
		case TIMEOUT:
		case 1:
			return(ii);
			break;
	}

	return(0);
} /* BckSimplify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Factoring for resolution inferences)  OCJ
 *
 *    This function performs the factoring required for resolution
 *    inferences. It factors the given literal with all unifiable
 *    selected literals in the formula of the binprefix structure.
 *    The given literal must also be in the  formula of the binprefix
 *    structure.
 *
 *    This function allocates memory for the result. It is the caller
 *    responsibility to free that memory.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure to be factored.
 *    literal: Pointer to literal in binp formula.
 *    result: Address of pointer where the address of the resulting
 *            binprefix structure will be placed.
 *    factlit: Address of pointer where the address of factored
 *             literal will be placed.
 *    Subst: Pointer to the initial substitution of the resolution inference.
 *           It will be upgraded to a compatible factoring substitution.
 *    flag: If not zero then only the literals following the given literal
 *          are checked for factoring, as it is assumed that the previous
 *          literals have been already checked for resolution and factored
 *          if possible with the given literal. Also it means that the
 *          binp clause is the primary parent and also the primary clause
 *          in the substitution. Therefore if marker is zero it means that
 *          the binp clause is the secondary clause in the substitution.
 *
 *  RETURNS:
 *
 *    0 if no factors were found.
 *    1 if factors were found.
 *    NOMEMORY if not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t ResolFactor(binprefix *binp,uint8_t *literal,binprefix **result,
		uint8_t **factlit,subst *Subst,int32_t flag){
	auto cmprefix auxcl;                                 /* Auxiliary cmprefix structure. */
	auto uint8_t *ptr1,*ptr2,*ptr4;                      /* Auxiliary pointers */
	auto binprefix *ptr3;                                /* Auxiliary pointer */
	auto int32_t ii,jj,kk,nn,ss,rr;                      /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Auxiliary */
	#endif

	/* Allocate storage for result and copy original clause. */
	if (NULL==(auxcl.part2.bin=mallocmn(binpsize+2*binp->size))){
		return(NOMEMORY);
	}
	memcpy(auxcl.part2.bin,binp,binpsize+binp->size);
	ptr2=&auxcl.part2.bin->formula[0]+(literal-&binp->formula[0]);
	ii=GetLiteralNbr(auxcl.part2.bin,ptr2);

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile(auxcl.part2.bin);
	MYFREE(ptr10);
	#endif

	/* Loop through literals. */
	rr=0;
	for (ptr1=(flag?NextItem(ptr2,OVERSUBTERMS):&auxcl.part2.bin->formula[0]),
			ptr4=(flag?NextItem(literal,OVERSUBTERMS):NULL),jj=(flag?ii+1:0);
			ptr1[0]!=UNITEND;ptr1=(kk?ptr1:NextItem(ptr1,OVERSUBTERMS)),
			ptr4=(flag?NextItem(ptr4,OVERSUBTERMS):NULL),jj=(kk?jj:jj+1)){
		kk=0;

		/* Literal is selected and is not the primary literal. */
		if ((ii!=jj)&&(SELECTED&ptr1[0])){

			/* Remember substitution status. */
			if (Subst->buffer!=NULL){
				ss=Subst->substsize;
			} else {
				ss=-1;
			}

			/* Check unification of terms. */
			if (0==(nn=Unify(ptr2,ptr1,Subst,flag?0:ITEM1SEC|ITEM2SEC))){
				if (ptr1[0]&EQUALITY){
					if (ss>0){
						Subst->substsize=ss;
						Subst->buffer[ss-1]=UNITEND;
					} else if (Subst->buffer!=NULL){
						Subst->substsize=1;
						Subst->buffer[0]=UNITEND;
					} else {
						Subst->substsize=0;
					}
					nn=SymEquUnify(ptr2,ptr1,Subst,flag?0:ITEM1SEC|ITEM2SEC);
				}
			}
			switch (nn){

				/* Not enough memory. */
				case NOMEMORY:
					if (auxcl.part2.bin->asserts!=NULL){
						MYFREE(auxcl.part2.bin->asserts);
					}
					MYFREE(auxcl.part2.bin);
					return(NOMEMORY);
					break;

				/* Literals unify. */
				case 1:

					/* Delete ptr1 literal. */
					DeleteLiteral(auxcl.part2.bin,ptr1);

					/* Update pointers and literal counters and indicate */
					/* that there was a literal deletion in this iteration. */
					/* Also mark the factored literal if appropriate. */
					if (ii>jj){
						ii--;
						ptr2=GetLiteralPtr(&auxcl,ii);
					}
					kk=rr=1;
					break;

				/* Literals don't unify. Reset substitution to its previous state. */
				default:
					if (ss>0){
						Subst->substsize=ss;
						Subst->buffer[ss-1]=UNITEND;
					} else if (Subst->buffer!=NULL){
						Subst->substsize=1;
						Subst->buffer[0]=UNITEND;
					} else {
						Subst->substsize=0;
					}
					break;
			}
		}
	}

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile(auxcl.part2.bin);
	MYFREE(ptr10);
	#endif

	/* Reallocate result to exact size, set factlit and return. */
	if (NULL==(ptr3=reallocmn(auxcl.part2.bin,binpsize+auxcl.part2.bin->size))){
		if (auxcl.part2.bin->asserts!=NULL){
			MYFREE(auxcl.part2.bin->asserts);
		}
		MYFREE(auxcl.part2.bin);
		return(NOMEMORY);
	}
	*result=ptr3;
	ptr3->equflg=0;
	*factlit=&ptr3->formula[0]+(ptr2-&auxcl.part2.bin->formula[0]);
	return(rr);
} /* ResolFactor */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform forward demodulation)  OCJ
 *
 *    This function tries to simplify the given clause in compiled or binary
 *    format with demodulation inferences by clauses in the indicated KB
 *    by calling Demodulate() function for every simplifier candidate clause
 *    in the requested KB. It must also be specified the type of clauses that
 *    will be considered, either active and/or passive clauses.
 *
 *    Candidate clauses are selected by inspecting every term in the
 *    argument clause and selecting clauses with a generalization term that
 *    is a maximum equality root term in a single literal clause.
 *
 *    The clause storage is freed in case of a NOMEMORY error condition.
 *
 *
 *  ARGUMENTS:
 *
 *    pclause: Address of pointer of clause to be simplified. The pointer to
 *             the simplified clause will be placed here on exit.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Clause was not simplified.
 *    1 -> Clause was simplified and the result was NOT a tautology.
 *    2 -> Clause was simplified and the result was a tautology.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FwdDemodulation(cmprefix **pclause,int32_t type){
	auto cmprefix *clause;                               /* Pointer to clause to be simplified */
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto subst Subst;                                    /* Substitution for calls to Demodulate() */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto binprefix *ptr4;                                /* Auxiliary pointer */
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	auto int32_t ii,kk,nn,oo;                            /* Auxiliary */

	/* Initialize clause pointer and substitution for Generalize() calls. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Renumber variables. This is necessary in order for SetEquData() */
	/* and CompareEquTerms() to generate the correct weight polynomials. */
	/* As the clause has been just inferred it may have shifted variable */
	/* numbers. */
	clause=*pclause;
	if (NOMEMORY==RenumberVars(clause->part2.bin)){
		FreeEquData(clause->part2.bin);
		if (clause->part2.bin->asserts!=NULL){
			MYFREE(clause->part2.bin->asserts);
		}
		MYFREE(clause->part2.bin);
		MYFREE(clause);
		return(NOMEMORY);
	}

	/* Loop through clause items. */
	for (ptr1=NextItem(&clause->part2.bin->formula[0],IMMED);
			ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,IMMED)){

		/* Iterate if item is not a function. */
		if (0==(FUNCTION&ptr1[0])){
			continue;
		}

		/* Loop through ACTIVE and PASSIVE clauses as appropriate. */
		for (kk=ACTIVE;kk!=-1;kk=((kk==ACTIVE)&&(type&PASSIVE)?PASSIVE:-1)){

			/* Get pointer to top tree node. */
			switch (kk){
				case ACTIVE:
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr[3];
					break;
				case PASSIVE:
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr[4];
					break;
			}

			/* Loop through generalization candidates. */
			nn=0;
			do {

				/* Check timeout and solution by other process. */
				gettimeofday(&timestr,NULL);
				time=timestr.tv_sec-kbset.time.tv_sec
						+(timestr.tv_usec-kbset.time.tv_usec)/1000000.0;
				if ((time>=kbset.opts.timeout)||(procctl->status==UNSATISFIABLE)
						||(procctl->status==SATISFIABLE)){
					FreeEquData(clause->part2.bin);
					if (clause->part2.bin->asserts!=NULL){
						MYFREE(clause->part2.bin->asserts);
					}
					MYFREE(clause->part2.bin);
					MYFREE(clause);
					if (Subst.buffer!=NULL){
						MYFREE(Subst.buffer);
					}
					return(TIMEOUT);
				}
				if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
					FreeEquData(clause->part2.bin);
					if (clause->part2.bin->asserts!=NULL){
						MYFREE(clause->part2.bin->asserts);
					}
					MYFREE(clause->part2.bin);
					MYFREE(clause);
					if (Subst.buffer!=NULL){
						MYFREE(Subst.buffer);
					}
					return(procctl->status);
				}

				/* Get a term in a candidate simplifying clause. */
				if (0!=(ii=QueryDscTree(ptr1,&ptr2,ptr3,GENERALIZATION,nn))){

					/* Check generalization. */
					nn=1;
					switch (Generalize(ptr2,ptr1,&Subst,INITSUBST)){

						/* Not enough memory. */
						case NOMEMORY:
							FreeEquData(clause->part2.bin);
							if (clause->part2.bin->asserts!=NULL){
								MYFREE(clause->part2.bin->asserts);
							}
							MYFREE(clause->part2.bin);
							MYFREE(clause);
							if (Subst.buffer!=NULL){
								MYFREE(Subst.buffer);
							}
							return(NOMEMORY);
							break;

						/* Valid generalization. */
						case 1:

							/* Get pointer to candidate binprefix structure. */
							binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

							/* Debug. */
							#ifdef DEBUGCODE
							ptr10=Decompile(binp);
							MYFREE(ptr10);
							#endif

							/* Perform demodulation and check if it is successful. */
							if (1==(oo=Demodulate(clause,ptr1,binp,ptr2,pclause,&Subst))){

								/* Debug. */
								#ifdef DEBUGCODE
								ptr10=Decompile(binp);
								#ifdef VERBOSE
								if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
									printf("Forward demodulated from:\n%s\n",ptr10);
								}
								#endif
								MYFREE(ptr10);
								ptr10=Decompile((*pclause)->part2.bin);
								#ifdef VERBOSE
								if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
									printf("Result [%ld]:\n%s\n",kbset.nformulas,ptr10);
								}
								#endif
								MYFREE(ptr10);
								#endif

								/* Check if result from demodulation is a tautology. */
								if (IsTautology(*pclause)){
									kbset.prstats.tautologies++;
									oo=2;
									ptr4=NULL;
									MYFREE((*pclause)->part2.bin);
									MYFREE(*pclause);

								/* Result from demodulation is not a tautology. */
								/* Set type of inference, flags and dist2conj in new clause. */
								} else {
									ptr4=(*pclause)->part2.bin;
									(*pclause)->inference=FWDEMODULATION;
									(*pclause)->flags=UNPROC;
									(*pclause)->part2.bin->dist2conj=clause->part2.bin->dist2conj;
								}

								/* Set locks of simplified clause and asserts of demodulation result. */
								if (NOMEMORY==GetLocksAsserts(binp,clause->part2.bin,ptr4)){
									FreeEquData(clause->part2.bin);
									if (clause->part2.bin->asserts!=NULL){
										MYFREE(clause->part2.bin->asserts);
									}
									MYFREE(clause->part2.bin);
									MYFREE(clause);
									if ((*pclause)->part2.bin->asserts!=NULL){
										MYFREE((*pclause)->part2.bin->asserts);
									}
									MYFREE((*pclause)->part2.bin);
									MYFREE((*pclause));
									if (Subst.buffer!=NULL){
										MYFREE(Subst.buffer);
									}
									return(NOMEMORY);
								}

								/* If original clause is not locked then convert it to text form. */
								if (0==(LOCKED&clause->flags)){
									if (NULL==(ptr10=Decompile(clause->part2.bin))){
										FreeEquData(clause->part2.bin);
										if (clause->part2.bin->asserts!=NULL){
											MYFREE(clause->part2.bin->asserts);
										}
										MYFREE(clause->part2.bin);
										MYFREE(clause);
										if (oo==1){
											if ((*pclause)->part2.bin->asserts!=NULL){
												MYFREE((*pclause)->part2.bin->asserts);
											}
											MYFREE((*pclause)->part2.bin);
											MYFREE((*pclause));
										}
										if (Subst.buffer!=NULL){
											MYFREE(Subst.buffer);
										}
										return(NOMEMORY);
									}
									FreeEquData(clause->part2.bin);
									if (clause->part2.bin->asserts!=NULL){
										MYFREE(clause->part2.bin->asserts);
									}
									MYFREE(clause->part2.bin);
									clause->part2.text=ptr10;
								}
							}

							/* If clause was simplified or we got a NOMEMORY condition */
							/* then free memory and return. */
							if (oo!=0){
								switch (oo){

									/* Not enough memory or timeout. */
									case NOMEMORY:
									case TIMEOUT:
										FreeEquData(clause->part2.bin);
										if (clause->part2.bin->asserts!=NULL){
											MYFREE(clause->part2.bin->asserts);
										}
										MYFREE(clause->part2.bin);
										MYFREE(clause);
										break;

									/* The clause was simplified. */
									case 1:
									case 2:

										/* If original clause is not locked then add it in */
										/* text form to KB. */
										if (0==(clause->flags&LOCKED)){
											AddTxtFormula2KB(clause,clause->inference,&kbset,SETFNUMBER);

										/* If original clause is locked then add it to LOCKED queue. */
										} else {
											clause->flags&=(~(UNPROC|PASSIVE|ACTIVE));
											if (NOMEMORY==AddBinClause2KB(clause,&kbset,SETFNUMBER)){
												MYFREE(clause->part2.bin->lockasserts);
												FreeEquData(clause->part2.bin);
												if (clause->part2.bin->asserts!=NULL){
													MYFREE(clause->part2.bin->asserts);
												}
												MYFREE(clause->part2.bin);
												MYFREE(clause);
												oo=NOMEMORY;
											}
										}

										/* Account for this inference. */
										kbset.prstats.fwdemodulations++;
										break;
								}
								if (Subst.buffer!=NULL){
									MYFREE(Subst.buffer);
								}
								return(oo);
							}
							break;
					}
				}
			} while(ii);
		}
	}

	/* Restore pointer to original clause, free memory and return clause not simplified. */
	*pclause=clause;
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}
	return(0);
} /* FwdDemodulation */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform backward demodulation)  OCJ
 *
 *    This function tries to simplify the active and/or passive clauses
 *    in the indicated KB by the given clause in compiled or binary format
 *    with demodulation inferences by calling Demodulate() function for
 *    every candidate clause for simplification in the requested KB. It
 *    must also be specified the type of clauses that will be simplified,
 *    either active and/or passive clauses.
 *
 *    The given clause must be suitable for performing demodulation: It must
 *    be a single literal positive equality with a maximum equality root term.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Process performed and no empty clause without assertions generated.
 *    1 -> Process performed and an empty clause without assertions was generated.
 *         This function never returns 1 but this is documented here as a possible
 *         return code from functions called by BckSimplify() function.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t BckDemodulation(cmprefix *clause,int32_t type){
	auto cmprefix *newcl;                                /* Pointer to simplified clause cmprefix structure */
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto subst Subst;                                    /* Substitution for calls to Demodulate() */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *equality;                              /* Pointer to equality literal in clause */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto int32_t ii,jj,kk,nn,oo;                         /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Check if clause is suitable for performing demodulation from it. */
	if ((clause->part2.bin->literals>1)||(EQUALITY!=((EQUALITY|NEGATED)&clause->part2.bin->formula[0]))){
		return(0);
	} else {
		equality=&clause->part2.bin->formula[0];
	}

	/* Get pointers to equality root terms and reset MAXIMUM */
	/* and SELECTED flags in each of them. */
	ptr1=NextItem(equality,IMMED);
	ptr2=NextItem(ptr1,OVERSUBTERMS);
	ptr1[0]&=(~(MAXIMUM|SELECTED));
	ptr2[0]&=(~(MAXIMUM|SELECTED));

	/* Renumber variables. This is necessary in order for SetEquData() */
	/* and CompareEquTerms() to generate the correct weight polynomials. */
	/* As the clause has been just inferred it may have shifted variable */
	/* numbers. */
	if (NOMEMORY==RenumberVars(clause->part2.bin)){
		return(NOMEMORY);
	}

	/* Calculate equality ordering data and remember that */
	/* this was done for a potentially demodulating clause. */
	/* Equality ordering data may be already calculated because */
	/* a clause can go several times as unprocessed due to being */
	/* previously locked. */
	if (clause->part2.bin->equflg==0){
		if (NOMEMORY==SetEquData(clause->part2.bin,equality,ptr1,ptr2)){
			return(NOMEMORY);
		}
		clause->part2.bin->equflg=2;
	}

	/* Set the SELECTED and MAXIMUM flags for equality root terms */
	/* as appropriate. Maximal equality root terms are also */
	/* candidates for demodulating from them as they may become */
	/* maximum after the substitution. If both terms are equal */
	/* then demodulation is not possible as there is no possible */
	/* maximum term even after a substitution. */
	/* Release equality tree data if the root terms ordering */
	/* is completely defined. */
	switch (CompareEquTerms(clause->part2.bin,ptr1,ptr2,&((symbol *)&equality[1])->leftnode,
			&((symbol *)&equality[1])->rightnode,kbset.opts.termord)){
		case NOMEMORY:
			return(NOMEMORY);
			break;
		case EQUAL:
			FreeEquNode(ptr1,((symbol *)&equality[1])->leftnode);
			FreeEquNode(ptr2,((symbol *)&equality[1])->rightnode);
			((symbol *)&equality[1])->leftnode=NULL;
			return(0);
			break;
		case FAILURE:
			ptr1[0]|=SELECTED;
			ptr2[0]|=SELECTED;
			return(0);
			break;
		case GREATER:
			ptr1[0]|=(SELECTED|MAXIMUM);
			FreeEquNode(ptr1,((symbol *)&equality[1])->leftnode);
			FreeEquNode(ptr2,((symbol *)&equality[1])->rightnode);
			((symbol *)&equality[1])->leftnode=NULL;
			break;
		case LESSER:
			ptr2[0]|=(SELECTED|MAXIMUM);
			FreeEquNode(ptr1,((symbol *)&equality[1])->leftnode);
			FreeEquNode(ptr2,((symbol *)&equality[1])->rightnode);
			((symbol *)&equality[1])->leftnode=NULL;
			break;
	}

	/* Initialize substitution for Generalize() calls. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop through equality root terms. */
	for (ptr1=NextItem(equality,IMMED),jj=0;jj<2;ptr1=NextItem(ptr1,OVERSUBTERMS),jj++){

		/* Iterate if the term is not selected (that is, maximal) */
		/* or it is a variable. Variables are discarded because if */
		/* they are also in the other root term then the variable */
		/* can never be maximum even after a substitution. Also, */
		/* if the variable is not part of the other root term the */
		/* equality makes no sense as it that can be unified with */
		/* everything). */
		if ((0==(SELECTED&ptr1[0]))||(VARIABLE&ptr1[0])){
			continue;
		}

		/* Loop through ACTIVE and PASSIVE clauses as appropriate. */
		for (kk=ACTIVE;kk!=-1;kk=((kk==ACTIVE)&&(type&PASSIVE)?PASSIVE:-1)){

			/* Get pointer to top tree node. */
			switch (kk){
				case ACTIVE:
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr[2];
					break;
				case PASSIVE:
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr[5];
					break;
			}

			/* Loop through candidates that are instances of maximum equality root term. */
			nn=0;
			do {

				/* Check timeout and solution by other process. */
				gettimeofday(&timestr,NULL);
				time=timestr.tv_sec-kbset.time.tv_sec
						+(timestr.tv_usec-kbset.time.tv_usec)/1000000.0;
				if ((time>=kbset.opts.timeout)||(procctl->status==UNSATISFIABLE)
						||(procctl->status==SATISFIABLE)){
					if (Subst.buffer!=NULL){
						MYFREE(Subst.buffer);
					}
					return(TIMEOUT);
				}
				if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
					if (Subst.buffer!=NULL){
						MYFREE(Subst.buffer);
					}
					return(procctl->status);
				}

				/* Get a term in a candidate clause to be simplified. */
				if (0!=(ii=QueryDscTree(ptr1,&ptr2,ptr3,INSTANCE,nn))){

					/* Get pointer to binprefix structure of clause to be simplified. */
					nn=1;
					binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

					/* Clause is not in disposal queue therefore */
					/* it has not been already simplified. */
					if (0==(binp->clause->flags&INDISPOSALQUEUE)){

						/* Check generalization. */
						switch (Generalize(ptr1,ptr2,&Subst,INITSUBST)){

							/* Not enough memory. */
							case NOMEMORY:
								if (Subst.buffer!=NULL){
									MYFREE(Subst.buffer);
								}
								return(NOMEMORY);
								break;

							/* Valid generalization. */
							case 1:

								/* Debug. */
								#ifdef DEBUGCODE
								ptr10=Decompile(binp);
								MYFREE(ptr10);
								#endif

								/* Perform demodulation. */
								if (NOMEMORY==(oo=Demodulate(binp->clause,ptr2,clause->part2.bin,
										ptr1,&newcl,&Subst))){
									if (Subst.buffer!=NULL){
										MYFREE(Subst.buffer);
									}
									return(NOMEMORY);
								}

								/* Successful simplification. */
								if (oo){

									/* Set distance to conjecture of new simplified clause. */
									newcl->part2.bin->dist2conj=binp->dist2conj;

									/* Set locks of simplified clause and asserts of demodulation result. */
									if (NOMEMORY==GetLocksAsserts(clause->part2.bin,binp,newcl->part2.bin)){
										if (newcl->part2.bin->asserts!=NULL){
											MYFREE(newcl->part2.bin->asserts);
										}
										MYFREE(newcl->part2.bin);
										MYFREE(newcl);
										if (Subst.buffer!=NULL){
											MYFREE(Subst.buffer);
										}
										return(NOMEMORY);
									}

									/* Debug. */
									#ifdef DEBUGCODE
									ptr10=Decompile(binp);
									#ifdef VERBOSE
									if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
										printf("Backward demodulating to:\n%s\n",ptr10);
									}
									#endif
									MYFREE(ptr10);
									ptr10=Decompile(newcl->part2.bin);
									#ifdef VERBOSE
									if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
										printf("Result [%ld]:\n%s\n",kbset.nformulas,ptr10);
									}
									#endif
									MYFREE(ptr10);
									#endif

									/* Add original simplified clause to disposal queue. This queue interconnects */
									/* clauses from other queues. The clause must not be removed from any queue */
									/* to which it is linked. */
									binp->clause->flags|=INDISPOSALQUEUE;
									binp->clause->prevdisp=kbset.lastdisp;
									kbset.lastdisp=binp->clause;
									kbset.prstats.bkdemodulations++;

									/* Add new simplified clause to KB as UNPROC. A new number is not assigned. */
									/* It will be assigned when moving this clause to passive, so that it */
									/* has a higher number than its parents. */
									newcl->flags=UNPROC;
									newcl->inference=BKDEMODULATION;
									if (NOMEMORY==AddBinClause2KB(newcl,&kbset,0)){
										if (newcl->part2.bin->asserts!=NULL){
											MYFREE(newcl->part2.bin->asserts);
										}
										MYFREE(newcl->part2.bin);
										MYFREE(newcl);
										if (Subst.buffer!=NULL){
											MYFREE(Subst.buffer);
										}
										return(NOMEMORY);
									}
								}
								break;
						}
					}
				}
			} while(ii);
		}
	}

	/* Free memory and return. */
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}
	return(0);
} /* BckDemodulation */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Demodulate a clause)  OCJ
 *
 *    This function demodulates a clause in internal compiled binary
 *    format. The new clause is placed in new allocated storage and
 *    its parents are set, but no other data is included in the
 *    clause cmprefix.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by demodulation.
 *    term1: Pointer to term that will be replaced in clause.
 *    binp: Pointer to binprefix structure of clause from where
 *          demodulation is performed.
 *    term2: Pointer to the term in "from" clause that is a generalization
 *           of term1.
 *    pnewcl: Address of pointer to cmprefix structure of new clause to be
 *            allocated for the result of demodulation.
 *    Subst: Pointer to substitution to be applied to term2.
 *
 *  RETURNS:
 *
 *    0 -> The clause was not simplified because ordering constraints
 *         were not met or redundancy check detected that demodulation
 *         does not preserve completeness.
 *    1 -> The clause was successfully simplified.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t Demodulate(cmprefix *clause,uint8_t *term1,binprefix *binp,
		uint8_t *term2,cmprefix **pnewcl,subst *Subst){
	auto uint8_t *equality;                              /* Pointer to equality literal in binp */
	auto binprefix *binp2;                               /* Auxiliary binary prefix pointer */
	auto uint8_t *ptr1,*ptr2;                            /* Auxiliary pointers */
	auto int32_t ii,jj,kk,nn;                            /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                     /* Auxiliary pointer */

	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	ptr10=Decompile(binp);
	MYFREE(ptr10);
	#endif

	/* Get pointer to equality from which demodulation is performed. */
	equality=&binp->formula[0];

	/* Set ptr1 to the other root term in the "from" equality. If Next item */
	/* of term2 is the end of clause then it is the second root term in the */
	/* "from" equality. Otherwise term2 is the first root term in the "from" */
	/* equality. Take into account that the "from" clause has only one literal. */
	ptr1=NextItem(term2,OVERSUBTERMS);
	if (ptr1[0]&UNITEND){
		ptr1=NextItem(equality,IMMED);
	}

	/* Check ordering constraint of root terms in equality */
	/* "from" literal. */
	switch (ii=CheckEqRootOrder(binp,equality,term2,ptr1,Subst,binp->maxvarnb+1,MAXIMUM)){
		case NOMEMORY:
		case 0:
			return(ii);
			break;
	}

	/* Allocate storage for inferred clause. */
	if (NULL==(*pnewcl=mallocmn(sizeof(cmprefix)))){
		return(NOMEMORY);
	}
	nn=binpsize+2*(clause->part2.bin->size+binp->size);
	if (NULL==(((*pnewcl)->part2.bin)=mallocmn(nn))){
		MYFREE(*pnewcl);
		return(NOMEMORY);
	}
	(*pnewcl)->part2.bin->clause=*pnewcl;
	(*pnewcl)->part2.bin->equflg=0;

	/* Copy formula of "to" parent clause up to and not including term1. */
	ii=term1-&clause->part2.bin->formula[0];
	memcpy(&(*pnewcl)->part2.bin->formula[0],&clause->part2.bin->formula[0],ii);

	/* Shift the numbers of variables of copied items. */
	kk=binp->maxvarnb+1;
	(*pnewcl)->part2.bin->maxvarnb=-1;
	for (ptr2=NextItem(&(*pnewcl)->part2.bin->formula[0],IMMED);
			ptr2<&(*pnewcl)->part2.bin->formula[ii];
			ptr2=NextItem(ptr2,IMMED)){
		if (VARIABLE&ptr2[0]){
			*((uint16_t *)&ptr2[1])+=kk;
			if ((*pnewcl)->part2.bin->maxvarnb<*((uint16_t *)&ptr2[1])){
				(*pnewcl)->part2.bin->maxvarnb=*((uint16_t *)&ptr2[1]);
			}
		}
	}

	/* Add the other equality root term of clause from */
	/* where demodulation is performed. */
	jj=NextItem(ptr1,OVERSUBTERMS)-ptr1;
	memcpy(ptr2=&(*pnewcl)->part2.bin->formula[ii],ptr1,jj);
	ii+=jj;

	/* Add remaining items of clause being simplified */
	/* including the UNITEND marker byte. */
	ptr1=NextItem(term1,OVERSUBTERMS);
	jj=clause->part2.bin->size-(ptr1-&clause->part2.bin->formula[0]);
	memcpy(&(*pnewcl)->part2.bin->formula[ii],ptr1,jj);

	/* Shift the numbers of variables of last copied items. */
	for (ptr1=&(*pnewcl)->part2.bin->formula[ii];ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,IMMED)){
		if (VARIABLE&ptr1[0]){
			*((uint16_t *)&ptr1[1])+=kk;
			if ((*pnewcl)->part2.bin->maxvarnb<*((uint16_t *)&ptr1[1])){
				(*pnewcl)->part2.bin->maxvarnb=*((uint16_t *)&ptr1[1]);
			}
		}
	}

	/* Apply substitution to inferred clause. Binary clause */
	/* size field is set by this function, so there is no need */
	/* to call UpdateParams() function here. In case of NOMEMORY */
	/* there is no need to free asserts as they are not set yet. */
	if (NOMEMORY==ApplySubst2Clause(*pnewcl,&nn,Subst,kk)){
		MYFREE((*pnewcl)->part2.bin);
		MYFREE(*pnewcl);
		return(NOMEMORY);
	}

	/* Perform redundancy check if completeness must be preserved. */
	if ((kbset.opts.demodulation==DEMODCPL)&&(0!=(ii=RedundancyCheck(term2,ptr2,(*pnewcl)->part2.bin,binp)))){
		MYFREE((*pnewcl)->part2.bin);
		MYFREE(*pnewcl);
		if (ii==NOMEMORY){
			return(NOMEMORY);
		}
		return(0);
	}

	/* Reallocate the formula buffer to the exact length needed. */
	if (NULL==(binp2=reallocmn((*pnewcl)->part2.bin,binpsize+(*pnewcl)->part2.bin->size))){
		if ((*pnewcl)->part2.bin->asserts!=NULL){
			MYFREE((*pnewcl)->part2.bin->asserts);
		}
		MYFREE((*pnewcl)->part2.bin);
		MYFREE(*pnewcl);
		return(NOMEMORY);
	}
	(*pnewcl)->part2.bin=binp2;

	/* Debug. */
	#ifdef DEBUGCODE
	(*pnewcl)->part2.bin->asserts=NULL;
	ptr10=Decompile((*pnewcl)->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Set simplified clause cmprefix data. */
	(*pnewcl)->parent1=binp->clause;
	(*pnewcl)->parent2=clause;

	return(1);
} /* Demodulate */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform demodulation redundancy check)  OCJ
 *
 *    This function checks if a demodulation inference preserves
 *    completion or not. This check is experimental and may be
 *    not correct, so it may have to be corrected in the future.
 *
 *    The check is based on a free interpretation of information
 *    from two sources:
 *    - Document "Implementing an efficient theorem prover" by
 *      Alexandre Riazanov.
 *    - Help text for --demodulation_redundancy_check option of
 *      Vampire V4.0 program.
 *
 *    The check is performed as follows:
 *    - if the two following conditions are met then the demodulation
 *      doesn't preserve completeness.
 *    - The term replaced in the demodulated clause must be an
 *      equality root term.
 *    - The demodulating clause must be greater than the resulting
 *      clause from demodulation, according to the multiset and
 *      literal orderings described in chapter 2 and section 3.1.1
 *      of the above mentioned document.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    termfr: Pointer to equality root term in demodulating clause that
 *            is a generalization of a term in demodulated clause.
 *    repterm: Pointer to term in clause resulting from demodulation that
 *             replaced the term generalized by termfr.
 *    binprslt: Pointer to binprefix structure of clause resulting from
 *              demodulation.
 *    binpfr: Pointer to binprefix structure of demodulating clause.
 *
 *  RETURNS:
 *
 *    0 -> The demodulation preserves completeness.
 *    1 -> The demodulation DOES NOT preserve completeness.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
inline int32_t RedundancyCheck(uint8_t *termfr,uint8_t *repterm,binprefix *binprslt,binprefix *binpfr){
	auto coeff coef1,coef2;                              /* Coeficients for equality root term comparison */
	auto uint8_t *ptr1,*ptr2;                            /* Auxiliary pointers */
	auto int32_t ii,jj,kk;                               /* Auxiliary */

	/* Loop through literals in resulting clause from demodulation */
	/* to check that the term replaced in the demodulated clause */
	/* is an equality root term and that all literals in clause */
	/* are equalities or inequalities. This last check is a */
	/* prerequisite to check the clauses comparison. */
	for (ptr1=&binprslt->formula[0],ii=1;ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){
		if (0==(EQUALITY&ptr1[0])){
			return(0);
		}
		if (ii){
			if (repterm==(ptr2=NextItem(ptr1,IMMED))){
				ii=0;
			} else if (repterm==NextItem(ptr2,OVERSUBTERMS)){
				ii=0;
			}
		}
	}
	if (ii){
		return(0);
	}

	/* If we are here then all literals in resulting clause are equalities */
	/* or inequalities, so the demodulating clause is possibly greater */
	/* than the resulting clause. Calculate coefficients for termfr. */
	coef1.eqterm1=coef1.eqterm2=coef2.eqterm1=coef2.eqterm2=NULL;
	jj=binpfr->maxvarnb;
	kk=binprslt->maxvarnb;
	if (jj>kk){
		binprslt->maxvarnb=jj;
	} else {
		binpfr->maxvarnb=kk;
	}
	if (NOMEMORY==CalcCoeffs(termfr,&coef1,binpfr->maxvarnb)){
		binpfr->maxvarnb=jj;
		binprslt->maxvarnb=kk;
		return(NOMEMORY);
	}

	/* Check clause ordering constraint. To check this it is enough to */
	/* check that termfr is greater than all equality root terms in the */
	/* clause resulting from the demodulation clause. */
	/* Loop through literals in resulting clause from demodulation. */
	ii=kbset.opts.termord&STANDARD?STANDARD:NONRECURSIVE;
	for (ptr1=&binprslt->formula[0];ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Compare with the first equality root term. */
		ptr2=NextItem(ptr1,IMMED);
		if (NOMEMORY==CalcCoeffs(ptr2,&coef2,binprslt->maxvarnb)){
			MYFREE(coef1.alpha);
			binpfr->maxvarnb=jj;
			binprslt->maxvarnb=kk;
			return(NOMEMORY);
		}
		switch (CompareItems(termfr,ptr2,&coef1,&coef2,ii)){
			case NOMEMORY:
				MYFREE(coef1.alpha);
				MYFREE(coef2.alpha);
				binpfr->maxvarnb=jj;
				binprslt->maxvarnb=kk;
				return(NOMEMORY);
				break;
			case GREATER:
				break;
			default:
				MYFREE(coef1.alpha);
				MYFREE(coef2.alpha);
				binpfr->maxvarnb=jj;
				binprslt->maxvarnb=kk;
				return(0);
				break;
		}
		MYFREE(coef2.alpha);

		/* Compare with the second equality root term. */
		ptr2=NextItem(ptr2,OVERSUBTERMS);
		if (NOMEMORY==CalcCoeffs(ptr2,&coef2,binprslt->maxvarnb)){
			MYFREE(coef1.alpha);
			binpfr->maxvarnb=jj;
			binprslt->maxvarnb=kk;
			return(NOMEMORY);
		}
		switch (CompareItems(termfr,ptr2,&coef1,&coef2,ii)){
			case NOMEMORY:
				MYFREE(coef1.alpha);
				MYFREE(coef2.alpha);
				binpfr->maxvarnb=jj;
				binprslt->maxvarnb=kk;
				return(NOMEMORY);
				break;
			case GREATER:
				break;
			default:
				MYFREE(coef1.alpha);
				MYFREE(coef2.alpha);
				binpfr->maxvarnb=jj;
				binprslt->maxvarnb=kk;
				return(0);
				break;
		}
		MYFREE(coef2.alpha);
	}

	/* Free coefficients memory. */
	MYFREE(coef1.alpha);
	binpfr->maxvarnb=jj;
	binprslt->maxvarnb=kk;

	return(1);
} /* RedundancyCheck */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform forward subsumption resolution)  OCJ
 *
 *    This function tries to simplify the given clause in compiled or binary
 *    format with subsumption resolution inferences by clauses in the indicated
 *    KB by calling SubsResltn() function for every simplifier candidate clause
 *    in the requested KB. It must also be specified the type of clauses that
 *    will be considered, either active and/or passive clauses.
 *
 *    Simplifier candidate clauses are selected from those that have a literal
 *    whose negation is a generalization of a literal in the clause to be
 *    simplified.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    pnewcl: Address of pointer where the simplified clause will be placed.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Clause was not simplified.
 *    1 -> Clause was simplified.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FwSubsResltn(cmprefix *clause,cmprefix **pnewcl,int32_t type){
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto uint8_t *equality;                              /* Pointer to equality with root terms in reverse order */
	auto uint8_t *eqterms;                               /* Pointer to equality root terms in reverse order */
	auto uint8_t *queryitm;                              /* Pointer to QueryDscTree() query item */
	auto int32_t size;                                   /* Size of eqterms buffer */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3,*ptr4,*ptr5;          /* Auxiliary pointers */
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	auto int32_t ii,jj,kk,mm,nn,rr,ss;                   /* Auxiliary */

	/* Loop through clause literals. */
	eqterms=equality=NULL;
	size=0;
	for (ptr1=&clause->part2.bin->formula[0];ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Loop for equality symmetry management. */
		ptr4=NextItem(ptr1,IMMED);
		mm=(ptr1[0]&EQUALITY?2:1);
		for (jj=0;jj<mm;jj++){

			/* This is the first iteration. */
			if (jj==0){
				queryitm=ptr1;

			/* This is the second iteration for an equality literal. */
			/* Build the set of root terms in reverse order. */
			} else {
				ptr3=NextItem(ptr4,OVERSUBTERMS);
				rr=ptr3-ptr4;
				ss=NextItem(ptr3,OVERSUBTERMS)-ptr3;
				if (size==0){
					if (NULL==(equality=mallocmn(size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
						return(NOMEMORY);
					}
					eqterms=&equality[1+sizeof(symbol)];
				} else if (size<(2+sizeof(symbol)+rr+ss)){
					if (NULL==(ptr5=reallocmn(equality,size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
						MYFREE(equality);
						return(NOMEMORY);
					}
					equality=ptr5;
					eqterms=&equality[1+sizeof(symbol)];
				}
				memcpy(equality,ptr1,1+sizeof(symbol));
				memcpy(eqterms,ptr3,ss);
				memcpy(&eqterms[ss],ptr4,rr);
				eqterms[rr+ss]=UNITEND;
				queryitm=equality;
			}

			/* Loop through ACTIVE and PASSIVE clauses. ACTIVE clauses have */
			/* selected and non selected literals. */
			for (kk=ACTIVE;kk!=-1;
					kk=(kk==ACTIVE?ACTIVE|SELECTED:(kk==PASSIVE?-1:(type&PASSIVE?PASSIVE:-1)))){

				/* Get pointer to top tree node. */
				switch (kk){
					case ACTIVE:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[2];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[3];
						}
						break;
					case PASSIVE:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[4];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[5];
						}
						break;
					default:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[0];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[1];
						}
						break;
				}

				/* Loop through generalization candidates. */
				nn=0;
				do {

					/* Check timeout and solution by other process. */
					gettimeofday(&timestr,NULL);
					time=timestr.tv_sec-kbset.time.tv_sec
							+(timestr.tv_usec-kbset.time.tv_usec)/1000000.0;
					if ((time>=kbset.opts.timeout)||(procctl->status==UNSATISFIABLE)
							||(procctl->status==SATISFIABLE)){
						MYFREE(equality);
						return(TIMEOUT);
					}
					if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
						MYFREE(equality);
						return(procctl->status);
					}

					/* Get a candidate negative generalization literal. */
					if (0!=(ii=QueryDscTree(queryitm,&ptr2,ptr3,GENERALIZATION,nn))){

						/* Get pointer to candidate binprefix structure. */
						nn=1;
						binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

						/* Debug. */
						#ifdef DEBUGCODE
						ptr10=Decompile(binp);
						MYFREE(ptr10);
						#endif

						/* Perform subsumption resolution and check if it is successful. */
						switch (SubsResltn(clause,binp,ptr2,pnewcl)){

							/* Not enough memory. */
							case NOMEMORY:
								MYFREE(equality);
								return(NOMEMORY);
								break;

							/* Successful simplification. */
							case 1:

								/* Set locks of simplified clause and asserts of demodulation result. */
								if (NOMEMORY==GetLocksAsserts(binp,clause->part2.bin,(*pnewcl)->part2.bin)){
									if ((*pnewcl)->part2.bin->asserts!=NULL){
										MYFREE((*pnewcl)->part2.bin->asserts);
									}
									MYFREE((*pnewcl)->part2.bin);
									MYFREE((*pnewcl));
									MYFREE(equality);
									return(NOMEMORY);
								}

								/* Set type of inference, flags and dist2conj in new clause. */
								(*pnewcl)->inference=FWSUBSRESLTNS;
								(*pnewcl)->flags=UNPROC;
								(*pnewcl)->part2.bin->dist2conj=clause->part2.bin->dist2conj;

								/* If original clause is not locked then convert it to text form. */
								if (0==(LOCKED&clause->flags)){
									if (NULL==(ptr10=Decompile(clause->part2.bin))){
										if ((*pnewcl)->part2.bin->asserts!=NULL){
											MYFREE((*pnewcl)->part2.bin->asserts);
										}
										MYFREE((*pnewcl)->part2.bin);
										MYFREE((*pnewcl));
										MYFREE(equality);
										return(NOMEMORY);
									}
									FreeEquData(clause->part2.bin);
									if (clause->part2.bin->asserts!=NULL){
										MYFREE(clause->part2.bin->asserts);
									}
									MYFREE(clause->part2.bin);
									clause->part2.text=ptr10;
								}

								/* Debug. */
								#ifdef DEBUGCODE
								ptr10=Decompile(binp);
								#ifdef VERBOSE
								if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
									printf("Subsumption resolution simplification by:\n%s\n",ptr10);
								}
								#endif
								MYFREE(ptr10);
								ptr10=Decompile((*pnewcl)->part2.bin);
								#ifdef VERBOSE
								if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
									printf("Result [%ld]:\n%s\n",kbset.nformulas,ptr10);
								}
								#endif
								MYFREE(ptr10);
								#endif
								MYFREE(equality);
								return(1);
								break;
						}
					}
				} while(ii);
			}
		}
	}

	/* Free memory and return unsuccessful simplification. */
	MYFREE(equality);
	return(0);
} /* FwSubsResltn */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform backward subsumption resolution)  OCJ
 *
 *    This function tries to simplify clauses in KB by the given clause in
 *    compiled or binary format with subsumption resolution inferences by
 *    calling SubsResltn() function for every simplifier candidate clause
 *    in the requested KB. It must also be specified the type of clauses that
 *    will be considered, either active and/or passive clauses.
 *
 *    Clauses candidate to be simplified are selected from those that have
 *    a literal whose negation is an instance of a literal in the clause
 *    to be simplified.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Process performed and no empty clause without assertions generated.
 *    1 -> Process performed and an empty clause without assertions was generated.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t BkSubsResltn(cmprefix *clause,int32_t type){
	auto cmprefix *newcl;                                /* Pointer to simplified clause cmprefix structure */
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto uint8_t *equality;                              /* Pointer to equality with root terms in reverse order */
	auto uint8_t *eqterms;                               /* Pointer to equality root terms in reverse order */
	auto uint8_t *queryitm;                              /* Pointer to QueryDscTree() query item */
	auto int32_t size;                                   /* Size of eqterms buffer */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3,*ptr4,*ptr5;          /* Auxiliary pointers */
	auto int32_t ii,jj,kk,mm,nn,rr,ss;                   /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Loop through clause literals. */
	eqterms=equality=NULL;
	size=0;
	for (ptr1=&clause->part2.bin->formula[0];ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Loop for equality symmetry management. */
		ptr4=NextItem(ptr1,IMMED);
		mm=(ptr1[0]&EQUALITY?2:1);
		for (jj=0;jj<mm;jj++){

			/* This is the first iteration. */
			if (jj==0){
				queryitm=ptr1;

			/* This is the second iteration for an equality literal. */
			/* Build the set of root terms in reverse order. */
			} else {
				ptr3=NextItem(ptr4,OVERSUBTERMS);
				rr=ptr3-ptr4;
				ss=NextItem(ptr3,OVERSUBTERMS)-ptr3;
				if (size==0){
					if (NULL==(equality=mallocmn(size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
						return(NOMEMORY);
					}
					eqterms=&equality[1+sizeof(symbol)];
				} else if (size<(2+sizeof(symbol)+rr+ss)){
					if (NULL==(ptr5=reallocmn(equality,size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
						MYFREE(equality);
						return(NOMEMORY);
					}
					equality=ptr5;
					eqterms=&equality[1+sizeof(symbol)];
				}
				memcpy(equality,ptr1,1+sizeof(symbol));
				memcpy(eqterms,ptr3,ss);
				memcpy(&eqterms[ss],ptr4,rr);
				eqterms[rr+ss]=UNITEND;
				queryitm=equality;
			}

			/* Loop through ACTIVE and PASSIVE clauses. ACTIVE clauses have */
			/* selected and non selected literals. */
			for (kk=ACTIVE;kk!=-1;
					kk=(kk==ACTIVE?ACTIVE|SELECTED:(kk==PASSIVE?-1:(type&PASSIVE?PASSIVE:-1)))){

				/* Get pointer to top tree node. */
				switch (kk){
					case ACTIVE:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[2];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[3];
						}
						break;
					case PASSIVE:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[4];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[5];
						}
						break;
					default:
						if (ptr1[0]&NEGATED){
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[0];
						} else {
							ptr3=(((symbol *)&ptr1[1])->symbol)->discr[1];
						}
						break;
				}

				/* Loop through candidate instances. */
				nn=0;
				do {

					/* Check timeout and solution by other process. */
					gettimeofday(&timestr,NULL);
					time=timestr.tv_sec-kbset.time.tv_sec
							+(timestr.tv_usec-kbset.time.tv_usec)/1000000.0;
					if ((time>=kbset.opts.timeout)||(procctl->status==UNSATISFIABLE)
							||(procctl->status==SATISFIABLE)){
						MYFREE(equality);
						return(TIMEOUT);
					}
					if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
						MYFREE(equality);
						return(procctl->status);
					}

					/* Get a candidate negative instance literal. */
					if (0!=(ii=QueryDscTree(queryitm,&ptr2,ptr3,INSTANCE,nn))){

						/* Get pointer to candidate binprefix structure. */
						nn=1;
						binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

						/* Debug. */
						#ifdef DEBUGCODE
						ptr10=Decompile(binp);
						MYFREE(ptr10);
						#endif

						/* Clause is not in disposal queue therefore */
						/* it has not been already simplified. */
						if (0==(binp->clause->flags&INDISPOSALQUEUE)){

							/* Perform subsumption resolution and check if it is successful. */
							switch (SubsResltn(binp->clause,clause->part2.bin,ptr1,&newcl)){

								/* Not enough memory. */
								case NOMEMORY:
									MYFREE(equality);
									return(NOMEMORY);
									break;

								/* Successful simplification. */
								case 1:

									/* Set distance to conjecture of new simplified clause. */
									newcl->part2.bin->dist2conj=binp->dist2conj;

									/* Set locks of simplified clause and asserts of simplification result. */
									if (NOMEMORY==GetLocksAsserts(clause->part2.bin,binp,newcl->part2.bin)){
										if (newcl->part2.bin->asserts!=NULL){
											MYFREE(newcl->part2.bin->asserts);
										}
										MYFREE(newcl->part2.bin);
										MYFREE(newcl);
										MYFREE(equality);
										return(NOMEMORY);
									}

									/* Debug. */
									#ifdef DEBUGCODE
									ptr10=Decompile(binp);
									#ifdef VERBOSE
									if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
										printf("Backward subsumption resolution simplification to:\n%s\n",ptr10);
									}
									#endif
									MYFREE(ptr10);
									ptr10=Decompile(newcl->part2.bin);
									#ifdef VERBOSE
									if ((active_cores==1)&&(verbfrom<=kbset.nformulas)&&(verbto>=kbset.nformulas)){
										printf("Result [%ld]:\n%s\n",kbset.nformulas,ptr10);
									}
									#endif
									MYFREE(ptr10);
									#endif

									/* Add original simplified clause to disposal queue for later */
									/* deletion or locking. This queue interconnects clauses from other */
									/* queues. The clause must not be removed from any queue to which */
									/* it is linked. */
									binp->clause->flags|=INDISPOSALQUEUE;
									binp->clause->prevdisp=kbset.lastdisp;
									kbset.lastdisp=binp->clause;
									kbset.prstats.bksubsresltns++;

									/* Add new simplified clause to KB as UNPROC. */
									newcl->flags=UNPROC;
									newcl->inference=BKSUBSRESLTNS;
									if (NOMEMORY==AddBinClause2KB(newcl,&kbset,0)){
										if (newcl->part2.bin->asserts!=NULL){
											MYFREE(newcl->part2.bin->asserts);
										}
										MYFREE(newcl->part2.bin);
										MYFREE(newcl);
										MYFREE(equality);
										return(NOMEMORY);
									}

									/* Check empty clause without assertions. */
									if ((newcl->part2.bin->formula[0]==UNITEND)&&(newcl->part2.bin->asserts==NULL)){
										MYFREE(equality);
										if (NULL==(kbset.frstprfnode=mallocmn(sizeof(proofnode)))){
											return(NOMEMORY);
										}
										kbset.frstprfnode->type=ACLAUSE;
										kbset.frstprfnode->ptr.Aclause=newcl;
										return(1);
									}
									break;
							}
						}
					}
				} while(ii);
			}
		}
	}

	/* Free memory and return unsuccessful simplification. */
	MYFREE(equality);
	return(0);
} /* BkSubsResltn */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform subsumption resolution inference)  OCJ
 *
 *    This function tries to perform subsumption resolution simplification
 *    of a clause by another clause. The simplified clause is placed in
 *    new allocated storage and its parents are set, but no other data is
 *    included in the clause cmprefix.
 *
 *    The check for a valid subsumption resolution is done by negating
 *    literal2 and checking that the clause in binp subsumes the input
 *    clause.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified.
 *    binp: Pointer to binprefix structure of clause that simplifies
 *          input clause.
 *    literal2: Pointer to the term whose negation must be a generalization
 *              of a literal in clause.
 *    pnewcl: Address of pointer to cmprefix structure of new clause to be
 *            allocated for the clause that results from simplification.
 *
 *  RETURNS:
 *
 *    0 -> The clause was not simplified.
 *    1 -> The clause was successfully simplified.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t SubsResltn(cmprefix *clause,binprefix *binp,uint8_t *literal2,cmprefix **pnewcl){
	auto uint8_t *literal1;                              /* Pointer to literal that will be deleted in clause */
	auto subst Subst;                                    /* Auxiliary substitution */
	auto binprefix *ptr1;                                /* Auxiliary pointer */
	auto int32_t ii,nn;                                  /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                     /* Auxiliary pointer */

	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	ptr10=Decompile(binp);
	MYFREE(ptr10);
	#endif

	/* Initialize substitution. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Check if this is a valid subsumption resolution. */
	literal2[0]^=NEGATED;
	literal1=literal2;
	ii=IsSubsumedByClause(&clause->part2.bin->formula[0],clause->part2.bin->literals,
			binp,&Subst,&literal1);
	MYFREE(Subst.buffer);
	literal2[0]^=NEGATED;
	if (ii!=1){
		return(ii);
	}

	/* Allocate storage for new clause and copy input binary clause. */
	if (NULL==((*pnewcl)=mallocmn(sizeof(cmprefix)))){
		return(NOMEMORY);
	}
	nn=clause->part2.bin->size+binpsize;
	if (NULL==((*pnewcl)->part2.bin=mallocmn(nn))){
		MYFREE(*pnewcl);
		return(NOMEMORY);
	}
	memcpy((*pnewcl)->part2.bin,clause->part2.bin,nn);

	/* Delete literal in new clause. */
	DeleteLiteral((*pnewcl)->part2.bin,
			((uint8_t *)(*pnewcl)->part2.bin)+(literal1-(uint8_t *)clause->part2.bin));

	/* Reallocate the new clause to the exact length needed. */
	if (NULL==(ptr1=reallocmn((*pnewcl)->part2.bin,binpsize+(*pnewcl)->part2.bin->size))){
		MYFREE((*pnewcl)->part2.bin);
		MYFREE(*pnewcl);
		return(NOMEMORY);
	}

	/* Set simplified clause binprefix data. */
	(*pnewcl)->part2.bin=ptr1;
	ptr1->equflg=0;
	ptr1->clause=*pnewcl;

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile((*pnewcl)->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Set simplified clause cmprefix data. */
	(*pnewcl)->parent1=clause;
	(*pnewcl)->parent2=binp->clause;

	return(1);
} /* SubsResltn */

#ifdef UNITCLRESOLUTION
/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform unit clause resolution inference)  OCJ
 *
 *    This function performs the resolution inference between the
 *    unit clause passed as argument and another matching active
 *    or passive unit clause in the working KB. However, as in this
 *    case a successful resolution may mean solving the problem
 *    or at least generate a contradiction clause this resolution
 *    is exceptionally considered a simplification inference.
 *
 *    It is assumed that the clause is a unit clause but this
 *    is not verified.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause with which inferences will be made.
 *
 *  RETURNS:
 *
 *    0 -> An empty clause without assertions was NOT inferred.
 *    1 -> An empty clause without assertions was inferred.
 *    TIMEOUT -> Time has expired or other process has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *
 *--------------------------------------------------------------*/
int32_t UCResolution(cmprefix *clause){
	auto cmprefix *newclause;                            /* Inferred clause */
	auto cmprefix *parent;                               /* Second parent clause */
	auto subst Subst;                                    /* Substitution */
	auto uint8_t *equality;                              /* Pointer to equality with root terms in reverse order */
	auto uint8_t *eqterms;                               /* Pointer to equality root terms in reverse order */
	auto uint8_t *queryitm;                              /* Pointer to QueryDscTree() query item */
	auto int32_t size;                                   /* Size of eqterms buffer */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3,*ptr4,*ptr5;          /* Auxiliary pointers */
	auto int32_t ii,jj,kk,mm,nn,rr,ss;                   /* Auxiliary */

	/* Initialize substitution, stack and equality root terms buffer. */
	Subst.buffer=NULL;
	eqterms=equality=NULL;
	size=0;

	/* Get pointer to appropriate discrimination tree. */
	ptr1=&clause->part2.bin->formula[0];
	if (ptr1[0]&PREDICATE){
		if (ptr1[0]&NEGATED){
			ptr2=(((symbol *)&ptr1[1])->symbol)->discr[7];
		} else {
			ptr2=(((symbol *)&ptr1[1])->symbol)->discr[6];
		}
	} else {
		if (ptr1[0]&NEGATED){
			ptr2=equkey.discr[7];
		} else {
			ptr2=equkey.discr[6];
		}
	}

	/* Loop for equality symmetry management. */
	ptr4=NextItem(ptr1,IMMED);
	mm=(ptr1[0]&EQUALITY?2:1);
	for (kk=0;kk<mm;kk++){

		/* This is the first iteration. */
		if (kk==0){
			queryitm=ptr1;

		/* This is the second iteration for an equality literal. */
		/* Build the set of root terms in reverse order. */
		} else {
			ptr3=NextItem(ptr4,OVERSUBTERMS);
			rr=ptr3-ptr4;
			ss=NextItem(ptr3,OVERSUBTERMS)-ptr3;
			if (size==0){
				if (NULL==(equality=mallocmn(size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
					MYFREE(Subst.buffer);
					MYFREE(equality);
					return(NOMEMORY);
					eqterms=&equality[1+sizeof(symbol)];
				}
			} else if (size<(2+sizeof(symbol)+rr+ss)){
				if (NULL==(ptr5=reallocmn(equality,size=2+sizeof(symbol)+rr+ss+CLAUSE_CHUNK_SIZE))){
					MYFREE(Subst.buffer);
					MYFREE(equality);
					return(NOMEMORY);
				}
				equality=ptr5;
				eqterms=&equality[1+sizeof(symbol)];
			}
			memcpy(equality,ptr1,1+sizeof(symbol));
			memcpy(eqterms,ptr3,ss);
			memcpy(&eqterms[ss],ptr4,rr);
			eqterms[rr+ss]=UNITEND;
			ptr4=eqterms;
			queryitm=equality;
		}

		/* Loop through candidates for unification. */
		jj=1;
		nn=0;
		while (jj){

			/* Check timeout and solution by other process. */
			gettimeofday(&timestr,NULL);
			time=timestr.tv_sec-kbset.time.tv_sec
					+(timestr.tv_usec-kbset.time.tv_usec)/1000000.0;
			if ((time>=kbset.opts.timeout)||(procctl->status==UNSATISFIABLE)
					||(procctl->status==SATISFIABLE)){
				MYFREE(Subst.buffer);
				MYFREE(equality);
				return(TIMEOUT);
			}
			if ((procctl->status==NOMEMORY)||(procctl->status==TIMEOUT)){
				MYFREE(Subst.buffer);
				MYFREE(equality);
				return(procctl->status);
			}

			/* Get candidate for unification. */
			if (0!=(jj=QueryDscTree(queryitm,&ptr3,ptr2,UNIFICATION,nn))){

				/* Check unification. */
				nn=1;
				switch (Unify(ptr1,ptr3,&Subst,ITEM2SEC|NEGATED|INITSUBST)){

					/* Not enough memory. */
					case NOMEMORY:
						MYFREE(Subst.buffer);
						MYFREE(equality);
						return(NOMEMORY);
						break;

					/* Candidate unifies with literal. */
					case 1:

						/* Generate empty clause. */
						if (NULL==(newclause=mallocmn(sizeof(cmprefix)))){
							return(NOMEMORY);
						}
						if (NULL==((newclause->part2.bin)=mallocmn(binpsize+1))){
							MYFREE(newclause);
							return(NOMEMORY);
						}
						newclause->part2.bin->formula[0]=UNITEND;
						newclause->part2.bin->clause=newclause;
						newclause->part2.bin->equflg=0;
						parent=((binprefix *)(ptr3-(binpsize+((symbol *)&ptr3[1])->offset)))->clause;
						/*if (NOMEMORY==AddAssertions(clause->part2.bin,parent->part2.bin,newclause->part2.bin)){
							MYFREE(newclause->part2.bin);
							MYFREE(newclause);
							return(NOMEMORY);
						}*/
						newclause->parent1=clause;
						newclause->parent2=parent;
						newclause->inference=RESOLUTION;
						ii=MIN(clause->part2.bin->dist2conj,parent->part2.bin->dist2conj);
						if (ii<MAXDIST2CONJ){
							newclause->part2.bin->dist2conj=ii+1;
						} else {
							newclause->part2.bin->dist2conj=MAXDIST2CONJ;
						}

						/* Add empty clause as UNPROC and return. */
						newclause->flags=UNPROC;
						if (NOMEMORY==AddBinClause2KB(newclause,&kbset,SETFNUMBER)){
							/*if (newclause->part2.bin->asserts!=NULL){
								MYFREE(newxlause->part2.bin->asserts);
							}*/
							MYFREE(newclause->part2.bin);
							MYFREE(newclause);
							return(NOMEMORY);
						}
						if (NULL==(kbset.frstprfnode=mallocmn(sizeof(proofnode)))){
							/*if (newclause->part2.bin->asserts!=NULL){
								MYFREE(newxlause->part2.bin->asserts);
							}*/
							MYFREE(newclause->part2.bin);
							MYFREE(newclause);
							return(NOMEMORY);
						}
						kbset.frstprfnode->type=ACLAUSE;
						kbset.frstprfnode->ptr.Aclause=newclause;
						kbset.status=UNSATISFIABLE;
						return(1);
						break;
				}
			}
		}
	}

	/* Free memory and return. */
	MYFREE(Subst.buffer);
	MYFREE(equality);
	return(0);
} /* UCResolution */
#endif
